------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a GNU GPL real time scheduling analysis tool.
-- This program provides services to automatically check performances
-- of real time architectures.
--
-- Copyright (C) 2002-2010,  by Frank Singhoff, Alain Plantec, Jerome Legrand
--
-- The Cheddar project was started in 2002 by
-- the LISyC Team, University of Western Britanny.
--
-- Since 2008, Ellidiss technologies also contributes to the development of
-- Cheddar and provides industrial support.
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : cheddar@listes.univ-brest.fr
--
-----------------------------------------------------------------------------
-- Last update :
--    $Rev: 209 $
--    $Date: 2008-06-29 14:47:31 +0200 (dim, 29 jun 2008) $
--    $Author: singhoff $
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Ada.Strings.Unbounded;     use Ada.Strings.Unbounded;
with Parameters;                use Parameters;
with Parameters.extended;       use Parameters.extended;
use Parameters.extended.Framework_Parameters_Table_Package;
with Framework_Config;          use Framework_Config;
with Systems;                   use Systems;
with Tables;
with Convert_Strings;
with Convert_Unbounded_Strings;

package Call_Framework is

   type Framework_Statement_Type is (  -- Compute a set of basic simulation
                                       --services
                                       --
      Scheduling_Simulation_Basics,

   -- Compute the scheduling time line
   --
      Scheduling_Simulation_Time_Line,

   -- Compute from a time line, the number
   -- of preemption
   --
      Scheduling_Simulation_Preemption_Number,

   -- Compute from a time line, the number
   -- of context switch
   --
      Scheduling_Simulation_Context_Switch_Number,

   -- Compute maximum bound on task response
   -- time from a simulation
   --
      Scheduling_Simulation_Response_Time,

   -- Compute from a simulation, for each task,
   -- response times for all its activation
   --
      Scheduling_Simulation_All_Response_Time,

   -- Compute from a simulation, for each task,
   -- its blocking time on shared resources
   --
      Scheduling_Simulation_Blocking_Time,

   -- Find in a simulation, if priority
   -- inversions exist
   --
      Scheduling_Simulation_Priority_Inversion,

   -- Find in a simulation, if
   -- deadlocks exist
   --
      Scheduling_Simulation_Deadlock,

   -- Run user's defined event handlers.
   -- This service allows users to dfine their
   -- own scheduling data inspectors
   --
      Scheduling_Simulation_Run_Event_Handler,

   -- Compute a set of basic feasibility tests services
   --
      Scheduling_Feasibility_Basics,

   -- Compute bound on task response time
   --
      Scheduling_Feasibility_Response_Time,

   -- Compute feasibility test based on cpu utilization factor
   --
      Scheduling_Feasibility_Cpu_Utilization,

   -- Set task priority according to Deadline Monotonic
   -- scheduler
   --
      Scheduling_Set_Priorities_According_To_Deadline_Monotonic,

   -- Set task priority according to Rate Monotonic
   -- scheduler
   --
      Scheduling_Set_Priorities_According_To_Rate_Monotonic,

   -- Compute shared resources blocking time
   --
      Scheduling_Feasibility_Compute_Blocking_Time,

   -- Compute shared resources blocking time
   -- and update the task set wxith the resulting blocking time
   --
      Scheduling_Feasibility_Compute_And_Set_Blocking_Time,

   -- Apply several multiprocessor partition rules
   --
      Scheduling_Feasibility_First_Fit,
      Scheduling_Feasibility_General_Task,
      Scheduling_Feasibility_Next_Fit,
      Scheduling_Feasibility_Best_Fit,
      Scheduling_Feasibility_Small_Task,

   -- Compute the global memory requirement
   --
      Global_Memory_Set_Requirement_Analysis,
      Global_Memory_Compute_Requirement_Analysis,

   -- Compute bound on buffer size
   --
      Buffer_Feasibility_Tests,

   -- Compute buffer utilization factor from a scheduling
   -- simulation
   --
      Buffer_Scheduling_Simulation,

   -- Compute a random law mean response time and its
   --  variance from a given scheduling
   --
      Random_Response_Time_Density,

   -- Compute end to end response time
   -- response time between several tasks which can
   -- be hostedon different processors
   --
      Dependency_Compute_End_To_End_Response_Time_One_Step,
      Dependency_Set_End_To_End_Response_Time_One_Step,
      Dependency_Compute_End_To_End_Response_Time_All_Steps,
      Dependency_Set_End_To_End_Response_Time_All_Steps,

   -- Assign deadlines or fixed priority
   -- to a set of tasks depending the precedency relatioship
   --
      Dependency_Compute_Chetto_Blazewicz_Priority,
      Dependency_Compute_Chetto_Blazewicz_Deadline,
      Dependency_Set_Chetto_Blazewicz_Priority,
      Dependency_Set_Chetto_Blazewicz_Deadline);

   pragma Convention (C, Framework_Statement_Type);

   procedure To_Framework_Statement is new Convert_Strings (
      Framework_Statement_Type,
      Scheduling_Simulation_Basics);
   procedure To_Framework_Statement is new Convert_Unbounded_Strings (
      Framework_Statement_Type,
      Scheduling_Simulation_Basics);

   -- Store a computing request.
   -- If "target" contains something, the computing should only
   -- done on one processor which is referecendd in "target"
   -- if "target" is a empty string, we compute for all processors
   --
   type Framework_Request is record
      Statement : Framework_Statement_Type;
      Param     : Framework_Parameters_Table;
      Target    : Unbounded_String;
   end record;

   type Framework_Request_Ptr is access Framework_Request;

   Invalid_Number_Of_Parameter : exception;
   Unexpected_Parameter : exception;
   Parameter_Type_Error : exception;

   procedure Put (F : in Framework_Request);
   procedure Initialize (F : in out Framework_Request);

   package Framework_Request_Package is new Tables (
      Framework_Request,
      Max_Requests,
      Put);
   use Framework_Request_Package;

   subtype Framework_Request_Range is Framework_Request_Package.Table_Range;
   subtype Framework_Request_Table is Framework_Request_Package.Table;

   type Framework_Request_Table_Ptr is access Framework_Request_Table;

   type Framework_Response is record
      Title : Unbounded_String;
      Text  : Unbounded_String;
   end record;

   type Framework_Response_Ptr is access Framework_Response;

   procedure Put (F : in Framework_Response);
   procedure Initialize (F : in out Framework_Response);

   package Framework_Response_Package is new Tables (
      Framework_Response,
      Max_Requests,
      Put);
   use Framework_Response_Package;

   subtype Framework_Response_Range is Framework_Response_Package.Table_Range;
   subtype Framework_Response_Table is Framework_Response_Package.Table;

   type Framework_Response_Table_Ptr is access Framework_Response_Table;

   type Output_Format is (Xml_Output, String_Output, Raw_Output);
   pragma Convention (C, Output_Format);

   type Perform_Order is (Total_Order, Causal_Order);
   pragma Convention (C, Perform_Order);

   -- Initialize the framework. Should be called before any
   -- use of the framework
   --
   procedure initialize (Aadl_Labels : in Boolean);

   procedure Sequential_Framework_Request
     (Sys      : in out System;
      Request  : in Framework_Request_Table;
      Response : in out Framework_Response_Table;
      Order    : in Perform_Order;
      Output   : in Output_Format);

   procedure Sequential_Framework_Request
     (Sys      : in out System_Ptr;
      Request  : in Framework_Request_Table;
      Response : in out Framework_Response_Table;
      Order    : in Perform_Order;
      Output   : in Output_Format);

   procedure Sequential_Framework_Request
     (Sys      : in out System;
      Request  : in Framework_Request_Table;
      Response : in out Framework_Response_Table);

   procedure Parallel_Framework_Request
     (Sys      : in out System;
      Request  : in Framework_Request_Table;
      Response : in out Framework_Response_Table;
      Order    : in Perform_Order := Total_Order;
      Output   : in Output_Format := String_Output);

   procedure Distributed_Framework_Request
     (Sys      : in out System;
      Request  : in Framework_Request_Table;
      Response : in out Framework_Response_Table;
      Order    : in Perform_Order := Total_Order;
      Output   : in Output_Format := String_Output);

end Call_Framework;
