------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a GNU GPL real time scheduling analysis tool.
-- This program provides services to automatically check performances
-- of real time architectures.
--
-- Copyright (C) 2002-2010,  by Frank Singhoff, Alain Plantec, Jerome Legrand
--
-- The Cheddar project was started in 2002 by
-- the LISyC Team, University of Western Britanny.
--
-- Since 2008, Ellidiss technologies also contributes to the development of
-- Cheddar and provides industrial support.
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : cheddar@listes.univ-brest.fr
--
-----------------------------------------------------------------------------
-- Last update :
--    $Rev: 209 $
--    $Date: 2008-06-29 14:47:31 +0200 (dim, 29 jun 2008) $
--    $Author: singhoff $
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with interpreter;            use interpreter;
use interpreter.Sets_Type_Package;
with Events;                 use Events;
with Systems;                use Systems;
with Processors;             use Processors;
with Buffer_Set;             use Buffer_Set;
use Buffer_Set.Generic_Buffer_Set;
with Message_Set;            use Message_Set;
use Message_Set.Generic_Message_Set;
with Event_Analyzers;        use Event_Analyzers;
with Ada.Strings.Unbounded;  use Ada.Strings.Unbounded;
with Framework_Config;       use Framework_Config;
with Scheduling_Sequences;   use Scheduling_Sequences;
with Task_Dependencies;      use Task_Dependencies;
with Indexed_Tables;
with Unchecked_Deallocation;
with Processor_Set;          use Processor_Set;
use Processor_Set.Generic_Processor_Set;
with Address_Space_Set;      use Address_Space_Set;
use Address_Space_Set.Generic_Address_Space_Set;
with Time_Units;             use Time_Units;
use Time_Units.A_Time_Unit;
with Task_Set;               use Task_Set;
use Task_Set.Generic_Task_Set;
with Resources;              use Resources;
use Resources.Task_Lists;
with Resource_Set;           use Resource_Set;
use Resource_Set.Generic_Resource_Set;

package Multiprocessor_Services is

   --------------------------------------------------
   -- The types below define data layout to
   -- store information computed at simulation time
   --------------------------------------------------

   type Scheduling_Result is record
      Scheduling_Msg : Unbounded_String;
      Has_Error      : Boolean;
      Error_Msg      : Unbounded_String;
      Result         : Scheduling_Sequence_Ptr;
   end record;

   type Scheduling_Result_Ptr is access all Scheduling_Result;

   procedure Free is new Unchecked_Deallocation (
      Scheduling_Result,
      Scheduling_Result_Ptr);

   procedure Put (R : in Scheduling_Result);

   procedure Initialize (R : out Scheduling_Result);

   package Scheduling_Result_Per_Processor is new Indexed_Tables (
      Scheduling_Result,
      Processor_Ptr,
      Framework_Config.Max_Processors,
      0,
      Put,
      Initialize,
      Processors.Put_Name);
   use Scheduling_Result_Per_Processor;

   subtype Scheduling_Table is Scheduling_Result_Per_Processor.Indexed_Table;
   subtype Scheduling_Table_Range is
     Scheduling_Result_Per_Processor.Indexed_Table_Range;

   type Scheduling_Table_Ptr is access all Scheduling_Table;

   procedure Free (Sched : in out Scheduling_Table_Ptr);

   ---------------------------------------
   -- Put the event table to the screen
   ---------------------------------------

   procedure Put (Sched : in Scheduling_Table_Ptr);

   ------------------------------------------------------
   -- Write/read the event table into/from a XML file
   ------------------------------------------------------

   procedure Write_To_Xml_File
     (Sched     : in Scheduling_Table_Ptr;
      Sys       : in System;
      File_Name : in String);

   procedure Write_To_Xml_File
     (Sched     : in Scheduling_Table_Ptr;
      Sys       : in System;
      File_Name : in Unbounded_String);

   procedure Read_From_Xml_File
     (Sched     : in out Scheduling_Table_Ptr;
      Sys       : in System;
      File_Name : in String);

   procedure Read_From_Xml_File
     (Sched     : in out Scheduling_Table_Ptr;
      Sys       : in System;
      File_Name : in Unbounded_String);

   -----------------------------------------------------------------
   --  This function is the major entry of the scheduling simulator
   -----------------------------------------------------------------

   procedure Build_Multiprocessor_Scheduling
     (My_Processors           : in Processors_Set;
      My_Address_Spaces       : in Address_Spaces_Set;
      My_Tasks                : in Tasks_Set;
      My_Resources            : in Resources_Set;
      My_Buffers              : in Buffers_Set;
      My_Messages             : in Messages_Set;
      My_Dependencies         : in Tasks_Dependencies_Ptr;
      Result                  : in out Scheduling_Table_Ptr;
      Event_To_Generate       : in Time_Unit_Event_Type_Boolean_Table;
      Last_Time               : in Natural;
      With_Offsets            : in Boolean := True;
      With_Precedencies       : in Boolean := True;
      With_Resources          : in Boolean := True;
      With_Task_Specific_Seed : in Boolean := True;
      Global_Seed_Value       : in Integer := 0;
      Predictable_Global_Seed : in Boolean := True);

   --------------------------------------------------------------
   -- User's defined event anayser
   -- From a scheduling table, this function do an
   -- user's defined analysis (computed with "event analyzers"
   --------------------------------------------------------------

   procedure Run_An_Event_Analyzer
     (My_Tasks          : in Tasks_Set;
      My_Processors     : in Processors_Set;
      My_Address_Spaces : in Address_Spaces_Set;
      My_Resources      : in Resources_Set;
      My_Buffers        : in Buffers_Set;
      My_Messages       : in Messages_Set;
      Sched             : in Scheduling_Table_Ptr;
      An_Event_Analyzer : in Event_Analyzer_Ptr;
      Result            : in out Unbounded_String);

   --------------------------------------------------------------
   -- Partitionning tools  : these subprograms allow users to
   -- split tasks on a set of processors
   --------------------------------------------------------------

   -- All processors should have the same scheduler
   --
   Processors_Should_Have_The_Same_Scheduler : exception;

   -- We need more processors to do partition of this task set
   --
   No_Such_Processors : exception;

   -- can we applied scheduling partition algorithms with this scheduler ?
   --
   Invalid_Scheduler_Multiprocessors : exception;

   procedure Partition_General_Task
     (Src_Processors : in Processors_Set;
      Src_Tasks      : in Tasks_Set;
      Msg            : in out Unbounded_String;
      Result_Tasks   : in out Tasks_Set);

   procedure Partition_Small_Task
     (Src_Processors : in Processors_Set;
      Src_Tasks      : in Tasks_Set;
      Msg            : in out Unbounded_String;
      Result_Tasks   : in out Tasks_Set);

   procedure Partition_Small_Task
     (Src_Processors   : in Processors_Set;
      Src_Tasks        : in Tasks_Set;
      Msg              : in out Unbounded_String;
      Result_Tasks     : in out Tasks_Set;
      Processor_Result : out Processors_Iterator);

   procedure Partition_Small_Task
     (Src_Processors   : in Processors_Set;
      Result_Tasks     : in out Tasks_Set;
      Processor_Result : out Processors_Iterator);

   procedure Rm_General_Tasks
     (Src_Processors : in Processors_Set;
      Result_Tasks   : in out Tasks_Set);

   procedure Partition_Next_Fit
     (Src_Processors : in Processors_Set;
      Src_Tasks      : in Tasks_Set;
      Msg            : in out Unbounded_String;
      Result_Tasks   : in out Tasks_Set);

   procedure Partition_First_Fit
     (Src_Processors : in Processors_Set;
      Src_Tasks      : in Tasks_Set;
      Msg            : in out Unbounded_String;
      Result_Tasks   : in out Tasks_Set);

   procedure Partition_Best_Fit
     (Src_Processors : in Processors_Set;
      Src_Tasks      : in Tasks_Set;
      Msg            : in out Unbounded_String;
      Result_Tasks   : in out Tasks_Set);

   procedure Validate_Multiprocessor_Tasks
     (Src_Processors : in Processors_Set;
      Src_Tasks      : in Tasks_Set);

end Multiprocessor_Services;
