------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a GNU GPL real time scheduling analysis tool.
-- This program provides services to automatically check performances
-- of real time architectures.
--
-- Copyright (C) 2002-2010,  by Frank Singhoff, Alain Plantec, Jerome Legrand
--
-- The Cheddar project was started in 2002 by
-- the LISyC Team, University of Western Britanny.
--
-- Since 2008, Ellidiss technologies also contributes to the development of
-- Cheddar and provides industrial support.
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : cheddar@listes.univ-brest.fr
--
-----------------------------------------------------------------------------
-- Last update :
--    $Rev: 209 $
--    $Date: 2008-06-29 14:47:31 +0200 (dim, 29 jun 2008) $
--    $Author: singhoff $
------------------------------------------------------------------------------
------------------------------------------------------------------------------
with Unbounded_Strings; use Unbounded_Strings;
with Ada.Exceptions;    use Ada.Exceptions;
with Translate;         use Translate;

package body Network_Set is

   procedure Check_Network
     (My_Networks  : in Networks_Set;
      Name         : in Unbounded_String;
      Network_Type : in Networks_Type)
   is

   begin

      if Name = "" then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Network (Current_Language) &
                " " &
                Name &
                " : " &
                Lb_Networks_Name (Current_Language) &
                Lb_Mandatory (Current_Language)));
      end if;

   end Check_Network;

   procedure Add_Network
     (My_Networks  : in out Networks_Set;
      Name         : in Unbounded_String;
      Network_Type : in Networks_Type)
   is

      My_Iterator : Iterator;
      A_Network   : Network_Ptr;

   begin

      Check_network (My_Networks, Name, Network_Type);

      if (Get_Number_Of_Elements (My_Networks) > 0) then
         Reset_Iterator (My_Networks, My_Iterator);

         loop
            Current_Element (My_Networks, A_Network, My_Iterator);

            if (Name = A_Network.Name) then
               raise Network_Already_Defined;
            end if;

            exit when Is_Last_Element (My_Networks, My_Iterator);

            Next_Element (My_Networks, My_Iterator);

         end loop;
      end if;

      A_Network := new Network;

      A_Network.Name         := Name;
      A_Network.Network_Type := Network_Type;

      Add (My_Networks, A_Network);

   exception
      when Generic_Network_Set.Full_Set =>
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String (Lb_Can_Not_Define_More_Networks (Current_Language)));

   end Add_Network;

   procedure Update_Network
     (My_Networks  : in out Networks_Set;
      Name         : in Unbounded_String;
      Network_Type : in Networks_Type)
   is

      A_Network : Network_Ptr;

   begin

      A_Network := Search_Network (My_Networks, Name);
      Delete (My_Networks, A_Network);
      Add_Network (My_Networks, Name, Network_Type);

   end Update_Network;

   function Search_Network
     (My_Networks : in Networks_Set;
      Name        : in Unbounded_String)
      return        Network_Ptr
   is
      My_Iterator : Iterator;
      A_Network   : Network_Ptr;
      Result      : Network_Ptr;

      Found : Boolean := False;

   begin

      Reset_Iterator (My_Networks, My_Iterator);

      loop
         Current_Element (My_Networks, A_Network, My_Iterator);

         if (A_Network.Name = Name) then
            Found  := True;
            Result := A_Network;
         end if;

         exit when Is_Last_Element (My_Networks, My_Iterator);

         Next_Element (My_Networks, My_Iterator);

      end loop;

      if not Found then
         raise Network_Not_Found;
      end if;

      return Result;

   end Search_Network;

   function Export_Xml
     (My_Networks : in Networks_Set)
      return        Unbounded_String
   is
      My_Iterator : Network_Iterator;
      A_Network   : Network_Ptr;

      Result : Unbounded_String := Empty_String;

   begin

      if not Is_Empty (My_Networks) then

         Result := Result &
                   To_Unbounded_String ("<networks>") &
                   Unbounded_Lf;

         Reset_Iterator (My_Networks, My_Iterator);

         loop
            Current_Element (My_Networks, A_Network, My_Iterator);

            Result := Result &
                      To_Unbounded_String (ASCII.HT & "<network>") &
                      Unbounded_Lf;

            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "<name>" &
                  To_String (A_Network.Name) &
                  "</name>") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "<type>" &
                  A_Network.Network_Type'Img &
                  "</type>") &
              Unbounded_Lf;

            Result := Result &
                      To_Unbounded_String (ASCII.HT & "</network>") &
                      Unbounded_Lf;

            exit when Is_Last_Element (My_Networks, My_Iterator);

            Next_Element (My_Networks, My_Iterator);

         end loop;
         Result := Result &
                   To_Unbounded_String ("</networks>") &
                   Unbounded_Lf;
      end if;

      return Result;

   end Export_Xml;

   function Export_Aadl_Implementations
     (My_Networks : in Networks_Set)
      return        Unbounded_String
   is
      My_Iterator : Iterator;
      A_Network   : Network_Ptr;

      Result : Unbounded_String := Empty_String;

   begin

      if not Is_Empty (My_Networks) then

         Reset_Iterator (My_Networks, My_Iterator);

         loop
            Current_Element (My_Networks, A_Network, My_Iterator);
            exit when Is_Last_Element (My_Networks, My_Iterator);

            Result := Result &
                      To_Unbounded_String
                         ("bus " & To_String (A_Network.Name));
            Result := Result &
                      To_Unbounded_String
                         ("end " & To_String (A_Network.Name) & ";");

            Result := Result &
                      To_Unbounded_String
                         ("bus implementation " &
                          To_String (A_Network.Name) &
                          ".Impl");
            Result := Result &
                      To_Unbounded_String
                         ("end " & To_String (A_Network.Name) & ".Impl");

            Next_Element (My_Networks, My_Iterator);

         end loop;
      end if;

      return Result;

   end Export_Aadl_Implementations;

   function Export_Aadl_Declarations
     (My_Networks  : in Networks_Set;
      Number_Of_Ht : in Natural)
      return         Unbounded_String
   is
      My_Iterator : Iterator;
      A_Network   : Network_Ptr;

      Result : Unbounded_String := Empty_String;

   begin

      if not Is_Empty (My_Networks) then

         Reset_Iterator (My_Networks, My_Iterator);

         loop
            Current_Element (My_Networks, A_Network, My_Iterator);
            exit when Is_Last_Element (My_Networks, My_Iterator);

            for I in 1 .. Number_Of_Ht loop
               Result := Result & ASCII.HT;
            end loop;
            Result :=
              Result &
              To_Unbounded_String
                 ("instancied_" &
                  To_String (A_Network.Name) &
                  " : bus " &
                  To_String (A_Network.Name) &
                  ".Impl;");

            Next_Element (My_Networks, My_Iterator);

         end loop;
      end if;

      return Result;

   end Export_Aadl_Declarations;

   procedure Delete_network
     (My_networks    : in out Networks_Set;
      A_network_name : in Unbounded_String)
   is

      Tmp         : Networks_Set;
      A_Network   : Network_Ptr;
      My_Iterator : Network_Iterator;

   begin

      if (Get_Number_Of_Elements (My_networks) > 0) then

         Reset_Iterator (My_networks, My_Iterator);
         loop
            Current_Element (My_networks, A_Network, My_Iterator);

            if (A_Network.Name = A_network_name) then
               Add (Tmp, A_Network);
            end if;

            exit when Is_Last_Element (My_networks, My_Iterator);
            Next_Element (My_networks, My_Iterator);
         end loop;

         if not Is_Empty (Tmp) then
            Reset_Iterator (My_networks, My_Iterator);
            loop
               Current_Element (Tmp, A_Network, My_Iterator);

               Delete (My_networks, A_Network);

               exit when Is_Last_Element (Tmp, My_Iterator);
               Next_Element (Tmp, My_Iterator);
            end loop;

            Reset (Tmp, False);
         end if;
      end if;

   end Delete_network;

   --

end Network_Set;
