------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a GNU GPL real time scheduling analysis tool.
-- This program provides services to automatically check performances
-- of real time architectures.
--
-- Copyright (C) 2002-2010,  by Frank Singhoff, Alain Plantec, Jerome Legrand
--
-- The Cheddar project was started in 2002 by
-- the LISyC Team, University of Western Britanny.
--
-- Since 2008, Ellidiss technologies also contributes to the development of
-- Cheddar and provides industrial support.
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : cheddar@listes.univ-brest.fr
--
-----------------------------------------------------------------------------
-- Last update :
--    $Rev: 97 $
--    $Date: 2007-07-20 15:17:15 +0200 (Fri, 20 Jul 2007) $
--    $Author: singhoff $
------------------------------------------------------------------------------
------------------------------------------------------------------------------
with Io_Tools;                      use Io_Tools;
with Ada.IO_Exceptions;             use Ada.IO_Exceptions;
with Ada.Strings.Unbounded.Text_IO; use Ada.Strings.Unbounded.Text_IO;
with Ada.Exceptions;                use Ada.Exceptions;
with Translate;                     use Translate;
with Parameters;                    use Parameters;
use Parameters.User_Defined_Parameters_Table_Package;
with Debug;                         use Debug;

package body Scheduler.user_defined.Generated.Compiled is

   function Export_Aadl_Properties
     (My_Scheduler : in compiled_user_defined_Scheduler;
      Number_Of_Ht : in Natural)
      return         Unbounded_String
   is
      Result : Unbounded_String :=
         Export_Aadl_Properties
           (Generic_Scheduler (My_Scheduler),
            Number_Of_Ht);

   begin

      for I in 1 .. Number_Of_Ht loop
         Result := Result & ASCII.HT;
      end loop;

      Result := Result &
                To_Unbounded_String
                   ("Source_Text => """ &
                    To_String (Get_Behavior_File_Name (My_Scheduler)) &
                    """;") &
                Unbounded_Lf;

      return Result;

   end Export_Aadl_Properties;

   function Export_Xml
     (My_Scheduler : in compiled_user_defined_Scheduler)
      return         Unbounded_String
   is
      Result        : Unbounded_String         := Empty_String;
      File          : Unbounded_String         := Empty_String;
      Quantum       : constant Natural         := Get_Quantum (My_Scheduler);
      Is_Preemptive : constant Preemptive_Type :=
         Get_Preemptive (My_Scheduler);
      Name          : Unbounded_String         := Get_Name (My_Scheduler);

   begin

      Result := Result &
                To_Unbounded_String (ASCII.HT & ASCII.HT & "<scheduler");

      if Quantum /= 0 then
         Result := Result &
                   To_Unbounded_String (" quantum=""" & Quantum'Img & """");
      end if;

      if Is_Preemptive /= Preemptive then
         Result := Result &
                   To_Unbounded_String
                      (" is_preemptive=""" & Is_Preemptive'Img & """");
      end if;

      File := Get_Behavior_File_Name (My_Scheduler);

      Result := Result &
                To_Unbounded_String
                   (" parametric_file_name=""" & To_String (File) & """");

      Result := Result & To_Unbounded_String (">");
      Result := Result &
                To_Unbounded_String (" ") &
                Name &
                To_Unbounded_String (" ");
      Result := Result &
                To_Unbounded_String ("</scheduler>") &
                Unbounded_Lf;

      return Result;

   end Export_Xml;

   procedure Initialize
     (A_Scheduler : in out compiled_user_defined_Scheduler)
   is
   begin
      Reset (A_Scheduler);
      A_Scheduler.Name            := Compiled_User_Defined_Protocol;
      A_Scheduler.String_Behavior := new Unbounded_String;
   end Initialize;

   function Copy
     (A_Scheduler : in compiled_user_defined_Scheduler)
      return        Generic_Scheduler_Ptr
   is
      Ptr : compiled_user_defined_Scheduler_Ptr;

   begin

      Ptr := new compiled_user_defined_Scheduler;

      Ptr.Name               := A_Scheduler.Name;
      Ptr.Is_Preemptive      := A_Scheduler.Is_Preemptive;
      Ptr.Quantum            := A_Scheduler.Quantum;
      Ptr.Previously_Elected := A_Scheduler.Previously_Elected;

      Ptr.Behavior_File_Name  := A_Scheduler.Behavior_File_Name;
      Ptr.String_Behavior     := new Unbounded_String;
      Ptr.String_Behavior.all := A_Scheduler.String_Behavior.all;
      return Generic_Scheduler_Ptr (Ptr);

   end Copy;

   procedure Put (My_Scheduler : in compiled_user_defined_Scheduler) is
   begin
      Put (Generic_Scheduler (My_Scheduler));
      Put_Line ("Behavior file  name : " & My_Scheduler.Behavior_File_Name);
      Put_Line ("Behavior file  : " & My_Scheduler.String_Behavior.all);
   end Put;

   procedure Do_Election
     (My_Scheduler      : in out compiled_user_defined_Scheduler;
      Si                : in out Scheduling_Information;
      Result            : in out Scheduling_Sequence_Ptr;
      Msg               : in out Unbounded_String;
      Current_Time      : in Natural;
      Processor_Name    : in Unbounded_String;
      My_Dependencies   : in Tasks_Dependencies_Ptr;
      With_Offsets      : in Boolean;
      With_Precedencies : in Boolean;
      With_Resources    : in Boolean;
      Event_To_Generate : in Time_Unit_Event_Type_Boolean_Table;
      Elected           : in out Tasks_Range;
      No_Task           : in out Boolean)
   is

   begin
      Run_Priority_Section
        (My_Scheduler,
         Si,
         Current_Time,
         Processor_Name,
         My_Dependencies,
         Msg);

      Run_Election_Section
        (My_Scheduler,
         Si,
         Current_Time,
         Processor_Name,
         My_Dependencies,
         With_Offsets,
         With_Precedencies,
         With_Resources,
         Elected,
         No_Task);
   end Do_Election;

   procedure Specific_Scheduler_Initialization
     (My_Scheduler   : in out compiled_user_defined_Scheduler;
      Si             : in out Scheduling_Information;
      Processor_Name : in Unbounded_String;
      My_Tasks       : in out Tasks_Set;
      My_Schedulers  : in Scheduler_table;
      My_Resources   : in out Resources_Set;
      My_Buffers     : in out Buffers_Set;
      My_Messages    : in Messages_Set;
      Msg            : in out Unbounded_String)
   is
   begin
      Run_Start_Section
        (Generated_User_Defined_Scheduler (My_Scheduler),
         Processor_Name,
         Msg);
   end Specific_Scheduler_Initialization;

end Scheduler.User_Defined.Generated.Compiled;
