------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a GNU GPL real time scheduling analysis tool.
-- This program provides services to automatically check performances
-- of real time architectures.
--
-- Copyright (C) 2002-2010,  by Frank Singhoff, Alain Plantec, Jerome Legrand
--
-- The Cheddar project was started in 2002 by
-- the LISyC Team, University of Western Britanny.
--
-- Since 2008, Ellidiss technologies also contributes to the development of
-- Cheddar and provides industrial support.
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : cheddar@listes.univ-brest.fr
--
-----------------------------------------------------------------------------
-- Last update :
--    $Rev: 209 $
--    $Date: 2008-06-29 14:47:31 +0200 (dim, 29 jun 2008) $
--    $Author: singhoff $
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;
with Framework_Config;      use Framework_Config;
with Tasks;                 use Tasks;
with Task_Set;              use Task_Set;
use Task_Set.Generic_Task_Set;
with Buffers;               use Buffers;
use Buffers.Buffer_Roles;
with Buffer_Set;            use Buffer_Set;
use Buffer_Set.Generic_Buffer_Set;
with Messages;              use Messages;
with Message_Set;           use Message_Set;
use Message_Set.Generic_Message_Set;
with Dependencies;          use Dependencies;
with Sets;

package Task_Dependencies is

   -- Raised when a dependency accessor do not find the
   -- required dependency
   --
   Dependency_Not_Found : exception;

   --------------------------------------------
   -- A set of half dependencies
   --------------------------------------------
   package Half_Dep_Set is new Sets (
      Max_Element => Framework_Config.Max_Tasks_Dependencies,
      Element => Dependency_Ptr,
      Free => Free,
      Copy => Copy,
      Put => Put);
   use Half_Dep_Set;

   type Tasks_Dependencies is new Half_Dep_Set.Set with private;

   type Tasks_Dependencies_Ptr is access all Tasks_Dependencies'Class;

   subtype Tasks_Dependencies_Range is Half_Dep_Set.Element_Range;
   subtype Tasks_Dependencies_Iterator is Half_Dep_Set.Iterator;

   ----------------------------------------------------------
   --  Basic services
   ----------------------------------------------------------

   procedure Free (My_Dependencies : in out Tasks_Dependencies_Ptr);

   function Export_Xml
     (My_Dependencies : in Tasks_Dependencies_Ptr)
      return            Unbounded_String;

   function Export_Aadl_Properties
     (My_Dependencies : in Tasks_Dependencies_Ptr;
      Number_Of_Ht    : in Natural)
      return            Unbounded_String;

   procedure Reset
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      Free_Object     : in Boolean := False);

   procedure Duplicate
     (Src  : in Tasks_Dependencies_Ptr;
      Dest : in out Tasks_Dependencies_Ptr);

   procedure Put (My_Dependencies : in Tasks_Dependencies_Ptr);

   -----------------------------------------------------------------------
   --  Services below allow us to test
   --  existing dependencies.
   -----------------------------------------------------------------------

   -- Sub programs to test precedency constraints : dependency between two
   --tasks
   --

   function Get_A_Successor
     (My_Dependencies : in Tasks_Dependencies_Ptr;
      My_Task         : in Generic_Task_Ptr)
      return            Generic_Task_Ptr;

   function Get_A_Predecessor
     (My_Dependencies : in Tasks_Dependencies_Ptr;
      My_Task         : in Generic_Task_Ptr)
      return            Generic_Task_Ptr;

   function Get_Successors_List
     (My_Dependencies : in Tasks_Dependencies_Ptr;
      My_Task         : in Generic_Task_Ptr)
      return            Tasks_Set;

   function Get_Predecessors_List
     (My_Dependencies : in Tasks_Dependencies_Ptr;
      My_Task         : in Generic_Task_Ptr)
      return            Tasks_Set;

   function Has_Predecessor
     (My_Dependencies : in Tasks_Dependencies_Ptr;
      My_Task         : in Generic_Task_Ptr)
      return            Boolean;

   function Has_Successor
     (My_Dependencies : in Tasks_Dependencies_Ptr;
      My_Task         : in Generic_Task_Ptr)
      return            Boolean;

   -- Get a task without predecessor
   --
   function Get_A_Root_Task
     (My_Dependencies : in Tasks_Dependencies_Ptr)
      return            Generic_Task_Ptr;

   -- Get a task without successor
   --
   function Get_A_Leaf_Task
     (My_Dependencies : in Tasks_Dependencies_Ptr)
      return            Generic_Task_Ptr;

   -- Get all tasks without predecessor
   --
   function Get_Root_Tasks
     (My_Dependencies : in Tasks_Dependencies_Ptr)
      return            Tasks_Set;

   -- Get all tasks without successor
   --
   function Get_Leaf_Tasks
     (My_Dependencies : in Tasks_Dependencies_Ptr)
      return            Tasks_Set;

   -- Check is a task dependencies is acyclic or not
   --
   function Is_Acyclic
     (My_Dependencies : in Tasks_Dependencies_Ptr)
      return            Boolean;

   function Is_Cyclic
     (My_Dependencies : in Tasks_Dependencies_Ptr)
      return            Boolean;

   ----------------------------------------------------------------------
   --  CURRENTLY, IN THIS PACKAGE, THERE IS NO SUB PROGRAM TO TEST
   --    DEPENDENCIES WITH MESSAGES AND BUFFERS
   --
   --  IN THE FUTURE, SOME SUB-PROGRAMS AS BELOW SHOULD BE ADDED
   --   (ex : Has_Predecessor,  Get_A_Successor  Get_A_Predecessor
   --       Get_Successors_List  Get_Predecessors_List Has_Predecessor
   --       Has_Successor  Get_A_Root_Task  Get_A_Leaf_Task
   --       Get_Root_Tasks Get_Leaf_Tasks
   --
   --        but with "Message_Ptr" and "Buffer_Ptr" instead of "Task_Ptr"
   --        return values to test such properties
   ----------------------------------------------------------------------

   -- Sub programs to test buffer constraints : dependency between a task
   -- and a buffer
   --

   function Has_Buffer_To_Read
     (My_Dependencies : in Tasks_Dependencies_Ptr;
      My_Task         : in Generic_Task_Ptr)
      return            Boolean;

   function Has_Buffer_To_Write
     (My_Dependencies : in Tasks_Dependencies_Ptr;
      My_Task         : in Generic_Task_Ptr)
      return            Boolean;

   ----------------------------------------------------------------------
   --  Services below allow you to delete
   --  dependencies.
   ----------------------------------------------------------------------

   -- Delete ALL dependencies  of a
   -- processor (parameter "A_Processor")
   --
   procedure Delete_Processor_All_Task_Dependencies
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      A_Processor     : in Unbounded_String);

   -- Delete ALL dependencies  of an
   -- address space (parameter "An_Address_Space")
   --
   procedure Delete_Address_Space_All_Task_Dependencies
     (My_Dependencies  : in out Tasks_Dependencies_Ptr;
      An_Address_Space : in Unbounded_String);

   -- Delete ALL dependencies :
   --  1) of a task (if My_Dep_Name is a Task_ptr)
   --  2) of a message  (if My_Dep_Name is a Message_ptr)
   --  3) of a buffer  (if My_Dep_Name is a Buffer_ptr)
   --
   procedure Delete_All_Task_Dependencies
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      A_Task          : in Generic_Task_Ptr);
   procedure Delete_All_Task_Dependencies
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      A_Buffer        : in Buffer_Ptr);
   procedure Delete_All_Task_Dependencies
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      A_Message       : in Message_Ptr);

   -- Remove ONE dependncy between two tasks
   -- (remove a precedency relationship, not with bufer of message)
   --
   procedure Delete_One_Task_Dependency
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      From_Task       : in Generic_Task_Ptr;
      To_Task         : in Generic_Task_Ptr);

   -- Remove ONE dependency between a task and
   --   a message or a buffer
   --
   procedure Delete_One_Task_Dependency
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      A_Task          : in Generic_Task_Ptr;
      A_Buffer        : in Buffer_Ptr;
      A_Type          : in Orientation_Dependency_Type);
   procedure Delete_One_Task_Dependency
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      A_Task          : in Generic_Task_Ptr;
      A_Message       : in Message_Ptr;
      A_Type          : in Orientation_Dependency_Type);

   ---------------------------------------
   --  Services below allow to update
   --  dependencies.
   ---------------------------------------

   -- Update ALL dependencies :
   --  1) of a task (if My_Dep_Name is a Task_ptr)
   --  2) of a message  (if My_Dep_Name is a Message_ptr)
   --  3) of a buffer  (if My_Dep_Name is a Buffer_ptr)
   --
   procedure Update_All_Task_Dependencies
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      A_Task          : in Generic_Task_Ptr);
   procedure Update_All_Task_Dependencies
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      A_Buffer        : in Buffer_Ptr);
   procedure Update_All_Task_Dependencies
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      A_Message       : in Message_Ptr);

   ---------------------------------------
   --  Services below allow to add
   --  dependencies.
   ---------------------------------------

   -- Add ALL dependencies taken from the definition
   -- of a given buffer
   --
   procedure Add_All_Task_Dependencies
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      My_Tasks        : in Tasks_Set;
      A_Buffer        : in Buffer_Ptr);

   -- Add a dependency between a task and
   --   a message or a buffer
   --
   procedure Add_One_Task_Dependency
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      A_Task          : in Generic_Task_Ptr;
      A_Dep           : in Message_Ptr;
      A_Type          : in Orientation_Dependency_Type);

   procedure Add_One_Task_Dependency
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      A_Task          : in Generic_Task_Ptr;
      A_Dep           : in Buffer_Ptr;
      A_Type          : in Orientation_Dependency_Type);

   -- Add a precedency constraint between two tasks
   -- (not a dependency with message or buffer)
   --
   procedure Add_One_Task_Dependency
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      From_Task       : in Generic_Task_Ptr;
      To_Task         : in Generic_Task_Ptr);

private

   type Tasks_Dependencies is new Half_Dep_Set.Set with record
      Dependent_Tasks : Tasks_Set;
   end record;

end Task_Dependencies;
