------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a GNU GPL real time scheduling analysis tool.
-- This program provides services to automatically check performances
-- of real time architectures.
--
-- Copyright (C) 2002-2010,  by Frank Singhoff, Alain Plantec, Jerome Legrand
--
-- The Cheddar project was started in 2002 by
-- the LISyC Team, University of Western Britanny.
--
-- Since 2008, Ellidiss technologies also contributes to the development of
-- Cheddar and provides industrial support.
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : cheddar@listes.univ-brest.fr
--
-----------------------------------------------------------------------------
-- Last update :
--    $Rev: 209 $
--    $Date: 2008-06-29 14:47:31 +0200 (dim, 29 jun 2008) $
--    $Author: singhoff $
------------------------------------------------------------------------------
------------------------------------------------------------------------------
with Framework_Config;               use Framework_Config;
with Translate;                      use Translate;
with Project;                        use Project;
with References;
with Manual;
with Address_Spaces;                 use Address_Spaces;
with Address_Space_Set;              use Address_Space_Set;
use Address_Space_Set.Generic_Address_Space_Set;
with Processors;                     use Processors;
with Processor_Set;                  use Processor_Set;
use Processor_Set.Generic_Processor_Set;
with Resources;                      use Resources;
with Resource_Set;                   use Resource_Set;
use Resource_Set.Generic_Resource_Set;
with Networks;                       use Networks;
with Network_Set;                    use Network_Set;
use Network_Set.Generic_Network_Set;
with Time_Units;                     use Time_Units;
with Tasks;                          use Tasks;
with Task_Set;                       use Task_Set;
use Task_Set.Generic_Task_Set;
with Scheduling_Sequences;           use Scheduling_Sequences;
use Scheduling_Sequences.Density_Table_Package;
use Scheduling_Sequences.Densities_Table_Package;
with Buffers;                        use Buffers;
with Buffer_Set;                     use Buffer_Set;
use Buffer_Set.Generic_Buffer_Set;
with Ada.Strings.Unbounded;          use Ada.Strings.Unbounded;
with Messages;                       use Messages;
with Message_Set;                    use Message_Set;
use Message_Set.Generic_Message_Set;
with Event_Analyzers;                use Event_Analyzers;
with Event_Analyzer_Set;             use Event_Analyzer_Set;
use Event_Analyzer_Set.Generic_Event_Analyzer_Set;
with Call_Framework;                 use Call_Framework;
use Call_Framework.Framework_Request_Package;
use Call_Framework.Framework_Response_Package;
with Call_Scheduling_Framework;      use Call_Scheduling_Framework;
with Call_Memory_Framework;          use Call_Memory_Framework;
use Call_Memory_Framework.Buffer_Result_Package;
with Parameters;                     use Parameters;
with Parameters.extended;            use Parameters.extended;
use Parameters.extended.Framework_Parameters_Table_Package;
with Glib;                           use Glib;
with Mmi_Tools;                      use Mmi_Tools;
with Update_Event_Analyzers_Set_Pkg; use Update_Event_Analyzers_Set_Pkg;
with Update_Buffers_Set_Pkg;         use Update_Buffers_Set_Pkg;
with Update_Tasks_Set_Pkg;           use Update_Tasks_Set_Pkg;
with Update_Address_Spaces_Set_Pkg;  use Update_Address_Spaces_Set_Pkg;
with Update_networks_Set_Pkg;        use Update_networks_Set_Pkg;
with Update_Processors_Set_Pkg;      use Update_Processors_Set_Pkg;
with Update_Resources_Set_Pkg;       use Update_Resources_Set_Pkg;
with Update_Messages_Set_Pkg;        use Update_Messages_Set_Pkg;
with Draw_Scheduling_Pkg;            use Draw_Scheduling_Pkg;
with Scrolling_Draw_Buffer_Pkg;      use Scrolling_Draw_Buffer_Pkg;
with Scrolling_Text_Pkg;             use Scrolling_Text_Pkg;
with Question_Pkg;                   use Question_Pkg;
with Do_Scheduling_Simulation_Pkg;   use Do_Scheduling_Simulation_Pkg;
with Do_Scheduling_Feasibility_Pkg;  use Do_Scheduling_Feasibility_Pkg;
with Graph_Window_Pkg;               use Graph_Window_Pkg;
with Scheduling_Options_Pkg;         use Scheduling_Options_Pkg;
with Aadl_Options_Pkg;               use Aadl_Options_Pkg;
with Duplicate_Pkg;                  use Duplicate_Pkg;
with Schedule_From_To_Pkg;           use Schedule_From_To_Pkg;
with Draw_Time_Line_From_To_Pkg;     use Draw_Time_Line_From_To_Pkg;
with Event_Table_Analysis_Pkg;       use Event_Table_Analysis_Pkg;
with Systems;                        use Systems;
with open_Event_Table_Pkg;           use open_Event_Table_Pkg;
with Save_Event_Table_Pkg;           use Save_Event_Table_Pkg;
with Text_Tools;                     use Text_Tools;
with Multiprocessor_Services;        use Multiprocessor_Services;
use Multiprocessor_Services.Scheduling_Result_Per_Processor;

package body Root_Window_Pkg.Callbacks is

   use Gtk.Arguments;

   ----------------------------------------------------------------------
   --   Callbacks which call the framework
   ----------------------------------------------------------------------

   procedure Check_Buffer is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;

   begin

      if (Get_Number_Of_Elements (Sys.Buffers) = 0) then
         Show_Message_Box (Lb_Define_Buffers_Before (Current_Language));
         return;
      end if;

      Initialize (A_Request);
      A_Request.Statement := Buffer_Feasibility_Tests;
      Add (Request_List, A_Request);

      Sequential_Framework_Request (Sys, Request_List, Response_List);

      Write_Title_And_Text (Root_Window.Text, Response_List);

   end Check_Buffer;

   procedure On_Draw_Buffer_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;

   begin

      if (Get_Number_Of_Elements (Sys.Buffers) = 0) then
         Show_Message_Box (Lb_Define_Buffers_Before (Current_Language));
      else

         -- Check if scheduling is computed or not
         --
         if Call_Scheduling_Framework.Sched.Nb_Entries = 0 then
            Show_Message_Box (Lb_Schedule_Before (Current_Language));
            return;
         end if;

         -- For each buffer, compute and draw its size from
         -- 0 to the base period
         --
         Initialize (A_Request);
         A_Request.Statement := Buffer_Scheduling_Simulation;
         Add (Request_List, A_Request);

         Sequential_Framework_Request (Sys, Request_List, Response_List);

         Write_Title_And_Text (Root_Window.Text, Response_List);

         -- Now, draw the result on buffer utilization
         --
         for I in 0 .. Buff.Nb_Entries - 1 loop

            Gtk_New
              (Scrolling_Draw_Buffer,
               To_String (Buff.Entries (I).Item.Name));
            Show_All (Scrolling_Draw_Buffer);
            Clear (Scrolling_Draw_Buffer);
            Draw
              (Scrolling_Draw_Buffer,
               Buff.Entries (I).Item,
               Buff.Entries (I).Data,
               Call_Scheduling_Framework.Scheduling_Base_Period,
               Time_Unit_Range (Buff.Entries (I).Data.Nb_Entries));
         end loop;

      end if;
   end On_Draw_Buffer_Activate;

   -- Resource tools : blocking time evaluation
   --

   procedure On_Blocking_Time_Display_Callback
     (Object : access Gtk_Menu_Item_Record'Class)
   is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;

   begin
      if (Get_Number_Of_Elements (Sys.Resources) = 0) then
         Show_Message_Box (Lb_Define_Resources_Before (Current_Language));
      else
         Initialize (A_Request);
         A_Request.Statement := Scheduling_Feasibility_Compute_Blocking_Time;
         Add (Request_List, A_Request);

         Sequential_Framework_Request (Sys, Request_List, Response_List);

         Write_Title_And_Text (Root_Window.Text, Response_List);

      end if;
   end On_Blocking_Time_Display_Callback;

   procedure On_Blocking_Time_Update_Callback
     (Object : access Gtk_Menu_Item_Record'Class)
   is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;

   begin
      if (Get_Number_Of_Elements (Sys.Resources) = 0) then
         Show_Message_Box (Lb_Define_Resources_Before (Current_Language));
      else
         Initialize (A_Request);
         A_Request.Statement :=
           Scheduling_Feasibility_Compute_And_Set_Blocking_Time;
         Add (Request_List, A_Request);

         Sequential_Framework_Request (Sys, Request_List, Response_List);

         Write_Title_And_Text (Root_Window.Text, Response_List);

         -- Do not forget to save project
         -- since tasks are modified !!
         --
         Is_New   := False;
         Is_Saved := False;

      end if;
   end On_Blocking_Time_Update_Callback;

   -- Precedency tools : deadline, jitter and priority modifications
   --
   procedure On_Chetto_Priority_Callback
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin
      null;
   end On_Chetto_Priority_Callback;

   procedure On_Chetto_Priority_Display_Callback
     (Object : access Gtk_Menu_Item_Record'Class)
   is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;

   begin
      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));

      else
         Initialize (A_Request);
         A_Request.Statement := Dependency_Compute_Chetto_Blazewicz_Priority;
         Add (Request_List, A_Request);

         Sequential_Framework_Request (Sys, Request_List, Response_List);

         Write_Title_And_Text (Root_Window.Text, Response_List);

      end if;
   end On_Chetto_Priority_Display_Callback;

   procedure On_Chetto_Priority_Update_Callback
     (Object : access Gtk_Menu_Item_Record'Class)
   is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;

   begin
      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));

      else
         Initialize (A_Request);
         A_Request.Statement := Dependency_Set_Chetto_Blazewicz_Priority;
         Add (Request_List, A_Request);

         Sequential_Framework_Request (Sys, Request_List, Response_List);

         Write_Title_And_Text (Root_Window.Text, Response_List);

         -- Do not forget to save project
         -- since tasks are modified !!
         --
         Is_New   := False;
         Is_Saved := False;

      end if;
   end On_Chetto_Priority_Update_Callback;

   procedure On_Chetto_Deadline_Callback
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin
      null;
   end On_Chetto_Deadline_Callback;

   procedure On_Chetto_Deadline_Display_Callback
     (Object : access Gtk_Menu_Item_Record'Class)
   is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;

   begin
      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));

      else
         Initialize (A_Request);
         A_Request.Statement := Dependency_Compute_Chetto_Blazewicz_Deadline;
         Add (Request_List, A_Request);

         Sequential_Framework_Request (Sys, Request_List, Response_List);

         Write_Title_And_Text (Root_Window.Text, Response_List);

      end if;
   end On_Chetto_Deadline_Display_Callback;

   procedure On_Chetto_Deadline_Update_Callback
     (Object : access Gtk_Menu_Item_Record'Class)
   is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;

   begin
      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));

      else
         Initialize (A_Request);
         A_Request.Statement := Dependency_Set_Chetto_Blazewicz_Deadline;
         Add (Request_List, A_Request);

         Sequential_Framework_Request (Sys, Request_List, Response_List);

         Write_Title_And_Text (Root_Window.Text, Response_List);

         -- Do not forget to save project
         -- since tasks are modified !!
         --
         Is_New   := False;
         Is_Saved := False;

      end if;
   end On_Chetto_Deadline_Update_Callback;

   procedure On_Jitter_From_Response_Time_Callback
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin
      null;
   end On_Jitter_From_Response_Time_Callback;

   procedure On_Jitter_From_Response_Time_Display_Callback
     (Object : access Gtk_Menu_Item_Record'Class)
   is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;

   begin
      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));

      else
         Initialize (A_Request);
         A_Request.Statement :=
           Dependency_Compute_End_To_End_Response_Time_All_Steps;
         Add (Request_List, A_Request);

         Sequential_Framework_Request (Sys, Request_List, Response_List);

         Write_Title_And_Text (Root_Window.Text, Response_List);

      end if;
   end On_Jitter_From_Response_Time_Display_Callback;

   procedure On_Jitter_From_Response_Time_Update_Callback
     (Object : access Gtk_Menu_Item_Record'Class)
   is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;

   begin
      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));

      else
         Initialize (A_Request);
         A_Request.Statement :=
           Dependency_Set_End_To_End_Response_Time_All_Steps;
         Add (Request_List, A_Request);

         Sequential_Framework_Request (Sys, Request_List, Response_List);

         Write_Title_And_Text (Root_Window.Text, Response_List);

         -- Do not forget to save project
         -- since tasks are modified !!
         --
         Is_New   := False;
         Is_Saved := False;

      end if;
   end On_Jitter_From_Response_Time_Update_Callback;

   procedure On_Jitter_From_Response_Time_Display_One_Step_Callback
     (Object : access Gtk_Menu_Item_Record'Class)
   is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;

   begin
      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));

      else
         Initialize (A_Request);
         A_Request.Statement :=
           Dependency_Compute_End_To_End_Response_Time_One_Step;
         Add (Request_List, A_Request);

         Sequential_Framework_Request (Sys, Request_List, Response_List);

         Write_Title_And_Text (Root_Window.Text, Response_List);

      end if;
   end On_Jitter_From_Response_Time_Display_One_Step_Callback;

   procedure On_Jitter_From_Response_Time_Update_One_Step_Callback
     (Object : access Gtk_Menu_Item_Record'Class)
   is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;

   begin

      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));

      else
         Initialize (A_Request);
         A_Request.Statement :=
           Dependency_Set_End_To_End_Response_Time_One_Step;
         Add (Request_List, A_Request);

         Sequential_Framework_Request (Sys, Request_List, Response_List);

         Write_Title_And_Text (Root_Window.Text, Response_List);

         -- Do not forget to save project
         -- since tasks are modified !!
         --
         Is_New   := False;
         Is_Saved := False;

      end if;
   end On_Jitter_From_Response_Time_Update_One_Step_Callback;

   -- Set priorities according to a fixed priority scheduler
   --

   function Rm_Priorities_Callback
     (Value : Unbounded_String)
      return  Boolean
   is
      My_Iterator   : Processors_Iterator;
      A_Processor   : Processor_Ptr;
      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;

   begin

      loop

         Current_Element (Sys.Processors, A_Processor, My_Iterator);

         if A_Processor.Name = Value then

            Initialize (A_Request);
            A_Request.Target    := Value;
            A_Request.Statement :=
              Scheduling_Set_Priorities_According_To_Rate_Monotonic;
            Add (Request_List, A_Request);

            Sequential_Framework_Request (Sys, Request_List, Response_List);

            Write_Title_And_Text (Root_Window.Text, Response_List);

            -- Do not forget to save project
            -- since tasks are modified !!
            --
            Is_New   := False;
            Is_Saved := False;

            return True;
         end if;

         exit when Is_Last_Element (Sys.Processors, My_Iterator);
         Next_Element (Sys.Processors, My_Iterator);

      end loop;

      Show_Message_Box (Lb_Processor_Not_Found (Current_Language));
      return False;

   end Rm_Priorities_Callback;

   function Dm_Priorities_Callback
     (Value : Unbounded_String)
      return  Boolean
   is
      My_Iterator   : Processors_Iterator;
      A_Processor   : Processor_Ptr;
      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;

   begin

      loop

         Current_Element (Sys.Processors, A_Processor, My_Iterator);

         if A_Processor.Name = Value then

            Initialize (A_Request);
            A_Request.Target    := Value;
            A_Request.Statement :=
              Scheduling_Set_Priorities_According_To_Deadline_Monotonic;
            Add (Request_List, A_Request);

            Sequential_Framework_Request (Sys, Request_List, Response_List);

            Write_Title_And_Text (Root_Window.Text, Response_List);

            -- Do not forget to save project
            -- since tasks are modified !!
            --
            Is_New   := False;
            Is_Saved := False;

            return True;
         end if;

         exit when Is_Last_Element (Sys.Processors, My_Iterator);
         Next_Element (Sys.Processors, My_Iterator);

      end loop;

      Show_Message_Box (Lb_Processor_Not_Found (Current_Language));
      return False;

   end Dm_Priorities_Callback;

   procedure On_Dm_Priorities_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is

   begin

      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));
      else
         Gtk_New
           (Question,
            Lb_Set_Priorities_Dm (Current_Language),
            Lb_Processor_Name (Current_Language) & " : ",
            To_Unbounded_String (""),
            Dm_Priorities_Callback'Access);
         Show_All (Question);
      end if;

   end On_Dm_Priorities_Activate;

   procedure On_Rm_Priorities_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin

      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));
      else
         Gtk_New
           (Question,
            Lb_Set_Priorities_Rm (Current_Language),
            Lb_Processor_Name (Current_Language) & " : ",
            To_Unbounded_String (""),
            Rm_Priorities_Callback'Access);
         Show_All (Question);
      end if;

   end On_Rm_Priorities_Activate;

   -- Random services
   --

   procedure On_Compute_Response_Time_Density_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;

   begin

      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));
         return;
      end if;

      -- Check if scheduling is computed or not
      --
      if Call_Scheduling_Framework.Sched.Nb_Entries = 0 then
         Show_Message_Box (Lb_Schedule_Before (Current_Language));
         return;
      end if;

      Initialize (A_Request);
      A_Request.Statement := Random_Response_Time_Density;
      Add (Request_List, A_Request);

      Sequential_Framework_Request (Sys, Request_List, Response_List);

      Write_Title_And_Text (Root_Window.Text, Response_List);

   end On_Compute_Response_Time_Density_Activate;

   ----------------------------------------------------------------------
   --   Menu callbacks without calls to the framework
   ----------------------------------------------------------------------

   procedure Clear_Work_Space is
      Size  : Guint;
      Dummy : Boolean;
   begin
      Size := Get_Length (Root_Window.Text);
      Set_Point (Root_Window.Text, 0);
      Dummy := Forward_Delete (Root_Window.Text, Size);
      Clear_Time_Line (Draw_Scheduling);
   end Clear_Work_Space;

   function On_Root_Window_Delete_Event
     (Object : access Gtk_Widget_Record'Class;
      Params : Gtk.Arguments.Gtk_Args)
      return   Boolean
   is
   begin
      Do_Exit;
      return True;
   end On_Root_Window_Delete_Event;

   procedure On_File_Activate (Object : access Gtk_Menu_Item_Record'Class) is
   begin
      null;
   end On_File_Activate;

   procedure On_New_Activate (Object : access Gtk_Menu_Item_Record'Class) is
   begin
      Do_New;
      Clear_Work_Space;
      Rebuild_Canvas (Graph_Window);
   end On_New_Activate;

   procedure On_Open_Activate (Object : access Gtk_Menu_Item_Record'Class) is
   begin
      Do_Open;
   end On_Open_Activate;

   procedure On_Save_Activate (Object : access Gtk_Menu_Item_Record'Class) is
   begin
      Do_Save;
   end On_Save_Activate;

   procedure On_Save_As_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin
      Do_Save_As;
   end On_Save_As_Activate;

   procedure On_Exit_Activate (Object : access Gtk_Menu_Item_Record'Class) is
   begin
      Do_Exit;
   end On_Exit_Activate;

   procedure On_Update_Task_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin
      if (Get_Number_Of_Elements (Sys.Address_Spaces) = 0) then
         Show_Message_Box
           (Lb_Define_Address_Spaces_Before (Current_Language));
      else
         Gtk_New (Update_Tasks_Set, Lb_Update_Task (Current_Language));
         Show_Advanced (Update_Tasks_Set, True);
      end if;
   end On_Update_Task_Activate;

   procedure On_Update_Processor_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin
      Gtk_New (Update_Processors_Set, Lb_Update_Processor (Current_Language));
      Show_Advanced (Update_Processors_Set, True);
   end On_Update_Processor_Activate;

   procedure On_Update_Network_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin
      Gtk_New (Update_networks_Set, Lb_Update_Network (Current_Language));
      Show_Advanced (Update_networks_Set, True);
   end On_Update_Network_Activate;

   procedure On_Tools_Activate (Object : access Gtk_Menu_Item_Record'Class) is
   begin
      null;
   end On_Tools_Activate;

   procedure On_Clear_Work_Space_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin
      Clear_Work_Space;
   end On_Clear_Work_Space_Activate;

   procedure Call_Schedule_From_To_After_Do_Scheduling is
   begin

      Gtk_New
        (Schedule_From_To,
         Root_Window.Text,
         Draw_Scheduling,
         Sys,
         True);
      Show_All (Schedule_From_To);

   end Call_Schedule_From_To_After_Do_Scheduling;

   procedure On_Draw_Scheduling_Graph_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is

   begin

      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));
         return;
      end if;

      Gtk_New
        (Do_Scheduling_Simulation,
         Call_Schedule_From_To_After_Do_Scheduling'Access);
      Show_All (Do_Scheduling_Simulation);
   end On_Draw_Scheduling_Graph_Activate;

   procedure On_Options_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin
      Gtk_New (Scheduling_Options);
      Show_All (Scheduling_Options);
   end On_Options_Activate;

   procedure On_Check_Scheduling_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is

   begin

      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));
         return;
      end if;

      Gtk_New (Do_Scheduling_Feasibility, Root_Window.Text);
      Show_All (Do_Scheduling_Feasibility);
   end On_Check_Scheduling_Activate;

   procedure On_Help_Activate (Object : access Gtk_Menu_Item_Record'Class) is
   begin
      null;
   end On_Help_Activate;

   procedure On_Update_Resource_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin
      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));
      else
         Gtk_New
           (Update_Resources_Set,
            Lb_Update_Resource (Current_Language));
         Show_Advanced (Update_Resources_Set, True);
      end if;
   end On_Update_Resource_Activate;

   procedure On_About_Cheddar_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is

   begin
      Show_Message_Box (Lb_About_Text (Current_Language));
   end On_About_Cheddar_Activate;

   procedure On_Manual_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin
      Gtk_New (Scrolling_Text, Lb_Manual (Current_Language));
      Set_USize (Scrolling_Text, 690, 425);

      Write_Title_Small (Scrolling_Text, Manual.Ustr_File_Menu);
      Write_Text (Scrolling_Text, Manual.Ustr_File_Menu_Content);

      Write_Title_Small (Scrolling_Text, Manual.Ustr_Edit_Menu);
      Write_Text (Scrolling_Text, Manual.Ustr_Edit_Menu_Content);

      Write_Title_Small (Scrolling_Text, Manual.Ustr_View_Menu);
      Write_Text (Scrolling_Text, Manual.Ustr_View_Menu_Content);

      Write_Title_Small (Scrolling_Text, Manual.Ustr_Tools_Menu);
      Write_Text (Scrolling_Text, Manual.Ustr_Tools_Menu_Content);

      Write_Title_Small (Scrolling_Text, Manual.Ustr_Help_Menu);
      Write_Text (Scrolling_Text, Manual.Ustr_Help_Menu_Content);

      Show_All (Scrolling_Text);
   end On_Manual_Activate;

   procedure On_Bibliography_References_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin
      Gtk_New (Scrolling_Text, Lb_Ref (Current_Language));
      Set_USize (Scrolling_Text, 535, 440);
      Write_Text (Scrolling_Text, References.Ref_Text);
      Show_All (Scrolling_Text);
   end On_Bibliography_References_Activate;

   procedure On_Update_Message_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin
      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));
      else
         Gtk_New
           (Update_Messages_Set,
            To_Unbounded_String ("cheddar_window"));
         Show_All (Update_Messages_Set);
      end if;
   end On_Update_Message_Activate;

   procedure On_Update_Buffer_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin
      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));
      else
         Gtk_New
           (Update_Buffers_Set,
            Lb_Update_Buffer (Current_Language),
            To_Unbounded_String ("cheddar_window"));
         Show_Advanced (Update_Buffers_Set, True);
      end if;

   end On_Update_Buffer_Activate;

   procedure On_Check_Buffer_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin
      Check_Buffer;
   end On_Check_Buffer_Activate;

   procedure On_Blocking_Time_Callback
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin
      null;
   end On_Blocking_Time_Callback;

   procedure On_Duplicate_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is

   begin
      Gtk_New (Duplicate_Pkg.Duplicate);
      Show_All (Duplicate_Pkg.Duplicate);
   end On_Duplicate_Activate;

   ----------------------------------------------------------------------
   --   Pixmap callbacks without calls to the framework
   ----------------------------------------------------------------------

   procedure On_Pixmap_Draw_Scheduling_Clicked
     (Object : access Gtk_Toolbar_Record'Class)
   is

   begin

      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));
         return;
      end if;

      Gtk_New (Schedule_From_To, Root_Window.Text, Draw_Scheduling, Sys);
      Show_All (Schedule_From_To);

   end On_Pixmap_Draw_Scheduling_Clicked;

   procedure On_Pixmap_New_Clicked
     (Object : access Gtk_Toolbar_Record'Class)
   is
   begin
      Do_New;
      Clear_Work_Space;
      Rebuild_Canvas (Graph_Window);
   end On_Pixmap_New_Clicked;

   procedure On_Pixmap_Open_Clicked
     (Object : access Gtk_Toolbar_Record'Class)
   is
   begin
      Do_Open;
   end On_Pixmap_Open_Clicked;

   procedure On_Pixmap_Save_Clicked
     (Object : access Gtk_Toolbar_Record'Class)
   is
   begin
      Do_Save;
   end On_Pixmap_Save_Clicked;

   procedure On_Pixmap_Save_As_Clicked
     (Object : access Gtk_Toolbar_Record'Class)
   is
   begin
      Do_Save_As;
   end On_Pixmap_Save_As_Clicked;

   procedure On_Pixmap_Exit_Clicked
     (Object : access Gtk_Toolbar_Record'Class)
   is
   begin
      Do_Exit;
   end On_Pixmap_Exit_Clicked;

   procedure On_Pixmap_Clear_Work_Space_Clicked
     (Object : access Gtk_Toolbar_Record'Class)
   is
   begin
      Clear_Work_Space;
   end On_Pixmap_Clear_Work_Space_Clicked;

   procedure On_Pixmap_Dependencies_Clicked
     (Object : access Gtk_Toolbar_Record'Class)
   is
   begin
      if (Get_Number_Of_Elements (Sys.Address_Spaces) = 0) then
         Show_Message_Box
           (Lb_Define_Address_Spaces_Before (Current_Language));
         return;
      end if;

      Gtk_New (Graph_Window);
      Show (Graph_Window);
   end On_Pixmap_Dependencies_Clicked;

   procedure On_Pixmap_Check_Scheduling_Clicked
     (Object : access Gtk_Toolbar_Record'Class)
   is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;

   begin

      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));
         return;
      end if;

      Initialize (A_Request);
      A_Request.Statement := Scheduling_Feasibility_Basics;
      A_Request.Target    := To_Unbounded_String ("");
      Add (Request_List, A_Request);

      Sequential_Framework_Request (Sys, Request_List, Response_List);

      Write_Title_And_Text (Root_Window.Text, Response_List);

   end On_Pixmap_Check_Scheduling_Clicked;

   procedure On_Deadlock_Callback
     (Object : access Gtk_Menu_Item_Record'Class)
   is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;

   begin
      -- check if there are resources
      --
      if (Get_Number_Of_Elements (Sys.Resources) = 0) then
         Show_Message_Box (Lb_Define_Resources_Before (Current_Language));
         return;
      end if;

      -- check if tasks exist
      --
      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));
         return;
      end if;

      -- Check if scheduling is computed or not
      --
      if Call_Scheduling_Framework.Sched.Nb_Entries = 0 then
         Show_Message_Box (Lb_Schedule_Before (Current_Language));
         return;
      end if;

      Initialize (A_Request);
      A_Request.Statement := Scheduling_Simulation_Deadlock;
      Add (Request_List, A_Request);

      Sequential_Framework_Request (Sys, Request_List, Response_List);

      Write_Title_And_Text (Root_Window.Text, Response_List);

   end On_Deadlock_Callback;

   procedure On_Priority_Inversion_Callback
     (Object : access Gtk_Menu_Item_Record'Class)
   is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;

   begin
      -- check if there are resources
      --
      if (Get_Number_Of_Elements (Sys.Resources) = 0) then
         Show_Message_Box (Lb_Define_Resources_Before (Current_Language));
         return;
      end if;

      -- check if tasks exist
      --
      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));
         return;
      end if;

      -- Check if scheduling is computed or not
      --
      if Call_Scheduling_Framework.Sched.Nb_Entries = 0 then
         Show_Message_Box (Lb_Schedule_Before (Current_Language));
         return;
      end if;

      Initialize (A_Request);
      A_Request.Statement := Scheduling_Simulation_Priority_Inversion;
      Add (Request_List, A_Request);

      Sequential_Framework_Request (Sys, Request_List, Response_List);

      Write_Title_And_Text (Root_Window.Text, Response_List);

   end On_Priority_Inversion_Callback;

   procedure On_Update_Event_Analyzer_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin
      Gtk_New
        (Update_Event_Analyzers_Set,
         Lb_Update_Event_Analyzer (Current_Language));
      Show_Advanced (Update_Event_Analyzers_Set, True);
   end On_Update_Event_Analyzer_Activate;

   procedure On_Partition_Small_Task_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;

   begin

      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));
         return;
      end if;
      if (Get_Number_Of_Elements (Sys.Processors) = 0) then
         Show_Message_Box (Lb_Define_Processors_Before (Current_Language));
         return;
      end if;
      if (Get_Number_Of_Elements (Sys.Address_Spaces) = 0) then
         Show_Message_Box
           (Lb_Define_Address_Spaces_Before (Current_Language));
         return;
      end if;

      Initialize (A_Request);
      A_Request.Statement := Scheduling_Feasibility_Small_Task;
      Add (Request_List, A_Request);

      Sequential_Framework_Request (Sys, Request_List, Response_List);

      Write_Title_And_Text (Root_Window.Text, Response_List);

   end On_Partition_Small_Task_Activate;

   procedure On_Partition_Best_Fit_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;

   begin

      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));
         return;
      end if;
      if (Get_Number_Of_Elements (Sys.Processors) = 0) then
         Show_Message_Box (Lb_Define_Processors_Before (Current_Language));
         return;
      end if;
      if (Get_Number_Of_Elements (Sys.Address_Spaces) = 0) then
         Show_Message_Box
           (Lb_Define_Address_Spaces_Before (Current_Language));
         return;
      end if;

      Initialize (A_Request);
      A_Request.Statement := Scheduling_Feasibility_Best_Fit;
      Add (Request_List, A_Request);

      Sequential_Framework_Request (Sys, Request_List, Response_List);

      Write_Title_And_Text (Root_Window.Text, Response_List);

   end On_Partition_Best_Fit_Activate;

   procedure On_Partition_Next_Fit_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;

   begin

      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));
         return;
      end if;
      if (Get_Number_Of_Elements (Sys.Processors) = 0) then
         Show_Message_Box (Lb_Define_Processors_Before (Current_Language));
         return;
      end if;
      if (Get_Number_Of_Elements (Sys.Address_Spaces) = 0) then
         Show_Message_Box
           (Lb_Define_Address_Spaces_Before (Current_Language));
         return;
      end if;

      Initialize (A_Request);
      A_Request.Statement := Scheduling_Feasibility_Next_Fit;
      Add (Request_List, A_Request);

      Sequential_Framework_Request (Sys, Request_List, Response_List);

      Write_Title_And_Text (Root_Window.Text, Response_List);

   end On_Partition_Next_Fit_Activate;

   procedure On_Partition_General_Task_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;

   begin

      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));
         return;
      end if;
      if (Get_Number_Of_Elements (Sys.Processors) = 0) then
         Show_Message_Box (Lb_Define_Processors_Before (Current_Language));
         return;
      end if;
      if (Get_Number_Of_Elements (Sys.Address_Spaces) = 0) then
         Show_Message_Box
           (Lb_Define_Address_Spaces_Before (Current_Language));
         return;
      end if;

      Initialize (A_Request);
      A_Request.Statement := Scheduling_Feasibility_General_Task;
      Add (Request_List, A_Request);

      Sequential_Framework_Request (Sys, Request_List, Response_List);

      Write_Title_And_Text (Root_Window.Text, Response_List);

   end On_Partition_General_Task_Activate;

   procedure On_Partition_First_Fit_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;

   begin

      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));
         return;
      end if;
      if (Get_Number_Of_Elements (Sys.Processors) = 0) then
         Show_Message_Box (Lb_Define_Processors_Before (Current_Language));
         return;
      end if;
      if (Get_Number_Of_Elements (Sys.Address_Spaces) = 0) then
         Show_Message_Box
           (Lb_Define_Address_Spaces_Before (Current_Language));
         return;
      end if;

      Initialize (A_Request);
      A_Request.Statement := Scheduling_Feasibility_First_Fit;
      Add (Request_List, A_Request);

      Sequential_Framework_Request (Sys, Request_List, Response_List);

      Write_Title_And_Text (Root_Window.Text, Response_List);

   end On_Partition_First_Fit_Activate;

   procedure On_Load_Event_Table_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is

   begin

      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));
         return;
      end if;

      Gtk_New (open_event_table);
      Show_All (open_event_table);
   end On_Load_Event_Table_Activate;

   procedure On_Save_Event_Table_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin

      -- Check if scheduling is computed or not
      --
      if Call_Scheduling_Framework.Sched.Nb_Entries = 0 then
         Show_Message_Box (Lb_Schedule_Before (Current_Language));
         return;
      end if;

      Gtk_New (save_event_table);
      Show_All (save_event_table);

   end On_Save_Event_Table_Activate;

   procedure On_Event_Table_Services_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin
      null;
   end On_Event_Table_Services_Activate;

   procedure On_Draw_Time_Line_From_Event_Table_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin

      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));
         return;
      end if;

      -- Check if scheduling is computed or not
      --
      if Call_Scheduling_Framework.Sched.Nb_Entries = 0 then
         Show_Message_Box (Lb_Schedule_Before (Current_Language));
         return;
      end if;

      Gtk_New (Draw_Time_Line_From_To, Draw_Scheduling, Sys);
      Show_All (Draw_Time_Line_From_To);
   end On_Draw_Time_Line_From_Event_Table_Activate;

   procedure On_Compute_Scheduling_And_Generate_Event_Table_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin

      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));
         return;
      end if;

   end On_Compute_Scheduling_And_Generate_Event_Table_Activate;

   procedure On_Run_Analysis_On_Event_Table_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin

      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));
         return;
      end if;

      -- Check if scheduling is computed or not
      --
      if Call_Scheduling_Framework.Sched.Nb_Entries = 0 then
         Show_Message_Box (Lb_Schedule_Before (Current_Language));
         return;
      end if;

      Gtk_New (Event_Table_Analysis, Root_Window.Text, Sys);
      Show_All (Event_Table_Analysis);

   end On_Run_Analysis_On_Event_Table_Activate;

   procedure On_Update_Address_Space_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin
      if (Get_Number_Of_Elements (Sys.Processors) = 0) then
         Show_Message_Box (Lb_Define_Processors_Before (Current_Language));
      else
         Gtk_New
           (Update_Address_Spaces_Set,
            Lb_Update_Address_Space (Current_Language));
         Show_Advanced (Update_Address_Spaces_Set, True);
      end if;
   end On_Update_Address_Space_Activate;

   procedure On_Aadl_Activate (Object : access Gtk_Menu_Item_Record'Class) is
   begin
      null;
   end On_Aadl_Activate;

   procedure On_Export_Aadl_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin
      Do_Aadl_Export;
   end On_Export_Aadl_Activate;

   procedure On_Import_Aadl_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin
      Do_Aadl_Import;
   end On_Import_Aadl_Activate;

   procedure On_Check_Address_Space_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;

   begin

      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));
         return;
      end if;
      if (Get_Number_Of_Elements (Sys.Address_Spaces) = 0) then
         Show_Message_Box
           (Lb_Define_Address_Spaces_Before (Current_Language));
         return;
      end if;

      Initialize (A_Request);
      A_Request.Statement := Global_Memory_Set_Requirement_Analysis;
      Add (Request_List, A_Request);

      Sequential_Framework_Request (Sys, Request_List, Response_List);

      Write_Title_And_Text (Root_Window.Text, Response_List);

   end On_Check_Address_Space_Activate;

   procedure On_Check_And_Update_Address_Space_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;

   begin

      if (Get_Number_Of_Elements (Sys.Tasks) = 0) then
         Show_Message_Box (Lb_Define_Tasks_Before (Current_Language));
         return;
      end if;
      if (Get_Number_Of_Elements (Sys.Address_Spaces) = 0) then
         Show_Message_Box
           (Lb_Define_Address_Spaces_Before (Current_Language));
         return;
      end if;

      Initialize (A_Request);
      A_Request.Statement := Global_Memory_Compute_Requirement_Analysis;
      Add (Request_List, A_Request);

      Sequential_Framework_Request (Sys, Request_List, Response_List);

      Write_Title_And_Text (Root_Window.Text, Response_List);

   end On_Check_And_Update_Address_Space_Activate;

   procedure On_Export_Aadl_Standard_Properties_Set_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin
      Write_Aadl_Standard_Properties_Set_To_File (Sys);
   end On_Export_Aadl_Standard_Properties_Set_Activate;

   procedure On_Export_Cheddar_Property_Sets_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin
      Write_Cheddar_Property_Sets_To_File (Sys);
   end On_Export_Cheddar_Property_Sets_Activate;

   procedure On_Aadl_Options_Activate
     (Object : access Gtk_Menu_Item_Record'Class)
   is
   begin
      Gtk_New (Aadl_Options);
      Show_All (Aadl_Options);
   end On_Aadl_Options_Activate;

end Root_Window_Pkg.Callbacks;
