--******************************************************************************
--                                                                            **
--  File:        BIGNUMBE.ADS                                                 **
--  Description: Big_Number package specification                             **
--  Revision:    0.28 (beta version)                                          **
--  Date:        July 8, 2001                                                 **
--  Author:      Jrme Delcourt                                              **
--  Mail:        sikander@club-internet.fr                                    **
--                                                                            **
--  Copyright (c) Jrme Delcourt, 1998, 1999, 2000, 2001                     **
--  62, rue Ngrier                                                           **
--  59000 Lille                                                               **
--  FRANCE                                                                    **
--                                                                            **
--  Permission granted to use for any purpose, provided this copyright        **
--  remains attached and unmodified.                                          **
--                                                                            **
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR            **
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED            **
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       **
--                                                                            **
--******************************************************************************
--                                                                            **
-- Please, don't care of my poor english in the comments (I'm french...).     **
-- Don't hesitate to report any bug, comment, idea or improvement to my       **
-- e-mail, which is sikander@club-internet.fr.                                **
--                                                                            **
-- I tried to give a understandable name to all the functions or variables    **
-- used in this package (the beta-test functions were initially written with  **
-- some names in french). In case of problem with the choosen names, tell it  **
-- to me.                                                                     **
--                                                                            **
-- You can distribute this package, but at no charge at all, and only if      **
-- you specify where you modify it (if you do so). Don't remove the comments. **
--                                                                            **
--******************************************************************************

WITH SYSTEM;
WITH CALENDAR;
WITH TEXT_IO;
WITH ADA.STRINGS.UNBOUNDED;
WITH UStrings;


GENERIC
  TYPE Index_Type IS MOD <>;
  Nb_Item : Index_Type;


PACKAGE Big_Number IS

  --
  -- This package contains two main "sub-packages", one for big unsigned
  -- integer numbers (package Unsigned_Number), another for big signed
  -- numbers (package Signed_Number).
  --
  -- Seven other "sub-packages" are defined :
  --  * Package Conversion, for the conversion of big integers, signed or
  --    unsigned : conversion of one of this type of integers to the other or
  --    to a "LONG_LONG_FLOAT" variable;
  --  * Generic package Generic_Conversion for the conversion of a predefined
  --    type of integer in ADA (one function can be used to convert a modular
  --    type) to a big integer type (signed or unsigned);
  --  * String_Conversion, for the conversion between string of numeric
  --    characters and the types Big_Signed or Big_Unsigned;
  --  * Big_Signed_IO : for the output of a Big_Signed on the screen;
  --  * Big_Unsigned_IO : for the output of a Big_Unsigned on the screen;
  --  * Big_Fraction (complete version only);
  --  * Big_Unsigned_Random (complete version only).
  --
  -- This package is generic. It allows you to use big integers, signed or
  -- unsigned.The mantissa of these numbers is represented by an array of
  -- unsigned integers (the size of these integers is defined by My_Type_Size :
  -- see the note below).
  --
  -- For the instanciation of this package, Index_Type represents the type
  -- of array index, and Nb_Item represents the number of elements in this
  -- array. So you will be able to use numbers coded on My_Type_Size*Nb_Item
  -- bits. Even if Nb_Item can be 1, it should be greater or equal to 3, because
  -- TypeSize is set to the half of the greatest size for an integer :
  -- if Nb_Item is equal to 1 or 2, the corresponding types are usually
  -- already defined by the compiler.
  --
  -- The elements of this array can take every value allowed for a TypeSize
  -- bit number : they are not limited to a power of 10. Some other packages
  -- could use the limitation to a power of 10 in order to facilitate
  -- the IOs of these big numbers, but this has the inconvenient to waste
  -- some memory.
  --
  -- Example of instanciation :
  --
  -- WITH ModInt;     USE ModInt;
  -- WITH BigNumber;
  --
  -- PACKAGE Big_Integers IS NEW BigNumber(Byte, 64);
  --
  -- I suppose that the type Byte is defined in the package ModInt by
  -- TYPE Byte IS MOD 2**8;
  -- FOR Byte'SIZE USE 8;
  --
  -- Then, the package Big_Integers will allow you to use big integers with
  -- a mantissa of (64 * My_Type_Size / 8) bytes.
  --
  --
  -- Remark : be careful when you instanciate this package.
  --
  --  1) Choose very well the parameters of instanciation, because each
  --     big number can occupy a lot of memory. For example, if Nb_Item = 8000,
  --     the mantissa of each variable of the type Big_Unsigned will
  --     take 8000 * (My_Type_Size/8) bytes (the sign is a separate field of
  --     the record).
  --  2) Ideally, Nb_Item should be greater than or equal to 3. Otherwise,
  --     the big integers managed by this package have the same size than
  --     types predefined by your compiler : there will be only one advantage,
  --     the type Fraction.
  --

  TYPE Largest_Unsigned IS MOD SYSTEM.Max_Binary_Modulus;

  My_Type_Size : CONSTANT := Largest_Unsigned'SIZE / 2;

  TYPE My_Type IS MOD 2**My_Type_Size;
  FOR My_Type'SIZE USE My_Type_Size;


--******************************************************************************
--******************************************************************************

  TYPE Big_Signed IS PRIVATE;
  TYPE Big_Unsigned IS PRIVATE;
  TYPE Fraction IS PRIVATE;
 
  ------------------------------
  -- Definition of exceptions --
  ------------------------------

  Big_Number_Overflow : EXCEPTION;
     -- This exception is raised is raised in case of overflow during
     -- an operation on 2 big numbers.

  Division_By_Zero : EXCEPTION;
     -- This exception is raised during a tentative of division by 0 in
     -- a calculus on big numbers.

  Parameter_Error : EXCEPTION;
     -- This exception is raised in case of a bad parameter (for example,
     -- a parameter below 0 with the function Factorial(x : IN Big_Signed).

  
  --------------------------
  -- Constant definitions --
  --------------------------

  Big_Unsigned_Zero : CONSTANT Big_Unsigned;   -- = 0
  Big_Unsigned_One : CONSTANT Big_Unsigned;    -- = 1
  Big_Unsigned_Two : CONSTANT Big_Unsigned;    -- = 2
  Big_Unsigned_First : CONSTANT Big_Unsigned;  -- = 0
  Big_Unsigned_Last : CONSTANT Big_Unsigned;   -- = "Big_Unsigned'LAST"

  Big_Signed_Zero : CONSTANT Big_Signed;   -- = 0
  Big_Signed_One : CONSTANT Big_Signed;    -- = 1
  Big_Signed_First : CONSTANT Big_Signed;  -- = "Big_Signed'FIRST"
  Big_Signed_Last : CONSTANT Big_Signed;   -- = "Big_Signed'LAST"


--******************************************************************************
--******************************************************************************

PACKAGE Conversion IS

  FUNCTION Unsigned2Signed(x : IN Big_Unsigned;
                           Positive : IN BOOLEAN := TRUE) RETURN Big_Signed;

  FUNCTION Signed2Unsigned(x : IN Big_Signed) RETURN Big_Unsigned;


  FUNCTION Big_Unsigned2Long_Long_Float(x : IN Big_Unsigned)
                                        RETURN LONG_LONG_FLOAT;

  FUNCTION Big_Signed2Long_Long_Float(x : IN Big_Signed)
                                      RETURN LONG_LONG_FLOAT;

END Conversion;

 
--******************************************************************************
--******************************************************************************

USE Conversion;

PACKAGE Generic_Conversion IS

  Conversion_Error : EXCEPTION;
     -- This exception is raised if an error occurs during a conversion
     -- (for example if you try to convert an integer below 0 to a
     -- Big_Unsigned).

  ------------------------------------------------------------------------------

  GENERIC
    TYPE Mod_Generic IS MOD <>;

  FUNCTION Mod_Number2Big_Unsigned(x : IN Mod_Generic) RETURN Big_Unsigned;

  ------------------------------------------------------------------------------

  GENERIC
    TYPE Mod_Generic IS MOD <>;

  FUNCTION Mod_Number2Big_Signed(x : IN Mod_Generic;
                                 Positive : IN BOOLEAN := TRUE) RETURN Big_Signed;

  ------------------------------------------------------------------------------

  GENERIC
    TYPE Int_Generic IS RANGE <>;

  FUNCTION Int_Number2Big_Unsigned(x : IN Int_Generic) RETURN Big_Unsigned;

  ------------------------------------------------------------------------------

  GENERIC
    TYPE Int_Generic IS RANGE <>;

  FUNCTION Int_Number2Big_Signed(x : IN Int_Generic) RETURN Big_Signed;

  ------------------------------------------------------------------------------

END Generic_Conversion;

--******************************************************************************
--******************************************************************************

USE Conversion;

PACKAGE Unsigned_Number IS

  --------------------------
  -- Exception definition --
  --------------------------
  
  Big_Unsigned_Substraction_Error : EXCEPTION;
    -- This exception is raised when you try to compute x - y, where y is
    -- greater than x (a Big_Unsigned variable can't have a negative value).

  -----------------
  -- Comparisons --
  -----------------

  FUNCTION "<"(Left, Right : IN Big_Unsigned) RETURN BOOLEAN;
  FUNCTION ">"(Left, Right : IN Big_Unsigned) RETURN BOOLEAN;
  FUNCTION "<="(Left, Right : IN Big_Unsigned) RETURN BOOLEAN;
  FUNCTION ">="(Left, Right : IN Big_Unsigned) RETURN BOOLEAN;

  FUNCTION Is_Null(x : IN Big_Unsigned) RETURN BOOLEAN;

  FUNCTION Min(x, y : IN Big_Unsigned) RETURN Big_Unsigned;
  FUNCTION Max(x, y : IN Big_Unsigned) RETURN Big_Unsigned;

  ----------------------
  -- Common functions --
  ----------------------

  PROCEDURE Inc(x : IN OUT Big_Unsigned);  -- Set x to x + 1.
  PROCEDURE Dec(x : IN OUT Big_Unsigned);  -- Set x to x - 1.

  PROCEDURE Swap1(x, y : IN OUT Big_Unsigned);
  PROCEDURE Swap2(x, y : IN OUT Big_Unsigned);
    -- Swap1 and Swap2 swaps the contents of x and y.
    -- Swap2 is faster than Swap1 when x and y are slightly lower
    -- than Big_Unsigned_Last. Otherwise, you should prefer Swap1.
    -- So examine the usual conditions in your program in order
    -- choose the right function.
  FUNCTION "+"(x, y : IN Big_Unsigned) RETURN Big_Unsigned;
  FUNCTION "-"(x, y : IN Big_Unsigned) RETURN Big_Unsigned;
  FUNCTION "*"(x : IN Big_Unsigned; y : IN My_Type) RETURN Big_Unsigned;
  FUNCTION "*"(x, y : IN Big_Unsigned) RETURN Big_Unsigned;
  FUNCTION "**"(x : IN Big_Unsigned; y : IN Index_Type) RETURN Big_Unsigned;
  FUNCTION "**"(x : IN Big_Unsigned; y : IN My_Type) RETURN Big_Unsigned;
    -- This function returns (x ** y).
    -- If the result is too big, the function raises the
    -- exception Big_Number_Overflow.
    -- If y = 0, this function returns Big_Unsigned_One.
  FUNCTION "**"(x, y : IN Big_Unsigned) RETURN Big_Unsigned;
    -- This function returns (x ** y).
    -- If the result is too big, the function raises the
    -- exception Big_Number_Overflow.
    -- If y = Big_Unsigned_Zero, this function returns Big_Unsigned_One.
  FUNCTION "**"(x : IN Big_Unsigned;
                y : IN Big_Signed) RETURN Big_Unsigned;
    -- This function returns (x ** y).
    -- If the result is too big, the function raises the
    -- exception Big_Number_Overflow.
    -- If y = Big_Signed_Zero, this function returns Big_Unsigned_One.
    -- If y < Big_Signed_Zero, this function returns Big_Unsigned_Zero.
  FUNCTION "/"(x, y : IN Big_Unsigned) RETURN Big_Unsigned;
    -- Returns x/y
  FUNCTION "MOD"(x, y : IN Big_Unsigned) RETURN Big_Unsigned;
    -- Returns x MOD y
  FUNCTION "/"(x : IN Big_Unsigned; y : IN My_Type) RETURN Big_Unsigned;
    -- Returns x/y
  FUNCTION "MOD"(x : IN Big_Unsigned; y : IN My_Type) RETURN My_Type;
    -- Returns x MOD y
  PROCEDURE Big_Div(x, y : IN Big_Unsigned;
                    Quot, Remainder : OUT Big_Unsigned);
    -- This procedure returns Quot and Remainder, so :
    -- x = Quot * y + Remainder, and Quot * y < x < (Quot + 1) * y.
  PROCEDURE Short_Div(x : IN Big_Unsigned;
                      y : IN My_Type;
                      Quot : OUT Big_Unsigned;
                      Remainder : OUT My_Type);
    -- This procedure returns Quot and Remainder, so :
    -- x = Quot * y + Remainder, and Quot * y < x < (Quot + 1) * y.

  -----------------------
  -- Special functions --
  -----------------------

  FUNCTION Number_of_Bytes(x : IN Big_Unsigned) RETURN Largest_Unsigned;
    -- This function returns the number of bytes in the mantissa of x.
  FUNCTION Number_of_Bits(x : IN Big_Unsigned) RETURN Largest_Unsigned;
    -- This function returns the number of bits in the mantissa of x.

  FUNCTION Square_Root(x : IN Big_Unsigned) RETURN Big_Unsigned;
  FUNCTION Factorial(x : IN Big_Unsigned) RETURN Big_Unsigned;
    -- This function returns x!.
  FUNCTION GCF(x, y : IN Big_Unsigned) RETURN Big_Unsigned;
    -- This function returns the greatest common factor of x and y.
  FUNCTION LCM(x, y : IN Big_Unsigned) RETURN Big_Unsigned;
    -- This function returns the least common factor of x and y.

  FUNCTION "AND"(x, y : IN Big_Unsigned) RETURN Big_Unsigned;
    -- This function returns x and y.
  FUNCTION "XOR"(x, y : IN Big_Unsigned) RETURN Big_Unsigned;
    -- This function returns x xor y.


  FUNCTION NCR(N, R : IN Big_Unsigned) RETURN Big_Unsigned;
    -- NCR = N! / [R! * (N-R)!]
  FUNCTION NPR(N, R : IN Big_Unsigned) RETURN Big_Unsigned;
  FUNCTION K(N, P : IN Big_Unsigned) RETURN Big_Unsigned;
    -- K(N, P) = NCR(N+P-1,P)

END Unsigned_Number;

--******************************************************************************
--******************************************************************************

USE Conversion;


PACKAGE Signed_Number IS

  -----------------
  -- Comparisons --
  -----------------

  FUNCTION "<"(Left, Right : IN Big_Signed) RETURN BOOLEAN;
  FUNCTION ">"(Left, Right : IN Big_Signed) RETURN BOOLEAN;
  FUNCTION "<="(Left, Right : IN Big_Signed) RETURN BOOLEAN;
  FUNCTION ">="(Left, Right : IN Big_Signed) RETURN BOOLEAN;

  FUNCTION Is_Null(x : IN Big_Signed) RETURN BOOLEAN;

  FUNCTION Min(x, y : IN Big_Signed) RETURN Big_Signed;
  FUNCTION Max(x, y : IN Big_Signed) RETURN Big_Signed;

  ----------------------
  -- Common functions --
  ----------------------

  PROCEDURE Inc(x : IN OUT Big_Signed);  -- Set x to x + 1.
  PROCEDURE Dec(x : IN OUT Big_Signed);  -- Set x to x - 1.

  PROCEDURE Swap1(x, y : IN OUT Big_Signed);
  PROCEDURE Swap2(x, y : IN OUT Big_Signed);
    -- Swap1 and Swap2 swaps the contents of x and y.
    -- Swap2 is faster than Swap1 when the absolute value of x and y are
    -- slightly lower than Big_Signed_Last. Otherwise, you should prefer Swap1.
    -- So examine the usual conditions in your program in order
    -- choose the right function.
  FUNCTION "+"(x, y : IN Big_Signed) RETURN Big_Signed;
  FUNCTION "-"(x, y : IN Big_Signed) RETURN Big_Signed;
  FUNCTION "*"(x, y : IN Big_Signed) RETURN Big_Signed;
  FUNCTION "**"(x : IN Big_Signed; y : IN Index_Type) RETURN Big_Signed;
  FUNCTION "**"(x : IN Big_Signed; y : IN My_Type) RETURN Big_Signed;
    -- This function returns (x ** y).
    -- If the result is too big, the function raises the
    -- exception Big_Number_Overflow.
    -- If y = 0, this function returns Big_Signed_One.
    -- By definition, Index_Type is a modular type, so y can't be negative.
  FUNCTION "**"(x : IN Big_Signed;
                y : IN Big_Unsigned) RETURN Big_Signed;
    -- This function returns (x ** y).
    -- If the result is too big, the function raised the
    -- exception Big_Number_Overflow.
    -- If y = Big_Unsigned_Zero, this function returns Big_Signed_One.
    -- By definition, y can't be negative.
  FUNCTION "**"(x, y : IN Big_Signed) RETURN Big_Signed;
    -- This function returns (x ** y) if y is positive.
    -- If y is < 0, the function returns Big_Unsigned_Zero.
    -- If y = Big_Signed_Zero, this function returns Big_Signed_One;
    -- If the result is too big, the function raises the
    -- exception Big_Number_Overflow.
  FUNCTION "/"(x, y : IN Big_Signed) RETURN Big_Signed;
  FUNCTION "MOD"(x, y : IN Big_Signed) RETURN Big_Signed;
    -- This function returns a number, always positive, so that
    -- (x - Result) is an exact multiple of y.
  FUNCTION "/"(x : IN Big_Signed; y : IN My_Type) RETURN Big_Signed;
  FUNCTION "MOD"(x : IN Big_Signed; y : IN My_Type) RETURN My_Type;
    -- This function returns a number, always positive, so that
    -- (x - Result) is an exact multiple of y.
  FUNCTION Opposite(x : IN Big_Signed) RETURN Big_Signed;
    -- Returns -x.
  FUNCTION Absolute(x : IN Big_Signed) RETURN Big_Signed;
    -- Returns -x if x < 0.
    -- Returns x if x >=0.
  PROCEDURE Big_Div(x, y : IN Big_Signed;
                    Quot, Remainder : OUT Big_Signed);
    -- This procedure returns Quot and Remainder so that :
    -- x = Quot * y + Remainder,
    -- and  Absolute(Quot * y) < Absolute(x) < Absolute(Quot + 1) * y)

  -----------------------
  -- Special functions --
  -----------------------

  FUNCTION Number_of_Bytes(x : IN Big_Signed) RETURN Largest_Unsigned;
    -- This function returns the number of bytes in the mantissa of x.
  FUNCTION Number_of_Bits(x : IN Big_Signed) RETURN Largest_Unsigned;
    -- This function returns the number of bits in the mantissa of x.

  FUNCTION Square_Root(x : IN Big_Signed) RETURN Big_Signed;
  FUNCTION Factorial(x : IN Big_Signed) RETURN Big_Signed;
    -- This function returns x!.
    -- If x is strictly lower than 0, the exception Parameter_Error is raised.
  FUNCTION GCF(x, y : IN Big_Signed) RETURN Big_Signed;
    -- This function returns the greatest common factor of x and y.
  FUNCTION LCM(x, y : IN Big_Signed) RETURN Big_Signed;
    -- This function returns the least common factor of x and y.


END Signed_Number;

--******************************************************************************
--******************************************************************************

USE Generic_Conversion;
USE Unsigned_Number;
USE ADA.STRINGS.UNBOUNDED;
USE UStrings;


PACKAGE String_Conversion IS

  Conversion_Error : EXCEPTION;
     -- This exception is raised if an error occurs during a conversion.

  Base_Error : EXCEPTION;
  -- This exception is raised when the base parameter is greater than 36,
  -- or when this parameter is lower than 2.

  FUNCTION String2Unsigned(x : IN STRING) RETURN Big_Unsigned;
  FUNCTION String2Signed(x : IN STRING) RETURN Big_Signed;
  -- These two functions convert an integer in base 10, given as a string,
  -- respectively in a Big_Unsigned or a Big_Signed.

  FUNCTION UString2Unsigned(x : IN UNBOUNDED_STRING) RETURN Big_Unsigned;
  FUNCTION UString2Signed(x : IN UNBOUNDED_STRING) RETURN Big_Signed;
  -- These two functions convert an integer in base 10, given as a string,
  -- respectively in a Big_Unsigned or a Big_Signed.


  FUNCTION Big_Unsigned2UString(x : IN Big_Unsigned;
                               Base : IN My_Type := 10)
                               RETURN UNBOUNDED_STRING;
    -- This function returns the UNBOUNDED_STRING corresponding to
    -- the Big_Unsigned number x, using the base Base. By default, Base = 10.
  FUNCTION Big_Signed2UString(x : IN Big_Signed;
                             Base : IN My_Type := 10)
                             RETURN UNBOUNDED_STRING;
    -- This function returns the UNBOUNDED_STRING corresponding to
    -- the Big_Signed number x, using the base Base. By default, Base = 10.

END String_Conversion;


--******************************************************************************
--******************************************************************************

USE Unsigned_Number;
USE String_Conversion;
USE TEXT_IO;
USE UStrings;


PACKAGE Big_Unsigned_IO IS

  PROCEDURE GET_LINE(x : OUT Big_Unsigned);
    -- This procedure waits for a Big_Unsigned, to be entered in base 10 on the
    -- keyboard.
  PROCEDURE GET_LINE(File : IN FILE_TYPE; x : OUT Big_Unsigned);
    -- This procedure reads a Big_Unsigned in a text file.

  PROCEDURE PUT(x : IN Big_Unsigned;
                Base : IN My_Type := 10);
    -- This procedure writes on screen the variable x, using the
    -- indicated base (by default, it's base 10).
  PROCEDURE PUT(File : IN FILE_TYPE;
                x : IN Big_Unsigned;
                Base : IN My_Type := 10);
    -- This procedure writes the variable x in a text file, using the
    -- indicated base (by default, it's base 10).

  PROCEDURE PUT_LINE(x : IN Big_Unsigned;
                     Base : IN My_Type := 10);
    -- This procedure writes on screen the variable x, using the
    -- indicated base (by default, it's base 10), and then go to the next line.
  PROCEDURE PUT_LINE(File : IN FILE_TYPE;
                     x : IN Big_Unsigned;
                     Base : IN My_Type := 10);
    -- This procedure writes the variable x in a text file, using the
    -- indicated base (by default, it's base 10), and then go to the next line.


END Big_Unsigned_IO;

--******************************************************************************
--******************************************************************************

USE Signed_Number;
USE String_Conversion;
USE TEXT_IO;
USE UStrings;


PACKAGE Big_Signed_IO IS

  PROCEDURE GET_LINE(x : OUT Big_Signed);
    -- This procedure waits a Big_Signed, to be entered in base 10 on the
    -- keyboard.
  PROCEDURE GET_LINE(File : IN FILE_TYPE; x : OUT Big_Signed);
    -- This procedure reads a Big_signed in a text file.

  PROCEDURE PUT(x : IN Big_Signed;
                Base : IN My_Type := 10);
    -- This procedure writes on screen the variable x, using the
    -- indicated base (by default, it's base 10).
  PROCEDURE PUT(File : IN FILE_TYPE;
                x : IN Big_Signed;
                Base : IN My_Type := 10);
    -- This procedure writes the variable x in a text file, using the
    -- indicated base (by default, it's base 10).

  PROCEDURE PUT_LINE(x : IN Big_Signed;
                     Base : IN My_Type := 10);
    -- This procedure writes on screen the variable x, using the
    -- indicated base (by default, it's base 10), and then go to the next line.
  PROCEDURE PUT_LINE(File : IN FILE_TYPE;
                     x : IN Big_Signed;
                     Base : IN My_Type := 10);
    -- This procedure writes the variable x in a text file, using the
    -- indicated base (by default, it's base 10), and then go to the next line.

END Big_Signed_IO;

--******************************************************************************
--******************************************************************************

USE Unsigned_Number;
USE Big_Unsigned_IO;
USE ADA.STRINGS.UNBOUNDED;
USE UStrings;

PACKAGE Big_Fraction IS

-- Remark : for all the operations on fractions, the function or the procedure
-- return the simplified form of the result. For example, 1/4 + 1/4 = 2/4,
-- but the function "+" returns 1/2 (2/4 = 1/2). The simplification of the
-- result requires some more execution time, but it allows to avoid some
-- overflows.

  Fraction_Overflow : EXCEPTION;

  FUNCTION CreateFraction(Numerator, Denominator : IN Big_Unsigned;
                          Positive : IN BOOLEAN := TRUE) RETURN Fraction;
  FUNCTION CreateFraction(Numerator, Denominator : IN Big_Signed) RETURN Fraction;

  FUNCTION Get_Numerator(x : IN Fraction) RETURN Big_Unsigned;
  FUNCTION Get_Denominator(x : IN Fraction) RETURN Big_Unsigned;
  FUNCTION Is_Positive(x : IN Fraction) RETURN BOOLEAN;

  FUNCTION Inverse_Fraction(x : IN Fraction) RETURN Fraction;
    -- This function returns 1/x.

    -- No comments for the following functions...
  FUNCTION "*"(x, y : IN Fraction) RETURN Fraction;
  FUNCTION "/"(x, y : IN Fraction) RETURN Fraction;
  FUNCTION "+"(x, y : IN Fraction) RETURN Fraction;
  FUNCTION "-"(x, y : IN Fraction) RETURN Fraction;
  FUNCTION "**"(x : IN Fraction; y : IN Big_Unsigned) RETURN Fraction;
  FUNCTION "**"(x : IN Fraction; y : IN Big_Signed) RETURN Fraction;
  FUNCTION "**"(x : IN Fraction; y : IN Index_Type) RETURN Fraction;
  FUNCTION "**"(x : IN Fraction; y : IN My_Type) RETURN Fraction;
  FUNCTION "*"(x : IN Fraction; y : IN Big_Unsigned) RETURN Fraction;
  FUNCTION "*"(x : IN Fraction; y : IN Big_Signed) RETURN Fraction;
  FUNCTION "/"(x : IN Fraction; y : IN Big_Unsigned) RETURN Fraction;
  FUNCTION "/"(x : IN Fraction; y : IN Big_Signed) RETURN Fraction;

  FUNCTION "<"(Left, Right : IN Fraction) RETURN BOOLEAN;
  FUNCTION ">"(Left, Right : IN Fraction) RETURN BOOLEAN;
  FUNCTION "<="(Left, Right : IN Fraction) RETURN BOOLEAN;
  FUNCTION ">="(Left, Right : IN Fraction) RETURN BOOLEAN;

  FUNCTION Opposite(x : IN Fraction) RETURN Fraction;
    -- Returns -x.
  FUNCTION Absolute(x : IN Fraction) RETURN Fraction;
    -- Returns -x if x < 0.
    -- Returns x if x >=0.
  FUNCTION Is_Null(x : IN Fraction) RETURN BOOLEAN;

  PROCEDURE PUT(x : IN Fraction);
    -- This procedure puts x on the screen.
  PROCEDURE PUT_LINE(x : IN Fraction);
    -- This procedure puts x on the screen and then go to the next line.

  PROCEDURE PUT(File : IN FILE_TYPE; x : IN Fraction);
    -- This procedure puts x in a text file.
  PROCEDURE PUT_LINE(File : IN FILE_TYPE; x : IN Fraction);
    -- This procedure puts x in a text file and then go to the next line.

  FUNCTION UString2Fraction(x : IN UNBOUNDED_STRING) RETURN Fraction;

  PROCEDURE GET_LINE(x : OUT Fraction);
    -- This procedure gets x and then go to the next line.
    -- The exception Parameter_Error is raised in case of error
    -- (for example if there are 2 or more '/').
  PROCEDURE GET_LINE(File : IN FILE_TYPE; x : OUT Fraction);
    -- This procedure gets x in a text file and then go to the next line.
    -- The exception Parameter_Error is raised in case of error
    -- (for example if there are 2 or more '/').

  PROCEDURE Float_in_UString2Fraction(x : IN UNBOUNDED_STRING;
                                      Nb_Periods : IN POSITIVE;
                                      Result : OUT Fraction;
                                      Error : OUT BOOLEAN);
  -- This procedure tries to get the fraction corresponding to
  -- the float point number contained in the string x.
  -- INPUT :
  --    - x : the string
  --    - Nb_Periods :  the minimal number of periods to find in the string
  --                    if you want the function to consider that number as a
  --                    fraction. For example, the string "5.345345345345"
  --                    contains the period "345" four times.
  --                    The string "5.3453453453457" is considered to contain 
  --                    no period, even if there are 4 repetitions of "345",
  --                    because these repetitions are followed by "7", which
  --                    DOES NOT belong to this sequence.
  --
  -- OUTPUT :
  --    - Result : the fraction (only if Error = FALSE)
  --    - Error : TRUE if x isn't a valid string or if Nb_Periods = 1,
  --              FALSE otherwise.
  --
  -- An valid string is : "(+/-)nnn.nnnnn[(E:e(+/-))nnnnn]".
  -- The mantissa is composed of any decimal digit and may have only one sign
  -- at the beginning. The mantissa may be a real number.
  -- If there is an exponent, it must start with E or e, may have only one
  -- sign and MUST be an integer.
  -- In the mantissa or in the exponent, the spaces and the character "_"
  -- are accepted, and they will have no consequence concerning the result.
  --
  -- Remarks :
  -- 1) If Nb_Periods = 4, 1.353535 (3 repetitions of "35") will be considered
  --    as the exact value of the fraction 270_707 / 200_000, but with the
  --    number 1.35353535 (which has 4 repetitions of "35"), the result will be
  --    the fraction 134/99.
  -- 2) The parameter Nb_Periods must always be equal or greater than 2.
  --    There is no upper limit for Nb_Periods.
  -- 3) Be careful : if there is an error and if you don't check it
  --    (by regarding the boolean returned by the procedure), you
  --    may have a DIVISION_BY_ZERO exception in a further operation,
  --    because, in this case, the function sets the result to 0/0.
  --    That was not done to plague you, but just to help you to detect
  --    errors in your programs : I think that, during the tests,
  --    it's better to have an exception than a bad result which you may
  --    never detect. Furthermore, exceptions can be handled.
  --


END Big_Fraction;

--******************************************************************************
--******************************************************************************

USE CALENDAR;
USE Unsigned_Number;


PACKAGE Big_Unsigned_Random IS

Instanciation_Error : EXCEPTION;
  -- This exception is raised when Index_Type'LAST > (2**TypeSize - 1) / 125.
  -- It is due to the implementation : in the package Big_Unsigned_Random,
  -- ForMultiplication'LAST / 125 must be greater than Index_Type'LAST.

PROCEDURE Randomize;
  -- This procedure resets the pseudo-random number generator.
  -- It is automatically executed at the start of the program.

FUNCTION Random RETURN Big_Unsigned;
  -- This function returns a Big_Unsigned between Big_Unsigned_Zero
  -- and Big_Unsigned_Last.

FUNCTION Random(x : IN Big_Unsigned) RETURN Big_Unsigned;
  -- This function returns a Big_Unsigned between Big_Unsigned_Zero
  -- and x. If x = Big_Unsigned_Zero, it returns Big_Unsigned_Zero.

FUNCTION Random(x, y : IN Big_Unsigned) RETURN Big_Unsigned;
  -- This function returns a Big_Unsigned between Min(x,y) and Max(x,y).


END Big_Unsigned_Random;

--******************************************************************************
--******************************************************************************
--******************************************************************************
--******************************************************************************

PRIVATE

  TYPE Mantiss IS ARRAY(0..Nb_Item-1) OF My_Type;

  TYPE Big_Signed IS RECORD
                       Positive : BOOLEAN;
                       Last_Significant_Digit : Index_Type;
                       Mantissa : Mantiss;
                     END RECORD;

  TYPE Big_Unsigned IS RECORD
                         Last_Significant_Digit : Index_Type;
                         Mantissa : Mantiss;
                       END RECORD;

  TYPE Fraction IS RECORD
                     Positive    : BOOLEAN;
                     Numerator   : Big_Unsigned;
                     Denominator : Big_Unsigned;
                   END RECORD;

  --
  -- By convention, the lower digit is in Mantissa(0), and the greater one
  -- is in Mantissa(Nb_Elt-1). This looks like a "BigEndian" coding.
  --

  Big_Unsigned_Zero : CONSTANT Big_Unsigned
                                    := (Last_Significant_Digit => 0,
                                        Mantissa => (OTHERS => My_Type(0)));
  Big_Unsigned_One : CONSTANT Big_Unsigned
                                    := (Last_Significant_Digit => 0,
                                        Mantissa => (1, OTHERS => My_Type(0)));
  Big_Unsigned_Two : CONSTANT Big_Unsigned
                                    := (Last_Significant_Digit => 0,
                                        Mantissa => (2, OTHERS => My_Type(0)));
  Big_Unsigned_First : CONSTANT Big_Unsigned
                                    := Big_Unsigned_Zero;
  Big_Unsigned_Last : CONSTANT Big_Unsigned
                                    := (Last_Significant_Digit => Nb_Item - 1,
                                        Mantissa => (OTHERS => My_Type'LAST));

  Big_Signed_Zero : CONSTANT Big_Signed
                                    := (Positive => TRUE,
                                        Last_Significant_Digit => 0,
                                        Mantissa => (OTHERS => My_Type(0)));
  Big_Signed_One : CONSTANT Big_Signed
                                    := (Positive => TRUE,
                                        Last_Significant_Digit => 0,
                                        Mantissa => (1, OTHERS => My_Type(0)));
  Big_Signed_First : CONSTANT Big_Signed
                                    := (Positive => FALSE,
                                        Last_Significant_Digit => Nb_Item - 1,
                                        Mantissa => (OTHERS => My_Type'LAST));
  Big_Signed_Last : CONSTANT Big_Signed
                                    := (Positive => TRUE,
                                        Last_Significant_Digit => Nb_Item - 1,
                                        Mantissa => (OTHERS => My_Type'LAST));

--******************************************************************************
--******************************************************************************

END Big_Number;
--******************************************************************************
--                                                                            **
--  File:        BIGNUMBE.ADB                                                 **
--  Description: Big_Number package body                                      **
--  Revision:    0.28 (beta version)                                          **
--  Date:        July 8, 2001                                                 **
--  Author:      Jrme Delcourt                                              **
--  Mail:        sikander@club-internet.fr                                    **
--                                                                            **
--  Copyright (c) Jrme Delcourt, 1998, 1999, 2000, 2001                     **
--  62, rue Ngrier                                                           **
--  59800 Lille                                                               **
--  FRANCE                                                                    **
--                                                                            **
--  Permission granted to use for any purpose, provided this copyright        **
--  remains attached and unmodified.                                          **
--                                                                            **
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR            **
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED            **
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       **
--                                                                            **
--******************************************************************************

PACKAGE BODY Big_Number IS

  --****************************************************************************
  --**               Declarations of local types and constants                **
  --****************************************************************************

  ---------------------
  -- Local constants --
  ---------------------

  First_Indice : CONSTANT Index_Type := 0;
  Last_Indice : CONSTANT Index_Type := Nb_Item - 1;

  Base : CONSTANT Largest_Unsigned := 2**My_Type'SIZE;
  Base_Minus_One : CONSTANT Largest_Unsigned := Base - 1;
  LLF_Base : CONSTANT LONG_LONG_FLOAT := LONG_LONG_FLOAT(Base);
  LLF_Base_Square : CONSTANT LONG_LONG_FLOAT := LLF_Base * LLF_Base;
  Greatest_Digit : CONSTANT My_Type := My_Type'LAST;

  --****************************************************************************
  --****************************************************************************

  FUNCTION Min(x,y : IN Index_Type) RETURN Index_Type IS
  BEGIN
    IF (x < y) THEN
      RETURN x;
    ELSE
      RETURN y;
    END IF;
  END Min;

  --****************************************************************************

  FUNCTION Max(x,y : IN Index_Type) RETURN Index_Type IS
  BEGIN
    IF (x < y) THEN
      RETURN y;
    ELSE
      RETURN x;
    END IF;
  END Max;

  --****************************************************************************

  PACKAGE BODY Conversion IS SEPARATE;
  PACKAGE BODY Generic_Conversion IS SEPARATE;
  PACKAGE BODY String_Conversion IS SEPARATE;
  PACKAGE BODY Unsigned_Number IS SEPARATE;
  PACKAGE BODY Signed_Number IS SEPARATE;
  PACKAGE BODY Big_Unsigned_IO IS SEPARATE;
  PACKAGE BODY Big_Signed_IO IS SEPARATE;
  PACKAGE BODY Big_Fraction IS SEPARATE;
  PACKAGE BODY Big_Unsigned_Random IS SEPARATE;

  --****************************************************************************
  --****************************************************************************

BEGIN
  IF (My_Type_Size < 8) THEN
    TEXT_IO.PUT_LINE("The package Big_Number has been modified in a wrong way.");
    TEXT_IO.PUT_LINE("The type My_Type must have a size greater or equal to 8 bits.");
    RAISE PROGRAM_ERROR;
  END IF;
END Big_Number;
--******************************************************************************
--                                                                            **
--  File:        BINUMCON.ADB                                                 **
--  Description: Big_Number.Conversion package body                           **
--  Revision:    0.28 (beta version)                                          **
--  Date:        July 8, 2001                                                 **
--  Author:      Jrme Delcourt                                              **
--  Mail:        sikander@club-internet.fr                                    **
--                                                                            **
--  Copyright (c) Jrme Delcourt, 1998, 1999, 2000                           **
--  62, rue Ngrier                                                           **
--  59800 Lille                                                               **
--  FRANCE                                                                    **
--                                                                            **
--  Permission granted to use for any purpose, provided this copyright        **
--  remains attached and unmodified.                                          **
--                                                                            **
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR            **
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED            **
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       **
--                                                                            **
--******************************************************************************

SEPARATE(Big_Number)

PACKAGE BODY Conversion IS

--******************************************************************************
--******************************************************************************

FUNCTION Unsigned2Signed(x : IN Big_Unsigned; Positive : IN BOOLEAN := TRUE) RETURN Big_Signed IS
  Temp : Big_Signed;
BEGIN
  Temp.Last_Significant_Digit := x.Last_Significant_Digit;
  Temp.Mantissa := x.Mantissa;
  Temp.Positive := Positive;
  RETURN Temp;
END Unsigned2Signed;

--******************************************************************************
--******************************************************************************

FUNCTION Signed2Unsigned(x : IN Big_Signed) RETURN Big_Unsigned IS
  Temp : Big_Unsigned;
BEGIN
  Temp.Last_Significant_Digit := x.Last_Significant_Digit;
  Temp.Mantissa := x.Mantissa;
  RETURN Temp;
END Signed2Unsigned;

--******************************************************************************
--******************************************************************************

FUNCTION Big_Unsigned2Long_Long_Float(x : IN Big_Unsigned)
                                      RETURN LONG_LONG_FLOAT IS
  Temp_Result : LONG_LONG_FLOAT := 0.0;
BEGIN
  FOR I IN REVERSE First_Indice..x.Last_Significant_Digit LOOP
    Temp_Result := Temp_Result * LLF_Base + LONG_LONG_FLOAT(x.Mantissa(I));
  END LOOP;
  RETURN Temp_Result;
END Big_Unsigned2Long_Long_Float;

--******************************************************************************
--******************************************************************************

FUNCTION Big_Signed2Long_Long_Float(x : IN Big_Signed) RETURN LONG_LONG_FLOAT IS
  Temp_Result : LONG_LONG_FLOAT := 0.0;
BEGIN
  FOR I IN REVERSE First_Indice..x.Last_Significant_Digit LOOP
    Temp_Result := Temp_Result * LLF_Base + LONG_LONG_FLOAT(x.Mantissa(I));
  END LOOP;
  IF (x.Positive = FALSE) THEN
    Temp_Result := -Temp_Result;
  END IF;
  RETURN Temp_Result;
END Big_Signed2Long_Long_Float;

--******************************************************************************
--******************************************************************************

END Conversion;
--******************************************************************************
--                                                                            **
--  File:        BINUGECO.ADB                                                 **
--  Description: Big_Number.Generic_Conversion package body                   **
--  Revision:    0.28 (beta version)                                          **
--  Date:        July 8, 2001                                                 **
--  Author:      Jrme Delcourt                                              **
--  Mail:        sikander@club-internet.fr                                    **
--                                                                            **
--  Copyright (c) Jrme Delcourt, 1998, 1999, 2000                           **
--  62, rue Ngrier                                                           **
--  59800 Lille                                                               **
--  FRANCE                                                                    **
--                                                                            **
--  Permission granted to use for any purpose, provided this copyright        **
--  remains attached and unmodified.                                          **
--                                                                            **
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR            **
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED            **
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       **
--                                                                            **
--******************************************************************************

SEPARATE(Big_Number)

PACKAGE BODY Generic_Conversion IS

--******************************************************************************
--******************************************************************************

FUNCTION Mod_Number2Big_Unsigned(x : IN Mod_Generic) RETURN Big_Unsigned IS
  Result : Big_Unsigned := Big_Unsigned_Zero;
  FUNCTION Shift_Right(Value : IN Mod_Generic; Amount : IN NATURAL)
                      RETURN Mod_Generic;
  PRAGMA IMPORT(INTRINSIC, Shift_Right);
BEGIN
  IF (Mod_Generic'SIZE <= My_Type_Size) THEN
    -- No problem : My_Type'LAST >= Mod_Generic'LAST.
    -- Note that Result.Last_Significant_Digit was already initialized
    -- to First_Indice, since Result was initialized to Big_Unsigned_Zero.
    Result.Mantissa(First_Indice) := My_Type(x);
    RETURN Result;
  END IF;
  -- Now, we must check the case where the size of x is bigger than
  -- the size of My_Type (note : the mantisse of a Big_Unsigned variable
  -- is coded by an array of My_Type elements).
  DECLARE
    Remainder : Mod_Generic := x;
    Index : Index_Type := First_Indice;
  BEGIN
    LOOP
      Result.Mantissa(Index) := My_Type(Remainder AND Mod_Generic(My_Type'LAST));
      -- Note that My_Type'LAST = Base - 1
      Remainder := Shift_Right(Remainder, My_Type_Size);
      EXIT WHEN (Remainder = 0) OR ELSE (Index = Last_Indice);
      Index := Index + 1;
    END LOOP;
    IF (Remainder > 0) THEN
      -- You must think of a stupid situation, where the total number of bits
      -- in the mantissa of the result is smaller than the number of significant
      -- bits of x.
      -- [It doesn't worth to use this package, which may be buggy, if the ADA
      -- compiler can provide a type (predefined or defined by a user modular
      -- type) with more bits in the mantissa and with bug-free related
      -- functions.]
      RAISE Conversion_Error;
    END IF;
    Result.Last_Significant_Digit := Index;
    RETURN Result;
  END;
END Mod_Number2Big_Unsigned;

--******************************************************************************
--******************************************************************************

FUNCTION Mod_Number2Big_Signed(x : IN Mod_Generic;
                               Positive : IN BOOLEAN := TRUE) RETURN Big_Signed IS
  FUNCTION My_Conversion IS NEW Mod_Number2Big_Unsigned(Mod_Generic);
  Result : Big_Signed;
BEGIN
  Result := Unsigned2Signed(My_Conversion(x), Positive);
  IF (x = 0) THEN
    Result.Positive := TRUE;
  END IF;
  RETURN Result;
END Mod_Number2Big_Signed;

--******************************************************************************
--******************************************************************************

FUNCTION Int_Number2Big_Unsigned(x : IN Int_Generic) RETURN Big_Unsigned IS
  Result : Big_Unsigned := Big_Unsigned_Zero;
  FUNCTION Shift_Right(Value : IN Int_Generic; Amount : IN NATURAL)
                       RETURN Int_Generic;
  PRAGMA IMPORT(INTRINSIC, Shift_Right);
BEGIN
  IF (x < 0) THEN
    RAISE Conversion_Error;
  ELSIF (Int_Generic'SIZE <= My_Type_Size + 1) THEN
    -- No problem : My_Type'LAST > Int_Generic'LAST, because, for integer types,
    -- there is one bit reserved to store the sign.
    Result.Mantissa(First_Indice) := My_Type(x);
    RETURN Result;
  END IF;
  -- Now, we must check the case where the size of x is bigger than
  -- the size of My_Type (note : the mantissa of a Big_Unsigned variable
  -- is coded by an array of My_Type elements).
  DECLARE
    TmpBase : CONSTANT Int_Generic := Int_Generic(Base);
    -- Remark : Base = My_Type'LAST + 1;
    -- Here, Int_Generic'LAST > My_Type'LAST,
    -- So, Int_Generic'LAST >= Base
    Remainder : Int_Generic := x;
    Index : Index_Type := First_Indice;
  BEGIN
    LOOP
      Result.Mantissa(Index) := My_Type(Remainder MOD TmpBase);
      Remainder := Shift_Right(Remainder, My_Type_Size);
      EXIT WHEN (Remainder = 0) OR ELSE (Index = Last_Indice);
      Index := Index + 1;
    END LOOP;
    IF (Remainder > 0) THEN
      -- You must think of a stupid situation, where the total number of bits
      -- in the mantisse of the result is smaller than the number of significant
      -- bits of x.
      -- [It doesn't worth to use this package, which may be buggy, if the ADA
      -- compiler can provide a type (predefined or defined by a user modular
      -- type) with more bits in the mantissa and with bug-free related
      -- functions.]
      RAISE Conversion_Error;
    END IF;
    Result.Last_Significant_Digit := Index;
    RETURN Result;
  END;
END Int_Number2Big_Unsigned;

--******************************************************************************
--******************************************************************************

FUNCTION Int_Number2Big_Signed(x : IN Int_Generic) RETURN Big_Signed IS
  FUNCTION My_Conversion IS NEW Int_Number2Big_Unsigned(Int_Generic);
BEGIN
  RETURN Unsigned2Signed(My_Conversion(ABS(x)), x>=0);
END Int_Number2Big_Signed;

--******************************************************************************
--******************************************************************************

END Generic_Conversion;
--******************************************************************************
--                                                                            **
--  File:        BINUSTCO.ADB                                                 **
--  Description: Big_Number.String_Conversion package body                    **
--  Revision:    0.28 (beta version)                                          **
--  Date:        July 8, 2001                                                 **
--  Author:      Jrme Delcourt                                              **
--  Mail:        sikander@club-internet.fr                                    **
--                                                                            **
--  Copyright (c) Jrme Delcourt, 1998, 1999, 2000, 2001                     **
--  62, rue Ngrier                                                           **
--  59800 Lille                                                               **
--  FRANCE                                                                    **
--                                                                            **
--  Permission granted to use for any purpose, provided this copyright        **
--  remains attached and unmodified.                                          **
--                                                                            **
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR            **
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED            **
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       **
--                                                                            **
--******************************************************************************

SEPARATE(Big_Number)

PACKAGE BODY String_Conversion IS

--******************************************************************************
--******************************************************************************

  SUBTYPE Numeric_Digit IS CHARACTER RANGE '0'..'9';

--******************************************************************************
--******************************************************************************
--**           Conversion of strings to Big_Unsigned or Big_Signed            **
--******************************************************************************
--******************************************************************************

FUNCTION String_Size_Required(x : IN STRING) RETURN NATURAL IS
  Temp_Size : NATURAL := 0;
  x_First : CONSTANT POSITIVE := x'FIRST;
  Char : CHARACTER;
BEGIN
  FOR I IN x'RANGE LOOP
    Char := x(I);
    IF NOT(Char IN Numeric_Digit) THEN
      IF ((Char = '-') OR ELSE (Char = '+')) THEN
        -- A sign is authorized only for the first character
        IF (I /= x_First) THEN
          RAISE Conversion_Error;
        END IF;
      ELSIF (Char /= ' ') AND THEN (Char /= '_') THEN
        RAISE Conversion_Error;
      END IF;
    ELSE
      -- Now, Char is in '0'..'9'.
      Temp_Size := Temp_Size + 1;
    END IF;
  END LOOP;
  RETURN Temp_Size;
EXCEPTION
  WHEN OTHERS => RAISE Conversion_Error;
END String_Size_Required;

--******************************************************************************
--******************************************************************************

FUNCTION Str2Unsigned(x : IN STRING) RETURN Big_Unsigned IS
  Temp_Char : CHARACTER;
  Temp_Digit : Numeric_Digit;
  Temp : Big_Unsigned := Big_Unsigned_Zero;
  Pos0 : CONSTANT INTEGER := Numeric_Digit'POS('0');
  FUNCTION Integer2BU IS NEW Int_Number2Big_Unsigned(INTEGER);
BEGIN
  FOR I IN x'RANGE LOOP
    Temp_Char := x(I);
    Temp_Digit := Numeric_Digit(Temp_Char);
    Temp := (Temp * 10) + Integer2BU(Numeric_Digit'POS(Temp_Digit) - Pos0);
  END LOOP;
  RETURN Temp;
EXCEPTION
  WHEN OTHERS => RAISE Conversion_Error;
END Str2Unsigned;

--******************************************************************************
--******************************************************************************

PROCEDURE Del_Spaces_And_Underscore(x : IN STRING; y : OUT STRING) IS
  Dernier : CONSTANT POSITIVE := x'LAST;
  J : POSITIVE := y'FIRST;
  Temp_Char : CHARACTER;
BEGIN
  FOR I IN x'RANGE LOOP
    Temp_Char := x(I);
    IF (Temp_Char /= ' ') AND THEN (Temp_Char /= '_') THEN
      y(J) := Temp_Char;
      J := J + 1;
    END IF;
  END LOOP;
END Del_Spaces_And_Underscore;

--******************************************************************************
--******************************************************************************

FUNCTION String2Unsigned(x : IN STRING) RETURN Big_Unsigned IS
  String_Size : NATURAL;
  Temp_Positive, Sign_Present : BOOLEAN;
BEGIN
  String_Size := String_Size_Required(x);
  IF (String_Size = 0) THEN
    RETURN Big_Unsigned_Zero;
  END IF;
  DECLARE
    First_Char : CONSTANT CHARACTER := x(x'FIRST);
  BEGIN
    Sign_Present := (First_Char = '+') OR ELSE (First_Char = '-');
    Temp_Positive := (First_Char = '+') OR ELSE (First_Char IN Numeric_Digit);
  END;
  IF (Temp_Positive = FALSE) THEN
    RAISE Conversion_Error;
  END IF;
  DECLARE
    Temp_String : STRING(1..POSITIVE(String_Size));
  BEGIN
    IF (Sign_Present = TRUE) THEN
      Del_Spaces_And_Underscore(x(x'FIRST + 1..x'LAST), Temp_String);
    ELSE
      Del_Spaces_And_Underscore(x, Temp_String);
    END IF;
    RETURN Str2Unsigned(Temp_String);
  END;
EXCEPTION
  WHEN OTHERS => RAISE Conversion_Error;
END String2Unsigned;

--******************************************************************************
--******************************************************************************

FUNCTION String2Signed(x : IN STRING) RETURN Big_Signed IS
  String_Size : NATURAL;
  Sign_Present : BOOLEAN;
  Temp_Positive : BOOLEAN;
BEGIN
  String_Size := String_Size_Required(x);
  IF (String_Size = 0) THEN
    RETURN Big_Signed_Zero;
  END IF;
  DECLARE
    First_Char : CONSTANT CHARACTER := x(x'FIRST);
  BEGIN
    Sign_Present := (First_Char = '+') OR ELSE (First_Char = '-');
    Temp_Positive := (First_Char = '+') OR ELSE (First_Char IN Numeric_Digit);
  END;
  DECLARE
    Temp_String : STRING(1..POSITIVE(String_Size));
  BEGIN
    IF (Sign_Present = TRUE) THEN
      Del_Spaces_And_Underscore(x(x'FIRST + 1..x'LAST), Temp_String);
    ELSE
      Del_Spaces_And_Underscore(x, Temp_String);
    END IF;
    RETURN Unsigned2Signed(Str2Unsigned(Temp_String), Temp_Positive);
  END;
EXCEPTION
  WHEN OTHERS => RAISE Conversion_Error;
END String2Signed;

--******************************************************************************
--******************************************************************************
--**       Conversion of unbounded strings to Big_Unsigned or Big_Signed      **
--******************************************************************************
--******************************************************************************

FUNCTION UStr2Unsigned(x : IN UNBOUNDED_STRING) RETURN Big_Unsigned IS
  Temp_Char : CHARACTER;
  Temp_Digit : Numeric_Digit;
  Temp : Big_Unsigned := Big_Unsigned_Zero;
  Pos0 : CONSTANT INTEGER := Numeric_Digit'POS('0');
  FUNCTION Integer2BU IS NEW Int_Number2Big_Unsigned(INTEGER);
BEGIN
  FOR I IN 1..LENGTH(x) LOOP
    Temp_Char := ELEMENT(x, I);
    Temp_Digit := Numeric_Digit(Temp_Char);
    Temp := (Temp * 10) + Integer2BU(Numeric_Digit'POS(Temp_Digit) - Pos0);
  END LOOP;
  RETURN Temp;
EXCEPTION
  WHEN OTHERS => RAISE Conversion_Error;
END UStr2Unsigned;

--******************************************************************************
--******************************************************************************

FUNCTION UString2Unsigned(x : IN UNBOUNDED_STRING) RETURN Big_Unsigned IS
  Positive, Sign_Present : BOOLEAN;
  Temp_x : UNBOUNDED_STRING := Del_Spaces(x);
  Size1, Size2 : INTEGER;
BEGIN
  Temp_x := Del_Character(Temp_x, '_');
  IF (LENGTH(Temp_x) = 0) THEN
    RETURN Big_Unsigned_Zero;
  END IF;
  DECLARE
    First_Char : CONSTANT CHARACTER := ELEMENT(x, 1);
  BEGIN
    Sign_Present := (First_Char = '+') OR ELSE (First_Char = '-');
    Positive := (First_Char = '+') OR ELSE (First_Char IN Numeric_Digit);
  END;
  IF ((Sign_Present = TRUE) AND THEN (LENGTH(Temp_x) = 1))
    -- If x contains only a sign (+ or -)
    OR ELSE (Positive = FALSE) THEN
    -- or if the number is negative,
    -- this string can't be converted to a Big_Unsigned.
    RAISE Conversion_Error;
  END IF;
  -- Now, the number is positive.
  Size1 := LENGTH(Temp_x);
  Temp_x := Del_Character(Temp_x, '+');
  Temp_x := Del_Character(Temp_x, '-');
  Size2 := LENGTH(Temp_x);
  IF (Size1 - Size2 > 1) THEN
    -- If there are 2 or more signs in Temp_x...
    RAISE Conversion_Error;
  END IF;
  RETURN UStr2Unsigned(Temp_x);
EXCEPTION
  WHEN OTHERS => RAISE Conversion_Error;
END UString2Unsigned;

--******************************************************************************
--******************************************************************************

FUNCTION UString2Signed(x : IN UNBOUNDED_STRING) RETURN Big_Signed IS
  Positive, Sign_Present : BOOLEAN;
  Temp_x : UNBOUNDED_STRING := Del_Spaces(x);
  Size1, Size2 : INTEGER;
BEGIN
  Temp_x := Del_Character(Temp_x, '_');
  IF (LENGTH(Temp_x) = 0) THEN
    RETURN Big_Signed_Zero;
  END IF;
  DECLARE
    First_Char : CONSTANT CHARACTER := ELEMENT(x, 1);
  BEGIN
    Sign_Present := (First_Char = '+') OR ELSE (First_Char = '-');
    Positive := (First_Char = '+') OR ELSE (First_Char IN Numeric_Digit);
  END;
  IF ((Sign_Present = TRUE) AND THEN (LENGTH(Temp_x) = 1)) THEN
    -- If x contains only a sign (+ or -),
    -- this string can't be converted to a Big_Unsigned.
    RAISE Conversion_Error;
  END IF;
  -- Now, the sign is known.
  Size1 := LENGTH(Temp_x);
  Temp_x := Del_Character(Temp_x, '+');
  Temp_x := Del_Character(Temp_x, '-');
  Size2 := LENGTH(Temp_x);
  IF (Size1 - Size2 > 1) THEN
    -- If there are 2 or more signs in Temp_x...
    RAISE Conversion_Error;
  END IF;
  RETURN Unsigned2Signed(UStr2Unsigned(Temp_x), Positive);
EXCEPTION
  WHEN OTHERS => RAISE Conversion_Error;
END UString2Signed;

--******************************************************************************
--******************************************************************************
--**      Conversion of Big_Unsigned or Big_Signed to unbounded strings       **
--******************************************************************************
--******************************************************************************

FUNCTION Big_Unsigned2UString(x : IN Big_Unsigned;
                              Base : IN My_Type := 10)
                              RETURN UNBOUNDED_STRING IS
  TYPE Trans IS ARRAY(My_Type RANGE 0..35) OF CHARACTER;
  Table : CONSTANT Trans
                := ('0','1','2','3','4','5','6','7','8','9','A','B','C','D','E',
                    'F','G','H','I','J','K','L','M','N','O','P','Q','R','S','T',
                    'U','V','W','X','Y','Z');
  Result : UNBOUNDED_STRING := NULL_UNBOUNDED_STRING;
  Temp_x : Big_Unsigned := x;
  Remainder : My_Type;
BEGIN
  IF (Base > 35) OR ELSE (Base < 2) THEN
    RAISE Base_Error;
  END IF;
  IF (Temp_x.Last_Significant_Digit = First_Indice) AND THEN
     (Temp_x.Mantissa(First_Indice) = 0) THEN
    Result := TO_UNBOUNDED_STRING("0");
    RETURN Result;
  END IF;
  WHILE (Temp_x /= Big_Unsigned_Zero) LOOP
    Short_Div(Temp_x, Base, Temp_x, Remainder);
    Result := Table(Remainder) & Result;
  END LOOP;
  RETURN Result;
END Big_Unsigned2UString;

--******************************************************************************
--******************************************************************************

FUNCTION Big_Signed2UString(x : IN Big_Signed;
                            Base : IN My_Type := 10)
                            RETURN UNBOUNDED_STRING IS
  Temp : CONSTANT Big_Unsigned := Signed2Unsigned(x);
BEGIN
  IF (x.Positive = FALSE) THEN
    RETURN "-" & Big_Unsigned2UString(Temp, Base);
  ELSE
    RETURN Big_Unsigned2UString(Temp, Base);
  END IF;
END Big_Signed2UString;

--******************************************************************************
--******************************************************************************

END String_Conversion;
--******************************************************************************
--                                                                            **
--  File:        BINUUNNU.ADB                                                 **
--  Description: Big_Number.Unsigned_Number package body                      **
--  Revision:    0.28 (beta version)                                          **
--  Date:        July 8, 2001                                                 **
--  Author:      Jrme Delcourt                                              **
--  Mail:        sikander@club-internet.fr                                    **
--                                                                            **
--  Copyright (c) Jrme Delcourt, 1998, 1999, 2000, 2001                     **
--  62, rue Ngrier                                                           **
--  59800 Lille                                                               **
--  FRANCE                                                                    **
--                                                                            **
--  Permission granted to use for any purpose, provided this copyright        **
--  remains attached and unmodified.                                          **
--                                                                            **
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR            **
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED            **
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       **
--                                                                            **
--******************************************************************************

SEPARATE(Big_Number)

PACKAGE BODY Unsigned_Number IS

--******************************************************************************
--******************************************************************************
--**                              Comparisons                                 **
--**                                                                          **
--** The comparisons ">", "<=" and ">=" can be done by a call to "<"          **
--** (see below for explanations). The advantage is that we have only one     **
--** function to write and test, the 3 others depend on it.                   **
--******************************************************************************
--******************************************************************************

FUNCTION "<"(Left, Right : IN Big_Unsigned) RETURN BOOLEAN IS
  -- Remark : in order to understand this function, you must remind that the
  -- coding of the digits of x and y, in their mantissa, looks like the
  -- "Big_Endian" coding.
  Indice : Index_Type := Left.Last_Significant_Digit;
BEGIN
  IF (Indice /= Right.Last_Significant_Digit) THEN
    RETURN (Indice < Right.Last_Significant_Digit);
  END IF;
  -- Now, Left.Last_Significant_Digit = Right.Last_Significant_Digit
  WHILE (Left.Mantissa(Indice) = Right.Mantissa(Indice)) AND THEN
        (Indice > First_Indice) LOOP
    Indice := Indice - 1;
  END LOOP;
  RETURN (Left.Mantissa(Indice) < Right.Mantissa(Indice));
END "<";

--******************************************************************************

FUNCTION ">"(Left, Right : IN Big_Unsigned) RETURN BOOLEAN IS
BEGIN
  -- (Left > Right) is the same comparison than (Right < Left)...
  RETURN (Right < Left);
END ">";

--******************************************************************************

FUNCTION "<="(Left, Right : IN Big_Unsigned) RETURN BOOLEAN IS
  -- (Left <= Right) = NOT (Right < Left)
BEGIN
  RETURN NOT (Right < Left);
END "<=";

--******************************************************************************

FUNCTION ">="(Left, Right : IN Big_Unsigned) RETURN BOOLEAN IS
BEGIN
  -- (Left >= Right) = NOT (Left < Right).
  RETURN NOT (Left < Right);
END ">=";

--******************************************************************************

FUNCTION Is_Null(x : IN Big_Unsigned) RETURN BOOLEAN IS
BEGIN
  IF (x.Last_Significant_Digit = First_Indice) AND THEN
     (x.Mantissa(First_Indice) = 0) THEN
    RETURN TRUE;
  ELSE
    RETURN FALSE;
  END IF;
END Is_Null;

--******************************************************************************

FUNCTION Min(x, y : IN Big_Unsigned) RETURN Big_Unsigned IS
BEGIN
  IF (x < y) THEN
    RETURN x;
  ELSE
    RETURN y;
  END IF;
END Min;

--******************************************************************************

FUNCTION Max(x, y : IN Big_Unsigned) RETURN Big_Unsigned IS
BEGIN
  IF (x < y) THEN
    RETURN y;
  ELSE
    RETURN x;
  END IF;
END Max;

--******************************************************************************
--******************************************************************************
--**                            Common functions                              **
--******************************************************************************
--******************************************************************************

PROCEDURE Swap1(x, y : IN OUT Big_Unsigned) IS
  Temp : CONSTANT Big_Unsigned := x;
BEGIN
  x := y;
  y := Temp;
END Swap1;

--******************************************************************************

PROCEDURE Swap2(x, y : IN OUT Big_Unsigned) IS
  Max_Index : CONSTANT Index_Type
                     := Max(x.Last_Significant_Digit, y.Last_Significant_Digit);
  Temp : My_Type;
BEGIN
  FOR I IN First_Indice..Max_Index LOOP
    Temp := x.Mantissa(I);
    x.Mantissa(I) := y.Mantissa(I);
    y.Mantissa(I) := Temp;
  END LOOP;
END Swap2;

--******************************************************************************

FUNCTION Shift_Right(Value : IN Largest_Unsigned; Amount : IN NATURAL)
                     RETURN Largest_Unsigned;
FUNCTION Shift_Left(Value : IN Largest_Unsigned; Amount : IN NATURAL)
                    RETURN Largest_Unsigned;

FUNCTION Shift_Right(Value : IN My_Type; Amount : IN NATURAL)
                     RETURN My_Type;
PRAGMA IMPORT(INTRINSIC, Shift_Right);

FUNCTION Shift_Left(Value : IN My_Type; Amount : IN NATURAL)
                    RETURN My_Type;
PRAGMA IMPORT(INTRINSIC, Shift_Left);

--******************************************************************************

PROCEDURE Inc(x : IN OUT Big_Unsigned) IS
  Temp_Index : Index_Type := First_Indice;
BEGIN
  WHILE (x.Mantissa(Temp_Index) = Greatest_Digit) LOOP
    -- If x.Mantissa(Temp_Index) = Greatest_Digit, set it to 0
    -- and try to increment the next digit.
    IF (Temp_Index = Last_Indice) THEN
      -- In this case, the next digit can't be incremented.
      RAISE Big_Number_Overflow;
    END IF;
    x.Mantissa(Temp_Index) := 0;
    Temp_Index := Temp_Index + 1;
  END LOOP;
  x.Mantissa(Temp_Index) := x.Mantissa(Temp_Index) + 1;
  IF (Temp_Index > x.Last_Significant_Digit) THEN
    x.Last_Significant_Digit := Temp_Index;
  END IF;
END Inc;

--******************************************************************************

PROCEDURE Dec(x : IN OUT Big_Unsigned) IS
  Temp_Index : Index_Type := First_Indice;
BEGIN
  IF (x.Last_Significant_Digit = First_Indice) AND THEN
     (x.Mantissa(First_Indice) = 0) THEN
     -- If x = Big_Unsigned_Zero...
    RAISE Big_Unsigned_Substraction_Error;
  END IF;
  WHILE (x.Mantissa(Temp_Index) = 0) LOOP
    -- If x.Mantissa(Temp_Index) = 0, set it to Greatest_Digit
    -- and try to decrement the next digit.
    x.Mantissa(Temp_Index) := Greatest_Digit;
    Temp_Index := Temp_Index + 1;
  END LOOP;
  x.Mantissa(Temp_Index) := x.Mantissa(Temp_Index) - 1;
  IF (Temp_Index > First_Indice) AND THEN
     (x.Mantissa(Temp_Index) = 0) AND THEN
     (Temp_Index = x.Last_Significant_Digit) THEN
    x.Last_Significant_Digit := Temp_Index - 1;
  END IF;
END Dec;

--******************************************************************************

FUNCTION "+"(x, y : IN Big_Unsigned) RETURN Big_Unsigned IS
  Min_Index : CONSTANT Index_Type
                     := Min(x.Last_Significant_Digit, y.Last_Significant_Digit);
  Max_Index : CONSTANT Index_Type
                     := Max(x.Last_Significant_Digit, y.Last_Significant_Digit);
  Temp_Index : Index_Type := Min_Index;
  Carry : Largest_Unsigned := 0;
  Temp_Elt : Largest_Unsigned;
  Temp_Result : Big_Unsigned;
BEGIN
  FOR i IN (Max_Index + 1)..Last_Indice LOOP
    Temp_Result.Mantissa(i) := 0;
  END LOOP;
  Temp_Result.Last_Significant_Digit := Max_Index;
  FOR i IN First_Indice..Min_Index LOOP
    -- Remark : if i = First_Indice, Carry is 0.
    Temp_Elt :=   Largest_Unsigned(x.Mantissa(i))
                + Largest_Unsigned(y.Mantissa(i))
                + Carry;
    Carry := Shift_Right(Temp_Elt, My_Type_Size);  -- Carry := Temp_Elt / Base
    Temp_Result.Mantissa(i) := My_Type(Temp_Elt AND Base_Minus_One);
                         -- := Temp_Elt MOD Base
  END LOOP;
  IF (Min_Index /= Max_Index) THEN
    IF (x.Last_Significant_Digit > Min_Index) THEN
      Temp_Result.Mantissa(Min_Index+1..Max_Index)
                                          := x.Mantissa(Min_Index+1..Max_Index);
    ELSE
      Temp_Result.Mantissa(Min_Index+1..Max_Index)
                                          := y.Mantissa(Min_Index+1..Max_Index);
    END IF;
  END IF;
  WHILE (Carry > 0) AND THEN (Temp_Index < Last_Indice) LOOP
    Temp_Index := Temp_Index + 1;
    IF (Temp_Result.Mantissa(Temp_Index) < Greatest_Digit) THEN
      Carry := 0;
      Temp_Result.Mantissa(Temp_Index) := Temp_Result.Mantissa(Temp_Index) + 1;
      IF (Max_Index < Temp_Index) THEN
        Temp_Result.Last_Significant_Digit := Temp_Index;
      END IF;
    ELSE
      IF (Temp_Index = Last_Indice) THEN
        RAISE Big_Number_Overflow;
      END IF;
      Temp_Result.Mantissa(Temp_Index) := 0;
    END IF;
  END LOOP;
  RETURN Temp_Result;
END "+";

--******************************************************************************

FUNCTION "-"(x, y : IN Big_Unsigned) RETURN Big_Unsigned IS
BEGIN
  IF (x < y) THEN
    RAISE Big_Unsigned_Substraction_Error;
  END IF;
  --
  -- Common situation : x > y.
  -- We must calculate x - y.
  --
  DECLARE
    Nb_x : CONSTANT Index_Type := x.Last_Significant_Digit;
    Nb_y : CONSTANT Index_Type := y.Last_Significant_Digit;
    Temp_Index : Index_Type;
    Carry : My_Type := 0;
    Temp_Result : Big_Unsigned;
  BEGIN
    FOR i IN First_Indice..Nb_y LOOP
      IF (x.Mantissa(i) > y.Mantissa(i)) OR ELSE
         (x.Mantissa(i) = y.Mantissa(i) AND THEN Carry = 0) THEN
        -- This test is equivalent to (x.Mantissa(i) >= y.Mantissa(i) + Carry),
        -- but this last expression could raise a CONSTRAINT_ERROR if
        -- y.Mantissa(i) = My_Type'LAST and Carry = 1.
        Temp_Result.Mantissa(i) := x.Mantissa(i) - y.Mantissa(i) - Carry;
        Carry := 0;
      ELSE  -- x.Mantissa(i) < y.Mantissa(i) + Carry
        -- Temp_Result.Mantissa(i):=Base + x.Mantissa(i) - y.Mantissa(i) - Carry
        -- Base = My_Type'LAST + 1
        -- ==> Temp_Result.Mantissa(i)
        --                :=x.Mantissa(i)+(My_Type'LAST-y.Mantissa(i))+(1-Carry)
        -- My_Type'LAST = Greatest_Digit
        -- My_Type'LAST-y.Mantissa(i) = Greatest_Digit XOR y.Mantissa(i)
        -- ==> Temp_Result.Mantissa(i)
        --          :=x.Mantissa(i)+(Greatest_Digit XOR y.Mantissa(i))+(1-Carry)
        -- Don't change the order of operations, because that could cause
        -- overflows.
        Temp_Result.Mantissa(i) :=   x.Mantissa(i)
                                   + (Greatest_Digit XOR y.Mantissa(i))
                                   + (1 - Carry);
        -- Then Carry must be set to 1.
        Carry := 1;
      END IF;
    END LOOP;
    Temp_Index := Nb_y + 1;
    WHILE (Carry = 1) LOOP
      IF (x.Mantissa(Temp_Index) >= Carry) THEN
        Temp_Result.Mantissa(Temp_Index) := x.Mantissa(Temp_Index) - Carry;
        Carry := 0;
      ELSE
        -- x.Mantissa(Temp_Index) < Carry.
        -- This situation can occur only if x.Mantissa(i) = 0
        -- and if Carry is already equal to 1.
        Temp_Result.Mantissa(Temp_Index) := Greatest_Digit;
        -- Greatest_Digit := My_Type'LAST;
        -- It is not necessary to add "Carry := 1;",
        -- because Carry is already set to 1.
      END IF;
      Temp_Index := Temp_Index + 1;
    END LOOP;
    FOR i IN Temp_Index..Nb_x LOOP
      Temp_Result.Mantissa(i) := x.Mantissa(i);
    END LOOP;
    FOR i IN (Nb_x + 1)..Last_Indice LOOP
      Temp_Result.Mantissa(i) := 0;
    END LOOP;
    Temp_Index := Nb_x;
    WHILE (Temp_Result.Mantissa(Temp_Index) = 0) AND THEN
          (Temp_Index > First_Indice) LOOP
      Temp_Index := Temp_Index - 1;
    END LOOP;
    Temp_Result.Last_Significant_Digit := Temp_Index;
    RETURN Temp_Result;
  END;
END "-";

--******************************************************************************

FUNCTION Shift_Left(x : IN Big_Unsigned; Nb : IN Index_Type)
                   RETURN Big_Unsigned IS
BEGIN
  -- IF (Nb > Last_Indice) OR ELSE (x.Last_Significant_Digit + Nb > Last_Indice) THEN
  --   RAISE Big_Number_Overflow;
  -- END IF;
  -- In these lines of code, (x.Last_Significant_Digit + Nb) could raise
  -- an overflow, so they are replaced by the following lines :
  IF (Nb > Last_Indice) OR ELSE (x.Last_Significant_Digit > Last_Indice - Nb) THEN
    RAISE Big_Number_Overflow;
  END IF;
  DECLARE
    Result : Big_Unsigned;
    Last_Digit_Indice : CONSTANT Index_Type := x.Last_Significant_Digit + Nb;
  BEGIN
    IF (Nb /= 0) THEN
      FOR I IN First_Indice..(Nb - 1) LOOP
        Result.Mantissa(I) := 0;
      END LOOP;
    END IF;
    IF (Last_Digit_Indice /= Last_Indice) THEN
      FOR I IN (Last_Digit_Indice + 1)..Last_Indice LOOP
        Result.Mantissa(I) := 0;
      END LOOP;
    END IF;
    Result.Mantissa(Nb..Last_Digit_Indice)
                          := x.Mantissa(First_Indice..x.Last_Significant_Digit);
    Result.Last_Significant_Digit := Last_Digit_Indice;
    RETURN Result;
  END;
END Shift_Left;

--******************************************************************************

FUNCTION "*"(x : IN Big_Unsigned; y : IN My_Type) RETURN Big_Unsigned IS
BEGIN
  IF (y = 0) THEN
    RETURN Big_Unsigned_Zero;
  ELSIF (y = 1) THEN
    RETURN x;
  END IF;
  DECLARE
    Nb_x : CONSTANT Index_Type := x.Last_Significant_Digit;
    Temp_y : CONSTANT Largest_Unsigned := Largest_Unsigned(y);
    Carry : Largest_Unsigned := 0;
    Result : Big_Unsigned;
  BEGIN
    IF (Nb_x < Last_Indice) THEN
      FOR I IN (Nb_x + 1)..Last_Indice LOOP
        Result.Mantissa(I) := 0;
      END LOOP;
    END IF;
    FOR I IN First_Indice..Nb_x LOOP
      Carry := Largest_Unsigned(x.Mantissa(I)) * Temp_y + Carry;
      Result.Mantissa(I) := My_Type(Carry AND Largest_Unsigned(My_Type'LAST));
      Carry := Shift_Right(Carry, My_Type_Size);        -- := Carry DIV Base
    END LOOP;
    IF (Carry = 0) THEN
      Result.Last_Significant_Digit := Nb_x;
    ELSE
      IF (Nb_x < Last_Indice) THEN
        Result.Last_Significant_Digit := Nb_x + 1;
        Result.Mantissa(Nb_x + 1) := My_Type(Carry);
      ELSE
        RAISE Big_Number_Overflow;
      END IF;
    END IF;
    RETURN Result;
  END;             -- END of the DECLARE block
END "*";

--******************************************************************************

FUNCTION Special_Mul(x : IN Big_Unsigned;
                     y : IN My_Type;
                     Left_Shift : IN Index_Type) RETURN Big_Unsigned IS
  -- Internal function
  -- This function returns x * y, with a left shift of Left_Shift digits.
  -- algorithm similar to the one used for "*"(x : IN Big_Unsigned; y : IN My_Type)
BEGIN
  IF (y = 0) THEN
    RETURN Big_Unsigned_Zero;
  END IF;
  DECLARE
    Nb_x : CONSTANT Index_Type := x.Last_Significant_Digit;
    Expected_Indice : CONSTANT Index_Type := Nb_x + Left_Shift;
    Temp_y : CONSTANT Largest_Unsigned := Largest_Unsigned(y);
    Carry : Largest_Unsigned := 0;
    Result : Big_Unsigned;
  BEGIN
    IF (Left_Shift > First_Indice) THEN
      FOR I IN First_Indice..(Left_Shift - 1) LOOP
        Result.Mantissa(I) := 0;
      END LOOP;
    END IF;
    IF (Expected_Indice < Last_Indice) THEN
      FOR I IN (Expected_Indice + 1)..Last_Indice LOOP
        Result.Mantissa(I) := 0;
      END LOOP;
    END IF;
    FOR I IN First_Indice..Nb_x LOOP
      Carry := Largest_Unsigned(x.Mantissa(I)) * Temp_y + Carry;
      Result.Mantissa(I + Left_Shift)
                           := My_Type(Carry AND Largest_Unsigned(My_Type'LAST));
      Carry := Shift_Right(Carry, My_Type_Size);        -- := Carry DIV Base
    END LOOP;
    IF (Carry = 0) THEN
      Result.Last_Significant_Digit := Expected_Indice;
    ELSE
      IF (Expected_Indice < Last_Indice) THEN
        Result.Last_Significant_Digit := Expected_Indice + 1;
        Result.Mantissa(Expected_Indice + 1) := My_Type(Carry);
      ELSE
        RAISE Big_Number_Overflow;
      END IF;
    END IF;
    RETURN Result;
  END;             -- END of the DECLARE block
END Special_Mul;

PRAGMA INLINE(Special_Mul);

--******************************************************************************

FUNCTION "*"(x, y : IN Big_Unsigned) RETURN Big_Unsigned IS
  Lastx : CONSTANT Index_Type := x.Last_Significant_Digit;
  Lasty : CONSTANT Index_Type := y.Last_Significant_Digit;
  Result : Big_Unsigned;
BEGIN
  IF (Lastx = First_Indice) THEN
    RETURN (y * x.Mantissa(First_Indice));
  END IF;
  IF (Lastx > Last_Indice - Lasty) THEN
    RAISE Big_Number_Overflow;
  END IF;
  Result := x * y.Mantissa(First_Indice);
  -- Result is initialized here, because I don't want to waste
  -- execution time before, during the computation of the simple cases.
  FOR I IN 1..Lasty LOOP
    Result := Result + Special_Mul(x, y.Mantissa(I), I);
  END LOOP;
  RETURN Result;
EXCEPTION
  WHEN OTHERS => RAISE Big_Number_Overflow;
END "*";

--******************************************************************************

FUNCTION "**"(x : IN Big_Unsigned; y : IN My_Type) RETURN Big_Unsigned IS
BEGIN
  IF (y = 0) THEN
    RETURN Big_Unsigned_One;
  END IF;
  IF ((x.Last_Significant_Digit = First_Indice) AND THEN
     (x.Mantissa(First_Indice) <= 1))
     OR ELSE (y = 1) THEN
    -- Here, y /= 0.
    -- So, if y = 1 or x = 0 or x = 1, the result of (x ** y) is always x.
    RETURN x;
  END IF;
  DECLARE
    Temp_Result : Big_Unsigned := x;
  BEGIN
    FOR I IN 2..(y/2) LOOP
      Temp_Result := Temp_Result * x;
    END LOOP;
    Temp_Result := Temp_Result * Temp_Result;
    IF (y MOD 2 = 1) THEN
      Temp_Result := Temp_Result * x;
    END IF;
    RETURN Temp_Result;
  END;
END "**";

--******************************************************************************

FUNCTION "**"(x : IN Big_Unsigned; y : IN Index_Type) RETURN Big_Unsigned IS
BEGIN
  IF (y = 0) THEN
    RETURN Big_Unsigned_One;
  END IF;
  IF ((x.Last_Significant_Digit = First_Indice) AND THEN
     (x.Mantissa(First_Indice) <= 1))
     OR ELSE (y = 1) THEN
    -- Here, y /= 0.
    -- So, if y = 1 or x = 0 or x = 1, the result of (x ** y) is always x.
    RETURN x;
  END IF;
  DECLARE
    Temp_Result : Big_Unsigned := x;
  BEGIN
    FOR I IN 2..(y/2) LOOP
      Temp_Result := Temp_Result * x;
    END LOOP;
    Temp_Result := Temp_Result * Temp_Result;
    IF (y MOD 2 = 1) THEN
      Temp_Result := Temp_Result * x;
    END IF;
    RETURN Temp_Result;
  END;
END "**";

--******************************************************************************

FUNCTION "**"(x, y : IN Big_Unsigned) RETURN Big_Unsigned IS
  Temp_Result : Big_Unsigned := x;
  Indice : Big_Unsigned := Big_Unsigned_One;
BEGIN
  IF (y.Last_Significant_Digit = First_Indice) THEN
    IF (y.Mantissa(First_Indice) = 0) THEN
      -- If y = 0, x^y = x^0 = 1, even if x = 0.
      RETURN Big_Unsigned_One;
    ELSE
      RETURN x ** My_Type(y.Mantissa(First_Indice));
    END IF;
  END IF;
  WHILE (Indice < y) LOOP
    Inc(Indice);
    Temp_Result := Temp_Result * x;
  END LOOP;
  RETURN Temp_Result;
END "**";

--******************************************************************************

FUNCTION "**"(x : IN Big_Unsigned;
              y : IN Big_Signed) RETURN Big_Unsigned IS
BEGIN
  IF (y.Positive = FALSE) THEN    -- Here, we test if (y < Big_Signed_Zero).
    RETURN Big_Unsigned_Zero;
  END IF;
  DECLARE
    Temp_y : CONSTANT Big_Unsigned := Signed2Unsigned(y);
  BEGIN
    RETURN (x ** Temp_y);
  END;
END "**";

--******************************************************************************
--******************************************************************************

PROCEDURE Big_Div(x, y : IN Big_Unsigned;
                  Quot, Remainder : OUT Big_Unsigned) IS
  Lasty : CONSTANT Index_Type := y.Last_Significant_Digit;
BEGIN
  -- analysis of simple cases.
  IF (Lasty = First_Indice) THEN
    CASE y.Mantissa(First_Indice) IS
      WHEN 0 => RAISE Division_By_Zero;
      WHEN 1 => Quot := x;
                Remainder := Big_Unsigned_Zero;
                RETURN;
      WHEN OTHERS => DECLARE
                       Temp_Remainder : My_Type;
                       Temp_y : CONSTANT My_Type := y.Mantissa(First_Indice);
                     BEGIN
                       -- If y has only one significant item in its mantissa,
                       -- we use the function Short_Div, which is faster.
                       -- See below for the implementation of Short_Div.
                       Short_Div(x, Temp_y, Quot, Temp_Remainder);
                       Remainder := (Last_Significant_Digit => First_Indice,
                                     Mantissa => (Temp_Remainder, OTHERS => 0));
                       RETURN;
                     END;
    END CASE;
  END IF;
  IF (x.Last_Significant_Digit = First_Indice) AND THEN
     (x.Mantissa(First_Indice) = 0) THEN
    Quot := Big_Unsigned_Zero;
    Remainder := Big_Unsigned_Zero;
    RETURN;
  ELSIF (x < y) THEN
    Quot := Big_Unsigned_Zero;
    Remainder := x;
    RETURN;
  ELSIF (x = y) THEN
    Quot := Big_Unsigned_One;
    Remainder := Big_Unsigned_Zero;
    RETURN;
  END IF;
  -- Now, there is only the case where (x > y) and (y /= 0).
  DECLARE

    FUNCTION Two_Biggest_Digits(x : IN Big_Unsigned) RETURN Largest_Unsigned IS
      Lastx : CONSTANT Index_Type := x.Last_Significant_Digit;
    BEGIN
      -- We don't have to check if there is only one digit :
      -- because of the algorithm, there are at least two digits.
      RETURN Largest_Unsigned(x.Mantissa(Lastx-1))
                + Shift_Left(Largest_Unsigned(x.Mantissa(Lastx)), My_Type_Size);
    END Two_Biggest_Digits;

    ----------------------------------------------------------------------------

    Two_Biggest_Digits_of_y : CONSTANT Largest_Unsigned:= Two_Biggest_Digits(y);
    Temp_Unsigned : Big_Unsigned;
    Temp_Quot : Big_Unsigned := Big_Unsigned_Zero;
    Temp_Remainder : Big_Unsigned := x;
    Temp_Index, Last_Indice_Quot : Index_Type;
    Carry : My_Type;
    Guess_Digit : My_Type;

    ----------------------------------------------------------------------------

    FUNCTION Last_Indice_of_the_Quotient RETURN Index_Type IS
      -- The function Last_Indice_of_the_Quotient is used to calculate the
      -- number of items of the quotient and, so, the indice of the array (used
      -- to represent the mantissa) where we have to put the guessed digit.
      Diff : CONSTANT Index_Type := Temp_Remainder.Last_Significant_Digit-Lasty;
      Indice : Index_Type := Lasty;
    BEGIN
      WHILE (Temp_Remainder.Mantissa(Indice+Diff) = y.Mantissa(Indice)) AND THEN
            (Indice > First_Indice) LOOP
        Indice := Indice - 1;
      END LOOP;
      IF (Temp_Remainder.Mantissa(Indice+Diff) < y.Mantissa(Indice)) AND THEN
         (Diff > First_Indice) THEN
        RETURN Diff - 1;
      ELSE
        RETURN Diff;
      END IF;
    END Last_Indice_of_the_Quotient;

    ----------------------------------------------------------------------------

    FUNCTION Guess RETURN My_Type IS
      Last : CONSTANT Index_Type := Temp_Remainder.Last_Significant_Digit;
      Temp_Rem : CONSTANT Largest_Unsigned := Two_Biggest_Digits(Temp_Remainder);
      Temp_Digit : Largest_Unsigned;
    BEGIN
      -- In this function, Last >= Lasty, and lasty >= 1.
      IF (Last_Indice_Quot = Last - Lasty) THEN
        Temp_Digit := Temp_Rem / Two_Biggest_Digits_of_y;
        IF (Lasty >= 2) AND THEN
           (Temp_Rem MOD Two_Biggest_Digits_of_y < Shift_Right(Largest_Unsigned(y.Mantissa(Lasty-2))*Temp_Digit,My_Type_Size)) THEN
          Temp_Digit := Temp_Digit - 1;
        END IF;
        RETURN My_Type(Temp_Digit);
      END IF;
      -- We have now to study the case where Indice_Digit = Last - Lasty - 1.
      -- Search for Temp_Digit, so Temp_Remainder / y = Temp_Digit
      DECLARE
        Temp_y1 : CONSTANT Largest_Unsigned
                                         := Largest_Unsigned(y.Mantissa(Lasty));
        Temp_y2 : CONSTANT Largest_Unsigned
                                       := Largest_Unsigned(y.Mantissa(Lasty-1));
        Max_Digit : Largest_Unsigned := Temp_Rem / Temp_y1;
        Min_Digit : Largest_Unsigned := Temp_Rem / (Temp_y1 + 1);
        Temp : Largest_Unsigned;
      BEGIN
        IF (Max_Digit > Base_Minus_One) THEN
          Max_Digit := Base_Minus_One;
        END IF;
        LOOP
          IF (Max_Digit = Min_Digit) THEN
            RETURN My_Type(Min_Digit);
          END IF;
          IF (Max_Digit - Min_Digit = 1) THEN
            Temp := Shift_Right(Temp_y2 * Max_Digit, My_Type_Size)
                    + (Temp_y1 * Max_Digit);
            IF (Temp <= Temp_Rem) THEN
              RETURN My_Type(Max_Digit);
            ELSE
              RETURN My_Type(Min_Digit);
            END IF;
          END IF;
          -- Now, Max_Digit - Min_Digit >= 2
          Temp_Digit := Shift_Right(Max_Digit + Min_Digit, 1);
            -- Temp_Digit := (Max_Digit + Min_Digit) / 2;
          Temp := Shift_Right(Temp_y2 * Temp_Digit, My_Type_Size)
                  + (Temp_y1 * Temp_Digit);
          IF (Temp = Temp_Rem) THEN
            RETURN My_Type(Temp_Digit);
          ELSIF (Temp > Temp_Rem) THEN
            Max_Digit := Temp_Digit;
            IF ((Temp - Temp_Rem) MOD Temp_y1 > 0) THEN
              Temp := (Temp - Temp_Rem) / Temp_y1 + 1;
            ELSE
              Temp := (Temp - Temp_Rem) / Temp_y1;
            END IF;
            IF (Temp <= Temp_Digit) THEN
              Min_Digit := Temp_Digit - Temp;
            END IF;
          ELSE   -- Temp < Temp_Rem
            Min_Digit := Temp_Digit;
            Temp := (Temp_Rem - Temp) / Temp_y1;
            IF (Base_Minus_One - Temp_Digit >= Temp) THEN
              Max_Digit := Temp_Digit + Temp;
            END IF;
          END IF;
        END LOOP;
      END;
    END Guess;

    ----------------------------------------------------------------------------

  BEGIN
    Temp_Quot.Last_Significant_Digit := Last_Indice_of_the_Quotient;
    WHILE (y < Temp_Remainder) LOOP
      Last_Indice_Quot := Last_Indice_of_the_Quotient;
      Guess_Digit := Guess;
      --
      -- Guess_Digit is the real digit or the real digit plus one.
      -- Temp_Unsigned := y * Guess_Digit * (Base**Last_Indice_Quot)
      --
      Temp_Unsigned := Special_mul(y, Guess_Digit, Last_Indice_Quot);
      WHILE (Temp_Remainder < Temp_Unsigned) LOOP
        -- Decrement Guess_Digit and calculate the new value of Temp_Unsigned
        -- We have not to test if Guess_Digit = 0,
        -- because Temp_Unsigned would be equal to zero,
        -- so it could not be greater than Temp_Remainder.
        Guess_Digit := Guess_Digit - 1;
        -- The following lines are a kind of inlining for
        -- calculating Temp_Unsigned - y * (Base**Last_Indice_Quot)
        Carry := 0;
        FOR i IN First_Indice..Lasty LOOP
          Temp_Index := i + Last_Indice_Quot;
          IF (Temp_Unsigned.Mantissa(Temp_Index) > y.Mantissa(i)) OR ELSE
             (Temp_Unsigned.Mantissa(Temp_Index)=y.Mantissa(i) AND THEN Carry=0) THEN
            Temp_Unsigned.Mantissa(Temp_Index)
                  := Temp_Unsigned.Mantissa(Temp_Index) - y.Mantissa(i) - Carry;
            Carry := 0;
          ELSE  -- Temp_Unsigned.Mantissa(Temp_Index) < y.Mantissa(i) + Carry
            Temp_Unsigned.Mantissa(Temp_Index)
                                         := Temp_Unsigned.Mantissa(Temp_Index)
                                            + (Greatest_Digit XOR y.Mantissa(i))
                                            + (1 - Carry);
            Carry := 1;
          END IF;
        END LOOP;
        -- During the last preceding loop, Temp_Index was equal
        -- to y.Last_Significant_Digit + Last_Indice_Quot.
        IF (Carry = 1) THEN
          Temp_Index := Temp_Index + 1;
          Temp_Unsigned.Mantissa(Temp_Index)
                                  := Temp_Unsigned.Mantissa(Temp_Index) - Carry;
        END IF;
        WHILE (Temp_Unsigned.Mantissa(Temp_Unsigned.Last_Significant_Digit) = 0)
              AND THEN (Temp_Unsigned.Last_Significant_Digit > First_Indice) LOOP
          Temp_Unsigned.Last_Significant_Digit
                                    := Temp_Unsigned.Last_Significant_Digit - 1;
        END LOOP;
      END LOOP;
      Carry := 0;
      --
      -- Calculate the new value of Temp_Remainder, which will be
      -- Temp_Remainder - Temp_Unsigned
      --
      FOR i IN Last_Indice_Quot..Temp_Unsigned.Last_Significant_Digit LOOP
        IF (Temp_Remainder.Mantissa(i) > Temp_Unsigned.Mantissa(i)) OR ELSE
           (Temp_Remainder.Mantissa(i) = Temp_Unsigned.Mantissa(i) AND THEN Carry=0) THEN
          Temp_Remainder.Mantissa(i)
              := Temp_Remainder.Mantissa(i) - Temp_Unsigned.Mantissa(i) - Carry;
          Carry := 0;
        ELSE  -- Temp_Remainder.Mantissa(i) < Temp_Unsigned.Mantissa(i) + Carry
          Temp_Remainder.Mantissa(i)
                             := Temp_Remainder.Mantissa(i)
                                + (Greatest_Digit XOR Temp_Unsigned.Mantissa(i))
                                + (1 - Carry);
          Carry := 1;
        END IF;
      END LOOP;
      IF (Carry = 1) THEN
        Temp_Index := Temp_Unsigned.Last_Significant_Digit + 1;
        Temp_Remainder.Mantissa(Temp_Index)
                                 := Temp_Remainder.Mantissa(Temp_Index) - Carry;
      END IF;
      WHILE (Temp_Remainder.Mantissa(Temp_Remainder.Last_Significant_Digit) = 0)
            AND THEN (Temp_Remainder.Last_Significant_Digit > First_Indice) LOOP
        Temp_Remainder.Last_Significant_Digit
                                   := Temp_Remainder.Last_Significant_Digit - 1;
      END LOOP;
      --
      -- Now, we have the correct values of Guess_Digit and Temp_Remainder.
      --
      Temp_Quot.Mantissa(Last_Indice_Quot) := Guess_Digit;
    END LOOP;
    Quot := Temp_Quot;
    Remainder := Temp_Remainder;
  END;        -- DECLARE
END Big_Div;

--******************************************************************************
--******************************************************************************

FUNCTION "/"(x, y : IN Big_Unsigned) RETURN Big_Unsigned IS
  Quot, Remainder : Big_Unsigned;
BEGIN
  Big_Div(x, y, Quot, Remainder);
  RETURN Quot;
END "/";

--******************************************************************************

FUNCTION "MOD"(x, y : IN Big_Unsigned) RETURN Big_Unsigned IS
  Quot, Remainder : Big_Unsigned;
BEGIN
  Big_Div(x, y, Quot, Remainder);
  RETURN Remainder;
END "MOD";

--******************************************************************************

PROCEDURE Short_Div(x : IN Big_Unsigned;
                    y : IN My_Type;
                    Quot : OUT Big_Unsigned;
                    Remainder : OUT My_Type) IS
  Lastx : CONSTANT Index_Type := x.Last_Significant_Digit;
  Temp_Quot : Big_Unsigned := Big_Unsigned_Zero;
  Carry : Largest_Unsigned := 0;
  Temp : Largest_Unsigned;
  Temp_y : CONSTANT Largest_Unsigned := Largest_Unsigned(y);
BEGIN
  IF (y = 0) THEN
    RAISE Division_By_Zero;
  END IF;
  IF (y = 1) THEN
    Quot := x;
    Remainder := 0;
    RETURN;
  END IF;
  FOR Indice IN REVERSE First_Indice..Lastx LOOP
    Temp := Largest_Unsigned(x.Mantissa(Indice))
            + Shift_Left(Carry, My_Type_Size);
    Temp_Quot.Mantissa(Indice) := My_Type(Temp / Temp_y);
    Carry := Temp MOD Temp_y;
  END LOOP;
  IF (Lastx > 0) AND THEN
     (Temp_Quot.Mantissa(Lastx) = 0) THEN
    Temp_Quot.Last_Significant_Digit := Lastx - 1;
  ELSE
    Temp_Quot.Last_Significant_Digit := Lastx;
  END IF;
  Quot := Temp_Quot;
  Remainder := My_Type(Carry);
END Short_Div;

--******************************************************************************

FUNCTION "/"(x : IN Big_Unsigned; y : IN My_Type) RETURN Big_Unsigned IS
  Quot : Big_Unsigned;
  Remainder : My_Type;
BEGIN
  Short_Div(x, y, Quot, Remainder);
  RETURN Quot;
END "/";

--******************************************************************************

FUNCTION "MOD"(x : IN Big_Unsigned; y : IN My_Type) RETURN My_Type IS
  Quot : Big_Unsigned;
  Remainder : My_Type;
BEGIN
  Short_Div(x, y, Quot, Remainder);
  RETURN Remainder;
END "MOD";

--******************************************************************************
--******************************************************************************
--**                           Special functions                              **
--******************************************************************************
--******************************************************************************

FUNCTION Number_of_Bytes(x : IN Big_Unsigned) RETURN Largest_Unsigned IS
  Temp1 : Largest_Unsigned;
  Temp2 : My_Type := x.Mantissa(x.Last_Significant_Digit);
BEGIN
  Temp1 := Largest_Unsigned(x.Last_Significant_Digit) * (My_Type_Size / 8);
  WHILE (Temp2 /= 0) LOOP
    Temp2 := Shift_Right(Temp2, 8);
    Temp1 := Temp1 + 1;
  END LOOP;
  RETURN Temp1;
END Number_of_Bytes;

--******************************************************************************

FUNCTION Number_of_Bits(x : IN Big_Unsigned) RETURN Largest_Unsigned IS
  Temp1 : Largest_Unsigned;
  Temp2 : My_Type := x.Mantissa(x.Last_Significant_Digit);
BEGIN
  Temp1 := Largest_Unsigned(x.Last_Significant_Digit) * My_Type_Size;
  WHILE (Temp2 /= 0) LOOP
    Temp2 := Shift_Right(Temp2, 1);
    Temp1 := Temp1 + 1;
  END LOOP;
  RETURN Temp1;
END Number_of_Bits;

--******************************************************************************

FUNCTION Square_Root(x : IN Big_Unsigned) RETURN Big_Unsigned IS
  Result, Temp_Result : Big_Unsigned;
  -- Remark :
  --   In order to avoid the use of a fifth variable, Diff is initialized to 3.
  Diff : Big_Unsigned := (Last_Significant_Digit => First_Indice,
                          Mantissa => (3, OTHERS => 0));
BEGIN
  IF (x.Last_Significant_Digit = First_Indice) THEN
    IF (x.Mantissa(First_Indice) = 0) THEN
      RETURN Big_Unsigned_Zero;
    ELSIF (x.Mantissa(First_Indice) <= 3) THEN
      RETURN Big_Unsigned_One;
    END IF;
  END IF;

  -- Boucle de calcul
  Result := x / 2;
  WHILE (Diff.Last_Significant_Digit > First_Indice) OR ELSE
        ((Diff.Last_Significant_Digit = First_Indice)
         AND THEN (Diff.Mantissa(First_Indice) > 1)) LOOP
    Temp_Result := Result;
    Result := (Temp_Result + x/Temp_Result) / 2;
    IF (Temp_Result > Result) THEN
      Diff := Temp_Result - Result;
    ELSE
      Diff := Result - Temp_Result;
    END IF;
  END LOOP;
  RETURN Result;
END Square_Root;

--******************************************************************************

FUNCTION Factorial(x : IN Big_Unsigned) RETURN Big_Unsigned IS
  Indice, Result : Big_Unsigned := Big_Unsigned_One;
BEGIN
  IF (x.Last_Significant_Digit = First_Indice) AND THEN
     (x.Mantissa(First_Indice) <= 1) THEN
    RETURN Big_Unsigned_One;
  END IF;
  WHILE (Indice < x) LOOP
    Inc(Indice);
    Result := Result * Indice;
  END LOOP;
  RETURN Result;
END Factorial;

--******************************************************************************

FUNCTION GCF(x, y : IN Big_Unsigned) RETURN Big_Unsigned IS
  -- GCF = greatest common factor

  FUNCTION Temp_GCF(A, B : IN Big_Unsigned) RETURN Big_Unsigned IS
    N, Temp_A, Temp : Big_Unsigned;
    Temp_B : Big_Unsigned := B;
  BEGIN
    Big_Div(A, B, Temp, N);
    LOOP
      IF (N.Last_Significant_Digit = First_Indice) THEN
        CASE (N.Mantissa(First_Indice)) IS
          WHEN 0 => RETURN Temp_B;         -- IF (N = Big_Unsigned_Zero)...
          WHEN 1 => RETURN N;              -- IF (N = Big_Unsigned_One)...
          WHEN OTHERS => NULL;
        END CASE;
      END IF;
      Temp_A := Temp_B;
      Temp_B := N;
      Big_Div(Temp_A, Temp_B, Temp, N);
        -- In this loop, temp is only a temporary variable and hasn't any
        -- effect in this algorithm.
        -- Now, N = Temp_A MOD Temp_B.
    END LOOP;
  END Temp_GCF;

BEGIN    -- Function GCF
  IF (x.Last_Significant_Digit = First_Indice) AND THEN
     (x.Mantissa(First_Indice) = 0) THEN
     -- IF x = Big_Unsigned_Zero THEN
    RETURN y;
  ELSIF (x = y) OR ELSE
        ((y.Last_Significant_Digit = First_Indice) AND THEN
         (y.Mantissa(First_Indice) = 0)) THEN
    RETURN x;
  ELSIF (x < y) THEN
    RETURN Temp_GCF(y,x);
  ELSE
    RETURN Temp_GCF(x,y);
  END IF;
END GCF;

--******************************************************************************

FUNCTION LCM(x, y : IN Big_Unsigned) RETURN Big_Unsigned IS
  -- finds the Least Common Multiple between 2 arguments : LCM
BEGIN
  RETURN (x / GCF(x,y)) * y;
END LCM;

--******************************************************************************

FUNCTION "AND"(x, y : IN Big_Unsigned) RETURN Big_Unsigned IS
  Temp : Big_Unsigned := x;
  Temp_Index : Index_Type
                     := Max(x.Last_Significant_Digit, y.Last_Significant_Digit);
BEGIN
  FOR I IN First_Indice..y.Last_Significant_Digit LOOP
    Temp.Mantissa(I) := Temp.Mantissa(I) AND y.Mantissa(I);
  END LOOP;
  WHILE (Temp_Index > First_Indice) AND THEN (Temp.Mantissa(Temp_Index)=0) LOOP
    Temp_Index := Temp_Index - 1;
  END LOOP;
  Temp.Last_Significant_Digit := Temp_Index;
  RETURN Temp;
END "AND";

--******************************************************************************

FUNCTION "XOR"(x, y : IN Big_Unsigned) RETURN Big_Unsigned IS
  Temp : Big_Unsigned := x;
  Temp_Index : Index_Type
                     := Max(x.Last_Significant_Digit, y.Last_Significant_Digit);
BEGIN
  FOR I IN First_Indice..y.Last_Significant_Digit LOOP
    Temp.Mantissa(I) := Temp.Mantissa(I) XOR y.Mantissa(I);
  END LOOP;
  WHILE (Temp_Index > First_Indice) AND THEN (Temp.Mantissa(Temp_Index)=0) LOOP
    Temp_Index := Temp_Index - 1;
  END LOOP;
  Temp.Last_Significant_Digit := Temp_Index;
  RETURN Temp;
END "XOR";

--******************************************************************************

FUNCTION NCR(N, R : IN Big_Unsigned) RETURN Big_Unsigned IS
  -- NCR(N,R) = N!/[R!(N-R)!]
  -- In this algorythm, we research of the maximum (max_R) and the
  -- minimum (Min_R) of R and (N-R).
  -- We have : NCR(N,R) = [N! / Max_R!] / Min_R!;
  -- Then, nCr(N,R) is considered as a fraction, even if its result is an
  -- integer : Temp_NCR_Numerator / Temp_NCR_Denominator
  -- with Temp_NCR_Numerator = N! / Max_R!
  -- and Temp_NCR_Denominateur = Min_R!
  Temp_N : CONSTANT Big_Unsigned := N - R;
  Max_R : CONSTANT Big_Unsigned := Max(R, Temp_N);
  Min_R : CONSTANT Big_Unsigned := Min(R, Temp_N);
  Indice : Big_Unsigned := Big_Unsigned_One;
  Temp_NCR_Numerator : Big_Unsigned := Max_R + Big_Unsigned_One;
  Temp_NCR_Denominator : Big_Unsigned := Big_Unsigned_One;
  Temp_Num, Temp_Den : Big_Unsigned;

  PROCEDURE Simplify(x, y : IN OUT Big_Unsigned) IS
    Temp : CONSTANT Big_Unsigned := GCF(x,y);
  BEGIN
    x := x / Temp;
    y := y / Temp;
  END Simplify;

BEGIN
  IF (N = R) THEN
    RETURN Big_Unsigned_One;
  END IF;
  IF (N < R) THEN
    RETURN Big_Unsigned_Zero;
  END IF;
  -- Now, N > R.
  WHILE (Indice < Min_R) LOOP
    Temp_Num := Max_R + Indice;
    Inc(Temp_Num);
    Temp_Den := Indice;
    Inc(Temp_Den);
    Simplify(Temp_Den, Temp_Num);
    Simplify(Temp_NCR_Denominator, Temp_Num);
    Simplify(Temp_NCR_Numerator, Temp_Den);
    Temp_NCR_Numerator := Temp_NCR_Numerator * Temp_Num;
    Temp_NCR_Denominator := Temp_NCR_Denominator * Temp_Den;
    Inc(Indice);
  END LOOP;
  RETURN Temp_NCR_Numerator;
END NCR;

--******************************************************************************

FUNCTION NPR(N, R : IN Big_Unsigned) RETURN Big_Unsigned IS
  Temp_Result : Big_Unsigned := Big_Unsigned_One;
  Index : Big_Unsigned;
BEGIN
  IF (N < R) THEN
    RETURN Big_Unsigned_Zero;
  END IF;
  IF (R.Last_Significant_Digit = First_Indice) AND THEN
     (R.Mantissa(First_Indice)  = 0) THEN
    RETURN Big_Unsigned_One;
  END IF;
  Index := N - R;
  WHILE (Index < N) LOOP
    Inc(Index);
    Temp_Result := Temp_Result * Index;
  END LOOP;
  RETURN Temp_Result;
END NPR;

--******************************************************************************

FUNCTION K(N, P : IN Big_Unsigned) RETURN Big_Unsigned IS
  Temp : Big_Unsigned;
BEGIN
  -- K(N,P) = NCR(N+P-1);
  -- We have to calculate Temp = N+P-1.
  -- The 3 following lines allows to avoid an overflow, in the
  -- case if N+P = Big_Unsigned_Last + 1 : in this situation,
  -- Temp will be Big_Unsigned_Last.
  Temp := Max(N,P);
  Dec(Temp);
  Temp := Temp + Min(N,P);
  -- Now, Temp = N+P-1.
  RETURN NCR(Temp,P);
END K;

--******************************************************************************

END Unsigned_Number;
--******************************************************************************
--                                                                            **
--  File:        BINUSINU.ADB                                                 **
--  Description: Big_Number.Signed_Number package body                        **
--  Revision:    0.28 (beta version)                                          **
--  Date:        July 8, 2001                                                 **
--  Author:      Jrme Delcourt                                              **
--  Mail:        sikander@club-internet.fr                                    **
--                                                                            **
--  Copyright (c) Jrme Delcourt, 1998, 1999, 2000, 2001                     **
--  62, rue Ngrier                                                           **
--  59800 Lille                                                               **
--  FRANCE                                                                    **
--                                                                            **
--  Permission granted to use for any purpose, provided this copyright        **
--  remains attached and unmodified.                                          **
--                                                                            **
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR            **
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED            **
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       **
--                                                                            **
--******************************************************************************

SEPARATE(Big_Number)

PACKAGE BODY Signed_Number IS

--******************************************************************************
--******************************************************************************
--**                              Comparisons                                 **
--******************************************************************************
--******************************************************************************

FUNCTION "<"(Left, Right : IN Big_Signed) RETURN BOOLEAN IS
BEGIN
  IF (Left.Positive /= Right.Positive) THEN
    RETURN BOOLEAN(Right.Positive);
  ELSIF (Left = Right) THEN
    RETURN FALSE;
  ELSE
    -- Now, x and y have the same sign and are different.
    DECLARE
      Indice : Index_Type := Left.Last_Significant_Digit;
    BEGIN
      -- Left and Right have the same sign but their mantissas are different.
      IF (Indice /= Right.Last_Significant_Digit) THEN
        RETURN ((Indice < Right.Last_Significant_Digit) = Left.Positive);
      END IF;
      WHILE (Left.Mantissa(Indice) = Right.Mantissa(Indice))
            AND THEN (Indice > First_Indice) LOOP
        Indice := Indice - 1;
      END LOOP;
      RETURN ((Left.Mantissa(Indice) < Right.Mantissa(Indice)) = Left.Positive);
    END;
  END IF;
END "<";

--******************************************************************************

FUNCTION ">"(Left, Right : IN Big_Signed) RETURN BOOLEAN IS
  -- The comparison "Left > Right" is equivalent to "Right < Left"...
BEGIN
  RETURN Right < Left;
END ">";

--******************************************************************************

FUNCTION "<="(Left, Right : IN Big_Signed) RETURN BOOLEAN IS
BEGIN
  RETURN NOT (Right < Left);
END "<=";

--******************************************************************************

FUNCTION ">="(Left, Right : IN Big_Signed) RETURN BOOLEAN IS
  -- The comparison "Left >= Right" is the same test than "Right <= Left"...
BEGIN
  RETURN Right <= Left;
END ">=";

--******************************************************************************

FUNCTION Is_Null(x : IN Big_Signed) RETURN BOOLEAN IS
BEGIN
  IF (x.Last_Significant_Digit = First_Indice) AND THEN
     (x.Mantissa(First_Indice) = 0) THEN
    RETURN TRUE;
  ELSE
    RETURN FALSE;
  END IF;
END Is_Null;

--******************************************************************************

FUNCTION Min(x, y : IN Big_Signed) RETURN Big_Signed IS
BEGIN
  IF (x < y) THEN
    RETURN x;
  ELSE
    RETURN y;
  END IF;
END Min;

--******************************************************************************

FUNCTION Max(x, y : IN Big_Signed) RETURN Big_Signed IS
BEGIN
  IF (x < y) THEN
    RETURN y;
  ELSE
    RETURN x;
  END IF;
END Max;

--******************************************************************************
--******************************************************************************
--**                            Binary operations                             **
--******************************************************************************
--******************************************************************************

PROCEDURE Swap1(x, y : IN OUT Big_Signed) IS
  Temp : CONSTANT Big_Signed := x;
BEGIN
  x := y;
  y := Temp;
END Swap1;

--******************************************************************************

PROCEDURE Swap2(x, y : IN OUT Big_Signed) IS
  Max_Index : CONSTANT Index_Type
                     := Max(x.Last_Significant_Digit, y.Last_Significant_Digit);
  Temp : My_Type;
  Temp_Positive : CONSTANT BOOLEAN := x.Positive;
BEGIN
  x.Positive := y.Positive;
  y.Positive := Temp_Positive;
  FOR I IN First_Indice..Max_Index LOOP
    Temp := x.Mantissa(I);
    x.Mantissa(I) := y.Mantissa(I);
    y.Mantissa(I) := Temp;
  END LOOP;
END Swap2;

--******************************************************************************

PROCEDURE Inc(x : IN OUT Big_Signed) IS
  Temp_x : Big_Unsigned := Signed2Unsigned(x);
BEGIN
  IF (x.Last_Significant_Digit = First_Indice) AND THEN
     (x.Mantissa(First_Indice) = 0) THEN
     -- If x = 0...
    x := Big_Signed_One;
    RETURN;
  ELSIF (x.Positive = TRUE) THEN
    -- x > 0
    Inc(Temp_x);
    x := Unsigned2Signed(Temp_x, x.Positive);
  ELSE
    -- x < 0
    IF (x.Last_Significant_Digit = First_Indice) AND THEN
       (x.Mantissa(First_Indice) = 1) THEN
      x := Big_Signed_Zero;
    ELSE
      Dec(Temp_x);
      x := Unsigned2Signed(Temp_x, x.Positive);
    END IF;
  END IF;
END Inc;

--******************************************************************************

PROCEDURE Dec(x : IN OUT Big_Signed) IS
  Temp_x : Big_Unsigned := Signed2Unsigned(x);
BEGIN
  IF (x.Last_Significant_Digit = First_Indice) AND THEN
     (x.Mantissa(First_Indice) = 0) THEN
     -- If x = 0...
    x := Big_Signed_One;
    x.Positive := FALSE;
  ELSIF (x.Positive = TRUE) THEN
    -- x > 0
    Dec(Temp_x);
    x := Unsigned2Signed(Temp_x, x.Positive);
  ELSE
    -- x < 0
    Inc(Temp_x);
    x := Unsigned2Signed(Temp_x, x.Positive);
  END IF;
END Dec;

--******************************************************************************

FUNCTION "+"(x, y : IN Big_Signed) RETURN Big_Signed IS
  Tmp_x : Big_Signed := x;
  Tmp_y : Big_Signed := y;
BEGIN
  IF (Tmp_x.Positive /= Tmp_y.Positive) THEN
    IF (Tmp_x.Positive = FALSE) THEN
      Swap1(Tmp_x, Tmp_y);
    END IF;
    Tmp_y.Positive := TRUE;
    RETURN Tmp_x - Tmp_y;
  END IF;

  -- Now, the function studies the case where x and y have the same sign.
  DECLARE
    Temp_x : CONSTANT Big_Unsigned := Signed2Unsigned(Tmp_x);
    Temp_y : CONSTANT Big_Unsigned := Signed2Unsigned(Tmp_y);
    Result : CONSTANT Big_Signed
                            := Unsigned2Signed(Temp_x + Temp_y, Tmp_x.Positive);
  BEGIN
    RETURN Result;
  END;
END "+";

--******************************************************************************
--******************************************************************************

FUNCTION "-"(x, y : IN Big_Signed) RETURN Big_Signed IS
BEGIN
  IF (x.Positive /= y.Positive) THEN
    RETURN x + Opposite(y);
  ELSE              -- (x.Positive = y.Positive)
    DECLARE
      Temp_x : CONSTANT Big_Unsigned := Signed2Unsigned(x);
      Temp_y : CONSTANT Big_Unsigned := Signed2Unsigned(y);
      Result : Big_Signed;
    BEGIN
      IF (Absolute(x) >= Absolute(y)) THEN
        Result := Unsigned2Signed(Temp_x - Temp_y, x.Positive);
      ELSE
        Result := Unsigned2Signed(Temp_y - Temp_x, NOT x.Positive);
      END IF;
      RETURN Result;
    END;
  END IF;
END "-";

--******************************************************************************
--******************************************************************************

FUNCTION "*"(x, y : IN Big_Signed) RETURN Big_Signed IS
  Temp_x : CONSTANT Big_Unsigned := Signed2Unsigned(x);
  Temp_y : CONSTANT Big_Unsigned := Signed2Unsigned(y);
  Result : CONSTANT Big_Signed
             := Unsigned2Signed(Temp_x * Temp_y, (x.Positive = y.Positive));
BEGIN
  RETURN Result;
END "*";

--******************************************************************************
--******************************************************************************

FUNCTION "**"(x : IN Big_Signed;
              y : IN My_Type) RETURN Big_Signed IS
  Temp_Result : Big_Signed := x;
BEGIN
  IF (y = 0) THEN
    RETURN Big_Signed_One;
  END IF;
  IF (x.Last_Significant_Digit = First_Indice)
     AND THEN (x.Mantissa(First_Indice) <= 1) THEN
    IF (Temp_Result.Positive = FALSE) THEN
      -- In this case, the result is positive if y MOD 2 = 0,
      -- and negative if y MOD 2 = 0.
      Temp_Result.Positive := (y MOD 2) = 0;
    END IF;
    RETURN Temp_Result;
  END IF;
  FOR I IN 2..y LOOP
    Temp_Result := Temp_Result * x;
  END LOOP;
  RETURN Temp_Result;
END "**";

--******************************************************************************
--******************************************************************************

FUNCTION "**"(x : IN Big_Signed;
              y : IN Index_Type) RETURN Big_Signed IS
  Temp_Result : Big_Signed := x;
BEGIN
  IF (y = 0) THEN
    RETURN Big_Signed_One;
  END IF;
  IF (x.Last_Significant_Digit = First_Indice)
     AND THEN (x.Mantissa(First_Indice) <= 1) THEN
    IF (Temp_Result.Positive = FALSE) THEN
      -- In this case, the result is positive if y MOD 2 = 0,
      -- and negative if y MOD 2 = 0.
      Temp_Result.Positive := (y MOD 2) = 0;
    END IF;
    RETURN Temp_Result;
  END IF;
  FOR I IN 2..y LOOP
    Temp_Result := Temp_Result * x;
  END LOOP;
  RETURN Temp_Result;
END "**";

--******************************************************************************
--******************************************************************************

FUNCTION "**"(x : IN Big_Signed;
              y : IN Big_Unsigned) RETURN Big_Signed IS
  Temp_Result : Big_Signed := x;
  Indice : Big_Unsigned := Big_Unsigned_One;
BEGIN
  IF (y.Last_Significant_Digit = First_Indice) AND THEN
     (y.Mantissa(First_Indice) = 0) THEN
     -- These two tests are equivalent to the
     -- test "y = Big_Unsigned_Zero", but it's much faster that way.
    RETURN Big_Signed_One;
  END IF;
  IF (x.Last_Significant_Digit = First_Indice) AND THEN
     (x.Mantissa(First_Indice) <= 1) THEN
    IF (Temp_Result.Positive = FALSE) THEN
      -- In this case, the result is positive if y MOD 2 = 0,
      -- and negative if y MOD 2 = 0.
      -- We have just to check if the lowest digit is a multiple or not of 2.
      Temp_Result.Positive := (y.Mantissa(First_Indice) MOD 2) = 0;
    END IF;
    RETURN Temp_Result;
  END IF;
  WHILE (Indice < y) LOOP
    Inc(Indice);
    Temp_Result := Temp_Result * x;
  END LOOP;
  RETURN Temp_Result;
END "**";

--******************************************************************************
--******************************************************************************

FUNCTION "**"(x, y : IN Big_Signed) RETURN Big_Signed IS
  Temp_Result : Big_Signed := x;
  Indice : Big_Signed := Big_Signed_One;
  Temp_y : CONSTANT Big_Unsigned := Signed2Unsigned(y);
BEGIN
  IF (y.Positive = FALSE) THEN
    RETURN Big_Signed_Zero;
  END IF;
  RETURN (x ** Temp_y);
END "**";

--******************************************************************************
--******************************************************************************

PROCEDURE Big_Div(x, y : IN Big_Signed;
                  Quot, Remainder : OUT Big_Signed) IS
  Temp_x : CONSTANT Big_Unsigned := Signed2Unsigned(x);
  Temp_y : CONSTANT Big_Unsigned := Signed2Unsigned(y);
  Temp_Quot, Temp_Remainder : Big_Unsigned;
  Result_Positive : CONSTANT BOOLEAN := (x.Positive = y.Positive);
BEGIN
  Big_Div(Temp_x, Temp_y, Temp_Quot, Temp_Remainder);
  Quot := Unsigned2Signed(Temp_Quot, Result_Positive);
  Remainder := Unsigned2Signed(Temp_Remainder, Result_Positive);
END Big_Div;

--******************************************************************************
--******************************************************************************

PROCEDURE Short_Div(x : IN Big_Signed; y : IN My_Type;
                    Quot : OUT Big_Signed; Remainder : OUT My_Type) IS
  Temp_x : CONSTANT Big_Unsigned := Signed2Unsigned(x);
  Temp_Quot : Big_Unsigned;
BEGIN
  Short_Div(Temp_x, y, Temp_Quot, Remainder);
  Quot := Unsigned2Signed(Temp_Quot, x.Positive);
END Short_Div;

--******************************************************************************
--******************************************************************************

FUNCTION "/"(x, y : IN Big_Signed) RETURN Big_Signed IS
  Quot, Remainder : Big_Signed;
BEGIN
  Big_Div(x, y, Quot, Remainder);
  RETURN Quot;
END "/";

--******************************************************************************
--******************************************************************************

FUNCTION "MOD"(x, y : IN Big_Signed) RETURN Big_Signed IS
  Quot, Remainder : Big_Signed;
BEGIN
  Big_Div(x, y, Quot, Remainder);
  IF (x > Big_Signed_Zero) THEN
    RETURN Absolute(Remainder);
  ELSE
    RETURN Absolute(Absolute(y) - Absolute(Remainder));
  END IF;
END "MOD";

--******************************************************************************
--******************************************************************************

FUNCTION "/"(x : IN Big_Signed; y : IN My_Type) RETURN Big_Signed IS
  Quot : Big_Signed;
  Remainder : My_Type;
BEGIN
  Short_Div(x, y, Quot, Remainder);
  RETURN Quot;
END "/";

--******************************************************************************
--******************************************************************************

FUNCTION "MOD"(x : IN Big_Signed; y : IN My_Type) RETURN My_Type IS
  Quot : Big_Signed;
  Remainder : My_Type;
BEGIN
  Short_Div(x, y, Quot, Remainder);
  IF (x > Big_Signed_Zero) THEN
    RETURN Remainder;
  ELSE
    RETURN (y - Remainder);
  END IF;
  RETURN Remainder;
END "MOD";

--******************************************************************************
--******************************************************************************

FUNCTION Opposite(x : IN Big_Signed) RETURN Big_Signed IS
  Temp_Result : Big_Signed := x;
BEGIN
  IF (Temp_Result.Last_Significant_Digit /= First_Indice) OR ELSE
     (Temp_Result.Mantissa(First_Indice) /= 0) THEN
    -- If Temp_Result /= 0, then inverse the sign.
    -- Note that if x = 0, by convention, the sign of x is always
    -- set to positive.
    Temp_Result.Positive := NOT Temp_Result.Positive;
  END IF;
  RETURN Temp_Result;
END Opposite;

--******************************************************************************
--******************************************************************************

FUNCTION Absolute(x : IN Big_Signed) RETURN Big_Signed IS
BEGIN
  IF (x.Positive = FALSE) THEN
    RETURN Opposite(x);
  ELSE
    RETURN x;
  END IF;
END Absolute;

--******************************************************************************
--******************************************************************************
--**                           Special functions                              **
--******************************************************************************
--******************************************************************************

FUNCTION Number_of_Bytes(x : IN Big_Signed) RETURN Largest_Unsigned IS
BEGIN
  RETURN Number_of_Bytes(Signed2Unsigned(x));
END Number_of_Bytes;

--******************************************************************************

FUNCTION Number_of_Bits(x : IN Big_Signed) RETURN Largest_Unsigned IS
BEGIN
  RETURN Number_of_Bits(Signed2Unsigned(x));
END Number_of_Bits;

--******************************************************************************

FUNCTION Square_Root(x : IN Big_Signed) RETURN Big_Signed IS
  Result, Temp_Result : Big_Signed;
  Two : CONSTANT Big_Signed := (Positive => TRUE,
                                Last_Significant_Digit => First_Indice,
                                Mantissa => (2, OTHERS => 0));
  -- Remark :
  --   In order to avoid the use of a fifth variable, Diff is initialized to 3.
  Diff : Big_Signed := (Positive => TRUE,
                        Last_Significant_Digit => First_Indice,
                        Mantissa => (3, OTHERS => 0));
BEGIN
  IF (x < Big_Signed_Zero) THEN
    RAISE Parameter_Error;
  ELSIF (x.Last_Significant_Digit = First_Indice) AND THEN
        (x.Mantissa(First_Indice) = 0) THEN
    RETURN Big_Signed_Zero;
  ELSIF (x <= Diff) THEN  -- On a provisoirement : Diff = 3.
    RETURN Big_Signed_One;
  END IF;

  -- Loop to calculate the square root
  Result := x / Two;
  WHILE (Diff > Big_Signed_One) LOOP
    Temp_Result := Result;
    Result := (Temp_Result + x/Temp_Result) / Two;
    IF (Temp_Result > Result) THEN
      Diff := Temp_Result - Result;
    ELSE
      Diff := Result - Temp_Result;
    END IF;
  END LOOP;
  RETURN Result;
END Square_Root;

--******************************************************************************

FUNCTION Factorial(x : IN Big_Signed) RETURN Big_Signed IS
  Temp_x : CONSTANT Big_Unsigned := Signed2Unsigned(x);
BEGIN
  IF (x.Positive = FALSE) THEN
    RAISE Parameter_Error;
  END IF;
  RETURN Unsigned2Signed(Factorial(Temp_x), TRUE);
END Factorial;

--******************************************************************************

FUNCTION GCF(x, y : IN Big_Signed) RETURN Big_Signed IS
  -- In english : GCF = greatest common factor
  Temp_x : CONSTANT Big_Unsigned := Signed2Unsigned(x);
  Temp_y : CONSTANT Big_Unsigned := Signed2Unsigned(y);
BEGIN
  RETURN Unsigned2Signed(GCF(Temp_x, Temp_y), TRUE);
END GCF;

--******************************************************************************

FUNCTION LCM(x, y : IN Big_Signed) RETURN Big_Signed IS
  -- finds the Least Common Multiple between 2 arguments : LCM
  Tmp : Big_Signed := (x / GCF(x,y)) * y;
BEGIN
  IF (Tmp.Positive = FALSE) THEN
    Tmp.Positive := TRUE;
  END IF;
  RETURN Tmp;
END LCM;

--******************************************************************************

END Signed_Number;
--******************************************************************************
--                                                                            **
--  File:        BNBIUNIO.ADB                                                 **
--  Description: Big_Number.Big_Unsigned_IO package body                      **
--  Revision:    0.28 (beta version)                                          **
--  Date:        July 8, 2001                                                 **
--  Author:      Jrme Delcourt                                              **
--  Mail:        sikander@club-internet.fr                                    **
--                                                                            **
--  Copyright (c) Jrme Delcourt, 1998, 1999, 2000, 2001                     **
--  62, rue Ngrier                                                           **
--  59800 Lille                                                               **
--  FRANCE                                                                    **
--                                                                            **
--  Permission granted to use for any purpose, provided this copyright        **
--  remains attached and unmodified.                                          **
--                                                                            **
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR            **
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED            **
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       **
--                                                                            **
--******************************************************************************

SEPARATE(Big_Number)

PACKAGE BODY Big_Unsigned_IO IS

--******************************************************************************
--******************************************************************************

PROCEDURE GET_LINE(x : OUT Big_Unsigned) IS
  Temp_String : UNBOUNDED_STRING;
BEGIN
  GET_LINE(Temp_String);
  x := UString2Unsigned(Temp_String);
END GET_LINE;

--******************************************************************************
--******************************************************************************

PROCEDURE GET_LINE(File : IN FILE_TYPE; x : OUT Big_Unsigned) IS
  Temp_String : UNBOUNDED_STRING;
BEGIN
  GET_LINE(File, Temp_String);
  x := UString2Unsigned(Temp_String);
END GET_LINE;

--******************************************************************************
--******************************************************************************

PROCEDURE PUT(x : IN Big_Unsigned;
              Base : IN My_Type := 10) IS
BEGIN
  PUT(Big_Unsigned2UString(x, Base));
END PUT;

--******************************************************************************
--******************************************************************************

PROCEDURE PUT(File : IN FILE_TYPE;
              x : IN Big_Unsigned;
              Base : IN My_Type := 10) IS
BEGIN
  PUT(File, Big_Unsigned2UString(x, Base));
END PUT;

--******************************************************************************
--******************************************************************************

PROCEDURE PUT_LINE(x : IN Big_Unsigned;
                   Base : IN My_Type := 10) IS
BEGIN
  PUT_LINE(Big_Unsigned2UString(x, Base));
END PUT_LINE;

--******************************************************************************
--******************************************************************************

PROCEDURE PUT_LINE(File : IN FILE_TYPE;
                   x : IN Big_Unsigned;
                   Base : IN My_Type := 10) IS
BEGIN
  PUT_LINE(File, Big_Unsigned2UString(x, Base));
END PUT_LINE;

--******************************************************************************
--******************************************************************************

END Big_Unsigned_IO;
--******************************************************************************
--                                                                            **
--  File:        BNBISIIO.ADB                                                 **
--  Description: Big_Number.Big_Signed_IO package body                        **
--  Revision:    0.28 (beta version)                                          **
--  Date:        July 8, 2001                                                 **
--  Author:      Jrme Delcourt                                              **
--  Mail:        sikander@club-internet.fr                                    **
--                                                                            **
--  Copyright (c) Jrme Delcourt, 1998, 1999, 2000, 2001                     **
--  62, rue Ngrier                                                           **
--  59800 Lille                                                               **
--  FRANCE                                                                    **
--                                                                            **
--  Permission granted to use for any purpose, provided this copyright        **
--  remains attached and unmodified.                                          **
--                                                                            **
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR            **
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED            **
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       **
--                                                                            **
--******************************************************************************

SEPARATE(Big_Number)

PACKAGE BODY Big_Signed_IO IS

--******************************************************************************
--******************************************************************************

PROCEDURE GET_LINE(x : OUT Big_Signed) IS
  Temp_String : UNBOUNDED_STRING;
BEGIN
  GET_LINE(Temp_String);
  x := UString2Signed(Temp_String);
END GET_LINE;

--******************************************************************************
--******************************************************************************

PROCEDURE GET_LINE(File : IN FILE_TYPE; x : OUT Big_Signed) IS
  Temp_String : UNBOUNDED_STRING;
BEGIN
  GET_LINE(File, Temp_String);
  x := UString2Signed(Temp_String);
END GET_LINE;

--******************************************************************************
--******************************************************************************

PROCEDURE PUT(x : IN Big_Signed;
              Base : IN My_Type := 10) IS
BEGIN
  PUT(Big_Signed2UString(x, Base));
END PUT;

--******************************************************************************
--******************************************************************************

PROCEDURE PUT(File : IN FILE_TYPE;
              x : IN Big_Signed;
              Base : IN My_Type := 10) IS
BEGIN
  PUT(File, Big_Signed2UString(x, Base));
END PUT;

--******************************************************************************
--******************************************************************************

PROCEDURE PUT_LINE(x : IN Big_Signed;
                   Base : IN My_Type := 10) IS
BEGIN
  PUT_LINE(Big_Signed2UString(x, Base));
END PUT_LINE;

--******************************************************************************
--******************************************************************************

PROCEDURE PUT_LINE(File : IN File_Type;
                   x : IN Big_Signed;
                   Base : IN My_Type := 10) IS
BEGIN
  PUT_LINE(File, Big_Signed2UString(x, Base));
END PUT_LINE;

--******************************************************************************
--******************************************************************************

END Big_Signed_IO;
--******************************************************************************
--                                                                            **
--  File:        BINUBIFR.ADB                                                 **
--  Description: Big_Number.Big_Fraction package body                         **
--  Revision:    0.28 (beta version)                                          **
--  Date:        July 8, 2001                                                 **
--  Author:      Jrme Delcourt                                              **
--  Mail:        sikander@club-internet.fr                                    **
--                                                                            **
--  Copyright (c) Jrme Delcourt, 1998, 1999, 2000, 2001                     **
--  62, rue Ngrier                                                           **
--  59800 Lille                                                               **
--  FRANCE                                                                    **
--                                                                            **
--  Permission granted to use for any purpose, provided this copyright        **
--  remains attached and unmodified.                                          **
--                                                                            **
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR            **
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED            **
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       **
--                                                                            **
--******************************************************************************

SEPARATE(Big_Number)

PACKAGE BODY Big_Fraction IS

--******************************************************************************
--******************************************************************************

FUNCTION Simplify_Fraction(x : IN Fraction) RETURN Fraction IS
  -- Here, we don't have to test for a denominator equal to zero,
  -- because it would have been already checked in the functions
  -- CreateFraction or InverseFraction.
BEGIN
  IF (x.Denominator.Last_Significant_Digit = First_Indice) AND THEN
     (x.Denominator.Mantissa(First_Indice) = 1) THEN
     -- If (x.Denominator = Big_Unsigned_One)...
    RETURN x;
  END IF;
  IF (x.Numerator.Last_Significant_Digit = First_Indice) THEN
    CASE x.Numerator.Mantissa(First_Indice) IS
      WHEN 0 => -- If (x.Numerator = Big_Unsigned_Zero)...
               RETURN (TRUE, Big_Unsigned_Zero, Big_Unsigned_One);
      WHEN 1 => -- If (x.Numerator = Big_Unsigned_One)...
               RETURN x;
      WHEN OTHERS => NULL;
    END CASE;
  END IF;
  DECLARE
    Temp : CONSTANT Big_Unsigned := GCF(x.Numerator, x.Denominator);
  BEGIN
    RETURN (x.Positive, x.Numerator / Temp, x.Denominator / Temp);
  END;
END Simplify_Fraction;

--******************************************************************************
--******************************************************************************

FUNCTION CreateFraction(Numerator, Denominator : IN Big_Unsigned;
                        Positive : IN BOOLEAN := TRUE) RETURN Fraction IS
BEGIN
  IF (Denominator.Last_Significant_Digit = First_Indice) AND THEN
     (Denominator.Mantissa(First_Indice) = 0) THEN
    -- These two tests are equivalent
    -- to the test Denominator = Big_Unsigned_Zero
    -- but it's much faster that way.
    RAISE Division_By_Zero;
  END IF;
  RETURN Simplify_Fraction((Positive, Numerator, Denominator));
END CreateFraction;

--******************************************************************************
--******************************************************************************

FUNCTION CreateFraction(Numerator, Denominator : IN Big_Signed) RETURN Fraction IS
BEGIN
  IF (Denominator.Last_Significant_Digit = First_Indice) AND THEN
     (Denominator.Mantissa(First_Indice) = 0) THEN
    -- These two tests are equivalent
    -- to the test Denominator = Big_Signed_Zero
    -- but it's much faster that way.
    RAISE Division_By_Zero;
  END IF;
  RETURN Simplify_Fraction((Numerator.Positive=Denominator.Positive,
                            Signed2Unsigned(Numerator),
                            Signed2Unsigned(Denominator)));
END CreateFraction;

--******************************************************************************
--******************************************************************************

FUNCTION Get_Numerator(x : IN Fraction) RETURN Big_Unsigned IS
BEGIN
  RETURN x.Numerator;
END Get_Numerator;

PRAGMA INLINE(Get_Numerator);

--******************************************************************************
--******************************************************************************

FUNCTION Get_Denominator(x : IN Fraction) RETURN Big_Unsigned IS
BEGIN
  RETURN x.Denominator;
END Get_Denominator;

PRAGMA INLINE(Get_Denominator);

--******************************************************************************
--******************************************************************************

FUNCTION Is_Positive(x : IN Fraction) RETURN BOOLEAN IS
BEGIN
  RETURN x.Positive;
END Is_Positive;

PRAGMA INLINE(Is_Positive);

--******************************************************************************
--******************************************************************************

PROCEDURE Swap_Fractions(x, y : IN OUT Fraction) IS
  Temp : CONSTANT Fraction := x;
BEGIN
  x := y;
  y := Temp;
END Swap_Fractions;

--******************************************************************************
--******************************************************************************

FUNCTION Inverse_Fraction(x : IN Fraction) RETURN Fraction IS
BEGIN
  -- This function returns 1/x.
  -- If the numerator of x is equal to 0, this function must raise
  -- the exception Division_By_Zero.
  IF (x.Numerator.Last_Significant_Digit = First_Indice) AND THEN
     (x.Numerator.Mantissa(First_Indice) = 0) THEN
     -- If (x.Numerator = Big_Unsigned_Zero) THEN...
     -- It's much faster that way.
    RAISE Division_By_Zero;
  ELSE
    RETURN (x.Positive, x.Denominator, x.Numerator);
  END IF;
END Inverse_Fraction;

--******************************************************************************
--******************************************************************************

FUNCTION "*"(x, y : IN Fraction) RETURN Fraction IS
  Temp1 : CONSTANT Big_Unsigned := GCF(x.Numerator, y.Denominator);
  Temp2 : CONSTANT Big_Unsigned := GCF(y.Numerator, x.Denominator);
BEGIN
  RETURN (x.Positive = y.Positive,
          (x.Numerator / Temp1) * (y.Numerator / Temp2),
          (x.Denominator / Temp2) * (y.Denominator / Temp1));
EXCEPTION
  WHEN OTHERS => RAISE Fraction_Overflow;
END "*";

--******************************************************************************
--******************************************************************************

FUNCTION "/"(x, y : IN Fraction) RETURN Fraction IS
BEGIN
  -- The function Inverse_Fraction can raise the exception "DIVISION_BY_ZERO"
  -- if y.Numerator = Big_Unsigned_Zero.
  RETURN (x * Inverse_Fraction(y));
END "/";

--******************************************************************************
--******************************************************************************

-- a/b + c/d = a/(x*b') + c/(x*d'),
-- with b = x*b' and d = x*d' and x = GCF(b,d)
-- So a/b + c/d = (a/b'+ c/d') * 1/x
-- a/b + c/d = (a*d' + c*b') / (x * b'* d')
-- x*b' = b
-- So a/b + c/d = (a*d' + c*b') / (b*d')
--
-- The fractions a/b and c/d are supposed to be already simplified,
-- so : - GCF(a, b) = 1 ==> GCF(a*d', b*d') = d'
--      - GCF(c, d) = 1 ==> GCF(c, d') = 1
--      - GCF(b', b) = b' ==> GCF(c*b', b*d') = b'
--
-- Furthermore, GCF(b', d') = 1
-- So, GCF(a*d' + c*b', b*d') = 1

FUNCTION "+"(x, y : IN Fraction) RETURN Fraction IS
  Result : Fraction;
  TempGCF : CONSTANT Big_Unsigned := GCF(x.Denominator, y.Denominator);
  Temp : CONSTANT Big_Unsigned := y.Denominator / TempGCF;
  Tmp1 : Big_Unsigned := x.Numerator * Temp;
  Tmp2 : CONSTANT Big_Unsigned := y.Numerator * (x.Denominator / TempGCF);
BEGIN
  IF (x.Positive = y.Positive) THEN
    -- The 2 fractions have the same sign.
    Result.Positive := x.Positive;
    Result.Numerator := Tmp1 + Tmp2;
  ELSE
    -- The 2 fractions have opposite signs.
    IF (Tmp1 < Tmp2) THEN
      Result.Positive := y.Positive;
      Result.Numerator := Tmp2 - Tmp1;
    ELSE   -- Tmp1 >= Tmp2
      Result.Positive := x.Positive;
      Result.Numerator := Tmp1 - Tmp2;
    END IF;
  END IF;
  IF (Result.Numerator.Last_Significant_Digit = First_Indice) THEN
    IF (Result.Numerator.Mantissa(First_Indice) = 0) THEN
       -- IF (Result.Numerator = Big_Unsigned_Zero) THEN...
      Result.Positive := TRUE;
      Result.Denominator := Big_Unsigned_One;
    ELSE
      Result.Denominator := x.Denominator * Temp;
      IF (Result.Numerator.Mantissa(First_Indice) > 1) THEN
        Tmp1 := GCF(Result.Numerator, Result.Denominator);
        Result.Numerator := Result.Numerator / Tmp1;
        Result.Denominator := Result.Denominator / Tmp1;
      END IF;
    END IF;
    RETURN Result;
  END IF;
  -- The three following lines allow to avoid some overflows.
  Tmp1 := GCF(Result.Numerator, x.Denominator);
  Result.Numerator := Result.Numerator / Tmp1;
  Result.Denominator := (x.Denominator / Tmp1) * Temp;
  RETURN Result;
EXCEPTION
  WHEN OTHERS => RAISE Fraction_Overflow;
END "+";

--******************************************************************************
--******************************************************************************

FUNCTION "-"(x, y : IN Fraction) RETURN Fraction IS
  Temp_y : Fraction := y;
  -- Just inverse the sign of y (Temp_y := -y) and then -y is added to x.
  -- So we have not to rewrite a new algorithm for this function.
BEGIN
  Temp_y.Positive := NOT (Temp_y.Positive);
  RETURN (x + Temp_y);
END "-";

--******************************************************************************
--******************************************************************************

PROCEDURE PUT(x : IN Fraction) IS
BEGIN
  IF (x.Positive = FALSE) THEN
    PUT('-');
  END IF;
  PUT(x.Numerator);
  IF (x.Denominator.Last_Significant_Digit /= First_Indice) OR ELSE
     (x.Denominator.Mantissa(First_Indice) /= 1) THEN
     -- IF (x.Denominator /= Big_Unsigned_One) THEN...
    PUT('/');
    PUT(x.Denominator);
  END IF;
END PUT;

--******************************************************************************
--******************************************************************************

PROCEDURE PUT_LINE(x : IN Fraction) IS
BEGIN
  PUT(x);
  NEW_LINE;
END PUT_LINE;

--******************************************************************************
--******************************************************************************

PROCEDURE PUT(File : IN FILE_TYPE; x : IN Fraction) IS
BEGIN
  IF (x.Positive = FALSE) THEN
    PUT(File, '-');
  END IF;
  PUT(File, x.Numerator);
  IF (x.Denominator.Last_Significant_Digit /= First_Indice) OR ELSE
     (x.Denominator.Mantissa(First_Indice) /= 1) THEN
     -- IF (x.Denominator /= Big_Unsigned_One) THEN...
    PUT(File, '/');
    PUT(File, x.Denominator);
  END IF;
END PUT;

--******************************************************************************
--******************************************************************************

PROCEDURE PUT_LINE(File : IN FILE_TYPE; x : IN Fraction) IS
BEGIN
  PUT(File, x);
  NEW_LINE(File);
END PUT_LINE;

--******************************************************************************
--******************************************************************************

FUNCTION UString2Fraction(x : IN UNBOUNDED_STRING) RETURN Fraction IS
  Numerator, Denominator : UNBOUNDED_STRING;
  Result : Fraction;
  Index : NATURAL;
BEGIN
  Index := Char_Count(x, '/');
  IF (Index = 0) THEN
    Result := CreateFraction(UString2Signed(x), Big_Signed_One);
    RETURN Result;
  ELSIF (Index = 1) THEN
    Index := First_Matching_Char_Position(x, '/');
    Numerator := Copy(x, 1, Index - 1);
    Denominator := Copy(x, Index + 1, Length(x)- Index);
    IF (LENGTH(Numerator) = 0) OR ELSE (LENGTH(Denominator) = 0) THEN
      RAISE Parameter_Error;
    END IF;
    Result := CreateFraction(UString2Signed(Numerator),
                             UString2Signed(Denominator));
    RETURN Result;
  ELSE
    RAISE Parameter_Error;
  END IF;
EXCEPTION
  WHEN OTHERS => RAISE Parameter_Error;
END UString2Fraction;

--******************************************************************************
--******************************************************************************

PROCEDURE GET_LINE(x : OUT Fraction) IS
  Temp_String : UNBOUNDED_STRING;
BEGIN
  GET_LINE(Temp_String);
  x := UString2Fraction(Temp_String);
END GET_LINE;

--******************************************************************************
--******************************************************************************

PROCEDURE GET_LINE(File : IN FILE_TYPE; x : OUT Fraction) IS
  Temp_String : UNBOUNDED_STRING;
BEGIN
  GET_LINE(File, Temp_String);
  x := UString2Fraction(Temp_String);
END GET_LINE;

--******************************************************************************
--******************************************************************************

FUNCTION "**"(x : IN Fraction; y : IN Big_Unsigned) RETURN Fraction IS
BEGIN
  IF (y.Last_Significant_Digit = First_Indice) THEN
    IF (y.Mantissa(First_Indice) = 0) THEN
      -- IF (y = Big_Unsigned_Zero) THEN ...
      RETURN (TRUE, Big_Unsigned_One, Big_Unsigned_One);     -- Returns 1/1.
    ELSIF (y.Mantissa(First_Indice) = 1) THEN
      RETURN x;
    END IF; 
  END IF;
  DECLARE
    Temp_Result : Fraction;
  BEGIN
    Temp_Result.Numerator := x.Numerator ** y;
    Temp_Result.Denominator := x.Denominator ** y;
    Temp_Result.Positive := x.Positive
                                   OR ELSE (y.Mantissa(First_Indice) MOD 2 = 0);
    -- Note : if x is positive, the result is always positive.
    -- If x is negative, the result is positive if the exponent is a multiple
    -- of 2.
    RETURN Temp_Result;
  END;
EXCEPTION
  WHEN OTHERS => RAISE Fraction_Overflow;
END "**";

--******************************************************************************
--******************************************************************************

FUNCTION "**"(x : IN Fraction; y : IN Big_Signed) RETURN Fraction IS
  Result : Fraction;
BEGIN
  Result := x ** Signed2Unsigned(y);
  IF (y.Positive = FALSE) THEN
    Result := Inverse_Fraction(Result);
  END IF;
  RETURN Result;
END "**";

--******************************************************************************
--******************************************************************************

FUNCTION "**"(x : IN Fraction; y : IN My_Type) RETURN Fraction IS
BEGIN
  IF (y = 0) THEN
    RETURN (TRUE, Big_Unsigned_One, Big_Unsigned_One);     -- Returns 1/1.
  ELSIF (y = 1) THEN
    RETURN x;
  END IF; 
  DECLARE
    Temp_Result : Fraction;
  BEGIN
    Temp_Result.Numerator := x.Numerator ** y;
    Temp_Result.Denominator := x.Denominator ** y;
    Temp_Result.Positive := x.Positive OR ELSE (y MOD 2 = 0);
    -- Note : if x is positive, the result is always positive.
    -- If x is negative, the result is positive if the exponent is a multiple
    -- of 2.
    RETURN Temp_Result;
  END;
EXCEPTION
  WHEN OTHERS => RAISE Fraction_Overflow;
END "**";

--******************************************************************************
--******************************************************************************

FUNCTION "**"(x : IN Fraction; y : IN Index_Type) RETURN Fraction IS
BEGIN
  IF (y = 0) THEN
    RETURN (TRUE, Big_Unsigned_One, Big_Unsigned_One);     -- Returns 1/1.
  ELSIF (y = 1) THEN
    RETURN x;
  END IF; 
  DECLARE
    Temp_Result : Fraction;
  BEGIN
    Temp_Result.Numerator := x.Numerator ** y;
    Temp_Result.Denominator := x.Denominator ** y;
    Temp_Result.Positive := x.Positive OR ELSE (y MOD 2 = 0);
    -- Note : if x is positive, the result is always positive.
    -- If x is negative, the result is positive if the exponent is a multiple
    -- of 2.
    RETURN Temp_Result;
  END;
EXCEPTION
  WHEN OTHERS => RAISE Fraction_Overflow;
END "**";

--******************************************************************************
--******************************************************************************

FUNCTION "*"(x : IN Fraction; y : IN Big_Unsigned) RETURN Fraction IS
BEGIN
  IF (y.Last_Significant_Digit = First_Indice) THEN
    IF (y.Mantissa(First_Indice) = 0) THEN
      RETURN (x.Positive, Big_Unsigned_Zero, Big_Unsigned_One);
    ELSIF (y.Mantissa(First_Indice) = 1) THEN
      RETURN x;
    END IF;
  END IF;
  DECLARE
    Temp : CONSTANT Big_Unsigned := GCF(y, x.Denominator);
  BEGIN
    RETURN (x.Positive, (y / Temp) * x.Numerator, x.Denominator / Temp);
  END;
EXCEPTION
  WHEN OTHERS => RAISE Fraction_Overflow;
END "*";

--******************************************************************************
--******************************************************************************

FUNCTION "*"(x : IN Fraction; y : IN Big_Signed) RETURN Fraction IS
  Result : Fraction;
  Temp_y : CONSTANT Big_Unsigned := Signed2Unsigned(y);
  Temp : CONSTANT Big_Unsigned := GCF(Temp_y, x.Denominator);
BEGIN
  IF (y.Last_Significant_Digit = First_Indice) THEN
    IF (y.Mantissa(First_Indice) = 0) THEN
      RETURN (x.Positive, Big_Unsigned_Zero, Big_Unsigned_One);
    ELSIF (y.Mantissa(First_Indice) = 1) THEN
      RETURN ((x.Positive = y.Positive), x.Numerator, x.Denominator);
    END IF;
  END IF;
  Result.Positive := (x.Positive = y.Positive);
  Result.Numerator := x.Numerator * (Temp_y / Temp);
  Result.Denominator := x.Denominator / Temp;
  RETURN Result;
EXCEPTION
  WHEN OTHERS => RAISE Fraction_Overflow;
END "*";

--******************************************************************************
--******************************************************************************

FUNCTION "/"(x : IN Fraction; y : IN Big_Unsigned) RETURN Fraction IS
  Temp : Big_Unsigned;
BEGIN
  IF (y.Last_Significant_Digit = First_Indice) AND THEN
     (y.Mantissa(First_Indice) = 0) THEN
    RAISE Division_By_Zero;
  END IF;
  Temp := GCF(x.Numerator, y);
  RETURN (x.Positive, x.Numerator / Temp, (y / Temp) * x.Denominator);
EXCEPTION
  WHEN OTHERS => RAISE Fraction_Overflow;
END "/";

--******************************************************************************
--******************************************************************************

FUNCTION "/"(x : IN Fraction; y : IN Big_Signed) RETURN Fraction IS
  Temp : Big_Unsigned;
  Temp_y : CONSTANT Big_Unsigned := Signed2Unsigned(y);
BEGIN
  IF (y.Last_Significant_Digit = First_Indice) AND THEN
     (y.Mantissa(First_Indice) = 0) THEN
    RAISE Division_By_Zero;
  END IF;
  Temp := GCF(x.Numerator, Temp_y);
  RETURN (x.Positive = y.Positive, x.Numerator / Temp,
          (Temp_y / Temp) * x.Denominator);
EXCEPTION
  WHEN OTHERS => RAISE Fraction_Overflow;
END "/";

--******************************************************************************
--******************************************************************************

FUNCTION "<"(Left, Right : IN Fraction) RETURN BOOLEAN IS
  --
  -- Be careful : do not convert the fraction to a real type. If so, there will
  -- be a truncation, so the result of this function could be mathematically
  -- false.
  --
BEGIN
  IF (Left.Positive /= Right.Positive) THEN
    -- If the 2 fractions have different signs, the comparison is easy...
    RETURN Right.Positive;
  --
  -- Now, we must study the case where the two fractions have the same sign.
  --
  -- First, some simple cases ...
  ELSIF (Left.Denominator = Right.Denominator) THEN
    IF (Left.Numerator = Right.Numerator) THEN
      RETURN FALSE;
    ELSE
      RETURN (Left.Positive = (Left.Numerator < Right.Numerator));
    END IF;
  END IF;
  -- The case (Left.Denominator = Right.Denominator)
  -- has already been studied.
  IF (Left.Numerator <= Right.Numerator) THEN
    RETURN NOT (Left.Positive = (Left.Denominator < Right.Denominator));
  ELSIF (Left.Numerator.Last_Significant_Digit = First_Indice) AND THEN
        (Left.Numerator.Mantissa(First_Indice) = 0) THEN
    -- Remark : the situation where Left.Numerator = Right.Numerator
    -- has already been studied, so Right.Numerator /= Big_Unsigned_Zero.
    -- Furthermore, at this stade, the 2 fractions have the same sign.
    -- That means that Left < Right, so the result is always TRUE.
    -- Note : by default, a fraction with a numerator equal to zero
    -- has a sign set to Positive.
    RETURN TRUE;
  END IF;
  --
  -- Now, when the two fractions have the same sign, we must study the
  -- general case (Left.Numerator /= Right.Numerator
  -- AND Left.Denominator /= Right.Denominator).
  --
  IF (Left.Numerator < Left.Denominator) AND THEN
     (Right.Numerator < Right.Denominator) THEN
    -- If Left < Right, then Inverse_Fraction(Right) < Inverse_Fraction(Left).
    RETURN (Inverse_Fraction(Right) < Inverse_Fraction(Left));
  END IF;
  --
  -- Now, at least one of the 2 fractions has an absolute value greater than 1.
  --
  DECLARE
    Temp_x : Fraction := Left;
    Temp_y : Fraction := Right;
    Tempx, Tempy, Temp : Big_Unsigned;
  BEGIN
    --
    -- If Tempx = Tempy, then we must study further the fractions Temp_x
    -- and Temp_y.
    --
    LOOP
      Big_Div(Temp_x.Numerator, Temp_x.Denominator, Tempx, Temp_x.Numerator);
      Big_Div(Temp_y.Numerator, Temp_y.Denominator, Tempy, Temp_y.Numerator);
      -- Don't forget that Tempx and Tempy are "only" big integers, which
      -- represent the truncated value of Left and Right, respectively.
      IF (Tempx /= Tempy) THEN
        RETURN (Temp_x.Positive = (Tempx < Tempy));
      END IF;
      Temp_x := Simplify_Fraction(Temp_x);
      -- Now, Temp_x < 1
      Temp_y := Simplify_Fraction(Temp_y);
      -- Now, Temp_y < 1
      IF (Temp_x.Numerator > Temp_y.Numerator) THEN
        Temp := Big_Unsigned_Last / Temp_x.Numerator;
      ELSE
        -- Temp_x.Numerator <= Temp_y.Numerator
        Temp := Big_Unsigned_Last / Temp_y.Numerator;
      END IF;
      -- Temp_x.Numerator <= Big_Unsigned_Last and
      -- Temp_y.Numerator <= Big_Unsigned_Last,
      -- so Temp >= Big_Unsigned_One.
      IF (Temp.Last_Significant_Digit > First_Indice) OR ELSE
         (Temp.Mantissa(First_Indice) > 1) THEN
        -- These two tests are the same test than (Temp > Big_Unsigned_One),
        -- but they are much faster.
        -- In this case, we can multiply Temp_x and Temp_y by Temp,
        -- without raising Big_Number_Overflow.
        -- Then, we can calculate Tempx and Tempy
        -- for a more precise comparison.
        Temp_x := Temp_x * Temp;
        Temp_y := Temp_y * Temp;
      ELSE
        -- Here, Tempx = Tempy, and Temp = Big_Unsigned_One.
        -- If Left < Right,
        -- then Inverse_Fraction(Left) > Inverse_Fraction(Right).
        RETURN (Inverse_Fraction(Right) < Inverse_Fraction(Left));
      END IF;
    END LOOP;
  END;     -- END of DECLARE
END "<";

--******************************************************************************
--******************************************************************************

FUNCTION ">"(Left, Right : IN Fraction) RETURN BOOLEAN IS
BEGIN
  -- If you want to know the result of Left > Right,
  -- just return the result of Right < Left...
  RETURN (Left < Right);
END ">";

--******************************************************************************
--******************************************************************************

FUNCTION "<="(Left, Right : IN Fraction) RETURN BOOLEAN IS
  -- Left <= Right is the same than (NOT (Left > Right)),
  -- so it's the same than (NOT (Right < Left)).
  Temp_Result : CONSTANT BOOLEAN := NOT (Right < Left);
BEGIN
  RETURN Temp_Result;
END "<=";

--******************************************************************************
--******************************************************************************

FUNCTION ">="(Left, Right : IN Fraction) RETURN BOOLEAN IS
  -- Left >= Right is the same than (NOT (Left < Right)).
  Temp_Result : CONSTANT BOOLEAN := NOT (Left < Right);
BEGIN
  RETURN Temp_Result;
END ">=";

--******************************************************************************
--******************************************************************************

FUNCTION Opposite(x : IN Fraction) RETURN Fraction IS
  Temp_Result : Fraction := x;
BEGIN
  IF (Temp_Result.Numerator.Last_Significant_Digit /= First_Indice) OR ELSE
     (Temp_Result.Numerator.Mantissa(First_Indice) /= 0) THEN
    -- If Temp_Result /= 0, then inverse the sign.
    -- Note that if x = 0, by convention, the sign of x is always
    -- set to positive.
    Temp_Result.Positive := NOT(Temp_Result.Positive);
  END IF;
  RETURN Temp_Result;
END Opposite;

--******************************************************************************
--******************************************************************************

FUNCTION Absolute(x : IN Fraction) RETURN Fraction IS
BEGIN
  IF (x.Positive = TRUE) THEN
    RETURN x;
  ELSE
    RETURN Opposite(x);
  END IF;
END Absolute;

--******************************************************************************
--******************************************************************************

FUNCTION Is_Null(x : IN Fraction) RETURN BOOLEAN IS
  Temp_Result : BOOLEAN;
BEGIN
  Temp_Result := (x.Numerator.Last_Significant_Digit = First_Indice) AND THEN
                 (x.Numerator.Mantissa(First_Indice) = 0);
  RETURN Temp_Result;
END Is_Null;

--******************************************************************************
--******************************************************************************

PROCEDURE Float_in_UString2Fraction(x : IN UNBOUNDED_STRING;
                                    Nb_Periods : IN POSITIVE;
                                    Result : OUT Fraction;
                                    Error : OUT BOOLEAN) IS

  FUNCTION Convert IS NEW Int_Number2Big_Unsigned(NATURAL);

  FUNCTION Find_a_Period(s : IN UNBOUNDED_STRING) RETURN UNBOUNDED_STRING IS
    -- This function looks for a period in the number represented by
    -- the string s.
    -- If a period is found, it is returned in a string.
    -- If no period is found, the returned result is the string s.
    --
    -- Here, the period returned is not the smaller.
    -- For example, in the number "1345345345345", the returned period is
    -- "345345", even if the smaller period is "345".
    --
    s1, s2 : UNBOUNDED_STRING;
    Size_s : CONSTANT POSITIVE := LENGTH(s);
    Temp_Size, Index1, Index2 : NATURAL;
  BEGIN
    IF (Size_s < 2) THEN
      RETURN s;
    ELSE
      Temp_Size := Size_s / 2;
      -- Begin to look for the greatest period which could be found
      LOOP
        Index1 := Size_s - 2 * Temp_Size + 1;
        Index2 := Size_s - Temp_Size + 1;
        s1 := Copy(s, Index1, Temp_Size);
        s2 := Copy(s, Index2, Temp_Size);
        Temp_Size := Temp_Size - 1;
        EXIT WHEN (s1 = s2) OR ELSE (Temp_Size = 0);
      END LOOP;

      IF (s1 /= s2) THEN
        RETURN s;
      ELSE
        RETURN s2;
      END IF;
    END IF;
  END Find_a_Period;

  ------------------------------------------------------------------------------

  FUNCTION Count_the_Number_of_Periods(s, period : IN UNBOUNDED_STRING)
                                       RETURN NATURAL IS
  BEGIN
    IF (s = period) THEN
      RETURN 1;
    END IF;
    DECLARE
      Index : NATURAL;
      Number : NATURAL := 1;
      String_Size : CONSTANT NATURAL := LENGTH(s);
      Period_Size : CONSTANT NATURAL := LENGTH(period);
      Max_Number : CONSTANT NATURAL := String_Size / Period_Size;
      Temp_String : UNBOUNDED_STRING;
    BEGIN
      LOOP
        EXIT WHEN (Number > Max_Number);
        Index := String_Size - Number * Period_Size + 1;
        Temp_String := Copy(s,Index,Period_Size);
        IF (Temp_String = period) THEN
          Number := Number + 1;
        ELSE
          EXIT;
        END IF;
      END LOOP;
      RETURN (Number - 1);
    END;
  END Count_the_Number_of_Periods;

  ------------------------------------------------------------------------------

  FUNCTION Find_the_Smallest_Period(s : IN UNBOUNDED_STRING)
                                    RETURN UNBOUNDED_STRING IS
    Max_Period, s1, s2, temp : UNBOUNDED_STRING;
    Size_Max_Period : POSITIVE;
    Index1, Index2 : NATURAL;
    Temp_Size : NATURAL := 1;
    Count : NATURAL;
    -- Be careful. Suppose that s = "1.852333852333852333".
    -- "852333" is the real period, not "3".
  BEGIN
    Max_Period := Find_a_Period(s);
    IF (s = Max_Period) THEN
      RETURN s;
    ELSE
      Size_Max_Period := LENGTH(Max_Period);
      -- It is certain that Max_Period is a period.
      -- Try to find a smaller period (s2).
      LOOP
        Index1 := Size_Max_Period - 2 * Temp_Size + 1;
        Index2 := Size_Max_Period - Temp_Size + 1;
        s1 := Copy(Max_Period, Index1, Temp_Size);
        s2 := Copy(Max_Period, Index2, Temp_Size);
        IF (s1 = s2) THEN
          -- Verify that Max_Period is a repetition of s1.
          Count := Count_the_Number_of_Periods(Max_Period, s1);
          Temp := NULL_UNBOUNDED_STRING;
          FOR I IN 1..Count LOOP
            Temp := Temp & s1;
          END LOOP;
          -- If Max_Period is a repetition of s1, Temp should be equal to Max_Period.
          IF (Max_Period = Temp) THEN
            RETURN s1;
          END IF;
        END IF;
        IF (Temp_Size < Size_Max_Period / 2) THEN
          -- There must be at least 2 small periods in Max_Period,
          -- so the size of this small period is smaller or equal
          -- to Size_Max_Period / 2.
          Temp_Size := Temp_Size + 1;
        ELSE
          RETURN Max_Period;
        END IF;
      END LOOP;
    END IF;
  END Find_the_Smallest_Period;

  ------------------------------------------------------------------------------

  FUNCTION Prepare_String(Str : IN UNBOUNDED_STRING) RETURN UNBOUNDED_STRING IS
    Temp_Result : UNBOUNDED_STRING := Del_Spaces(Str);
  BEGIN
    Temp_Result := Char_Replace(Temp_Result, 'e', 'E');
    Temp_Result := Char_Replace(Temp_Result, ',', '.');
    RETURN Temp_Result;
  END Prepare_String;

  ------------------------------------------------------------------------------

  FUNCTION Numeric_String(Str : IN UNBOUNDED_STRING) RETURN BOOLEAN IS
    I : NATURAL := 1;
    Temp_Str : CONSTANT UNBOUNDED_STRING := Del_Character(Str, '.');
    String_Length : CONSTANT NATURAL := LENGTH(Temp_Str);
    -- The number of '.' character has been checked in
    -- the procedure "Verify_String_And_Separe_Exponent_And_Mantissa",
    -- so we don't have to check it now (we are sure that there is 0
    -- or 1 '.' in this string).
  BEGIN
    LOOP
      EXIT WHEN (I > String_Length);
      IF (ELEMENT(Temp_Str, I) NOT IN '0'..'9') THEN
        RETURN FALSE;
      END IF;
      I := I + 1;
    END LOOP;
    RETURN TRUE;
  END Numeric_String;

  ------------------------------------------------------------------------------

  FUNCTION Trim_Leading_Zero(Str : IN UNBOUNDED_STRING) RETURN UNBOUNDED_STRING IS
    Temp_String : UNBOUNDED_STRING := NULL_UNBOUNDED_STRING;
    I : NATURAL := 0;
    Length_Str : NATURAL := LENGTH(Str);
    Char : CHARACTER;
  BEGIN
    IF (Length_Str = 0) THEN
      RETURN Temp_String;
    END IF;
    LOOP
      I := I + 1;
      EXIT WHEN (I > Length_Str);
      Char := ELEMENT(Str, I);
      EXIT WHEN (Char /= '0');
    END LOOP;
    Temp_String := Copy(Str, I, Length_Str - I + 1);
    IF (ELEMENT(Temp_String, 1) = '.') THEN
      Temp_String := '0' & Temp_String;
    END IF;
    I := First_Matching_Char_Position(Str, '.');
    IF (I /= 0) THEN
      Length_Str := LENGTH(Temp_String);
      LOOP
        Char := ELEMENT(Str, Length_Str);
        IF (Char = '0') OR ELSE (Char = '.') THEN
          Length_Str := Length_Str - 1;
          Temp_String := Copy(Temp_String, 1, Length_Str);
        ELSE
          EXIT;
        END IF;
      END LOOP;
    END IF;
    RETURN Temp_String;
  END Trim_Leading_Zero;

  ------------------------------------------------------------------------------

  PROCEDURE Verify_String_And_Separe_Exponent_And_Mantissa
                             (Str : IN UNBOUNDED_STRING;
                              Exponent_Str, Mantissa_Str : OUT UNBOUNDED_STRING;
                              Positive, Valid : OUT BOOLEAN) IS
    -- Verify if Str is a valid string for conversion in a fraction.
    Mantissa, Exponent : UNBOUNDED_STRING := NULL_UNBOUNDED_STRING;
    Count_E : CONSTANT NATURAL := Char_Count(Str, 'E');
    I : NATURAL;
    Temp_Valid : BOOLEAN;
  BEGIN
    Mantissa_Str := Mantissa;
    Exponent_Str := Exponent;
    IF (Count_E > 1) THEN
      Valid := FALSE;
      RETURN;
    END IF;
    IF (Count_E = 1) THEN
      I := First_Matching_Char_Position(Str, 'E');
      Mantissa := Copy(Str, 1, I-1);
      Exponent := Copy(Str, I+1, LENGTH(Str) - I);
    ELSE
      Mantissa := Str;
    END IF;
    Mantissa := Trim_Leading_Zero(Mantissa);
    Exponent := Trim_Leading_Zero(Exponent);
    -- You can't have more than one '.', or more than 1 sign (+ or -)
    -- in the mantissa.
    IF         (Char_Count(Mantissa, '.') > 1)
       OR ELSE (Char_Count(Mantissa, '+') + Char_Count(Mantissa, '-') > 1) THEN
      Valid := FALSE;
      RETURN;
    END IF;
    -- You can't have more than 1 sign (+ or -) in the exponent, where
    -- no '.' is authorized.
    IF         (Char_Count(Exponent, '.') > 0)
       OR ELSE (Char_Count(Exponent, '+') + Char_Count(Exponent, '-') > 1) THEN
      Valid := FALSE;
      RETURN;
    END IF;
    IF (Char_Count(Mantissa, '-') = 1) THEN
      Positive := FALSE;
    ELSE
      Positive := TRUE;
    END IF;
    Mantissa := Del_Character(Mantissa, '-');
    Mantissa := Del_Character(Mantissa, '+');
    Exponent := Del_Character(Exponent, '-');
    Exponent := Del_Character(Exponent, '+');
    Temp_Valid := (Numeric_String(Mantissa) AND THEN Numeric_String(Exponent));
    IF (Temp_Valid = TRUE) THEN
      Mantissa_Str := Mantissa;
      Exponent_Str := Exponent;
    END IF;
    Valid := Temp_Valid;
  END Verify_String_And_Separe_Exponent_And_Mantissa;

  ------------------------------------------------------------------------------

  FUNCTION Calc_Value(x : IN UNBOUNDED_STRING) RETURN Big_Unsigned IS
    Temp_Result : Big_Unsigned := Big_Unsigned_Zero;
    Ten : CONSTANT Big_Unsigned := Convert(10);
  BEGIN
    FOR I IN 1..LENGTH(x) LOOP
      Temp_Result := Temp_Result * Ten
                    + Convert(CHARACTER'POS(ELEMENT(x,I)) - CHARACTER'POS('0'));
    END LOOP;
    RETURN Temp_Result;
  END Calc_Value;
  
  ------------------------------------------------------------------------------

  PROCEDURE Calculate_The_Exponent(Exponent_String : IN UNBOUNDED_STRING;
                                   Mantissa_String : IN OUT UNBOUNDED_STRING;
                                   Exponent_Positive : OUT BOOLEAN;
                                   Exponent_Absolute_Value : OUT Big_Unsigned) IS
    I : NATURAL;
    Temp_Exponent_Positive : BOOLEAN := TRUE;
    Temp_Unsigned, Temp_Exponent : Big_Unsigned;
    Exponent_Str : UNBOUNDED_STRING := Exponent_String;
  BEGIN
    -- Look for the sign and the value of the exponent
    --
    IF (LENGTH(Exponent_Str) /= 0) AND THEN (ELEMENT(Exponent_Str, 1) = '-') THEN
      Temp_Exponent_Positive := FALSE;
      Exponent_Str := Copy(Exponent_Str, 2, LENGTH(Exponent_Str) - 1);
    END IF;
    Temp_Exponent := Calc_Value(Exponent_Str);
    --
    -- If there is a "." in the mantissa, check its position and modify
    -- the exponent, because the '.' will be removed.
    --
    I := First_Matching_Char_Position(Mantissa_String, '.');
    IF (I /= 0) THEN                       -- There is a '.' in Mantissa_String
      Temp_Unsigned := Convert(LENGTH(Mantissa_String) - I);
      IF (Temp_Exponent_Positive = FALSE) THEN
        Temp_Exponent := Temp_Exponent + Temp_Unsigned;
      ELSE
        IF (Temp_Unsigned > Temp_Exponent) THEN
          Temp_Exponent_Positive := FALSE;
          Temp_Exponent := Temp_Unsigned - Temp_Exponent;
        ELSE
          Temp_Exponent := Temp_Exponent - Temp_Unsigned;
        END IF;
      END IF;
      --
      -- Now, the exponent is computed.
      -- The "." can be deleted in the mantissa.
      --
      Mantissa_String := Copy(Mantissa_String, 1, I-1) &
                         Copy(Mantissa_String, I+1, LENGTH(Mantissa_String) -I);
    END IF;
    Exponent_Positive := Temp_Exponent_Positive;
    Exponent_Absolute_Value := Temp_Exponent;
  END Calculate_The_Exponent;

  ------------------------------------------------------------------------------

BEGIN
  DECLARE
    -- In procedures which require a lot of lines of code, I prefer to declare
    -- the variables only when it is necessary. This method eliminates a cause
    -- of side effects during the sub-procedures.
    Temp_Str : UNBOUNDED_STRING := Prepare_String(x);
    I : NATURAL;
    Temp_Unsigned, Exponent : Big_Unsigned;
    Temp, Exponent_Str, Mantissa_Str, Period : UNBOUNDED_STRING;
    Valid, Exponent_Positive : BOOLEAN;
    Temp_Result : Fraction;
  BEGIN
    -- If Nb_Periods < 2 or if the string is not valid, then exit.
    -- Remark : the function
    Verify_String_And_Separe_Exponent_And_Mantissa(Temp_Str, Exponent_Str,
                                                   Mantissa_Str,
                                                   Temp_Result.Positive, Valid);
    ------------------------------------------
    -- In case the parameters are not valid --
    ------------------------------------------
    IF (Valid = FALSE) THEN
      Error := TRUE;
      -- Set Result to an impossible value (0/0),
      -- so, in case of error, and if the parameter Error is not checked
      -- after a call to this procedure, the program will hang.
      Result.Numerator := Big_Unsigned_Zero;
      Result.Denominator := Big_Unsigned_Zero;
      Result.Positive := TRUE;
      RETURN;
    ELSE
      Error := FALSE;
    END IF;

    -- Now, Error = FALSE, and there is no situation that could change it.
    -- Note that the sign of the result has already been set.
    ---------------------------------
    -- Computation of the exponent --
    ---------------------------------
    Calculate_The_Exponent(Exponent_Str, Mantissa_Str,
                           Exponent_Positive, Exponent);
    -------------------------------------------------------------
    -- Now, search for the period                              --
    -- and calculate the number of repetitions of this period. --
    -------------------------------------------------------------
    Period := Find_the_Smallest_Period(Mantissa_Str);
    I := Count_the_Number_of_Periods(Mantissa_Str, Period);
    -----------------------------------------------  
    -- Now, we can really calculate the fraction --
    -----------------------------------------------  
    IF (Period = NULL_UNBOUNDED_STRING) OR ELSE (I < Nb_Periods) THEN
      Temp_Result.Numerator := Calc_Value(Mantissa_Str);
      Temp_Result.Denominator := Convert(1);
    ELSE
      Temp := Copy(Mantissa_Str, 1, LENGTH(Mantissa_Str) - I * LENGTH(Period));
      Temp_Result.Numerator := Calc_Value(Temp & Period) - Calc_Value(Temp);
      Temp := NULL_UNBOUNDED_STRING;
      FOR J IN 1..LENGTH(Period) LOOP
        Temp := Temp & "9";
      END LOOP;
      Temp_Result.Denominator := Calc_Value(Temp);
      Temp_Result := Simplify_Fraction(Temp_Result);
      --
      -- We must modify the exponent :
      --   - the length of the numerator has (I-1)*LENGTH(Period) digits less;
      --   - the length of the denominator has (LENGTH(Period)) digits more
      -- So we have to add I*LENGTH(Period) to the exponent.
      --
      Temp_Unsigned := Convert(I) * Convert(LENGTH(Period));
      IF (Exponent_Positive = TRUE) THEN
        Exponent := Exponent + Temp_Unsigned;
      ELSE
        IF (Temp_Unsigned > Exponent) THEN
          Exponent_Positive := TRUE;
          Exponent := Temp_Unsigned - Exponent;
        ELSE
          Exponent := Exponent - Temp_Unsigned;
        END IF;
      END IF;
    END IF;
    --
    -- If Exponent /= 0 ...
    --
    IF (Exponent.Last_Significant_Digit /= First_Indice)
       OR ELSE (Exponent.Mantissa(First_Indice) /= 0) THEN
       -- This combination of tests is equivalent to the
       -- test (Exponent /= Big_Unsigned_Zero), but it is
       -- executed faster.
      Exponent := Convert(10) ** Exponent;
      IF (Exponent_Positive = TRUE) THEN
        Temp_Result.Numerator := Temp_Result.Numerator * Exponent;
      ELSE
        Temp_Result.Denominator := Temp_Result.Denominator * Exponent;
      END IF;
      -- The fraction has been modified, so enventually it could be simplified.
      Temp_Result := Simplify_Fraction(Temp_Result);
    END IF;
    Result := Temp_Result;
  END;                   -- END DECLARE
END Float_in_UString2Fraction;

--******************************************************************************
--******************************************************************************

END Big_Fraction;
--
-- Permission to use, copy, modify, and distribute this software and its
-- documentation for any purpose and without fee is hereby granted,
-- provided that the above copyright and authorship notice appear in all
-- copies and that both that copyright notice and this permission notice
-- appear in supporting documentation.
-- 
-- The ARA makes no representations about the suitability of this software
-- for any purpose.  It is provided "as is" without express
-- or implied warranty.
-- 
--******************************************************************************
--                                                                            **
--  File:        BNBIUNRA.ADB                                                 **
--  Description: Big_Number.Big_Unsigned_RANDOM package body                  **
--  Revision:    0.28 (beta version)                                          **
--  Date:        July 8, 2001                                                 **
--  Author:      Jrme Delcourt                                              **
--  Mail:        sikander@club-internet.fr                                    **
--                                                                            **
--  Copyright (c) Jrme Delcourt, 1998, 1999, 2000, 2001                     **
--  62, rue Ngrier                                                           **
--  59800 Lille                                                               **
--  FRANCE                                                                    **
--                                                                            **
--  Permission granted to use for any purpose, provided this copyright        **
--  remains attached and unmodified.                                          **
--                                                                            **
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR            **
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED            **
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       **
--                                                                            **
--******************************************************************************

SEPARATE(Big_Number)

PACKAGE BODY Big_Unsigned_Random IS


-- In this package, there are in fact two separate pseudo-random number
-- generators (PRNG) : one for index, one for digits. They are combined
-- to get a PRNG for Big_Unsigned : so, the last significant digit
-- of the Big_Unsigned is set independantly of the digits.

My_Type_Random_Seed : Largest_Unsigned;
Index_Random_Seed : Largest_Unsigned;

A : CONSTANT Largest_Unsigned := 125;      -- 125 = 5^3
B : CONSTANT Largest_Unsigned := Largest_Unsigned'LAST / A;
  -- Be careful : B must be greater or equal to Nb_Item and Base.

PROCEDURE Randomize IS
  t : TIME;
  Year : YEAR_NUMBER;
  Month : MONTH_NUMBER;    -- 1 to 12
  Day : DAY_NUMBER;        -- 0 to 31
  Seconds : DAY_DURATION;  -- 0 to 86399
BEGIN
  t := CLOCK;
  SPLIT(t, Year, Month, Day, Seconds);
  My_Type_Random_Seed := Largest_Unsigned(Seconds) + Largest_Unsigned(Day)
                             + Largest_Unsigned(Month) + Largest_Unsigned(Year);
  Index_Random_Seed := 2 * (2 * Largest_Unsigned(Seconds) +
                            (Largest_Unsigned(Day) MOD Largest_Unsigned(Month)))
                       + 1;
  IF ((My_Type_Random_Seed MOD 2) = 0) THEN
    My_Type_Random_Seed := My_Type_Random_Seed + 1;
  END IF;
END Randomize;


FUNCTION Random(x : IN Index_Type) RETURN Index_Type IS
  Max : CONSTANT Largest_Unsigned := Largest_Unsigned(x);
BEGIN
  Index_Random_Seed := (A * Index_Random_Seed) MOD B;
  IF (Max <= Largest_Unsigned'LAST / 4) THEN
    RETURN Index_Type(Index_Random_Seed MOD (Max+1));
  ELSE
    LOOP
      Index_Random_Seed := (A * Index_Random_Seed) MOD B;
      EXIT WHEN (Index_Random_Seed <= Max);
    END LOOP;
    RETURN Index_Type(Index_Random_Seed);
  END IF;
END Random;


FUNCTION Random RETURN My_Type IS
BEGIN
  My_Type_Random_Seed := (A * My_Type_Random_Seed) MOD B;
  RETURN My_Type(My_Type_Random_Seed AND Largest_Unsigned(Greatest_Digit));
    -- Returns My_Type_Random_Seed MOD Base
END Random;


FUNCTION Random(x : IN My_Type) RETURN My_Type IS
  Result : My_Type;
BEGIN
  IF (x = 0) THEN
    RETURN 0;
  ELSIF (x < 2**(My_Type_Size / 4)) THEN
    Result := Random;
    RETURN (Result MOD (x+1));
  END IF;
  LOOP
    Result := Random;
    IF (Result <= x) THEN
      EXIT;
    END IF;
  END LOOP;
  RETURN Result;
END Random;


FUNCTION Random(x : IN Index_Type) RETURN Big_Unsigned IS
  Temp_Result : Big_Unsigned;
  Temp : Index_Type := x;
BEGIN
  FOR I IN First_Indice..Temp LOOP
    Temp_Result.Mantissa(I) := Random;
  END LOOP;
  FOR I IN (Temp + 1)..Last_Indice LOOP
    Temp_Result.Mantissa(I) := 0;
  END LOOP;
  WHILE (Temp > First_Indice) AND THEN (Temp_Result.Mantissa(Temp) = 0) LOOP
    Temp := Temp - 1;
  END LOOP;
  Temp_Result.Last_Significant_Digit := Temp;
  RETURN Temp_Result;
END Random;


FUNCTION Random RETURN Big_Unsigned IS
  Temp : Index_Type := Random(Last_Indice);
BEGIN
  RETURN Random(Temp);
END Random;


FUNCTION Random(x : IN Big_Unsigned) RETURN Big_Unsigned IS
  Temp_Unsigned : Big_Unsigned;
  Temp_Index : Index_Type;
  Temp_Boolean : BOOLEAN := TRUE;
  Temp_MyType1, Temp_MyType2 : My_Type;
  Last_x : CONSTANT Index_Type := x.Last_Significant_Digit;
  -- Be careful, we can't use the following algorithm :
  --
  -- IF (x = Big_Unsigned_Zero) THEN
  --   RETURN Big_Unsigned_Zero;
  -- END IF;
  -- Temp : Big_Unsigned := Random;
  -- IF (x = Big_Unsigned_Last) THEN
  --   RETURN Temp;
  -- END IF;
  -- RETURN Temp MOD (x + Big_Unsigned_One)
  --
  -- If we do so, and if (x > Big_Unsigned_Last), the distribution won't be
  -- uniform, because all the numbers between 0 and (Big_Unsigned_Last - x)
  -- will have a greater probability than numbers between
  -- (Big_Unsigned_Last - x) and x.
  --
  -- So, we choose a random number for the index, between First_Indice and
  -- x.Last_Significant_Digit. Then, we loop in order to have a random number
  -- below or equal to x.
  --
  -- This algorithm is slower, but there is a better uniform distribution.
  --
BEGIN
  IF (x.Last_Significant_Digit = First_Indice) AND THEN
     (x.Mantissa(First_Indice) = 0) THEN
     -- If x = 0...
    RETURN Big_Unsigned_Zero;
  END IF;
  IF (Last_x = First_Indice) THEN
    Temp_Index := First_Indice;
  ELSE
    Temp_Index := Random(Last_x);
  END IF;
  IF (Temp_Index < Last_x) THEN
    Temp_Unsigned := Random(Temp_Index);
  ELSE
    -- Temp_Index = Last_x
    WHILE (Temp_Boolean = TRUE) LOOP
      Temp_MyType1 := x.Mantissa(Temp_Index);
      Temp_MyType2 := Random(Temp_MyType1);
      Temp_Unsigned.Mantissa(Temp_Index) := Temp_MyType2;
      IF (Temp_Index = First_Indice) THEN
        EXIT;
      END IF;
      Temp_Boolean := (Temp_MyType1 = Temp_MyType2);
      Temp_Index := Temp_Index - 1;
    END LOOP;
    IF (Temp_Index > First_Indice) THEN
      Temp_Unsigned := Temp_Unsigned + Random(Temp_Index - 1);
    END IF;
    Temp_Index := Last_x;
    WHILE (Temp_Index > First_Indice) AND THEN
          (Temp_Unsigned.Mantissa(Temp_Index) = 0) LOOP
      Temp_Index := Temp_Index - 1;
    END LOOP;
    Temp_Unsigned.Last_Significant_Digit := Temp_Index;
  END IF;
  RETURN Temp_Unsigned;
END Random;


FUNCTION Random(x, y : IN Big_Unsigned) RETURN Big_Unsigned IS
  Mini : CONSTANT Big_Unsigned := Min(x, y);
  Diff : CONSTANT Big_Unsigned := Max(x, y) - Mini;
BEGIN
  RETURN Mini + Random(Diff);
END Random;


BEGIN
  IF (Largest_Unsigned(Index_Type'LAST) >= B) THEN
    RAISE Instanciation_Error;
  END IF;
  Randomize;
END Big_Unsigned_Random;
--******************************************************************************
--******************************************************************************

--
-- Copyright (C) 1996 Ada Resource Association (ARA), Columbus, Ohio.
-- Author: David A. Wheeler
--
-- The function COPY and all the character related functions have been
-- added by Jerome DELCOURT.
--

WITH TEXT_IO, ADA.STRINGS.UNBOUNDED;
USE TEXT_IO, ADA.STRINGS.UNBOUNDED;

PACKAGE UStrings IS

  -- This package provides a simpler way to work with type
  -- Unbounded_String, since this type will be used very often.
  -- Most users will want to ALSO with "ADA.STRINGS.UNBOUNDED".
  -- Ideally this would be a child package of "ADA.STRINGS.UNBOUNDED".
  --

  -- This package provides the following simplifications:
  --  + Shortens the type name from "UNBOUNDED_STRING" to "Ustring".
  --  + Creates shorter function names for To_Unbounded_String, i.e.
  --    To_Ustring(U) and U(S).  "U" is not a very readable name, but
  --    it's such a common operation that a short name seems appropriate
  --    (this function is needed every time a String constant is used).
  --    It also creates S(U) as the reverse of U(S).
  --  + Adds other subprograms, currently just "Swap".
  --  + Other packages can use this package to provide other simplifications.

  SUBTYPE Ustring IS UNBOUNDED_STRING;

  FUNCTION To_Ustring(Source : IN STRING) RETURN UNBOUNDED_STRING
                                          RENAMES To_Unbounded_String;
  FUNCTION U(Source : IN STRING) RETURN UNBOUNDED_STRING
                                 RENAMES To_Unbounded_String;
  FUNCTION S(Source : IN UNBOUNDED_STRING) RETURN STRING
                                           RENAMES To_String;

  -- "Swap" is important for reuse in some other packages, so we'll define it.

  PROCEDURE Swap(Left, Right : IN OUT UNBOUNDED_STRING);


  FUNCTION Empty(S : UNBOUNDED_STRING) RETURN BOOLEAN;
   -- returns True if Length(S)=0.
  PRAGMA INLINE(Empty);

  --*****************
  --* I/O Routines. *
  --*****************

  PROCEDURE Get_Line(File : IN File_Type; Item : OUT UNBOUNDED_STRING);
  PROCEDURE Get_Line(Item : OUT UNBOUNDED_STRING);

  PROCEDURE Put(File : IN File_Type; Item : IN UNBOUNDED_STRING);
  PROCEDURE Put(Item : IN UNBOUNDED_STRING);

  PROCEDURE Put_Line(File : IN File_Type; Item : IN UNBOUNDED_STRING);
  PROCEDURE Put_Line(Item : IN UNBOUNDED_STRING);


  FUNCTION Copy(s : IN UNBOUNDED_STRING;
                Index : IN POSITIVE;
                Size : IN NATURAL)
                RETURN UNBOUNDED_STRING;
  -- This function returns the unbounded string contained in s, which starts
  -- at the character number Index, and which has Size character. If the string
  -- is not long enough, the result may have less the Size character.
  -- If s is empty, or if Index if greater than the size of the string s,
  -- the result is an empty string.

  -- *******************************
  -- * Character related functions *
  -- *******************************

  FUNCTION Del_Spaces(Item : IN UNBOUNDED_STRING) RETURN UNBOUNDED_STRING;
    -- Removes all the spaces from the strings.
  FUNCTION Del_Character(Item : IN UNBOUNDED_STRING;
                         Char : IN CHARACTER)
                         RETURN UNBOUNDED_STRING;
    -- Removes all the occurences of Char from the string.

  FUNCTION Char_Count(S : IN UNBOUNDED_STRING;
                      Char : IN CHARACTER)
                      RETURN NATURAL;
    -- Counts the number of Char in this string.

  FUNCTION First_Matching_Char_Position(S : IN UNBOUNDED_STRING;
                                        Char : IN CHARACTER)
                                        RETURN NATURAL;
    -- Finds the first occurence of the character in the string.
    -- Returns 0 if none found.

  FUNCTION Last_Matching_Char_Position(S : IN UNBOUNDED_STRING;
                                       Char : IN CHARACTER)
                                       RETURN NATURAL;
    -- Finds the last occurence of the character in the string.
    -- Returns 0 if none found.

  FUNCTION Char_Replace(S : IN UNBOUNDED_STRING;
                        Char_to_be_replaced, New_char : IN CHARACTER)
                        RETURN UNBOUNDED_STRING;
    -- Counts the number of Char in this string.

END UStrings;
--
-- Permission to use, copy, modify, and distribute this software and its
-- documentation for any purpose and without fee is hereby granted,
-- provided that the above copyright and authorship notice appear in all
-- copies and that both that copyright notice and this permission notice
-- appear in supporting documentation.
-- 
-- The ARA makes no representations about the suitability of this software
-- for any purpose.  It is provided "as is" without express
-- or implied warranty.
-- 
--******************************************************************************
--******************************************************************************

--
-- Copyright (C) 1996 Ada Resource Association (ARA), Columbus, Ohio.
-- Author: David A. Wheeler
--

PACKAGE BODY UStrings IS

  Input_Line_Buffer_Length : CONSTANT := 1024;
    -- If an input line is longer, Get_Line will recurse to read in the line.


PROCEDURE Swap(Left, Right : IN OUT UNBOUNDED_STRING) IS
  -- Implement Swap.  This is the portable but slow approach.
  Temporary : CONSTANT UNBOUNDED_STRING := Left;
BEGIN
  Left := Right;
  Right := Temporary;
END Swap;


FUNCTION Empty(S : IN UNBOUNDED_STRING) RETURN BOOLEAN IS
  -- returns True if Length(S)=0.
BEGIN
 RETURN (LENGTH(S) = 0);
END Empty;
 

------------------------------------------------------------------------
-- Implement UNBOUNDED_STRING I/O by calling Text_IO String routines. --
------------------------------------------------------------------------


  -- Get_Line gets a line of text, limited only by the maximum number of
  -- characters in an UNBOUNDED_STRING.  It reads characters into a buffer
  -- and if that isn't enough, recurses to read the rest.

PROCEDURE Get_Line(File : IN File_Type; Item : OUT UNBOUNDED_STRING) IS

  FUNCTION More_Input RETURN UNBOUNDED_STRING IS
    Input : STRING(1..Input_Line_Buffer_Length);
    Last  : NATURAL;
  BEGIN
    GET_LINE(File, Input, Last);
    IF (Last < Input'LAST) THEN
      RETURN To_Unbounded_String(Input(1..Last));
    ELSE
      RETURN To_Unbounded_String(Input(1..Last)) & More_Input;
    END IF;
  END More_Input;

BEGIN
  Item := More_Input;
END Get_Line;


PROCEDURE Get_Line(Item : OUT UNBOUNDED_STRING) IS
BEGIN
  Get_Line(Current_Input, Item);
END Get_Line;


PROCEDURE Put(File : IN File_Type; Item : IN UNBOUNDED_STRING) IS
BEGIN
  Put(File, To_String(Item));
END Put;


PROCEDURE Put(Item : IN UNBOUNDED_STRING) IS
BEGIN
  PUT(Current_Output, To_String(Item));
END Put;


PROCEDURE Put_Line(File : IN File_Type; Item : IN UNBOUNDED_STRING) IS
BEGIN
  PUT(File, Item);
  NEW_LINE(File);
END Put_Line;


PROCEDURE Put_Line(Item : IN UNBOUNDED_STRING) IS
BEGIN
  PUT(Current_Output, Item);
  NEW_LINE;
END Put_Line;


--*******************
--* Other functions *
--*******************

FUNCTION Copy(s : IN UNBOUNDED_STRING;
              Index : IN POSITIVE;
              Size : IN NATURAL)
              RETURN UNBOUNDED_STRING IS
  Temp_Result : UNBOUNDED_STRING := NULL_UNBOUNDED_STRING;
  Source_Length : CONSTANT NATURAL := LENGTH(s);
BEGIN
  IF (Size = 0) THEN
    RETURN Temp_Result;
  END IF;
  DECLARE
    Temp_Index : POSITIVE := Index;
    Result_Size : NATURAL := 0;
    Number : POSITIVE;
  BEGIN
    IF (POSITIVE(Size) + Index > Source_Length) THEN
      Number := Source_Length - Index + 1;
    ELSE
      Number := POSITIVE(Size);
    END IF;
    FOR I IN 1..Number LOOP
      Temp_Result := Temp_Result & Element(S, Temp_Index);
      Temp_Index := Temp_Index + 1;
    END LOOP;
    RETURN Temp_Result;
  END;
END Copy;

--*******************************
--* Character related functions *
--*******************************

FUNCTION Del_Spaces(Item : IN UNBOUNDED_STRING) RETURN UNBOUNDED_STRING IS
  Result : UNBOUNDED_STRING := Null_Unbounded_String;
  Item_Length : CONSTANT NATURAL := LENGTH(Item);
  Temp_Char : CHARACTER;
BEGIN
  FOR I IN 1..Item_Length LOOP
    Temp_Char := ELEMENT(Item, I);
    IF (Temp_Char /= ' ') THEN
      Result := Result & Temp_Char;
    END IF;
  END LOOP;
  RETURN Result;
END Del_Spaces;


FUNCTION Del_Character(Item : IN UNBOUNDED_STRING;
                       Char : IN CHARACTER)
                      RETURN UNBOUNDED_STRING IS
  Result : UNBOUNDED_STRING := Null_Unbounded_String;
  Item_Length : CONSTANT NATURAL := LENGTH(Item);
  Temp_Char : CHARACTER;
BEGIN
  FOR I IN 1..Item_Length LOOP
    Temp_Char := ELEMENT(Item, I);
    IF (Temp_Char /= Char) THEN
      Result := Result & Temp_Char;
    END IF;
  END LOOP;
  RETURN Result;
END Del_Character;


FUNCTION Char_Count(S : IN UNBOUNDED_STRING;
                    Char : IN CHARACTER)
                    RETURN NATURAL IS
  Temp_Result : NATURAL := 0;
  String_Length : CONSTANT NATURAL := LENGTH(S);
BEGIN
  FOR I IN 1..String_Length LOOP
    IF (ELEMENT(S, I) = Char) THEN
      Temp_Result := Temp_Result + 1;
    END IF;
  END LOOP;
  RETURN Temp_Result;
END Char_Count;


FUNCTION First_Matching_Char_Position(S : IN UNBOUNDED_STRING;
                                      Char : IN CHARACTER)
                                      RETURN NATURAL IS
  String_Length : CONSTANT NATURAL := LENGTH(S);
  I : NATURAL := 1;
BEGIN
  LOOP
    EXIT WHEN (I > String_Length);
    IF (ELEMENT(S, I) = Char) THEN
      RETURN I;
    END IF;
    I := I + 1;
  END LOOP;
  RETURN 0;
END First_Matching_Char_Position;


FUNCTION Last_Matching_Char_Position(S : IN UNBOUNDED_STRING;
                                     Char : IN CHARACTER)
                                     RETURN NATURAL IS
  String_Length : CONSTANT NATURAL := LENGTH(S);
  I : NATURAL := String_Length;
BEGIN
  LOOP
    EXIT WHEN (I = 0);
    IF (ELEMENT(S, I) = Char) THEN
      RETURN I;
    END IF;
    I := I - 1;
  END LOOP;
  RETURN 0;
END Last_Matching_Char_Position;


FUNCTION Char_Replace(S : IN UNBOUNDED_STRING;
                      Char_to_be_replaced, New_Char : IN CHARACTER)
                      RETURN UNBOUNDED_STRING IS
  Temp_Result : UNBOUNDED_STRING := NULL_UNBOUNDED_STRING;
  Temp : UNBOUNDED_STRING := S;
  I : NATURAL;
BEGIN
  I := First_Matching_Char_Position(S, Char_to_be_replaced);
  WHILE (I /= 0) LOOP
    Temp_Result := Copy(Temp, 1, I-1) & New_Char;
    Temp := Copy(Temp, I + 1, LENGTH(Temp) - I);
    I := First_Matching_Char_Position(Temp, Char_to_be_replaced);
  END LOOP;
  Temp_Result := Temp_Result & Temp;
  RETURN Temp_Result;
END Char_Replace;

--******************************************************************************
--******************************************************************************

END Ustrings;

--
-- Permission to use, copy, modify, and distribute this software and its
-- documentation for any purpose and without fee is hereby granted,
-- provided that the above copyright and authorship notice appear in all
-- copies and that both that copyright notice and this permission notice
-- appear in supporting documentation.
-- 
-- The ARA makes no representations about the suitability of this software
-- for any purpose.  It is provided "as is" without express
-- or implied warranty.
-- 
