--******************************************************************************
--******************************************************************************
--
-- The package Convert contains some functions for the conversion
-- of Big_Unsigned, Big_Signed or Fractions of different sizes.
--


GENERIC
  TYPE Index_Type IS MOD <>;
  Nb_Item_Src, Nb_Item_Dest : Index_Type;



PACKAGE Convert IS

  TYPE Big_Unsigned_Src IS PRIVATE;
  TYPE Big_Unsigned_Dest IS PRIVATE;
  TYPE Big_Signed_Src IS PRIVATE;
  TYPE Big_Signed_Dest IS PRIVATE;
  TYPE Fraction_Src IS PRIVATE;
  TYPE Fraction_Dest IS PRIVATE;


  PROCEDURE Convert_Unsigned(Src : IN Big_Unsigned_Src;
                             Dest : OUT Big_Unsigned_Dest;
                             OK : OUT BOOLEAN);
    -- This procedure converts a Big_Unsigned_Src variable into a 
    -- Big_Unsigned_Dest variable.
    -- OK is set to TRUE if there was no errors during the conversion.
    -- OK is set to FALSE if the number of significant digits of Src
    -- can't stand in Dest (if the TYPE Big_Unsigned_Dest is smaller
    -- than the TYPE Big_Unsigned_Src).


  PROCEDURE Convert_Signed(Src : IN Big_Signed_Src;
                           Dest : OUT Big_Signed_Dest;
                           OK : OUT BOOLEAN);
    -- This procedure converts a Big_Signed_Src variable into a 
    -- Big_Signed_Dest variable.
    -- OK is set to TRUE if there was no errors during the conversion.
    -- OK is set to FALSE if the number of significant digits of Src
    -- can't stand in Dest (if the TYPE Big_Signed_Dest is smaller
    -- than the TYPE Big_Signed_Src).


  PROCEDURE Convert_Fraction(Src : IN Fraction_Src;
                             Dest : OUT Fraction_Dest;
                             OK : OUT BOOLEAN);
    -- This procedure converts a Fraction_Src variable into
    -- a Fraction_Dest variable.
    -- OK is set to TRUE if there was no errors during the conversion.
    -- OK is set to FALSE if the number of significant digits of Src
    -- can't stand in Dest (if the TYPE Fraction_Dest is smaller
    -- than the TYPE Fraction_Src).


PRIVATE

  TYPE My_Type IS MOD 2**32;
  FOR My_Type'SIZE USE 32;

  TYPE Mantissa_Src IS ARRAY(0..Nb_Item_Src-1) OF My_Type;
  TYPE Mantissa_Dest IS ARRAY(0..Nb_Item_Dest-1) OF My_Type;

  TYPE Big_Unsigned_Src IS RECORD
                             Last_Significant_Digit : Index_Type;
                             Mantissa : Mantissa_Src;
                           END RECORD;
  TYPE Big_Unsigned_Dest IS RECORD
                              Last_Significant_Digit : Index_Type;
                              Mantissa : Mantissa_Dest;
                            END RECORD;

  TYPE Big_Signed_Src IS RECORD
                           Positive : BOOLEAN;
                           Last_Significant_Digit : Index_Type;
                           Mantissa : Mantissa_Src;
                         END RECORD;
  TYPE Big_Signed_Dest IS RECORD
                            Positive : BOOLEAN;
                            Last_Significant_Digit : Index_Type;
                            Mantissa : Mantissa_Dest;
                          END RECORD;

  TYPE Fraction_Src IS RECORD
                         Positive    : BOOLEAN;
                         Numerator   : Big_Unsigned_Src;
                         Denominator : Big_Unsigned_Src;
                       END RECORD;
  TYPE Fraction_Dest IS RECORD
                          Positive    : BOOLEAN;
                          Numerator   : Big_Unsigned_Dest;
                          Denominator : Big_Unsigned_Dest;
                        END RECORD;

END Convert;
--*******************************************************************************
--*******************************************************************************

PACKAGE BODY Convert IS


FUNCTION Min(x, y : IN Index_Type) RETURN Index_Type IS
BEGIN
  IF (x < y) THEN
    RETURN x;
  ELSE
    RETURN y;
  END IF;
END Min;


FUNCTION Convert_Mantissa(Src : IN Mantissa_Src) RETURN Mantissa_Dest IS
  Temp_Result : Mantissa_Dest := (OTHERS => 0);
BEGIN
  FOR I IN 0..Min(Nb_Item_Src-1, Nb_Item_Dest-1) LOOP
    Temp_Result(I) := Src(I);
  END LOOP;
  RETURN Temp_Result;
END Convert_Mantissa;


PROCEDURE Convert_Unsigned(Src : IN Big_Unsigned_Src;
                           Dest : OUT Big_Unsigned_Dest;
                           OK : OUT BOOLEAN) IS
  -- This procedure converts a Big_Unsigned_Src variable into a 
  -- Big_Unsigned_Dest variable.
  -- OK is set to TRUE if there was no errors during the conversion.
  -- OK is set to FALSE if the number of significant digits of Src
  -- can't stand in Dest (if the TYPE Big_Unsigned_Dest is smaller
  -- than the TYPE Big_Unsigned_Src).
BEGIN
  IF (Src.Last_Significant_Digit > Nb_Item_Dest - 1) THEN
    OK := FALSE;
  ELSE
    OK := TRUE;
    Dest.Last_Significant_Digit := Src.Last_Significant_Digit;
    Dest.Mantissa := Convert_Mantissa(Src.Mantissa);
  END IF;
END Convert_Unsigned;



PROCEDURE Convert_Signed(Src : IN Big_Signed_Src;
                         Dest : OUT Big_Signed_Dest;
                         OK : OUT BOOLEAN) IS
  -- This procedure converts a Big_Signed_Src variable into a 
  -- Big_Signed_Dest variable.
  -- OK is set to TRUE if there was no errors during the conversion.
  -- OK is set to FALSE if the number of significant digits of Src
  -- can't stand in Dest (if the TYPE Big_Signed_Dest is smaller
  -- than the TYPE Big_Signed_Src).
BEGIN
  IF (Src.Last_Significant_Digit > Nb_Item_Dest - 1) THEN
    OK := FALSE;
  ELSE
    OK := TRUE;
    Dest.Positive := Src.Positive;
    Dest.Last_Significant_Digit := Src.Last_Significant_Digit;
    Dest.Mantissa := Convert_Mantissa(Src.Mantissa);
  END IF;
END Convert_Signed;


PROCEDURE Convert_Fraction(Src : IN Fraction_Src;
                           Dest : OUT Fraction_Dest;
                           OK : OUT BOOLEAN) IS
  -- This procedure converts a Fraction_Src variable into
  -- a Fraction_Dest variable.
  -- OK is set to TRUE if there was no errors during the conversion.
  -- OK is set to FALSE if the number of significant digits of Src
  -- can't stand in Dest (if the TYPE Fraction_Dest is smaller
  -- than the TYPE Fraction_Src).
  Temp1, Temp2 : BOOLEAN := TRUE;
BEGIN
  Dest.Positive := Src.Positive;
  Convert_Unsigned(Src.Numerator, Dest.Numerator, Temp1);
  IF (Temp1 = TRUE) THEN
    Convert_Unsigned(Src.Denominator, Dest.Denominator, Temp2);
  END IF;
  OK := Temp1 AND Temp2;
END Convert_Fraction;



END Convert;
--*******************************************************************************
--*******************************************************************************

WITH DIRECT_IO;
WITH Convert;
WITH ModInt;    USE ModInt;
WITH Word_IO;   USE Word_IO;
WITH TEXT_IO;   USE TEXT_IO;


PROCEDURE Conversion_Fraction IS
  SUBTYPE FileName IS STRING(1..12);
  NbSrc, NbDest : Word;
  NameSrc, NameDest : FileName;


PROCEDURE Init(Nb_Src, Nb_Dest : OUT Word;
               Name_Src, Name_Dest : OUT FileName) IS
BEGIN
  PUT_LINE("Size of mantissa : number of 32 bit elements in mantissa");
  PUT("Size of the mantissa for big integers in the source file : ");
  GET(Nb_Src);
  PUT("Size of the mantissa for big integers in the destination file : ");
  GET(Nb_Dest);
  PUT("Source file name ? : ");
  GET(Name_Src);
  PUT("Target file name ? : ");
  GET(Name_Dest);
END Init;


PROCEDURE Realize_Conversion(Nb_Src, Nb_Dest : IN Word;
                             Name_Src, Name_Dest : IN FileName) IS
  PACKAGE My_Convert IS NEW Convert(Word, NbSrc, NbDest);
  USE My_Convert;
  PACKAGE Dest_IO IS NEW DIRECT_IO(My_Convert.Fraction_Dest);
  PACKAGE Src_IO IS NEW DIRECT_IO(My_Convert.Fraction_Src);
  USE Dest_IO, Src_IO;
  OK : BOOLEAN;
  File_Src : Src_IO.FILE_TYPE;
  File_Dest : Dest_IO.FILE_TYPE;
  Fract_Src : Fraction_Src;
  Fract_Dest : Fraction_Dest;
  Nombre : Src_IO.COUNT;
BEGIN
  CREATE(File_Dest, OUT_FILE, Name_Dest);
  OPEN(File_Src, IN_FILE, Name_Src);
  Nombre := SIZE(File_Src);
  FOR I IN 1..Nombre LOOP
    READ(File_Src, Fract_Src);
    Convert_Fraction(Fract_Src, Fract_Dest, OK);
    IF (OK = TRUE) THEN
      WRITE(File_Dest, Fract_Dest);
    ELSE
      PUT_LINE("An error of conversion occured.");
    END IF;
  END LOOP;
  CLOSE(File_Dest);
  CLOSE(File_Src);
END Realize_Conversion;


BEGIN
  Init(NbSrc, NbDest, NameSrc, NameDest);
  Realize_Conversion(NbSrc, NbDest, NameSrc, NameDest);
END Conversion_Fraction;
--*******************************************************************************
--*******************************************************************************

WITH DIRECT_IO;
WITH Convert;
WITH ModInt;    USE ModInt;
WITH Word_IO;   USE Word_IO;
WITH TEXT_IO;   USE TEXT_IO;


PROCEDURE Conversion_Unsigned IS
  SUBTYPE FileName IS STRING(1..12);
  NbSrc, NbDest : Word;
  NameSrc, NameDest : FileName;


PROCEDURE Init(Nb_Src, Nb_Dest : OUT Word;
               Name_Src, Name_Dest : OUT FileName) IS
BEGIN
  PUT_LINE("Size of mantissa : number of 32 bit elements in mantissa");
  PUT("Size of the mantissa for big integers in the source file : ");
  GET(Nb_Src);
  PUT("Size of the mantissa for big integers in the destination file : ");
  GET(Nb_Dest);
  PUT("Source file name ? : ");
  GET(Name_Src);
  PUT("Target file name ? : ");
  GET(Name_Dest);
END Init;


PROCEDURE Realize_Conversion(Nb_Src, Nb_Dest : IN Word;
                             Name_Src, Name_Dest : IN FileName) IS
  PACKAGE My_Convert IS NEW Convert(Word, NbSrc, NbDest);
  USE My_Convert;
  PACKAGE Dest_IO IS NEW DIRECT_IO(My_Convert.Big_Unsigned_Dest);
  PACKAGE Src_IO IS NEW DIRECT_IO(My_Convert.Big_Unsigned_Src);
  USE Dest_IO, Src_IO;
  OK : BOOLEAN;
  File_Src : Src_IO.FILE_TYPE;
  File_Dest : Dest_IO.FILE_TYPE;
  Unsigned_Src : Big_Unsigned_Src;
  Unsigned_Dest : Big_Unsigned_Dest;
  Nombre : Src_IO.COUNT;
BEGIN
  CREATE(File_Dest, OUT_FILE, Name_Dest);
  OPEN(File_Src, IN_FILE, Name_Src);
  Nombre := SIZE(File_Src);
  FOR I IN 1..Nombre LOOP
    READ(File_Src, Unsigned_Src);
    Convert_Unsigned(Unsigned_Src, Unsigned_Dest, OK);
    IF (OK = TRUE) THEN
      WRITE(File_Dest, Unsigned_Dest);
    ELSE
      PUT_LINE("An error of conversion occured.");
    END IF;
  END LOOP;
  CLOSE(File_Dest);
  CLOSE(File_Src);
END Realize_Conversion;


BEGIN
  Init(NbSrc, NbDest, NameSrc, NameDest);
  Realize_Conversion(NbSrc, NbDest, NameSrc, NameDest);
END Conversion_Unsigned;
--*******************************************************************************
--*******************************************************************************

WITH DIRECT_IO;
WITH Convert;
WITH ModInt;    USE ModInt;
WITH Word_IO;   USE Word_IO;
WITH TEXT_IO;   USE TEXT_IO;


PROCEDURE Conversion_Signed IS
  SUBTYPE FileName IS STRING(1..12);
  NbSrc, NbDest : Word;
  NameSrc, NameDest : FileName;


PROCEDURE Init(Nb_Src, Nb_Dest : OUT Word;
               Name_Src, Name_Dest : OUT FileName) IS
BEGIN
  PUT_LINE("Size of mantissa : number of 32 bit elements in mantissa");
  PUT("Size of the mantissa for big integers in the source file : ");
  GET(Nb_Src);
  PUT("Size of the mantissa for big integers in the destination file : ");
  GET(Nb_Dest);
  PUT("Source file name ? : ");
  GET(Name_Src);
  PUT("Target file name ? : ");
  GET(Name_Dest);
END Init;


PROCEDURE Realize_Conversion(Nb_Src, Nb_Dest : IN Word;
                             Name_Src, Name_Dest : IN FileName) IS
  PACKAGE My_Convert IS NEW Convert(Word, NbSrc, NbDest);
  USE My_Convert;
  PACKAGE Dest_IO IS NEW DIRECT_IO(My_Convert.Big_Signed_Dest);
  PACKAGE Src_IO IS NEW DIRECT_IO(My_Convert.Big_Signed_Src);
  USE Dest_IO, Src_IO;
  OK : BOOLEAN;
  File_Src : Src_IO.FILE_TYPE;
  File_Dest : Dest_IO.FILE_TYPE;
  Signed_Src : Big_Signed_Src;
  Signed_Dest : Big_Signed_Dest;
  Nombre : Src_IO.COUNT;
BEGIN
  CREATE(File_Dest, OUT_FILE, Name_Dest);
  OPEN(File_Src, IN_FILE, Name_Src);
  Nombre := SIZE(File_Src);
  FOR I IN 1..Nombre LOOP
    READ(File_Src, Signed_Src);
    Convert_Signed(Signed_Src, Signed_Dest, OK);
    IF (OK = TRUE) THEN
      WRITE(File_Dest, Signed_Dest);
    ELSE
      PUT_LINE("An error of conversion occured.");
    END IF;
  END LOOP;
  CLOSE(File_Dest);
  CLOSE(File_Src);
END Realize_Conversion;


BEGIN
  Init(NbSrc, NbDest, NameSrc, NameDest);
  Realize_Conversion(NbSrc, NbDest, NameSrc, NameDest);
END Conversion_Signed;
