PACKAGE ModInt IS

TYPE Byte IS MOD 2**8;
FOR Byte'SIZE USE 8;

TYPE My_Index IS MOD 2**16;
FOR My_Index'SIZE USE 16;
-- You can modify the declaration of My_Index as you want.

END ModInt;

--******************************************************************************
WITH ModInt;
WITH SEQUENTIAL_IO;

PACKAGE Byte_Sequential_IO IS NEW SEQUENTIAL_IO(ModInt.Byte);

--******************************************************************************
WITH ModInt;
WITH Direct_IO;

PACKAGE Index_Direct_IO IS NEW DIRECT_IO(ModInt.My_Index);

--******************************************************************************
WITH ModInt;
WITH TEXT_IO;


PACKAGE Index_IO IS NEW TEXT_IO.MODULAR_IO(ModInt.My_Index);

--******************************************************************************
PACKAGE File_Utils IS


SUBTYPE FileName IS STRING(1..12);


FUNCTION Does_File_Exist(Nom : IN FileName) RETURN BOOLEAN;

PROCEDURE Delete_File(Nom : FileName);


END File_Utils;

--******************************************************************************
WITH Byte_Sequential_IO;    USE Byte_Sequential_IO;


PACKAGE BODY File_Utils IS

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

FUNCTION Does_File_Exist(Nom : IN FileName) RETURN BOOLEAN IS
  Fic : Byte_Sequential_IO.FILE_TYPE;
BEGIN
  OPEN(Fic, IN_FILE, Nom);
  CLOSE(Fic);
  RETURN TRUE;
EXCEPTION
  WHEN OTHERS => RETURN FALSE;
END Does_File_Exist;

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

PROCEDURE Delete_File(Nom : FileName) IS
  Fic : Byte_Sequential_IO.FILE_TYPE;
BEGIN
  OPEN(Fic, IN_FILE, Nom);
  DELETE(Fic);             -- La procdure DELETE ferme aussi le fichier.
END Delete_File;

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

END File_Utils;

--******************************************************************************
-----------------------------------------------------------------------
--
--  File:        djgplibr.ads
--  Description: Interface to DJGPP library functions
--  Rev:         0.7
--  Date:        01-feb-98
--  Author:      Jerry van Dijk
--  Mail:        jdijk@acm.org
--
--  Copyright (c) Jerry van Dijk, 1996, 1997, 1998
--  Billie Holidaystraat 28
--  2324 LK Leiden
--  THE NETHERLANDS
--  tel int + 31 71 531 4365
--
--  Permission granted to use for any purpose, provided this copyright
--  remains attached and unmodified.
--
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
--
-----------------------------------------------------------------------

with Interfaces;   use Interfaces;
with Interfaces.C; use Interfaces.C;

package DJGPP_Library is

   ---------------------------------------------
   -- NAME:    Dpmi_Regs                      --
   --                                         --
   -- PURPOSE: Simplified processor registers --
   ---------------------------------------------
   type Dpmi_Regs is
      record
         Di     : Unsigned_16;
         Di_Hi  : Unsigned_16;
         Si     : Unsigned_16;
         Si_Hi  : Unsigned_16;
         Bp     : Unsigned_16;
         Bp_Hi  : Unsigned_16;
         Res    : Unsigned_16;
         Res_Hi : Unsigned_16;
         Bx     : Unsigned_16;
         Bx_Hi  : Unsigned_16;
         Dx     : Unsigned_16;
         Dx_Hi  : Unsigned_16;
         Cx     : Unsigned_16;
         Cx_Hi  : Unsigned_16;
         Ax     : Unsigned_16;
         Ax_Hi  : Unsigned_16;
         Flags  : Unsigned_16;
         Es     : Unsigned_16;
         Ds     : Unsigned_16;
         Fs     : Unsigned_16;
         Gs     : Unsigned_16;
         Ip     : Unsigned_16;
         Cs     : Unsigned_16;
         Sp     : Unsigned_16;
         Ss     : Unsigned_16;
      end record;
   pragma Convention(C, Dpmi_Regs);


   ----------------------------------------
   -- NAME:    Go32_Info_Block           --
   --                                    --
   -- PURPOSE: Defines the extender info --
   ----------------------------------------
   type Go32_Info_Block is
      record
         Size_Of_This_Structure_In_Bytes        : Unsigned_32;
         Linear_Address_Of_Primary_Screen       : Unsigned_32;
         Linear_Address_Of_Secondary_Screen     : Unsigned_32;
         Linear_Address_Of_Transfer_Buffer      : Unsigned_32;
         Size_Of_Transfer_Buffer                : Unsigned_32;
         Pid                                    : Unsigned_32;
         Master_Interrupt_Controller_Base       : Unsigned_8;
         Slave_Interrupt_Controller_Base        : Unsigned_8;
         Selector_For_Linear_Memory             : Unsigned_16;
         Linear_Address_Of_Stub_Info_Structure  : Unsigned_32;
         Linear_Address_Of_Original_Psp         : Unsigned_32;
         Run_Mode                               : Unsigned_16;
         Run_Mode_Info                          : Unsigned_16;
      end record;
   pragma Convention(C, Go32_Info_Block);


   -------------------------------------------
   -- NAME:    Dos_DS                       --
   --                                       --
   -- PURPOSE: Conventional memory selector --
   -------------------------------------------
   Current_Info : Go32_Info_Block;
   pragma Import(C, Current_Info, "_go32_info_block");

   Dos_DS : Unsigned_16 renames Current_Info.Selector_For_Linear_Memory;


   --------------------------------------------------
   -- NAME:    Dpmi_Int                            --
   --                                              --
   -- PURPOSE: Call a real-mode interrupt          --
   --                                              --
   -- INPUTS:  Vector - Interrupt number           --
   --          Regs   - Processor registers        --
   --                                              --
   -- OUTPUTS: Regs - Modified processor registers --
   --------------------------------------------------
   procedure Dpmi_Int(Vector : in     Unsigned_16;
                      Regs   : in out Dpmi_Regs);
   pragma Import(C, Dpmi_Int, "__dpmi_int");


   ------------------------------------------------------
   -- NAME:    Farsetsel                               --
   --                                                  --
   -- PURPOSE: Preset the conventional memory selector --
   ------------------------------------------------------
   procedure Farsetsel(Selector : in Unsigned_16);
   pragma Import(C, Farsetsel, "_farsetsel");


   -----------------------------------------------------
   -- NAME:    Farpeekb                               --
   --                                                 --
   -- PURPOSE: Reads a byte from memory               --
   --                                                 --
   -- INPUTS:  Selector - The memory segment selector --
   --          Offset   - The fysical address         --
   --                                                 --
   -- RETURNS: The byte at the address                --
   -----------------------------------------------------
   function Farpeekb(Selector : in Unsigned_16;
                     Offset   : in Unsigned_32) return Unsigned_8;
   pragma Import(C, Farpeekb, "_farpeekb");


   -----------------------------------------------
   -- NAME:    Farnspeekb                       --
   --                                           --
   -- PURPOSE: Reads a byte from memory using a --
   --          preset selector                  --
   --                                           --
   -- INPUTS:  Offset - The fysical address     --
   --                                           --
   -- RETURNS: The byte at the address          --
   -----------------------------------------------
   function Farnspeekb(Offset : in Unsigned_32) return Unsigned_8;
   pragma Import(C, Farnspeekb, "_farnspeekb");


   ---------------------------------------------
   -- NAME:    Farpokeb                       --
   --                                         --
   -- PURPOSE: Writes a byte to memory        --
   --                                         --
   -- INPUTS:  Selector - The memory selector --
   --          Offset   - The fysical address --
   --          Value    - The byte to write   --
   ---------------------------------------------
   procedure Farpokeb(Selector : in Unsigned_16;
                      Offset   : in Unsigned_32;
                      Value    : in Unsigned_8);
   pragma Import(C, Farpokeb, "_farpokeb");


   ----------------------------------------------
   -- NAME:    Farnspokeb                      --
   --                                          --
   -- PURPOSE: Writes a byte to memory using a --
   --          preset selector                 --
   --                                          --
   -- INPUTS:  Offset - The fysical address    --
   --          Value  - The byte to write      --
   ----------------------------------------------
   procedure Farnspokeb(Offset : in Unsigned_32;
                        Value  : in Unsigned_8);
   pragma Import(C, Farnspokeb, "_farnspokeb");


   ---------------------------------------------
   -- NAME:    Outportb                       --
   --                                         --
   -- PURPOSE: Outputs a byte to a I/O port   --
   --                                         --
   -- INPUT:   Port - The I/O port            --
   --          Data - The byte to write to it --
   ---------------------------------------------
   procedure Outportb(Port : in Unsigned_16;
                      Data : in Unsigned_8);
   pragma Import(C, Outportb, "outportb");


   ---------------------------------------------------
   -- NAME:    Dosmemget                            --
   --                                               --
   -- PURPOSE: Transfers data from conventional     --
   --          memory to the application            --
   --                                               --
   -- INPUTS:  Offset - Fysical memory address      --
   --          Length - Number of bytes to transfer --
   --          Buffer - The destination             --
   ---------------------------------------------------
   procedure Dosmemget(Offset : in Unsigned_32;
                       Length : in Unsigned_32;
                       Buffer : char_array);
   pragma Import(C, Dosmemget, "dosmemget");


   ---------------------------------------------------
   -- NAME:    Dosmemput                            --
   --                                               --
   -- PURPOSE: Transfers data from the application  --
   --          to conventional memory               --
   --                                               --
   -- INPUTS:  Offset - Fysical memory address      --
   --          Length - Number of bytes to transfer --
   --          Buffer - The destination             --
   ---------------------------------------------------
   procedure Dosmemput(Buffer : char_array;
                       Length : Unsigned_32;
                       Offset : Unsigned_32);
   pragma Import(C, Dosmemput, "dosmemput");


end DJGPP_Library;

--******************************************************************************
-----------------------------------------------------------------------
--
--  File:        pckeyboa.ads
--  Description: Basic keyboard interface
--  Rev:         0.7
--  Date:        01-feb-98
--  Author:      Jerry van Dijk
--  Mail:        jdijk@acm.org
--
--  Copyright (c) Jerry van Dijk, 1996, 1997, 1998
--  Billie Holidaystraat 28
--  2324 LK Leiden
--  THE NETHERLANDS
--  tel int + 31 71 531 4365
--
--  Permission granted to use for any purpose, provided this copyright
--  remains attached and unmodified.
--
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
--
-----------------------------------------------------------------------


   ------------------------------------------------------------
   --                       PACKAGE CONTAINS                 --
   ------------------------------------------------------------
   -- Types:                                                 --
   --    Keyboard_Scancode - Scancode of a keypress          --
   --                                                        --
   -- Functions:                                             --
   --    Is_Key_Waiting    - Check if a key has been pressed --
   --    Get_Scancode      - Waits for a key press           --
   --    Get_Char          - Waits for a key press           --
   ------------------------------------------------------------


package PC_Keyboard is

   -------------------------------------
   -- NAME:    Keyboard_Scancode      --
   --                                 --
   -- PURPOSE: Scancode of a keypress --
   -------------------------------------
   subtype Keyboard_Scancode is Integer range 0..2**16-1;


   ----------------------------------------------------
   -- NAME:    Is_Key_Waiting                        --
   --                                                --
   -- PURPOSE: Check if a key has been pressed       --
   --                                                --
   -- RETURNS: True if key pressed, False otherwise  --
   --                                                --
   -- NOTES:   1. Leaves keypress in keyboard buffer --
   ----------------------------------------------------
   function Is_Key_Waiting return Boolean;


   -------------------------------------------
   -- NAME:    Get_Scancode                 --
   --                                       --
   -- PURPOSE: Waits for a key press        --
   --                                       --
   -- RETURNS: Key Scancode                 --
   --                                       --
   -- NOTES:   1. Control-Break not noticed --
   -------------------------------------------
   function Get_Scancode return Keyboard_Scancode;


   --------------------------------------------------
   -- NAME:    Get_Char                            --
   --                                              --
   -- PURPOSE: Waits for a key press               --
   --                                              --
   -- RETURNS: Character pressed, NUL if not ASCII --
   --                                              --
   -- NOTES:   1. Control-Break not noticed        --
   --------------------------------------------------
   function Get_Char return Character;


   --------------------------------------------------
   -- NAME:    Clear_Keyboard_Buffer               --
   --                                              --
   -- PURPOSE: Clear the keyboard buffer           --
   --                                              --
   -- RETURNS: None                                --
   --                                              --
   -- NOTES:   None                                --
   --------------------------------------------------
   procedure Clear_Keyboard_Buffer;


end PC_Keyboard;

--******************************************************************************
-----------------------------------------------------------------------
--
--  File:        pckeyboa.adb
--  Description: Basic keyboard interface
--  Rev:         0.7
--  Date:        01-feb-98
--  Author:      Jerry van Dijk
--  Mail:        jdijk@acm.org
--
--  Copyright (c) Jerry van Dijk, 1996, 1997, 1998
--  Billie Holidaystraat 28
--  2324 LK Leiden
--  THE NETHERLANDS
--  tel int + 31 71 531 4365
--
--  Permission granted to use for any purpose, provided this copyright
--  remains attached and unmodified.
--
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
--
-----------------------------------------------------------------------


   ------------------------------------------------------------
   --                       PACKAGE CONTAINS                 --
   ------------------------------------------------------------
   -- Types:                                                 --
   --    Keyboard_Scancode - Scancode of a keypress          --
   --                                                        --
   -- Functions:                                             --
   --    Is_Key_Waiting    - Check if a key has been pressed --
   --    Get_Scancode      - Waits for a key press           --
   --    Get_Char          - Waits for a key press           --
   ------------------------------------------------------------


with Interfaces;    use Interfaces;
with DJGPP_Library; use DJGPP_Library;

package body PC_Keyboard is

   ----------------------------------------------------
   -- NAME:    Is_Key_Waiting                        --
   --                                                --
   -- PURPOSE: Check if a key has been pressed       --
   --                                                --
   -- RETURNS: True if key pressed, False otherwise  --
   --                                                --
   -- NOTES:   1. Leaves keypress in keyboard buffer --
   ----------------------------------------------------

   function Is_Key_Waiting return Boolean is
      Regs   : Dpmi_Regs;
      Result : Boolean;
   begin
      Regs.Ax := 16#1100#;
      Dpmi_Int(16#16#, Regs);
      Result := ((Regs.Flags and 16#40#) = 0);
      return Result;
   end Is_Key_Waiting;


   -------------------------------------------
   -- NAME:    Get_Scancode                 --
   --                                       --
   -- PURPOSE: Waits for a key press        --
   --                                       --
   -- RETURNS: Key Scancode                 --
   --                                       --
   -- NOTES:   1. Control-Break not noticed --
   -------------------------------------------
   function Get_Scancode return Keyboard_Scancode is
      Regs : Dpmi_Regs;
   begin
      Regs.Ax := 16#0000#;
      Dpmi_Int(16#16#, Regs);
      return Keyboard_Scancode(Regs.Ax);
   end Get_Scancode;


   --------------------------------------------------
   -- NAME:    Get_Char                            --
   --                                              --
   -- PURPOSE: Waits for a key press               --
   --                                              --
   -- RETURNS: Character pressed, NUL if not ASCII --
   --                                              --
   -- NOTES:   1. Control-Break not noticed        --
   --------------------------------------------------
   function Get_Char return Character is
   begin
      return Character'VAL(Unsigned_16(Get_Scancode) and 16#00FF#);
   end Get_Char;


   --------------------------------------------------
   -- NAME:    Clear_Keyboard_Buffer               --
   --                                              --
   -- PURPOSE: Clear the keyboard buffer           --
   --                                              --
   -- RETURNS: None                                --
   --                                              --
   -- NOTES:   None                                --
   --------------------------------------------------
   procedure Clear_Keyboard_Buffer is
     Code : Keyboard_Scancode;
   begin
     while (Is_Key_Waiting = TRUE) loop
       Code := Get_Scancode;
     end loop;
   end Clear_Keyboard_Buffer;
 
end PC_Keyboard;

--******************************************************************************
WITH ModInt;
WITH Big_Number;


PACKAGE My_Big IS NEW Big_Number(ModInt.My_Index, 1000);

--******************************************************************************
WITH My_Big;      USE My_Big;
WITH DIRECT_IO;


PACKAGE Fraction_Direct_IO IS NEW DIRECT_IO(Fraction);

--******************************************************************************
WITH My_Big;      USE My_Big;
WITH DIRECT_IO;


PACKAGE Big_Unsigned_Direct_IO IS NEW DIRECT_IO(Big_Unsigned);

--******************************************************************************
--                                                                            **
--  File:        UTILS_PI.ADS                                                 **
--  Description: Package specification of the package Utils_Pi                **
--  Revision:    1.11                                                         **
--  Date:        January 19, 2001                                             **
--  Author:      Jrme Delcourt                                              **
--  Mail:        sikander@club-internet.fr                                    **
--                                                                            **
--  Copyright (c) Jrme Delcourt, 1998, 1999, 2000, 2001                     **
--  62, rue Ngrier                                                           **
--  59800 Lille                                                               **
--  FRANCE                                                                    **
--                                                                            **
--  Permission granted to use for any purpose, provided this copyright        **
--  remains attached and unmodified.                                          **
--                                                                            **
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR            **
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED            **
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       **
--                                                                            **
--******************************************************************************

WITH ModInt;                  USE ModInt;
WITH File_Utils;              USE File_Utils;
WITH TEXT_IO;                 USE TEXT_IO;
WITH Index_IO;                USE Index_IO;
WITH Index_Direct_IO;         USE Index_Direct_IO;
WITH My_Big;                  USE My_Big;
                              USE My_Big.Generic_Conversion;
                              USE My_Big.Unsigned_Number;
                              USE My_Big.Big_Unsigned_IO;
                              USE My_Big.Big_Fraction;
WITH Fraction_Direct_IO;      USE Fraction_Direct_IO;
WITH Big_Unsigned_Direct_IO;  USE Big_Unsigned_Direct_IO;


PACKAGE Utils_Pi IS


FUNCTION Convert IS NEW Mod_Number2Big_Unsigned(ModInt.My_Index);

Big2 : CONSTANT Big_Unsigned := Convert(2);
Big5 : CONSTANT Big_Unsigned := Convert(5);
Big8 : CONSTANT Big_Unsigned := Convert(8);
Fract0 : CONSTANT Fraction := CreateFraction(Big_Unsigned_Zero, Big_Unsigned_One);
Fract2 : CONSTANT Fraction := CreateFraction(Big_Unsigned_One, Big2);
Fract5 : CONSTANT Fraction := CreateFraction(Big_Unsigned_One, Big5);
Fract8 : CONSTANT Fraction := CreateFraction(Big_Unsigned_One, Big8);


PROCEDURE Init(Inf, Sup : OUT My_Index;
               Result2, Result5, Result8 : OUT Fraction;
               Term2, Term5, Term8 : OUT Big_Unsigned);

PROCEDURE Init(Inf : OUT My_Index;
               Result2, Result5, Result8 : OUT Fraction;
               Term2, Term5, Term8 : OUT Big_Unsigned);

PROCEDURE Save(Sup : IN My_Index;
               Result2, Result5, Result8 : IN Fraction;
               Term2, Term5, Term8 : IN Big_Unsigned);

PROCEDURE Resultat(Arctan2, Arctan5, Arctan8 : IN Fraction; BorneSup : IN My_Index);


END Utils_Pi;
--******************************************************************************
--                                                                            **
--  File:        UTILS_PI.ADB                                                 **
--  Description: Package body of the package Utils_Pi                         **
--  Revision:    1.11                                                         **
--  Date:        January 19, 2001                                             **
--  Author:      Jrme Delcourt                                              **
--  Mail:        sikander@club-internet.fr                                    **
--                                                                            **
--  Copyright (c) Jrme Delcourt, 1998, 1999, 2000, 2001                     **
--  62, rue Ngrier                                                           **
--  59800 Lille                                                               **
--  FRANCE                                                                    **
--                                                                            **
--  Permission granted to use for any purpose, provided this copyright        **
--  remains attached and unmodified.                                          **
--                                                                            **
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR            **
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED            **
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       **
--                                                                            **
--******************************************************************************

PACKAGE BODY Utils_Pi IS

--****************************************************************************** 
--****************************************************************************** 

  Loop_FileName : CONSTANT FileName := "CALLOOP1.DAT";
  Result_FileName : CONSTANT FileName := "RESULTS1.DAT";
  Term_FileName : CONSTANT FileName := "TERMS__1.DAT";

--****************************************************************************** 
-- Declarations of functions
--****************************************************************************** 

FUNCTION Files_Exist RETURN BOOLEAN IS
  Result : BOOLEAN;
BEGIN
  Result := Does_File_Exist(Loop_FileName) AND THEN
            Does_File_Exist(Result_FileName) AND THEN
            Does_File_Exist(Term_FileName);
  RETURN Result;
END Files_Exist;

--****************************************************************************** 

PROCEDURE Init(Inf, Sup : OUT My_Index;
               Result2, Result5, Result8 : OUT Fraction;
               Term2, Term5, Term8 : OUT Big_Unsigned) IS
  File_NbLoop : Index_Direct_IO.FILE_TYPE;
  File_Result : Fraction_Direct_IO.FILE_TYPE;
  File_Term : Big_Unsigned_Direct_IO.FILE_TYPE;
  Nb_Loop, Old_Loop : My_Index;
BEGIN
  IF Files_Exist THEN
    OPEN(File_NbLoop, IN_FILE, Loop_FileName);
    READ(File_NbLoop, Old_Loop);
    CLOSE(File_NbLoop);
    --
    OPEN(File_Result, IN_FILE, Result_FileName);
    READ(File_Result, Result2);
    READ(File_Result, Result5);
    READ(File_Result, Result8);
    CLOSE(File_Result);
    --
    OPEN(File_Term, IN_FILE, Term_FileName);
    READ(File_Term, Term2);
    READ(File_Term, Term5);
    READ(File_Term, Term8);
    CLOSE(File_Term);
  ELSE
    Result2 := Fract2;
    Result5 := Fract5;
    Result8 := Fract8;
    Term2 := Big2;
    Term5 := Big5;
    Term8 := Big8;
    Old_Loop := 0;
  END IF;
  PUT("Number of loops already done : ");
  PUT(Old_Loop, WIDTH => 0);
  NEW_LINE;
  PUT("Number of loops to do ? ");
  GET(Nb_Loop);
  Inf := Old_Loop + 1;
  Sup := Old_Loop + Nb_Loop;
END Init;

--******************************************************************************
--******************************************************************************

PROCEDURE Init(Inf : OUT My_Index;
               Result2, Result5, Result8 : OUT Fraction;
               Term2, Term5, Term8 : OUT Big_Unsigned) IS
  File_NbLoop : Index_Direct_IO.FILE_TYPE;
  File_Result : Fraction_Direct_IO.FILE_TYPE;
  File_Term : Big_Unsigned_Direct_IO.FILE_TYPE;
  Old_Loop : My_Index;
BEGIN
  IF Does_File_Exist(Loop_FileName) AND THEN Does_File_Exist(Result_FileName) THEN
    OPEN(File_NbLoop, IN_FILE, Loop_FileName);
    READ(File_NbLoop, Old_Loop);
    CLOSE(File_NbLoop);
    --
    OPEN(File_Result, IN_FILE, Result_FileName);
    READ(File_Result, Result2);
    READ(File_Result, Result5);
    READ(File_Result, Result8);
    CLOSE(File_Result);
    --
    OPEN(File_Term, IN_FILE, Term_FileName);
    READ(File_Term, Term2);
    READ(File_Term, Term5);
    READ(File_Term, Term8);
    CLOSE(File_Term);
  ELSE
    Result2 := Fract2;
    Result5 := Fract5;
    Result8 := Fract8;
    Term2 := Big2;
    Term5 := Big5;
    Term8 := Big8;
    Old_Loop := 0;
  END IF;
  PUT("Number of loops already done : ");
  PUT(Old_Loop, WIDTH => 0);
  NEW_LINE;
  Inf := Old_Loop + 1;
END Init;

--******************************************************************************
--******************************************************************************

PROCEDURE Save(Sup : IN My_Index;
               Result2, Result5, Result8 : IN Fraction;
               Term2, Term5, Term8 : IN Big_Unsigned) IS
  File_NbLoop : Index_Direct_IO.FILE_TYPE;
  File_Result : Fraction_Direct_IO.FILE_TYPE;
  File_Term : Big_Unsigned_Direct_IO.FILE_TYPE;
BEGIN
  CREATE(File_NbLoop, OUT_FILE, Loop_FileName);
  WRITE(File_NbLoop, Sup);
  CLOSE(File_NbLoop);
  --
  CREATE(File_Result, OUT_FILE, Result_FileName);
  WRITE(File_Result, Result2);
  WRITE(File_Result, Result5);
  WRITE(File_Result, Result8);
  CLOSE(File_Result);
  --
  CREATE(File_Term, OUT_FILE, Term_FileName);
  WRITE(File_Term, Term2);
  WRITE(File_Term, Term5);
  WRITE(File_Term, Term8);
  CLOSE(File_Term);
END Save;

--******************************************************************************
--******************************************************************************

PROCEDURE Resultat(Arctan2, Arctan5, Arctan8 : IN Fraction; BorneSup : IN My_Index) IS
  --
  -- PI/4 = Arctan(1/2) + Arctan(1/5) + Arctan(1/8)
  --
  F4 : CONSTANT Big_Unsigned := Convert(4);
  Temp_3, Exponent : Big_Unsigned;
  Result1 : Fraction;
  Result2 : Big_Unsigned;
  Char : CHARACTER := 'Z';
  File : TEXT_IO.FILE_TYPE;
  Nb_Digits : My_Index;
  Nb_Loop : CONSTANT STRING(1..18) := "Number of loops : ";
BEGIN
  WHILE (Char /= 'Y') AND THEN (Char /= 'y') AND THEN
        (Char /= 'N') AND THEN (Char /= 'n') LOOP
    PUT("Do you want the output of the result in a file ? ");
    GET(Char);
  END LOOP;

  PUT("Number of decimal digits ? ");
  GET(Nb_Digits);
  Exponent := Convert(10) ** Nb_Digits;
  Temp_3 := Exponent * 3;
  Result1 := (Arctan2 + Arctan5 + Arctan8) * F4;
  Result2 := (Get_Numerator(Result1) * Exponent) / Get_Denominator(Result1)
             - Temp_3;

  IF (Char = 'N') OR ELSE (Char = 'n') THEN
    PUT("3.");
    PUT_LINE(Result2);
  ELSE
    CREATE(File, OUT_FILE, "RESULT.TXT");
    PUT(File, "3.");
    PUT_LINE(File, Result2);
    PUT(File, Nb_Loop);
    PUT(File, BorneSup, WIDTH => 0);
    NEW_LINE(File);
    CLOSE(File);
  END IF;
END Resultat;

END Utils_Pi;

--******************************************************************************
--                                                                            **
--  File:        CALCPI.ADB                                                   **
--  Description: Calculate the number PI                                      **
--  Revision:    1.11                                                         **
--  Date:        January 19, 2001                                             **
--  Author:      Jrme Delcourt                                              **
--  Mail:        sikander@club-internet.fr                                    **
--                                                                            **
--  Copyright (c) Jrme Delcourt, 1998, 1999, 2000, 2001                     **
--  62, rue Ngrier                                                           **
--  59800 Lille                                                               **
--  FRANCE                                                                    **
--                                                                            **
--  Permission granted to use for any purpose, provided this copyright        **
--  remains attached and unmodified.                                          **
--                                                                            **
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR            **
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED            **
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       **
--                                                                            **
--******************************************************************************

WITH ModInt;           USE ModInt;
WITH TEXT_IO;          USE TEXT_IO;
WITH PC_Keyboard;      USE PC_Keyboard;
WITH My_Big;           USE My_Big;
                       USE My_Big.Unsigned_Number;
                       USE My_Big.Big_Fraction;
WITH Index_IO;         USE Index_IO;
WITH Utils_Pi;         USE Utils_Pi;


PROCEDURE CalcPi IS

-- PI/4 = Arctan(1/2) + Arctan(1/5) + Arctan(1/8)
-- Arctan(x) = SUM[[(-1)^N * x^(2N+1)] / (2N+1)]  N=0 to  infinity.

--****************************************************************************** 
-- Declarations of variables
--****************************************************************************** 

  Borne_Inf, Borne_Sup : My_Index;
  Arctan_2, Arctan_5, Arctan_8 : Fraction;
  Term2, Term5, Term8 : Big_Unsigned;

--****************************************************************************** 
-- Declarations of functions
--****************************************************************************** 

PROCEDURE Calc(Inf, Sup : IN My_Index;
               Result_2, Result_5, Result_8 : IN OUT Fraction;
               Term2, Term5, Term8 : IN OUT Big_Unsigned) IS
  -- Remark :
  -- Arctan(x) = SUM[[(-1)^N * x^(2N+1)] / (2N+1)]  N=0 to infinity.
  -- So Arctan(x) = x * SUM[[(-1)^N * x^(2N)] / (2N+1)]  N=0 to infinity.
  Temp2, Temp5, Temp8 : Fraction := Fract0;
  Temp_Term2, Temp_Term5, Temp_Term8 : Big_Unsigned := Big_Unsigned_One;
  Big2_Square : CONSTANT Big_Unsigned := Big2 ** My_Index(2);
  Big5_Square : CONSTANT Big_Unsigned := Big5 ** My_Index(2);
  Big8_Square : CONSTANT Big_Unsigned := Big8 ** My_Index(2);
  Temp : Big_Unsigned;
BEGIN
  FOR I IN Inf..Sup LOOP
    Temp := Convert(I) * 2;
    Inc(Temp);
    Temp_Term2 := Temp_Term2 * Big2_Square;
    Temp_Term5 := Temp_Term5 * Big5_Square;
    Temp_Term8 := Temp_Term8 * Big8_Square;
    Temp2 := Temp2 +
               CreateFraction(Big_Unsigned_One, Temp_Term2 * Temp, (I MOD 2)=0);
    Temp5 := Temp5 +
               CreateFraction(Big_Unsigned_One, Temp_Term5 * Temp, (I MOD 2)=0);
    Temp8 := Temp8 +
               CreateFraction(Big_Unsigned_One, Temp_Term8 * Temp, (I MOD 2)=0);
  END LOOP;
  Result_2 := Result_2 + Temp2 / Term2;
  Result_5 := Result_5 + Temp5 / Term5;
  Result_8 := Result_8 + Temp8 / Term8;
  Term2 := Term2 * Temp_Term2;
  Term5 := Term5 * Temp_Term5;
  Term8 := Term8 * Temp_Term8;
  Clear_Keyboard_Buffer;
END Calc;

--******************************************************************************
--******************************************************************************

BEGIN
  -- Initialisation
  Init(Borne_Inf, Borne_Sup, Arctan_2, Arctan_5, Arctan_8, Term2, Term5, Term8);
  -- Computation
  Calc(Borne_Inf, Borne_Sup,
       Arctan_2, Arctan_5, Arctan_8, Term2, Term5, Term8);
  -- Save the results
  Save(Borne_Sup, Arctan_2, Arctan_5, Arctan_8, Term2, Term5, Term8);
END CalcPi;
--******************************************************************************
--                                                                            **
--  File:        CALCPI2.ADB                                                  **
--  Description: Calculate the number PI                                      **
--  Revision:    1.11                                                         **
--  Date:        January 19, 2001                                             **
--  Author:      Jrme Delcourt                                              **
--  Mail:        sikander@club-internet.fr                                    **
--                                                                            **
--  Copyright (c) Jrme Delcourt, 1998, 1999, 2000, 2001                     **
--  62, rue Ngrier                                                           **
--  59800 Lille                                                               **
--  FRANCE                                                                    **
--                                                                            **
--  Permission granted to use for any purpose, provided this copyright        **
--  remains attached and unmodified.                                          **
--                                                                            **
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR            **
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED            **
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       **
--                                                                            **
--******************************************************************************

WITH ModInt;           USE ModInt;
WITH TEXT_IO;          USE TEXT_IO;
WITH PC_Keyboard;      USE PC_Keyboard;
WITH My_Big;           USE My_Big;
                       USE My_Big.Unsigned_Number;
                       USE My_Big.Big_Fraction;
WITH Index_IO;         USE Index_IO;
WITH Utils_Pi;         USE Utils_Pi;


PROCEDURE CalcPi2 IS

-- PI/4 = Arctan(1/2) + Arctan(1/5) + Arctan(1/8)
-- Arctan(x) = SUM[[(-1)^N * x^(2N+1)] / (2N+1)]  N=0 to  infinity.

--****************************************************************************** 
-- Declarations of variables
--****************************************************************************** 

  Borne_Inf, Borne_Sup : My_Index;
  Arctan_2, Arctan_5, Arctan_8 : Fraction;
  Term2, Term5, Term8 : Big_Unsigned;

--****************************************************************************** 
-- Declarations of functions
--****************************************************************************** 

PROCEDURE Calc(Inf : IN My_Index; Sup : OUT My_Index;
               Result_2, Result_5, Result_8 : IN OUT Fraction;
               Term2, Term5, Term8 : IN OUT Big_Unsigned) IS
  -- Remark :
  -- Arctan(x) = SUM[[(-1)^N * x^(2N+1)] / (2N+1)]  N=0 to infinity.
  -- So Arctan(x) = x * SUM[[(-1)^N * x^(2N)] / (2N+1)]  N=0 to infinity.
  I : My_Index := Inf;
  Temp2, Temp5, Temp8 : Fraction := Fract0;
  Temp_Term2, Temp_Term5, Temp_Term8 : Big_Unsigned := Big_Unsigned_One;
  Big2_Square : CONSTANT Big_Unsigned := Big2 ** My_Index(2);
  Big5_Square : CONSTANT Big_Unsigned := Big5 ** My_Index(2);
  Big8_Square : CONSTANT Big_Unsigned := Big8 ** My_Index(2);
  Temp : Big_Unsigned;
BEGIN
  LOOP
    Temp := Convert(I) * 2;
    Inc(Temp);
    Temp_Term2 := Temp_Term2 * Big2_Square;
    Temp_Term5 := Temp_Term5 * Big5_Square;
    Temp_Term8 := Temp_Term8 * Big8_Square;
    Temp2 := Temp2 +
               CreateFraction(Big_Unsigned_One, Temp_Term2 * Temp, (I MOD 2)=0);
    Temp5 := Temp5 +
               CreateFraction(Big_Unsigned_One, Temp_Term5 * Temp, (I MOD 2)=0);
    Temp8 := Temp8 +
               CreateFraction(Big_Unsigned_One, Temp_Term8 * Temp, (I MOD 2)=0);
    IF (Is_Key_Waiting = TRUE) OR ELSE (I = My_Index'LAST) THEN
      Sup := I;
      EXIT;
    END IF;
    I := I + 1;
  END LOOP;
  Result_2 := Result_2 + Temp2 / Term2;
  Result_5 := Result_5 + Temp5 / Term5;
  Result_8 := Result_8 + Temp8 / Term8;
  Term2 := Term2 * Temp_Term2;
  Term5 := Term5 * Temp_Term5;
  Term8 := Term8 * Temp_Term8;
  Clear_Keyboard_Buffer;
EXCEPTION
  WHEN OTHERS => PUT("Problem with loop n");
                 PUT(I, WIDTH => 0);
                 RAISE;
END Calc;

--******************************************************************************
--******************************************************************************

BEGIN
  -- Initialisation
  Init(Borne_Inf, Arctan_2, Arctan_5, Arctan_8, Term2, Term5, Term8);
  -- Computation
  Calc(Borne_Inf, Borne_Sup,
       Arctan_2, Arctan_5, Arctan_8, Term2, Term5, Term8);
  -- Save the results
  Save(Borne_Sup, Arctan_2, Arctan_5, Arctan_8, Term2, Term5, Term8);
END CalcPi2;
--******************************************************************************
--                                                                            **
--  File:        RESULTAT.ADB                                                 **
--  Description: Output of the number PI                                      **
--  Revision:    1.11                                                         **
--  Date:        January 19, 2001                                             **
--  Author:      Jrme Delcourt                                              **
--  Mail:        sikander@club-internet.fr                                    **
--                                                                            **
--  Copyright (c) Jrme Delcourt, 1998, 1999, 2000, 2001                     **
--  62, rue Ngrier                                                           **
--  59800 Lille                                                               **
--  FRANCE                                                                    **
--                                                                            **
--  Permission granted to use for any purpose, provided this copyright        **
--  remains attached and unmodified.                                          **
--                                                                            **
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR            **
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED            **
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       **
--                                                                            **
--******************************************************************************

WITH ModInt;           USE ModInt;
WITH My_Big;           USE My_Big;
                       USE My_Big.Big_Fraction;
WITH Utils_Pi;         USE Utils_Pi;


PROCEDURE Resultat IS

-- PI/4 = Arctan(1/2) + Arctan(1/5) + Arctan(1/8)
-- Arctan(x) = SUM[[(-1)^N * x^(2N+1)] / (2N+1)]  N=0 to  infinity.

--****************************************************************************** 
-- Declarations of variables
--****************************************************************************** 

  Borne : My_Index;
  Arctan_2, Arctan_5, Arctan_8 : Fraction;
  Term2, Term5, Term8 : Big_Unsigned;

--******************************************************************************
--******************************************************************************

BEGIN
  -- Initialisation
  Init(Borne, Arctan_2, Arctan_5, Arctan_8, Term2, Term5, Term8);
  -- Output of the result
  Resultat(Arctan_2, Arctan_5, Arctan_8, Borne - 1);
END Resultat;
