PACKAGE ModInt IS

TYPE Byte IS MOD 2**8;
FOR Byte'SIZE USE 8;

TYPE My_Index IS MOD 2**16;
FOR My_Index'SIZE USE 16;
-- You can modify the declaration of My_Index as you want.

END ModInt;

--******************************************************************************
WITH ModInt;
WITH SEQUENTIAL_IO;

PACKAGE Byte_Sequential_IO IS NEW SEQUENTIAL_IO(ModInt.Byte);

--******************************************************************************
WITH ModInt;
WITH Direct_IO;

PACKAGE Index_Direct_IO IS NEW DIRECT_IO(ModInt.My_Index);

--******************************************************************************
WITH ModInt;
WITH TEXT_IO;


PACKAGE Index_IO IS NEW TEXT_IO.MODULAR_IO(ModInt.My_Index);

--******************************************************************************
PACKAGE File_Utils IS


SUBTYPE FileName IS STRING(1..12);


FUNCTION Does_File_Exist(Nom : IN FileName) RETURN BOOLEAN;

PROCEDURE Delete_File(Nom : FileName);


END File_Utils;

--******************************************************************************
WITH Byte_Sequential_IO;    USE Byte_Sequential_IO;


PACKAGE BODY File_Utils IS

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

FUNCTION Does_File_Exist(Nom : IN FileName) RETURN BOOLEAN IS
  Fic : Byte_Sequential_IO.FILE_TYPE;
BEGIN
  OPEN(Fic, IN_FILE, Nom);
  CLOSE(Fic);
  RETURN TRUE;
EXCEPTION
  WHEN OTHERS => RETURN FALSE;
END Does_File_Exist;

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

PROCEDURE Delete_File(Nom : FileName) IS
  Fic : Byte_Sequential_IO.FILE_TYPE;
BEGIN
  OPEN(Fic, IN_FILE, Nom);
  DELETE(Fic);             -- La procdure DELETE ferme aussi le fichier.
END Delete_File;

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

END File_Utils;

--******************************************************************************
-----------------------------------------------------------------------
--
--  File:        djgplibr.ads
--  Description: Interface to DJGPP library functions
--  Rev:         0.7
--  Date:        01-feb-98
--  Author:      Jerry van Dijk
--  Mail:        jdijk@acm.org
--
--  Copyright (c) Jerry van Dijk, 1996, 1997, 1998
--  Billie Holidaystraat 28
--  2324 LK Leiden
--  THE NETHERLANDS
--  tel int + 31 71 531 4365
--
--  Permission granted to use for any purpose, provided this copyright
--  remains attached and unmodified.
--
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
--
-----------------------------------------------------------------------

with Interfaces;   use Interfaces;
with Interfaces.C; use Interfaces.C;

package DJGPP_Library is

   ---------------------------------------------
   -- NAME:    Dpmi_Regs                      --
   --                                         --
   -- PURPOSE: Simplified processor registers --
   ---------------------------------------------
   type Dpmi_Regs is
      record
         Di     : Unsigned_16;
         Di_Hi  : Unsigned_16;
         Si     : Unsigned_16;
         Si_Hi  : Unsigned_16;
         Bp     : Unsigned_16;
         Bp_Hi  : Unsigned_16;
         Res    : Unsigned_16;
         Res_Hi : Unsigned_16;
         Bx     : Unsigned_16;
         Bx_Hi  : Unsigned_16;
         Dx     : Unsigned_16;
         Dx_Hi  : Unsigned_16;
         Cx     : Unsigned_16;
         Cx_Hi  : Unsigned_16;
         Ax     : Unsigned_16;
         Ax_Hi  : Unsigned_16;
         Flags  : Unsigned_16;
         Es     : Unsigned_16;
         Ds     : Unsigned_16;
         Fs     : Unsigned_16;
         Gs     : Unsigned_16;
         Ip     : Unsigned_16;
         Cs     : Unsigned_16;
         Sp     : Unsigned_16;
         Ss     : Unsigned_16;
      end record;
   pragma Convention(C, Dpmi_Regs);


   ----------------------------------------
   -- NAME:    Go32_Info_Block           --
   --                                    --
   -- PURPOSE: Defines the extender info --
   ----------------------------------------
   type Go32_Info_Block is
      record
         Size_Of_This_Structure_In_Bytes        : Unsigned_32;
         Linear_Address_Of_Primary_Screen       : Unsigned_32;
         Linear_Address_Of_Secondary_Screen     : Unsigned_32;
         Linear_Address_Of_Transfer_Buffer      : Unsigned_32;
         Size_Of_Transfer_Buffer                : Unsigned_32;
         Pid                                    : Unsigned_32;
         Master_Interrupt_Controller_Base       : Unsigned_8;
         Slave_Interrupt_Controller_Base        : Unsigned_8;
         Selector_For_Linear_Memory             : Unsigned_16;
         Linear_Address_Of_Stub_Info_Structure  : Unsigned_32;
         Linear_Address_Of_Original_Psp         : Unsigned_32;
         Run_Mode                               : Unsigned_16;
         Run_Mode_Info                          : Unsigned_16;
      end record;
   pragma Convention(C, Go32_Info_Block);


   -------------------------------------------
   -- NAME:    Dos_DS                       --
   --                                       --
   -- PURPOSE: Conventional memory selector --
   -------------------------------------------
   Current_Info : Go32_Info_Block;
   pragma Import(C, Current_Info, "_go32_info_block");

   Dos_DS : Unsigned_16 renames Current_Info.Selector_For_Linear_Memory;


   --------------------------------------------------
   -- NAME:    Dpmi_Int                            --
   --                                              --
   -- PURPOSE: Call a real-mode interrupt          --
   --                                              --
   -- INPUTS:  Vector - Interrupt number           --
   --          Regs   - Processor registers        --
   --                                              --
   -- OUTPUTS: Regs - Modified processor registers --
   --------------------------------------------------
   procedure Dpmi_Int(Vector : in     Unsigned_16;
                      Regs   : in out Dpmi_Regs);
   pragma Import(C, Dpmi_Int, "__dpmi_int");


   ------------------------------------------------------
   -- NAME:    Farsetsel                               --
   --                                                  --
   -- PURPOSE: Preset the conventional memory selector --
   ------------------------------------------------------
   procedure Farsetsel(Selector : in Unsigned_16);
   pragma Import(C, Farsetsel, "_farsetsel");


   -----------------------------------------------------
   -- NAME:    Farpeekb                               --
   --                                                 --
   -- PURPOSE: Reads a byte from memory               --
   --                                                 --
   -- INPUTS:  Selector - The memory segment selector --
   --          Offset   - The fysical address         --
   --                                                 --
   -- RETURNS: The byte at the address                --
   -----------------------------------------------------
   function Farpeekb(Selector : in Unsigned_16;
                     Offset   : in Unsigned_32) return Unsigned_8;
   pragma Import(C, Farpeekb, "_farpeekb");


   -----------------------------------------------
   -- NAME:    Farnspeekb                       --
   --                                           --
   -- PURPOSE: Reads a byte from memory using a --
   --          preset selector                  --
   --                                           --
   -- INPUTS:  Offset - The fysical address     --
   --                                           --
   -- RETURNS: The byte at the address          --
   -----------------------------------------------
   function Farnspeekb(Offset : in Unsigned_32) return Unsigned_8;
   pragma Import(C, Farnspeekb, "_farnspeekb");


   ---------------------------------------------
   -- NAME:    Farpokeb                       --
   --                                         --
   -- PURPOSE: Writes a byte to memory        --
   --                                         --
   -- INPUTS:  Selector - The memory selector --
   --          Offset   - The fysical address --
   --          Value    - The byte to write   --
   ---------------------------------------------
   procedure Farpokeb(Selector : in Unsigned_16;
                      Offset   : in Unsigned_32;
                      Value    : in Unsigned_8);
   pragma Import(C, Farpokeb, "_farpokeb");


   ----------------------------------------------
   -- NAME:    Farnspokeb                      --
   --                                          --
   -- PURPOSE: Writes a byte to memory using a --
   --          preset selector                 --
   --                                          --
   -- INPUTS:  Offset - The fysical address    --
   --          Value  - The byte to write      --
   ----------------------------------------------
   procedure Farnspokeb(Offset : in Unsigned_32;
                        Value  : in Unsigned_8);
   pragma Import(C, Farnspokeb, "_farnspokeb");


   ---------------------------------------------
   -- NAME:    Outportb                       --
   --                                         --
   -- PURPOSE: Outputs a byte to a I/O port   --
   --                                         --
   -- INPUT:   Port - The I/O port            --
   --          Data - The byte to write to it --
   ---------------------------------------------
   procedure Outportb(Port : in Unsigned_16;
                      Data : in Unsigned_8);
   pragma Import(C, Outportb, "outportb");


   ---------------------------------------------------
   -- NAME:    Dosmemget                            --
   --                                               --
   -- PURPOSE: Transfers data from conventional     --
   --          memory to the application            --
   --                                               --
   -- INPUTS:  Offset - Fysical memory address      --
   --          Length - Number of bytes to transfer --
   --          Buffer - The destination             --
   ---------------------------------------------------
   procedure Dosmemget(Offset : in Unsigned_32;
                       Length : in Unsigned_32;
                       Buffer : char_array);
   pragma Import(C, Dosmemget, "dosmemget");


   ---------------------------------------------------
   -- NAME:    Dosmemput                            --
   --                                               --
   -- PURPOSE: Transfers data from the application  --
   --          to conventional memory               --
   --                                               --
   -- INPUTS:  Offset - Fysical memory address      --
   --          Length - Number of bytes to transfer --
   --          Buffer - The destination             --
   ---------------------------------------------------
   procedure Dosmemput(Buffer : char_array;
                       Length : Unsigned_32;
                       Offset : Unsigned_32);
   pragma Import(C, Dosmemput, "dosmemput");


end DJGPP_Library;

--******************************************************************************
-----------------------------------------------------------------------
--
--  File:        pckeyboa.ads
--  Description: Basic keyboard interface
--  Rev:         0.7
--  Date:        01-feb-98
--  Author:      Jerry van Dijk
--  Mail:        jdijk@acm.org
--
--  Copyright (c) Jerry van Dijk, 1996, 1997, 1998
--  Billie Holidaystraat 28
--  2324 LK Leiden
--  THE NETHERLANDS
--  tel int + 31 71 531 4365
--
--  Permission granted to use for any purpose, provided this copyright
--  remains attached and unmodified.
--
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
--
-----------------------------------------------------------------------


   ------------------------------------------------------------
   --                       PACKAGE CONTAINS                 --
   ------------------------------------------------------------
   -- Types:                                                 --
   --    Keyboard_Scancode - Scancode of a keypress          --
   --                                                        --
   -- Functions:                                             --
   --    Is_Key_Waiting    - Check if a key has been pressed --
   --    Get_Scancode      - Waits for a key press           --
   --    Get_Char          - Waits for a key press           --
   ------------------------------------------------------------


package PC_Keyboard is

   -------------------------------------
   -- NAME:    Keyboard_Scancode      --
   --                                 --
   -- PURPOSE: Scancode of a keypress --
   -------------------------------------
   subtype Keyboard_Scancode is Integer range 0..2**16-1;


   ----------------------------------------------------
   -- NAME:    Is_Key_Waiting                        --
   --                                                --
   -- PURPOSE: Check if a key has been pressed       --
   --                                                --
   -- RETURNS: True if key pressed, False otherwise  --
   --                                                --
   -- NOTES:   1. Leaves keypress in keyboard buffer --
   ----------------------------------------------------
   function Is_Key_Waiting return Boolean;


   -------------------------------------------
   -- NAME:    Get_Scancode                 --
   --                                       --
   -- PURPOSE: Waits for a key press        --
   --                                       --
   -- RETURNS: Key Scancode                 --
   --                                       --
   -- NOTES:   1. Control-Break not noticed --
   -------------------------------------------
   function Get_Scancode return Keyboard_Scancode;


   --------------------------------------------------
   -- NAME:    Get_Char                            --
   --                                              --
   -- PURPOSE: Waits for a key press               --
   --                                              --
   -- RETURNS: Character pressed, NUL if not ASCII --
   --                                              --
   -- NOTES:   1. Control-Break not noticed        --
   --------------------------------------------------
   function Get_Char return Character;


   --------------------------------------------------
   -- NAME:    Clear_Keyboard_Buffer               --
   --                                              --
   -- PURPOSE: Clear the keyboard buffer           --
   --                                              --
   -- RETURNS: None                                --
   --                                              --
   -- NOTES:   None                                --
   --------------------------------------------------
   procedure Clear_Keyboard_Buffer;


end PC_Keyboard;

--******************************************************************************
-----------------------------------------------------------------------
--
--  File:        pckeyboa.adb
--  Description: Basic keyboard interface
--  Rev:         0.7
--  Date:        01-feb-98
--  Author:      Jerry van Dijk
--  Mail:        jdijk@acm.org
--
--  Copyright (c) Jerry van Dijk, 1996, 1997, 1998
--  Billie Holidaystraat 28
--  2324 LK Leiden
--  THE NETHERLANDS
--  tel int + 31 71 531 4365
--
--  Permission granted to use for any purpose, provided this copyright
--  remains attached and unmodified.
--
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
--
-----------------------------------------------------------------------


   ------------------------------------------------------------
   --                       PACKAGE CONTAINS                 --
   ------------------------------------------------------------
   -- Types:                                                 --
   --    Keyboard_Scancode - Scancode of a keypress          --
   --                                                        --
   -- Functions:                                             --
   --    Is_Key_Waiting    - Check if a key has been pressed --
   --    Get_Scancode      - Waits for a key press           --
   --    Get_Char          - Waits for a key press           --
   ------------------------------------------------------------


with Interfaces;    use Interfaces;
with DJGPP_Library; use DJGPP_Library;

package body PC_Keyboard is

   ----------------------------------------------------
   -- NAME:    Is_Key_Waiting                        --
   --                                                --
   -- PURPOSE: Check if a key has been pressed       --
   --                                                --
   -- RETURNS: True if key pressed, False otherwise  --
   --                                                --
   -- NOTES:   1. Leaves keypress in keyboard buffer --
   ----------------------------------------------------

   function Is_Key_Waiting return Boolean is
      Regs   : Dpmi_Regs;
      Result : Boolean;
   begin
      Regs.Ax := 16#1100#;
      Dpmi_Int(16#16#, Regs);
      Result := ((Regs.Flags and 16#40#) = 0);
      return Result;
   end Is_Key_Waiting;


   -------------------------------------------
   -- NAME:    Get_Scancode                 --
   --                                       --
   -- PURPOSE: Waits for a key press        --
   --                                       --
   -- RETURNS: Key Scancode                 --
   --                                       --
   -- NOTES:   1. Control-Break not noticed --
   -------------------------------------------
   function Get_Scancode return Keyboard_Scancode is
      Regs : Dpmi_Regs;
   begin
      Regs.Ax := 16#0000#;
      Dpmi_Int(16#16#, Regs);
      return Keyboard_Scancode(Regs.Ax);
   end Get_Scancode;


   --------------------------------------------------
   -- NAME:    Get_Char                            --
   --                                              --
   -- PURPOSE: Waits for a key press               --
   --                                              --
   -- RETURNS: Character pressed, NUL if not ASCII --
   --                                              --
   -- NOTES:   1. Control-Break not noticed        --
   --------------------------------------------------
   function Get_Char return Character is
   begin
      return Character'VAL(Unsigned_16(Get_Scancode) and 16#00FF#);
   end Get_Char;


   --------------------------------------------------
   -- NAME:    Clear_Keyboard_Buffer               --
   --                                              --
   -- PURPOSE: Clear the keyboard buffer           --
   --                                              --
   -- RETURNS: None                                --
   --                                              --
   -- NOTES:   None                                --
   --------------------------------------------------
   procedure Clear_Keyboard_Buffer is
     Code : Keyboard_Scancode;
   begin
     while (Is_Key_Waiting = TRUE) loop
       Code := Get_Scancode;
     end loop;
   end Clear_Keyboard_Buffer;
 
end PC_Keyboard;

--******************************************************************************

WITH ModInt;        USE ModInt;
WITH Big_Number;


PACKAGE My_Big IS NEW Big_Number(My_Index, 2500);

--******************************************************************************

WITH My_Big;      USE My_Big;
WITH DIRECT_IO;


PACKAGE Fraction_Direct_IO IS NEW DIRECT_IO(Fraction);

--******************************************************************************
--                                                                            **
--  File:        UTILS_PI.ADS                                                 **
--  Description: Package specification of the package Utils_Pi                **
--  Revision:    1.23                                                         **
--  Date:        February 3, 2001                                             **
--  Author:      Jrme Delcourt                                              **
--  Mail:        sikander@club-internet.fr                                    **
--                                                                            **
--  Copyright (c) Jrme Delcourt, 1998, 1999, 2000, 2001                     **
--  62, rue Ngrier                                                           **
--  59800 Lille                                                               **
--  FRANCE                                                                    **
--                                                                            **
--  Permission granted to use for any purpose, provided this copyright        **
--  remains attached and unmodified.                                          **
--                                                                            **
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR            **
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED            **
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       **
--                                                                            **
--******************************************************************************

WITH ModInt;              USE ModInt;
WITH File_Utils;          USE File_Utils;
WITH TEXT_IO;             USE TEXT_IO;
WITH ADA.STRINGS.UNBOUNDED;   USE ADA.STRINGS.UNBOUNDED;
WITH UStrings;            USE UStrings;
WITH Index_IO;            USE Index_IO;
WITH Index_Direct_IO;     USE Index_Direct_IO;
WITH My_Big;              USE My_Big;
                          USE My_Big.String_Conversion;
                          USE My_Big.Generic_Conversion;
                          USE My_Big.Unsigned_Number;
                          USE My_Big.Big_Unsigned_IO;
                          USE My_Big.Big_Fraction;
WITH Fraction_Direct_IO;  USE Fraction_Direct_IO;


PACKAGE Utils_Pi IS

Fract_One : CONSTANT Fraction := CreateFraction(Big_Unsigned_One, Big_Unsigned_One);

FUNCTION Convert IS NEW Mod_Number2Big_Unsigned(My_Index);

PROCEDURE Init(Inf, Sup : OUT My_Index;
               Old_ArcSin, Temp_Fract : OUT Fraction);

PROCEDURE Init(Inf : OUT My_Index; Old_ArcSin, Temp_Fract : OUT Fraction);

PROCEDURE Save(Sup : IN My_Index; Old_Result, Temp_Fract : IN Fraction);

FUNCTION Result2UString(Result : IN Fraction;
                        Nb_Digits : IN My_Index) RETURN UString;

PROCEDURE Resultat(Result : IN Fraction; BorneSup : IN My_Index);


END Utils_Pi;

--******************************************************************************
--                                                                            **
--  File:        UTILS_PI.ADB                                                 **
--  Description: Package body of the package Utils_Pi                         **
--  Revision:    1.23                                                         **
--  Date:        February 3, 2001                                             **
--  Author:      Jrme Delcourt                                              **
--  Mail:        sikander@club-internet.fr                                    **
--                                                                            **
--  Copyright (c) Jrme Delcourt, 1998, 1999, 2000, 2001                     **
--  62, rue Ngrier                                                           **
--  59800 Lille                                                               **
--  FRANCE                                                                    **
--                                                                            **
--  Permission granted to use for any purpose, provided this copyright        **
--  remains attached and unmodified.                                          **
--                                                                            **
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR            **
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED            **
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       **
--                                                                            **
--******************************************************************************

PACKAGE BODY Utils_Pi IS

--****************************************************************************** 
--****************************************************************************** 

  Loop_FileName : CONSTANT FileName := "CALCLOOP.NWT";
  Result_FileName : CONSTANT FileName := "RESULTAT.NWT";
  Temp_Fract_FileName : CONSTANT FileName := "TEMPFRAC.NWT";

--****************************************************************************** 
-- Declarations of functions
--****************************************************************************** 

FUNCTION Files_Exist RETURN BOOLEAN IS
  Temp_Result : BOOLEAN;
BEGIN
  Temp_Result := Does_File_Exist(Loop_FileName)
                 AND THEN Does_File_Exist(Result_FileName)
                 AND THEN Does_File_Exist(Temp_Fract_FileName);
  RETURN Temp_Result;
END Files_Exist;

--****************************************************************************** 
--****************************************************************************** 

PROCEDURE Init(Inf, Sup : OUT My_Index; Old_ArcSin, Temp_Fract : OUT Fraction) IS
  File_NbLoop : Index_Direct_IO.FILE_TYPE;
  File_Result : Fraction_Direct_IO.FILE_TYPE;
  Nb_Loop, Old_Loop : My_Index;
BEGIN
  IF Files_Exist THEN
    OPEN(File_NbLoop, IN_FILE, Loop_FileName);
    READ(File_NbLoop, Old_Loop);
    CLOSE(File_NbLoop);
    OPEN(File_Result, IN_FILE, Result_FileName);
    READ(File_Result, Old_ArcSin);
    CLOSE(File_Result);
    OPEN(File_Result, IN_FILE, Temp_Fract_FileName);
    READ(File_Result, Temp_Fract);
    CLOSE(File_Result);
  ELSE
    Old_ArcSin := CreateFraction(Convert(1), Convert(2), TRUE);
    Temp_Fract := Fract_One;
    Old_Loop := 0;
  END IF;
  PUT("Number of loops already done : ");
  PUT(Old_Loop, WIDTH => 0);
  NEW_LINE;
  PUT("Number of loops to do ? ");
  GET(Nb_Loop);
  Inf := Old_Loop + 1;
  Sup := Old_Loop + Nb_Loop;
END Init;

--******************************************************************************
--******************************************************************************

PROCEDURE Init(Inf : OUT My_Index; Old_ArcSin, Temp_Fract : OUT Fraction) IS
  File_NbLoop : Index_Direct_IO.FILE_TYPE;
  File_Result : Fraction_Direct_IO.FILE_TYPE;
  Old_Loop : My_Index;
BEGIN
  IF Files_Exist THEN
    OPEN(File_NbLoop, IN_FILE, Loop_FileName);
    READ(File_NbLoop, Old_Loop);
    CLOSE(File_NbLoop);
    OPEN(File_Result, IN_FILE, Result_FileName);
    READ(File_Result, Old_ArcSin);
    CLOSE(File_Result);
    OPEN(File_Result, IN_FILE, Temp_Fract_FileName);
    READ(File_Result, Temp_Fract);
    CLOSE(File_Result);
  ELSE
    Old_ArcSin := CreateFraction(Convert(1), Convert(2), TRUE);
    Temp_Fract := Fract_One;
    Old_Loop := 0;
  END IF;
  PUT("Number of loops already done : ");
  PUT(Old_Loop, WIDTH => 0);
  NEW_LINE;
  Inf := Old_Loop + 1;
END Init;

--******************************************************************************
--******************************************************************************

PROCEDURE Save(Sup : IN My_Index; Old_Result, Temp_Fract : IN Fraction) IS
  File_NbLoop : Index_Direct_IO.FILE_TYPE;
  File_Result : Fraction_Direct_IO.FILE_TYPE;
BEGIN
  CREATE(File_NbLoop, OUT_FILE, Loop_FileName);
  WRITE(File_NbLoop, Sup);
  CLOSE(File_NbLoop);
  CREATE(File_Result, OUT_FILE, Result_FileName);
  WRITE(File_Result, Old_Result);
  CLOSE(File_Result);
  CREATE(File_Result, OUT_FILE, Temp_Fract_FileName);
  WRITE(File_Result, Temp_Fract);
  CLOSE(File_Result);
END Save;

--******************************************************************************
--******************************************************************************

FUNCTION Result2UString(Result : IN Fraction;
                        Nb_Digits : IN My_Index) RETURN UString IS
  --
  -- PI/6 = ArcSin(1/2)
  -- So PI # 6 * Result
  --
  Exponent : CONSTANT Big_Unsigned := Convert(10) ** Nb_Digits;
  Temp_3 : CONSTANT Big_Unsigned := Exponent * 3;
  Temp_Result1 : CONSTANT Big_Unsigned
        := ((Get_Numerator(Result)*Exponent*6)/Get_Denominator(Result))-Temp_3;
  Temp_Result2 : UString;
BEGIN
  Temp_Result2 := "3." & Big_Unsigned2UString(Temp_Result1);
  RETURN Temp_Result2;
END Result2UString;

--******************************************************************************
--******************************************************************************

PROCEDURE Resultat(Result : IN Fraction; BorneSup : IN My_Index) IS
  Char : CHARACTER := 'Z';
  Fichier : TEXT_IO.FILE_TYPE;
  Nb_Digits : My_Index;
  Nb_Boucles : CONSTANT STRING(1..18) := "Number of loops : ";
BEGIN
  WHILE (Char /= 'Y') AND THEN (Char /= 'y') AND THEN
        (Char /= 'N') AND THEN (Char /= 'n') LOOP
    PUT("Do you want the output of the result in a file ? ");
    GET(Char);
  END LOOP;
  PUT("Number of decimal digits ? ");
  GET(Nb_Digits);

  IF (Char = 'N') OR ELSE (Char = 'n') THEN
    PUT_LINE(Result2UString(Result, Nb_Digits));
  ELSE
    CREATE(Fichier, OUT_FILE, "RESULT.TXT");
    PUT_LINE(Fichier, Result2UString(Result, Nb_Digits));
    PUT(Fichier, Nb_Boucles);
    PUT(Fichier, BorneSup, WIDTH => 0);
    NEW_LINE(Fichier);
    CLOSE(Fichier);
  END IF;
END Resultat;

--******************************************************************************
--******************************************************************************

END Utils_Pi;

--******************************************************************************
--                                                                            **
--  File:        NEWTON.ADB                                                   **
--  Description: Calculate the number PI                                      **
--  Revision:    1.23                                                         **
--  Date:        February 3, 2001                                             **
--  Author:      Jrme Delcourt                                              **
--  Mail:        sikander@club-internet.fr                                    **
--                                                                            **
--  Copyright (c) Jrme Delcourt, 1998, 1999, 2000, 2001                     **
--  62, rue Ngrier                                                           **
--  59800 Lille                                                               **
--  FRANCE                                                                    **
--                                                                            **
--  Permission granted to use for any purpose, provided this copyright        **
--  remains attached and unmodified.                                          **
--                                                                            **
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR            **
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED            **
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       **
--                                                                            **
--******************************************************************************

WITH ModInt;           USE ModInt;
WITH TEXT_IO;          USE TEXT_IO;
WITH Index_IO;         USE Index_IO;
WITH My_Big;           USE My_Big;
                       USE My_Big.Unsigned_Number;
                       USE My_Big.Big_Fraction;
WITH Utils_Pi;         USE Utils_Pi;


PROCEDURE Newton IS

-- PI/6 = Arcsin(1/2)

--****************************************************************************** 
-- Declarations of variables
--****************************************************************************** 

  Borne_Inf, Borne_Sup : My_Index;
  Result, TempFract : Fraction;

--****************************************************************************** 
-- Declarations of functions
--****************************************************************************** 

PROCEDURE ArcSin(Result, Temp_Fract1 : IN OUT Fraction;
                 Loop_Inf, Loop_Sup : IN My_Index) IS
  Temp, Two_I_plus_1 : Big_Unsigned;
  Coeff : Big_Unsigned := (Big_Unsigned_Two ** 2) ** Loop_Inf;
  Temp_Result : Fraction := CreateFraction(Big_Unsigned_Zero, Big_Unsigned_One);
    -- Temp_Result = 0/1 = 0
  Temp_Fract2 : Fraction := Fract_One;  -- Temp_Fract = 1
BEGIN
  FOR I IN Loop_Inf..Loop_Sup LOOP
    Two_I_plus_1 := Convert(I) * 2;
    IF (I >= 2) THEN
      Temp := Two_I_plus_1;
      Temp_Fract2 := Temp_Fract2 / Temp;
      Dec(Temp);
      Temp_Fract2 := Temp_Fract2 * Temp;
    END IF;
    Inc(Two_I_plus_1);
    Coeff := Coeff * 4;     -- Now, Coeff := 2 ** (2*I + 1)
    Temp_Result := Temp_Result + (Temp_Fract2 / Two_I_plus_1) / Coeff;
  END LOOP;
  --
  Result := Result + Temp_Result * Temp_Fract1;
  Temp_Fract1 := Temp_Fract1 * Temp_Fract2;
END ArcSin;

--******************************************************************************
--******************************************************************************

BEGIN
  -- Initialisation
  Init(Borne_Inf, Borne_Sup, Result, TempFract);
  -- Computation
  ArcSin(Result, TempFract, Borne_Inf, Borne_Sup);
  -- Save the result
  Save(Borne_Sup, Result, TempFract);
END Newton;

--******************************************************************************
--                                                                            **
--  File:        NEWTON2.ADB                                                  **
--  Description: Calculate the number PI                                      **
--  Revision:    1.23                                                         **
--  Date:        February 3, 2001                                             **
--  Author:      Jrme Delcourt                                              **
--  Mail:        sikander@club-internet.fr                                    **
--                                                                            **
--  Copyright (c) Jrme Delcourt, 1998, 1999, 2000, 2001                     **
--  62, rue Ngrier                                                           **
--  59800 Lille                                                               **
--  FRANCE                                                                    **
--                                                                            **
--  Permission granted to use for any purpose, provided this copyright        **
--  remains attached and unmodified.                                          **
--                                                                            **
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR            **
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED            **
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       **
--                                                                            **
--******************************************************************************

WITH ModInt;           USE ModInt;
WITH TEXT_IO;          USE TEXT_IO;
WITH PC_Keyboard;      USE PC_Keyboard;
WITH My_Big;           USE My_Big;
                       USE My_Big.Unsigned_Number;
                       USE My_Big.Big_Fraction;
WITH Index_IO;         USE Index_IO;
WITH Utils_Pi;         USE Utils_Pi;


PROCEDURE Newton2 IS

-- PI/6 = Arcsin(1/2)

--****************************************************************************** 
-- Declarations of variables
--****************************************************************************** 

  Loop_Inf, Loop_Sup : My_Index;
  Result, TempFract : Fraction;

--****************************************************************************** 
-- Declarations of functions
--****************************************************************************** 

PROCEDURE Calc(Inf : IN My_Index; Sup : OUT My_Index;
               ArcSin, Temp_Fract1 : IN OUT Fraction) IS
  I : My_Index := Inf;
  Temp, Two_I_plus_1 : Big_Unsigned;
  Coeff : Big_Unsigned := Big_Unsigned_Two ** (2*Inf);
  Temp_Result : Fraction := CreateFraction(Big_Unsigned_Zero, Big_Unsigned_One);
    -- Temp_Result = 0/1 = 0
  Temp_Fract2 : Fraction := Fract_One;  -- Temp_Result = 1
BEGIN
  LOOP
    Two_I_plus_1 := Convert(I) * 2;
    IF (I >= 2) THEN
      Temp := Two_I_plus_1;
      Temp_Fract2 := Temp_Fract2 / Temp;
      Dec(Temp);
      Temp_Fract2 := Temp_Fract2 * Temp;
    END IF;
    Inc(Two_I_plus_1);
    Coeff := Coeff * 4;     -- Now, Coeff := 2 ** (2*I + 1);
    Temp_Result := Temp_Result + (Temp_Fract2 / Two_I_plus_1) / Coeff;
    IF (Is_Key_Waiting = TRUE) OR ELSE (I = My_Index'LAST) THEN
      Sup := I;
      EXIT;
    END IF;
    I := I + 1;
  END LOOP;
  --
  ArcSin := ArcSin + Temp_Result * Temp_Fract1;
  Temp_Fract1 := Temp_Fract1 * Temp_Fract2;
  Clear_Keyboard_Buffer;
EXCEPTION
  WHEN OTHERS => PUT("Problem with loop n");
                 PUT(I, WIDTH => 0);
                 RAISE;
END Calc;

--******************************************************************************
--******************************************************************************

BEGIN
  -- Initialisation
  Init(Loop_Inf, Result, TempFract);
  -- Computation
  Calc(Loop_Inf, Loop_Sup, Result, TempFract);
  -- Save the result
  Save(Loop_Sup, Result, TempFract);
END Newton2;

--******************************************************************************
--                                                                            **
--  File:        RESULTAT.ADB                                                 **
--  Description: Output of the number PI                                      **
--  Revision:    1.23                                                         **
--  Date:        February 3, 2001                                             **
--  Author:      Jrme Delcourt                                              **
--  Mail:        sikander@club-internet.fr                                    **
--                                                                            **
--  Copyright (c) Jrme Delcourt, 1998, 1999, 2000, 2001                     **
--  62, rue Ngrier                                                           **
--  59800 Lille                                                               **
--  FRANCE                                                                    **
--                                                                            **
--  Permission granted to use for any purpose, provided this copyright        **
--  remains attached and unmodified.                                          **
--                                                                            **
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR            **
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED            **
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       **
--                                                                            **
--******************************************************************************

WITH ModInt;           USE ModInt;
WITH TEXT_IO;          USE TEXT_IO;
WITH My_Big;           USE My_Big;
                       USE My_Big.Big_Fraction;
WITH Index_IO;         USE Index_IO;
WITH Utils_Pi;         USE Utils_Pi;


PROCEDURE Resultat IS

--****************************************************************************** 
-- Declarations of variables
--****************************************************************************** 

  Borne : My_Index;
  ArcSin, Temp_Fract : Fraction;

--******************************************************************************
--******************************************************************************

BEGIN
  -- Initialisation
  Init(Borne, ArcSin, Temp_Fract);
  -- Output of the result
  Resultat(ArcSin, Borne - 1);
END Resultat;
