--******************************************************************************
--                                                                            **
--  File:        BINUBIFR.ADB                                                 **
--  Description: Big_Number.Big_Fraction package body                         **
--  Revision:    0.28 (beta version)                                          **
--  Date:        July 8, 2001                                                 **
--  Author:      Jrme Delcourt                                              **
--  Mail:        sikander@club-internet.fr                                    **
--                                                                            **
--  Copyright (c) Jrme Delcourt, 1998, 1999, 2000, 2001                     **
--  62, rue Ngrier                                                           **
--  59800 Lille                                                               **
--  FRANCE                                                                    **
--                                                                            **
--  Permission granted to use for any purpose, provided this copyright        **
--  remains attached and unmodified.                                          **
--                                                                            **
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR            **
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED            **
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       **
--                                                                            **
--******************************************************************************

SEPARATE(Big_Number)

PACKAGE BODY Big_Fraction IS

--******************************************************************************
--******************************************************************************

FUNCTION Simplify_Fraction(x : IN Fraction) RETURN Fraction IS
  -- Here, we don't have to test for a denominator equal to zero,
  -- because it would have been already checked in the functions
  -- CreateFraction or InverseFraction.
BEGIN
  IF (x.Denominator.Last_Significant_Digit = First_Indice) AND THEN
     (x.Denominator.Mantissa(First_Indice) = 1) THEN
     -- If (x.Denominator = Big_Unsigned_One)...
    RETURN x;
  END IF;
  IF (x.Numerator.Last_Significant_Digit = First_Indice) THEN
    CASE x.Numerator.Mantissa(First_Indice) IS
      WHEN 0 => -- If (x.Numerator = Big_Unsigned_Zero)...
               RETURN (TRUE, Big_Unsigned_Zero, Big_Unsigned_One);
      WHEN 1 => -- If (x.Numerator = Big_Unsigned_One)...
               RETURN x;
      WHEN OTHERS => NULL;
    END CASE;
  END IF;
  DECLARE
    Temp : CONSTANT Big_Unsigned := GCF(x.Numerator, x.Denominator);
  BEGIN
    RETURN (x.Positive, x.Numerator / Temp, x.Denominator / Temp);
  END;
END Simplify_Fraction;

--******************************************************************************
--******************************************************************************

FUNCTION CreateFraction(Numerator, Denominator : IN Big_Unsigned;
                        Positive : IN BOOLEAN := TRUE) RETURN Fraction IS
BEGIN
  IF (Denominator.Last_Significant_Digit = First_Indice) AND THEN
     (Denominator.Mantissa(First_Indice) = 0) THEN
    -- These two tests are equivalent
    -- to the test Denominator = Big_Unsigned_Zero
    -- but it's much faster that way.
    RAISE Division_By_Zero;
  END IF;
  RETURN Simplify_Fraction((Positive, Numerator, Denominator));
END CreateFraction;

--******************************************************************************
--******************************************************************************

FUNCTION CreateFraction(Numerator, Denominator : IN Big_Signed) RETURN Fraction IS
BEGIN
  IF (Denominator.Last_Significant_Digit = First_Indice) AND THEN
     (Denominator.Mantissa(First_Indice) = 0) THEN
    -- These two tests are equivalent
    -- to the test Denominator = Big_Signed_Zero
    -- but it's much faster that way.
    RAISE Division_By_Zero;
  END IF;
  RETURN Simplify_Fraction((Numerator.Positive=Denominator.Positive,
                            Signed2Unsigned(Numerator),
                            Signed2Unsigned(Denominator)));
END CreateFraction;

--******************************************************************************
--******************************************************************************

FUNCTION Get_Numerator(x : IN Fraction) RETURN Big_Unsigned IS
BEGIN
  RETURN x.Numerator;
END Get_Numerator;

PRAGMA INLINE(Get_Numerator);

--******************************************************************************
--******************************************************************************

FUNCTION Get_Denominator(x : IN Fraction) RETURN Big_Unsigned IS
BEGIN
  RETURN x.Denominator;
END Get_Denominator;

PRAGMA INLINE(Get_Denominator);

--******************************************************************************
--******************************************************************************

FUNCTION Is_Positive(x : IN Fraction) RETURN BOOLEAN IS
BEGIN
  RETURN x.Positive;
END Is_Positive;

PRAGMA INLINE(Is_Positive);

--******************************************************************************
--******************************************************************************

PROCEDURE Swap_Fractions(x, y : IN OUT Fraction) IS
  Temp : CONSTANT Fraction := x;
BEGIN
  x := y;
  y := Temp;
END Swap_Fractions;

--******************************************************************************
--******************************************************************************

FUNCTION Inverse_Fraction(x : IN Fraction) RETURN Fraction IS
BEGIN
  -- This function returns 1/x.
  -- If the numerator of x is equal to 0, this function must raise
  -- the exception Division_By_Zero.
  IF (x.Numerator.Last_Significant_Digit = First_Indice) AND THEN
     (x.Numerator.Mantissa(First_Indice) = 0) THEN
     -- If (x.Numerator = Big_Unsigned_Zero) THEN...
     -- It's much faster that way.
    RAISE Division_By_Zero;
  ELSE
    RETURN (x.Positive, x.Denominator, x.Numerator);
  END IF;
END Inverse_Fraction;

--******************************************************************************
--******************************************************************************

FUNCTION "*"(x, y : IN Fraction) RETURN Fraction IS
  Temp1 : CONSTANT Big_Unsigned := GCF(x.Numerator, y.Denominator);
  Temp2 : CONSTANT Big_Unsigned := GCF(y.Numerator, x.Denominator);
BEGIN
  RETURN (x.Positive = y.Positive,
          (x.Numerator / Temp1) * (y.Numerator / Temp2),
          (x.Denominator / Temp2) * (y.Denominator / Temp1));
EXCEPTION
  WHEN OTHERS => RAISE Fraction_Overflow;
END "*";

--******************************************************************************
--******************************************************************************

FUNCTION "/"(x, y : IN Fraction) RETURN Fraction IS
BEGIN
  -- The function Inverse_Fraction can raise the exception "DIVISION_BY_ZERO"
  -- if y.Numerator = Big_Unsigned_Zero.
  RETURN (x * Inverse_Fraction(y));
END "/";

--******************************************************************************
--******************************************************************************

-- a/b + c/d = a/(x*b') + c/(x*d'),
-- with b = x*b' and d = x*d' and x = GCF(b,d)
-- So a/b + c/d = (a/b'+ c/d') * 1/x
-- a/b + c/d = (a*d' + c*b') / (x * b'* d')
-- x*b' = b
-- So a/b + c/d = (a*d' + c*b') / (b*d')
--
-- The fractions a/b and c/d are supposed to be already simplified,
-- so : - GCF(a, b) = 1 ==> GCF(a*d', b*d') = d'
--      - GCF(c, d) = 1 ==> GCF(c, d') = 1
--      - GCF(b', b) = b' ==> GCF(c*b', b*d') = b'
--
-- Furthermore, GCF(b', d') = 1
-- So, GCF(a*d' + c*b', b*d') = 1

FUNCTION "+"(x, y : IN Fraction) RETURN Fraction IS
  Result : Fraction;
  TempGCF : CONSTANT Big_Unsigned := GCF(x.Denominator, y.Denominator);
  Temp : CONSTANT Big_Unsigned := y.Denominator / TempGCF;
  Tmp1 : Big_Unsigned := x.Numerator * Temp;
  Tmp2 : CONSTANT Big_Unsigned := y.Numerator * (x.Denominator / TempGCF);
BEGIN
  IF (x.Positive = y.Positive) THEN
    -- The 2 fractions have the same sign.
    Result.Positive := x.Positive;
    Result.Numerator := Tmp1 + Tmp2;
  ELSE
    -- The 2 fractions have opposite signs.
    IF (Tmp1 < Tmp2) THEN
      Result.Positive := y.Positive;
      Result.Numerator := Tmp2 - Tmp1;
    ELSE   -- Tmp1 >= Tmp2
      Result.Positive := x.Positive;
      Result.Numerator := Tmp1 - Tmp2;
    END IF;
  END IF;
  IF (Result.Numerator.Last_Significant_Digit = First_Indice) THEN
    IF (Result.Numerator.Mantissa(First_Indice) = 0) THEN
       -- IF (Result.Numerator = Big_Unsigned_Zero) THEN...
      Result.Positive := TRUE;
      Result.Denominator := Big_Unsigned_One;
    ELSE
      Result.Denominator := x.Denominator * Temp;
      IF (Result.Numerator.Mantissa(First_Indice) > 1) THEN
        Tmp1 := GCF(Result.Numerator, Result.Denominator);
        Result.Numerator := Result.Numerator / Tmp1;
        Result.Denominator := Result.Denominator / Tmp1;
      END IF;
    END IF;
    RETURN Result;
  END IF;
  -- The three following lines allow to avoid some overflows.
  Tmp1 := GCF(Result.Numerator, x.Denominator);
  Result.Numerator := Result.Numerator / Tmp1;
  Result.Denominator := (x.Denominator / Tmp1) * Temp;
  RETURN Result;
EXCEPTION
  WHEN OTHERS => RAISE Fraction_Overflow;
END "+";

--******************************************************************************
--******************************************************************************

FUNCTION "-"(x, y : IN Fraction) RETURN Fraction IS
  Temp_y : Fraction := y;
  -- Just inverse the sign of y (Temp_y := -y) and then -y is added to x.
  -- So we have not to rewrite a new algorithm for this function.
BEGIN
  Temp_y.Positive := NOT (Temp_y.Positive);
  RETURN (x + Temp_y);
END "-";

--******************************************************************************
--******************************************************************************

PROCEDURE PUT(x : IN Fraction) IS
BEGIN
  IF (x.Positive = FALSE) THEN
    PUT('-');
  END IF;
  PUT(x.Numerator);
  IF (x.Denominator.Last_Significant_Digit /= First_Indice) OR ELSE
     (x.Denominator.Mantissa(First_Indice) /= 1) THEN
     -- IF (x.Denominator /= Big_Unsigned_One) THEN...
    PUT('/');
    PUT(x.Denominator);
  END IF;
END PUT;

--******************************************************************************
--******************************************************************************

PROCEDURE PUT_LINE(x : IN Fraction) IS
BEGIN
  PUT(x);
  NEW_LINE;
END PUT_LINE;

--******************************************************************************
--******************************************************************************

PROCEDURE PUT(File : IN FILE_TYPE; x : IN Fraction) IS
BEGIN
  IF (x.Positive = FALSE) THEN
    PUT(File, '-');
  END IF;
  PUT(File, x.Numerator);
  IF (x.Denominator.Last_Significant_Digit /= First_Indice) OR ELSE
     (x.Denominator.Mantissa(First_Indice) /= 1) THEN
     -- IF (x.Denominator /= Big_Unsigned_One) THEN...
    PUT(File, '/');
    PUT(File, x.Denominator);
  END IF;
END PUT;

--******************************************************************************
--******************************************************************************

PROCEDURE PUT_LINE(File : IN FILE_TYPE; x : IN Fraction) IS
BEGIN
  PUT(File, x);
  NEW_LINE(File);
END PUT_LINE;

--******************************************************************************
--******************************************************************************

FUNCTION UString2Fraction(x : IN UNBOUNDED_STRING) RETURN Fraction IS
  Numerator, Denominator : UNBOUNDED_STRING;
  Result : Fraction;
  Index : NATURAL;
BEGIN
  Index := Char_Count(x, '/');
  IF (Index = 0) THEN
    Result := CreateFraction(UString2Signed(x), Big_Signed_One);
    RETURN Result;
  ELSIF (Index = 1) THEN
    Index := First_Matching_Char_Position(x, '/');
    Numerator := Copy(x, 1, Index - 1);
    Denominator := Copy(x, Index + 1, Length(x)- Index);
    IF (LENGTH(Numerator) = 0) OR ELSE (LENGTH(Denominator) = 0) THEN
      RAISE Parameter_Error;
    END IF;
    Result := CreateFraction(UString2Signed(Numerator),
                             UString2Signed(Denominator));
    RETURN Result;
  ELSE
    RAISE Parameter_Error;
  END IF;
EXCEPTION
  WHEN OTHERS => RAISE Parameter_Error;
END UString2Fraction;

--******************************************************************************
--******************************************************************************

PROCEDURE GET_LINE(x : OUT Fraction) IS
  Temp_String : UNBOUNDED_STRING;
BEGIN
  GET_LINE(Temp_String);
  x := UString2Fraction(Temp_String);
END GET_LINE;

--******************************************************************************
--******************************************************************************

PROCEDURE GET_LINE(File : IN FILE_TYPE; x : OUT Fraction) IS
  Temp_String : UNBOUNDED_STRING;
BEGIN
  GET_LINE(File, Temp_String);
  x := UString2Fraction(Temp_String);
END GET_LINE;

--******************************************************************************
--******************************************************************************

FUNCTION "**"(x : IN Fraction; y : IN Big_Unsigned) RETURN Fraction IS
BEGIN
  IF (y.Last_Significant_Digit = First_Indice) THEN
    IF (y.Mantissa(First_Indice) = 0) THEN
      -- IF (y = Big_Unsigned_Zero) THEN ...
      RETURN (TRUE, Big_Unsigned_One, Big_Unsigned_One);     -- Returns 1/1.
    ELSIF (y.Mantissa(First_Indice) = 1) THEN
      RETURN x;
    END IF; 
  END IF;
  DECLARE
    Temp_Result : Fraction;
  BEGIN
    Temp_Result.Numerator := x.Numerator ** y;
    Temp_Result.Denominator := x.Denominator ** y;
    Temp_Result.Positive := x.Positive
                                   OR ELSE (y.Mantissa(First_Indice) MOD 2 = 0);
    -- Note : if x is positive, the result is always positive.
    -- If x is negative, the result is positive if the exponent is a multiple
    -- of 2.
    RETURN Temp_Result;
  END;
EXCEPTION
  WHEN OTHERS => RAISE Fraction_Overflow;
END "**";

--******************************************************************************
--******************************************************************************

FUNCTION "**"(x : IN Fraction; y : IN Big_Signed) RETURN Fraction IS
  Result : Fraction;
BEGIN
  Result := x ** Signed2Unsigned(y);
  IF (y.Positive = FALSE) THEN
    Result := Inverse_Fraction(Result);
  END IF;
  RETURN Result;
END "**";

--******************************************************************************
--******************************************************************************

FUNCTION "**"(x : IN Fraction; y : IN My_Type) RETURN Fraction IS
BEGIN
  IF (y = 0) THEN
    RETURN (TRUE, Big_Unsigned_One, Big_Unsigned_One);     -- Returns 1/1.
  ELSIF (y = 1) THEN
    RETURN x;
  END IF; 
  DECLARE
    Temp_Result : Fraction;
  BEGIN
    Temp_Result.Numerator := x.Numerator ** y;
    Temp_Result.Denominator := x.Denominator ** y;
    Temp_Result.Positive := x.Positive OR ELSE (y MOD 2 = 0);
    -- Note : if x is positive, the result is always positive.
    -- If x is negative, the result is positive if the exponent is a multiple
    -- of 2.
    RETURN Temp_Result;
  END;
EXCEPTION
  WHEN OTHERS => RAISE Fraction_Overflow;
END "**";

--******************************************************************************
--******************************************************************************

FUNCTION "**"(x : IN Fraction; y : IN Index_Type) RETURN Fraction IS
BEGIN
  IF (y = 0) THEN
    RETURN (TRUE, Big_Unsigned_One, Big_Unsigned_One);     -- Returns 1/1.
  ELSIF (y = 1) THEN
    RETURN x;
  END IF; 
  DECLARE
    Temp_Result : Fraction;
  BEGIN
    Temp_Result.Numerator := x.Numerator ** y;
    Temp_Result.Denominator := x.Denominator ** y;
    Temp_Result.Positive := x.Positive OR ELSE (y MOD 2 = 0);
    -- Note : if x is positive, the result is always positive.
    -- If x is negative, the result is positive if the exponent is a multiple
    -- of 2.
    RETURN Temp_Result;
  END;
EXCEPTION
  WHEN OTHERS => RAISE Fraction_Overflow;
END "**";

--******************************************************************************
--******************************************************************************

FUNCTION "*"(x : IN Fraction; y : IN Big_Unsigned) RETURN Fraction IS
BEGIN
  IF (y.Last_Significant_Digit = First_Indice) THEN
    IF (y.Mantissa(First_Indice) = 0) THEN
      RETURN (x.Positive, Big_Unsigned_Zero, Big_Unsigned_One);
    ELSIF (y.Mantissa(First_Indice) = 1) THEN
      RETURN x;
    END IF;
  END IF;
  DECLARE
    Temp : CONSTANT Big_Unsigned := GCF(y, x.Denominator);
  BEGIN
    RETURN (x.Positive, (y / Temp) * x.Numerator, x.Denominator / Temp);
  END;
EXCEPTION
  WHEN OTHERS => RAISE Fraction_Overflow;
END "*";

--******************************************************************************
--******************************************************************************

FUNCTION "*"(x : IN Fraction; y : IN Big_Signed) RETURN Fraction IS
  Result : Fraction;
  Temp_y : CONSTANT Big_Unsigned := Signed2Unsigned(y);
  Temp : CONSTANT Big_Unsigned := GCF(Temp_y, x.Denominator);
BEGIN
  IF (y.Last_Significant_Digit = First_Indice) THEN
    IF (y.Mantissa(First_Indice) = 0) THEN
      RETURN (x.Positive, Big_Unsigned_Zero, Big_Unsigned_One);
    ELSIF (y.Mantissa(First_Indice) = 1) THEN
      RETURN ((x.Positive = y.Positive), x.Numerator, x.Denominator);
    END IF;
  END IF;
  Result.Positive := (x.Positive = y.Positive);
  Result.Numerator := x.Numerator * (Temp_y / Temp);
  Result.Denominator := x.Denominator / Temp;
  RETURN Result;
EXCEPTION
  WHEN OTHERS => RAISE Fraction_Overflow;
END "*";

--******************************************************************************
--******************************************************************************

FUNCTION "/"(x : IN Fraction; y : IN Big_Unsigned) RETURN Fraction IS
  Temp : Big_Unsigned;
BEGIN
  IF (y.Last_Significant_Digit = First_Indice) AND THEN
     (y.Mantissa(First_Indice) = 0) THEN
    RAISE Division_By_Zero;
  END IF;
  Temp := GCF(x.Numerator, y);
  RETURN (x.Positive, x.Numerator / Temp, (y / Temp) * x.Denominator);
EXCEPTION
  WHEN OTHERS => RAISE Fraction_Overflow;
END "/";

--******************************************************************************
--******************************************************************************

FUNCTION "/"(x : IN Fraction; y : IN Big_Signed) RETURN Fraction IS
  Temp : Big_Unsigned;
  Temp_y : CONSTANT Big_Unsigned := Signed2Unsigned(y);
BEGIN
  IF (y.Last_Significant_Digit = First_Indice) AND THEN
     (y.Mantissa(First_Indice) = 0) THEN
    RAISE Division_By_Zero;
  END IF;
  Temp := GCF(x.Numerator, Temp_y);
  RETURN (x.Positive = y.Positive, x.Numerator / Temp,
          (Temp_y / Temp) * x.Denominator);
EXCEPTION
  WHEN OTHERS => RAISE Fraction_Overflow;
END "/";

--******************************************************************************
--******************************************************************************

FUNCTION "<"(Left, Right : IN Fraction) RETURN BOOLEAN IS
  --
  -- Be careful : do not convert the fraction to a real type. If so, there will
  -- be a truncation, so the result of this function could be mathematically
  -- false.
  --
BEGIN
  IF (Left.Positive /= Right.Positive) THEN
    -- If the 2 fractions have different signs, the comparison is easy...
    RETURN Right.Positive;
  --
  -- Now, we must study the case where the two fractions have the same sign.
  --
  -- First, some simple cases ...
  ELSIF (Left.Denominator = Right.Denominator) THEN
    IF (Left.Numerator = Right.Numerator) THEN
      RETURN FALSE;
    ELSE
      RETURN (Left.Positive = (Left.Numerator < Right.Numerator));
    END IF;
  END IF;
  -- The case (Left.Denominator = Right.Denominator)
  -- has already been studied.
  IF (Left.Numerator <= Right.Numerator) THEN
    RETURN NOT (Left.Positive = (Left.Denominator < Right.Denominator));
  ELSIF (Left.Numerator.Last_Significant_Digit = First_Indice) AND THEN
        (Left.Numerator.Mantissa(First_Indice) = 0) THEN
    -- Remark : the situation where Left.Numerator = Right.Numerator
    -- has already been studied, so Right.Numerator /= Big_Unsigned_Zero.
    -- Furthermore, at this stade, the 2 fractions have the same sign.
    -- That means that Left < Right, so the result is always TRUE.
    -- Note : by default, a fraction with a numerator equal to zero
    -- has a sign set to Positive.
    RETURN TRUE;
  END IF;
  --
  -- Now, when the two fractions have the same sign, we must study the
  -- general case (Left.Numerator /= Right.Numerator
  -- AND Left.Denominator /= Right.Denominator).
  --
  IF (Left.Numerator < Left.Denominator) AND THEN
     (Right.Numerator < Right.Denominator) THEN
    -- If Left < Right, then Inverse_Fraction(Right) < Inverse_Fraction(Left).
    RETURN (Inverse_Fraction(Right) < Inverse_Fraction(Left));
  END IF;
  --
  -- Now, at least one of the 2 fractions has an absolute value greater than 1.
  --
  DECLARE
    Temp_x : Fraction := Left;
    Temp_y : Fraction := Right;
    Tempx, Tempy, Temp : Big_Unsigned;
  BEGIN
    --
    -- If Tempx = Tempy, then we must study further the fractions Temp_x
    -- and Temp_y.
    --
    LOOP
      Big_Div(Temp_x.Numerator, Temp_x.Denominator, Tempx, Temp_x.Numerator);
      Big_Div(Temp_y.Numerator, Temp_y.Denominator, Tempy, Temp_y.Numerator);
      -- Don't forget that Tempx and Tempy are "only" big integers, which
      -- represent the truncated value of Left and Right, respectively.
      IF (Tempx /= Tempy) THEN
        RETURN (Temp_x.Positive = (Tempx < Tempy));
      END IF;
      Temp_x := Simplify_Fraction(Temp_x);
      -- Now, Temp_x < 1
      Temp_y := Simplify_Fraction(Temp_y);
      -- Now, Temp_y < 1
      IF (Temp_x.Numerator > Temp_y.Numerator) THEN
        Temp := Big_Unsigned_Last / Temp_x.Numerator;
      ELSE
        -- Temp_x.Numerator <= Temp_y.Numerator
        Temp := Big_Unsigned_Last / Temp_y.Numerator;
      END IF;
      -- Temp_x.Numerator <= Big_Unsigned_Last and
      -- Temp_y.Numerator <= Big_Unsigned_Last,
      -- so Temp >= Big_Unsigned_One.
      IF (Temp.Last_Significant_Digit > First_Indice) OR ELSE
         (Temp.Mantissa(First_Indice) > 1) THEN
        -- These two tests are the same test than (Temp > Big_Unsigned_One),
        -- but they are much faster.
        -- In this case, we can multiply Temp_x and Temp_y by Temp,
        -- without raising Big_Number_Overflow.
        -- Then, we can calculate Tempx and Tempy
        -- for a more precise comparison.
        Temp_x := Temp_x * Temp;
        Temp_y := Temp_y * Temp;
      ELSE
        -- Here, Tempx = Tempy, and Temp = Big_Unsigned_One.
        -- If Left < Right,
        -- then Inverse_Fraction(Left) > Inverse_Fraction(Right).
        RETURN (Inverse_Fraction(Right) < Inverse_Fraction(Left));
      END IF;
    END LOOP;
  END;     -- END of DECLARE
END "<";

--******************************************************************************
--******************************************************************************

FUNCTION ">"(Left, Right : IN Fraction) RETURN BOOLEAN IS
BEGIN
  -- If you want to know the result of Left > Right,
  -- just return the result of Right < Left...
  RETURN (Left < Right);
END ">";

--******************************************************************************
--******************************************************************************

FUNCTION "<="(Left, Right : IN Fraction) RETURN BOOLEAN IS
  -- Left <= Right is the same than (NOT (Left > Right)),
  -- so it's the same than (NOT (Right < Left)).
  Temp_Result : CONSTANT BOOLEAN := NOT (Right < Left);
BEGIN
  RETURN Temp_Result;
END "<=";

--******************************************************************************
--******************************************************************************

FUNCTION ">="(Left, Right : IN Fraction) RETURN BOOLEAN IS
  -- Left >= Right is the same than (NOT (Left < Right)).
  Temp_Result : CONSTANT BOOLEAN := NOT (Left < Right);
BEGIN
  RETURN Temp_Result;
END ">=";

--******************************************************************************
--******************************************************************************

FUNCTION Opposite(x : IN Fraction) RETURN Fraction IS
  Temp_Result : Fraction := x;
BEGIN
  IF (Temp_Result.Numerator.Last_Significant_Digit /= First_Indice) OR ELSE
     (Temp_Result.Numerator.Mantissa(First_Indice) /= 0) THEN
    -- If Temp_Result /= 0, then inverse the sign.
    -- Note that if x = 0, by convention, the sign of x is always
    -- set to positive.
    Temp_Result.Positive := NOT(Temp_Result.Positive);
  END IF;
  RETURN Temp_Result;
END Opposite;

--******************************************************************************
--******************************************************************************

FUNCTION Absolute(x : IN Fraction) RETURN Fraction IS
BEGIN
  IF (x.Positive = TRUE) THEN
    RETURN x;
  ELSE
    RETURN Opposite(x);
  END IF;
END Absolute;

--******************************************************************************
--******************************************************************************

FUNCTION Is_Null(x : IN Fraction) RETURN BOOLEAN IS
  Temp_Result : BOOLEAN;
BEGIN
  Temp_Result := (x.Numerator.Last_Significant_Digit = First_Indice) AND THEN
                 (x.Numerator.Mantissa(First_Indice) = 0);
  RETURN Temp_Result;
END Is_Null;

--******************************************************************************
--******************************************************************************

PROCEDURE Float_in_UString2Fraction(x : IN UNBOUNDED_STRING;
                                    Nb_Periods : IN POSITIVE;
                                    Result : OUT Fraction;
                                    Error : OUT BOOLEAN) IS

  FUNCTION Convert IS NEW Int_Number2Big_Unsigned(NATURAL);

  FUNCTION Find_a_Period(s : IN UNBOUNDED_STRING) RETURN UNBOUNDED_STRING IS
    -- This function looks for a period in the number represented by
    -- the string s.
    -- If a period is found, it is returned in a string.
    -- If no period is found, the returned result is the string s.
    --
    -- Here, the period returned is not the smaller.
    -- For example, in the number "1345345345345", the returned period is
    -- "345345", even if the smaller period is "345".
    --
    s1, s2 : UNBOUNDED_STRING;
    Size_s : CONSTANT POSITIVE := LENGTH(s);
    Temp_Size, Index1, Index2 : NATURAL;
  BEGIN
    IF (Size_s < 2) THEN
      RETURN s;
    ELSE
      Temp_Size := Size_s / 2;
      -- Begin to look for the greatest period which could be found
      LOOP
        Index1 := Size_s - 2 * Temp_Size + 1;
        Index2 := Size_s - Temp_Size + 1;
        s1 := Copy(s, Index1, Temp_Size);
        s2 := Copy(s, Index2, Temp_Size);
        Temp_Size := Temp_Size - 1;
        EXIT WHEN (s1 = s2) OR ELSE (Temp_Size = 0);
      END LOOP;

      IF (s1 /= s2) THEN
        RETURN s;
      ELSE
        RETURN s2;
      END IF;
    END IF;
  END Find_a_Period;

  ------------------------------------------------------------------------------

  FUNCTION Count_the_Number_of_Periods(s, period : IN UNBOUNDED_STRING)
                                       RETURN NATURAL IS
  BEGIN
    IF (s = period) THEN
      RETURN 1;
    END IF;
    DECLARE
      Index : NATURAL;
      Number : NATURAL := 1;
      String_Size : CONSTANT NATURAL := LENGTH(s);
      Period_Size : CONSTANT NATURAL := LENGTH(period);
      Max_Number : CONSTANT NATURAL := String_Size / Period_Size;
      Temp_String : UNBOUNDED_STRING;
    BEGIN
      LOOP
        EXIT WHEN (Number > Max_Number);
        Index := String_Size - Number * Period_Size + 1;
        Temp_String := Copy(s,Index,Period_Size);
        IF (Temp_String = period) THEN
          Number := Number + 1;
        ELSE
          EXIT;
        END IF;
      END LOOP;
      RETURN (Number - 1);
    END;
  END Count_the_Number_of_Periods;

  ------------------------------------------------------------------------------

  FUNCTION Find_the_Smallest_Period(s : IN UNBOUNDED_STRING)
                                    RETURN UNBOUNDED_STRING IS
    Max_Period, s1, s2, temp : UNBOUNDED_STRING;
    Size_Max_Period : POSITIVE;
    Index1, Index2 : NATURAL;
    Temp_Size : NATURAL := 1;
    Count : NATURAL;
    -- Be careful. Suppose that s = "1.852333852333852333".
    -- "852333" is the real period, not "3".
  BEGIN
    Max_Period := Find_a_Period(s);
    IF (s = Max_Period) THEN
      RETURN s;
    ELSE
      Size_Max_Period := LENGTH(Max_Period);
      -- It is certain that Max_Period is a period.
      -- Try to find a smaller period (s2).
      LOOP
        Index1 := Size_Max_Period - 2 * Temp_Size + 1;
        Index2 := Size_Max_Period - Temp_Size + 1;
        s1 := Copy(Max_Period, Index1, Temp_Size);
        s2 := Copy(Max_Period, Index2, Temp_Size);
        IF (s1 = s2) THEN
          -- Verify that Max_Period is a repetition of s1.
          Count := Count_the_Number_of_Periods(Max_Period, s1);
          Temp := NULL_UNBOUNDED_STRING;
          FOR I IN 1..Count LOOP
            Temp := Temp & s1;
          END LOOP;
          -- If Max_Period is a repetition of s1, Temp should be equal to Max_Period.
          IF (Max_Period = Temp) THEN
            RETURN s1;
          END IF;
        END IF;
        IF (Temp_Size < Size_Max_Period / 2) THEN
          -- There must be at least 2 small periods in Max_Period,
          -- so the size of this small period is smaller or equal
          -- to Size_Max_Period / 2.
          Temp_Size := Temp_Size + 1;
        ELSE
          RETURN Max_Period;
        END IF;
      END LOOP;
    END IF;
  END Find_the_Smallest_Period;

  ------------------------------------------------------------------------------

  FUNCTION Prepare_String(Str : IN UNBOUNDED_STRING) RETURN UNBOUNDED_STRING IS
    Temp_Result : UNBOUNDED_STRING := Del_Spaces(Str);
  BEGIN
    Temp_Result := Char_Replace(Temp_Result, 'e', 'E');
    Temp_Result := Char_Replace(Temp_Result, ',', '.');
    RETURN Temp_Result;
  END Prepare_String;

  ------------------------------------------------------------------------------

  FUNCTION Numeric_String(Str : IN UNBOUNDED_STRING) RETURN BOOLEAN IS
    I : NATURAL := 1;
    Temp_Str : CONSTANT UNBOUNDED_STRING := Del_Character(Str, '.');
    String_Length : CONSTANT NATURAL := LENGTH(Temp_Str);
    -- The number of '.' character has been checked in
    -- the procedure "Verify_String_And_Separe_Exponent_And_Mantissa",
    -- so we don't have to check it now (we are sure that there is 0
    -- or 1 '.' in this string).
  BEGIN
    LOOP
      EXIT WHEN (I > String_Length);
      IF (ELEMENT(Temp_Str, I) NOT IN '0'..'9') THEN
        RETURN FALSE;
      END IF;
      I := I + 1;
    END LOOP;
    RETURN TRUE;
  END Numeric_String;

  ------------------------------------------------------------------------------

  FUNCTION Trim_Leading_Zero(Str : IN UNBOUNDED_STRING) RETURN UNBOUNDED_STRING IS
    Temp_String : UNBOUNDED_STRING := NULL_UNBOUNDED_STRING;
    I : NATURAL := 0;
    Length_Str : NATURAL := LENGTH(Str);
    Char : CHARACTER;
  BEGIN
    IF (Length_Str = 0) THEN
      RETURN Temp_String;
    END IF;
    LOOP
      I := I + 1;
      EXIT WHEN (I > Length_Str);
      Char := ELEMENT(Str, I);
      EXIT WHEN (Char /= '0');
    END LOOP;
    Temp_String := Copy(Str, I, Length_Str - I + 1);
    IF (ELEMENT(Temp_String, 1) = '.') THEN
      Temp_String := '0' & Temp_String;
    END IF;
    I := First_Matching_Char_Position(Str, '.');
    IF (I /= 0) THEN
      Length_Str := LENGTH(Temp_String);
      LOOP
        Char := ELEMENT(Str, Length_Str);
        IF (Char = '0') OR ELSE (Char = '.') THEN
          Length_Str := Length_Str - 1;
          Temp_String := Copy(Temp_String, 1, Length_Str);
        ELSE
          EXIT;
        END IF;
      END LOOP;
    END IF;
    RETURN Temp_String;
  END Trim_Leading_Zero;

  ------------------------------------------------------------------------------

  PROCEDURE Verify_String_And_Separe_Exponent_And_Mantissa
                             (Str : IN UNBOUNDED_STRING;
                              Exponent_Str, Mantissa_Str : OUT UNBOUNDED_STRING;
                              Positive, Valid : OUT BOOLEAN) IS
    -- Verify if Str is a valid string for conversion in a fraction.
    Mantissa, Exponent : UNBOUNDED_STRING := NULL_UNBOUNDED_STRING;
    Count_E : CONSTANT NATURAL := Char_Count(Str, 'E');
    I : NATURAL;
    Temp_Valid : BOOLEAN;
  BEGIN
    Mantissa_Str := Mantissa;
    Exponent_Str := Exponent;
    IF (Count_E > 1) THEN
      Valid := FALSE;
      RETURN;
    END IF;
    IF (Count_E = 1) THEN
      I := First_Matching_Char_Position(Str, 'E');
      Mantissa := Copy(Str, 1, I-1);
      Exponent := Copy(Str, I+1, LENGTH(Str) - I);
    ELSE
      Mantissa := Str;
    END IF;
    Mantissa := Trim_Leading_Zero(Mantissa);
    Exponent := Trim_Leading_Zero(Exponent);
    -- You can't have more than one '.', or more than 1 sign (+ or -)
    -- in the mantissa.
    IF         (Char_Count(Mantissa, '.') > 1)
       OR ELSE (Char_Count(Mantissa, '+') + Char_Count(Mantissa, '-') > 1) THEN
      Valid := FALSE;
      RETURN;
    END IF;
    -- You can't have more than 1 sign (+ or -) in the exponent, where
    -- no '.' is authorized.
    IF         (Char_Count(Exponent, '.') > 0)
       OR ELSE (Char_Count(Exponent, '+') + Char_Count(Exponent, '-') > 1) THEN
      Valid := FALSE;
      RETURN;
    END IF;
    IF (Char_Count(Mantissa, '-') = 1) THEN
      Positive := FALSE;
    ELSE
      Positive := TRUE;
    END IF;
    Mantissa := Del_Character(Mantissa, '-');
    Mantissa := Del_Character(Mantissa, '+');
    Exponent := Del_Character(Exponent, '-');
    Exponent := Del_Character(Exponent, '+');
    Temp_Valid := (Numeric_String(Mantissa) AND THEN Numeric_String(Exponent));
    IF (Temp_Valid = TRUE) THEN
      Mantissa_Str := Mantissa;
      Exponent_Str := Exponent;
    END IF;
    Valid := Temp_Valid;
  END Verify_String_And_Separe_Exponent_And_Mantissa;

  ------------------------------------------------------------------------------

  FUNCTION Calc_Value(x : IN UNBOUNDED_STRING) RETURN Big_Unsigned IS
    Temp_Result : Big_Unsigned := Big_Unsigned_Zero;
    Ten : CONSTANT Big_Unsigned := Convert(10);
  BEGIN
    FOR I IN 1..LENGTH(x) LOOP
      Temp_Result := Temp_Result * Ten
                    + Convert(CHARACTER'POS(ELEMENT(x,I)) - CHARACTER'POS('0'));
    END LOOP;
    RETURN Temp_Result;
  END Calc_Value;
  
  ------------------------------------------------------------------------------

  PROCEDURE Calculate_The_Exponent(Exponent_String : IN UNBOUNDED_STRING;
                                   Mantissa_String : IN OUT UNBOUNDED_STRING;
                                   Exponent_Positive : OUT BOOLEAN;
                                   Exponent_Absolute_Value : OUT Big_Unsigned) IS
    I : NATURAL;
    Temp_Exponent_Positive : BOOLEAN := TRUE;
    Temp_Unsigned, Temp_Exponent : Big_Unsigned;
    Exponent_Str : UNBOUNDED_STRING := Exponent_String;
  BEGIN
    -- Look for the sign and the value of the exponent
    --
    IF (LENGTH(Exponent_Str) /= 0) AND THEN (ELEMENT(Exponent_Str, 1) = '-') THEN
      Temp_Exponent_Positive := FALSE;
      Exponent_Str := Copy(Exponent_Str, 2, LENGTH(Exponent_Str) - 1);
    END IF;
    Temp_Exponent := Calc_Value(Exponent_Str);
    --
    -- If there is a "." in the mantissa, check its position and modify
    -- the exponent, because the '.' will be removed.
    --
    I := First_Matching_Char_Position(Mantissa_String, '.');
    IF (I /= 0) THEN                       -- There is a '.' in Mantissa_String
      Temp_Unsigned := Convert(LENGTH(Mantissa_String) - I);
      IF (Temp_Exponent_Positive = FALSE) THEN
        Temp_Exponent := Temp_Exponent + Temp_Unsigned;
      ELSE
        IF (Temp_Unsigned > Temp_Exponent) THEN
          Temp_Exponent_Positive := FALSE;
          Temp_Exponent := Temp_Unsigned - Temp_Exponent;
        ELSE
          Temp_Exponent := Temp_Exponent - Temp_Unsigned;
        END IF;
      END IF;
      --
      -- Now, the exponent is computed.
      -- The "." can be deleted in the mantissa.
      --
      Mantissa_String := Copy(Mantissa_String, 1, I-1) &
                         Copy(Mantissa_String, I+1, LENGTH(Mantissa_String) -I);
    END IF;
    Exponent_Positive := Temp_Exponent_Positive;
    Exponent_Absolute_Value := Temp_Exponent;
  END Calculate_The_Exponent;

  ------------------------------------------------------------------------------

BEGIN
  DECLARE
    -- In procedures which require a lot of lines of code, I prefer to declare
    -- the variables only when it is necessary. This method eliminates a cause
    -- of side effects during the sub-procedures.
    Temp_Str : UNBOUNDED_STRING := Prepare_String(x);
    I : NATURAL;
    Temp_Unsigned, Exponent : Big_Unsigned;
    Temp, Exponent_Str, Mantissa_Str, Period : UNBOUNDED_STRING;
    Valid, Exponent_Positive : BOOLEAN;
    Temp_Result : Fraction;
  BEGIN
    -- If Nb_Periods < 2 or if the string is not valid, then exit.
    -- Remark : the function
    Verify_String_And_Separe_Exponent_And_Mantissa(Temp_Str, Exponent_Str,
                                                   Mantissa_Str,
                                                   Temp_Result.Positive, Valid);
    ------------------------------------------
    -- In case the parameters are not valid --
    ------------------------------------------
    IF (Valid = FALSE) THEN
      Error := TRUE;
      -- Set Result to an impossible value (0/0),
      -- so, in case of error, and if the parameter Error is not checked
      -- after a call to this procedure, the program will hang.
      Result.Numerator := Big_Unsigned_Zero;
      Result.Denominator := Big_Unsigned_Zero;
      Result.Positive := TRUE;
      RETURN;
    ELSE
      Error := FALSE;
    END IF;

    -- Now, Error = FALSE, and there is no situation that could change it.
    -- Note that the sign of the result has already been set.
    ---------------------------------
    -- Computation of the exponent --
    ---------------------------------
    Calculate_The_Exponent(Exponent_Str, Mantissa_Str,
                           Exponent_Positive, Exponent);
    -------------------------------------------------------------
    -- Now, search for the period                              --
    -- and calculate the number of repetitions of this period. --
    -------------------------------------------------------------
    Period := Find_the_Smallest_Period(Mantissa_Str);
    I := Count_the_Number_of_Periods(Mantissa_Str, Period);
    -----------------------------------------------  
    -- Now, we can really calculate the fraction --
    -----------------------------------------------  
    IF (Period = NULL_UNBOUNDED_STRING) OR ELSE (I < Nb_Periods) THEN
      Temp_Result.Numerator := Calc_Value(Mantissa_Str);
      Temp_Result.Denominator := Convert(1);
    ELSE
      Temp := Copy(Mantissa_Str, 1, LENGTH(Mantissa_Str) - I * LENGTH(Period));
      Temp_Result.Numerator := Calc_Value(Temp & Period) - Calc_Value(Temp);
      Temp := NULL_UNBOUNDED_STRING;
      FOR J IN 1..LENGTH(Period) LOOP
        Temp := Temp & "9";
      END LOOP;
      Temp_Result.Denominator := Calc_Value(Temp);
      Temp_Result := Simplify_Fraction(Temp_Result);
      --
      -- We must modify the exponent :
      --   - the length of the numerator has (I-1)*LENGTH(Period) digits less;
      --   - the length of the denominator has (LENGTH(Period)) digits more
      -- So we have to add I*LENGTH(Period) to the exponent.
      --
      Temp_Unsigned := Convert(I) * Convert(LENGTH(Period));
      IF (Exponent_Positive = TRUE) THEN
        Exponent := Exponent + Temp_Unsigned;
      ELSE
        IF (Temp_Unsigned > Exponent) THEN
          Exponent_Positive := TRUE;
          Exponent := Temp_Unsigned - Exponent;
        ELSE
          Exponent := Exponent - Temp_Unsigned;
        END IF;
      END IF;
    END IF;
    --
    -- If Exponent /= 0 ...
    --
    IF (Exponent.Last_Significant_Digit /= First_Indice)
       OR ELSE (Exponent.Mantissa(First_Indice) /= 0) THEN
       -- This combination of tests is equivalent to the
       -- test (Exponent /= Big_Unsigned_Zero), but it is
       -- executed faster.
      Exponent := Convert(10) ** Exponent;
      IF (Exponent_Positive = TRUE) THEN
        Temp_Result.Numerator := Temp_Result.Numerator * Exponent;
      ELSE
        Temp_Result.Denominator := Temp_Result.Denominator * Exponent;
      END IF;
      -- The fraction has been modified, so enventually it could be simplified.
      Temp_Result := Simplify_Fraction(Temp_Result);
    END IF;
    Result := Temp_Result;
  END;                   -- END DECLARE
END Float_in_UString2Fraction;

--******************************************************************************
--******************************************************************************

END Big_Fraction;
