--
-- Permission to use, copy, modify, and distribute this software and its
-- documentation for any purpose and without fee is hereby granted,
-- provided that the above copyright and authorship notice appear in all
-- copies and that both that copyright notice and this permission notice
-- appear in supporting documentation.
-- 
-- The ARA makes no representations about the suitability of this software
-- for any purpose.  It is provided "as is" without express
-- or implied warranty.
-- 
--******************************************************************************
--                                                                            **
--  File:        BNBIUNRA.ADB                                                 **
--  Description: Big_Number.Big_Unsigned_RANDOM package body                  **
--  Revision:    0.28 (beta version)                                          **
--  Date:        July 8, 2001                                                 **
--  Author:      Jrme Delcourt                                              **
--  Mail:        sikander@club-internet.fr                                    **
--                                                                            **
--  Copyright (c) Jrme Delcourt, 1998, 1999, 2000, 2001                     **
--  62, rue Ngrier                                                           **
--  59800 Lille                                                               **
--  FRANCE                                                                    **
--                                                                            **
--  Permission granted to use for any purpose, provided this copyright        **
--  remains attached and unmodified.                                          **
--                                                                            **
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR            **
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED            **
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       **
--                                                                            **
--******************************************************************************

SEPARATE(Big_Number)

PACKAGE BODY Big_Unsigned_Random IS


-- In this package, there are in fact two separate pseudo-random number
-- generators (PRNG) : one for index, one for digits. They are combined
-- to get a PRNG for Big_Unsigned : so, the last significant digit
-- of the Big_Unsigned is set independantly of the digits.

My_Type_Random_Seed : Largest_Unsigned;
Index_Random_Seed : Largest_Unsigned;

A : CONSTANT Largest_Unsigned := 125;      -- 125 = 5^3
B : CONSTANT Largest_Unsigned := Largest_Unsigned'LAST / A;
  -- Be careful : B must be greater or equal to Nb_Item and Base.

PROCEDURE Randomize IS
  t : TIME;
  Year : YEAR_NUMBER;
  Month : MONTH_NUMBER;    -- 1 to 12
  Day : DAY_NUMBER;        -- 0 to 31
  Seconds : DAY_DURATION;  -- 0 to 86399
BEGIN
  t := CLOCK;
  SPLIT(t, Year, Month, Day, Seconds);
  My_Type_Random_Seed := Largest_Unsigned(Seconds) + Largest_Unsigned(Day)
                             + Largest_Unsigned(Month) + Largest_Unsigned(Year);
  Index_Random_Seed := 2 * (2 * Largest_Unsigned(Seconds) +
                            (Largest_Unsigned(Day) MOD Largest_Unsigned(Month)))
                       + 1;
  IF ((My_Type_Random_Seed MOD 2) = 0) THEN
    My_Type_Random_Seed := My_Type_Random_Seed + 1;
  END IF;
END Randomize;


FUNCTION Random(x : IN Index_Type) RETURN Index_Type IS
  Max : CONSTANT Largest_Unsigned := Largest_Unsigned(x);
BEGIN
  Index_Random_Seed := (A * Index_Random_Seed) MOD B;
  IF (Max <= Largest_Unsigned'LAST / 4) THEN
    RETURN Index_Type(Index_Random_Seed MOD (Max+1));
  ELSE
    LOOP
      Index_Random_Seed := (A * Index_Random_Seed) MOD B;
      EXIT WHEN (Index_Random_Seed <= Max);
    END LOOP;
    RETURN Index_Type(Index_Random_Seed);
  END IF;
END Random;


FUNCTION Random RETURN My_Type IS
BEGIN
  My_Type_Random_Seed := (A * My_Type_Random_Seed) MOD B;
  RETURN My_Type(My_Type_Random_Seed AND Largest_Unsigned(Greatest_Digit));
    -- Returns My_Type_Random_Seed MOD Base
END Random;


FUNCTION Random(x : IN My_Type) RETURN My_Type IS
  Result : My_Type;
BEGIN
  IF (x = 0) THEN
    RETURN 0;
  ELSIF (x < 2**(My_Type_Size / 4)) THEN
    Result := Random;
    RETURN (Result MOD (x+1));
  END IF;
  LOOP
    Result := Random;
    IF (Result <= x) THEN
      EXIT;
    END IF;
  END LOOP;
  RETURN Result;
END Random;


FUNCTION Random(x : IN Index_Type) RETURN Big_Unsigned IS
  Temp_Result : Big_Unsigned;
  Temp : Index_Type := x;
BEGIN
  FOR I IN First_Indice..Temp LOOP
    Temp_Result.Mantissa(I) := Random;
  END LOOP;
  FOR I IN (Temp + 1)..Last_Indice LOOP
    Temp_Result.Mantissa(I) := 0;
  END LOOP;
  WHILE (Temp > First_Indice) AND THEN (Temp_Result.Mantissa(Temp) = 0) LOOP
    Temp := Temp - 1;
  END LOOP;
  Temp_Result.Last_Significant_Digit := Temp;
  RETURN Temp_Result;
END Random;


FUNCTION Random RETURN Big_Unsigned IS
  Temp : Index_Type := Random(Last_Indice);
BEGIN
  RETURN Random(Temp);
END Random;


FUNCTION Random(x : IN Big_Unsigned) RETURN Big_Unsigned IS
  Temp_Unsigned : Big_Unsigned;
  Temp_Index : Index_Type;
  Temp_Boolean : BOOLEAN := TRUE;
  Temp_MyType1, Temp_MyType2 : My_Type;
  Last_x : CONSTANT Index_Type := x.Last_Significant_Digit;
  -- Be careful, we can't use the following algorithm :
  --
  -- IF (x = Big_Unsigned_Zero) THEN
  --   RETURN Big_Unsigned_Zero;
  -- END IF;
  -- Temp : Big_Unsigned := Random;
  -- IF (x = Big_Unsigned_Last) THEN
  --   RETURN Temp;
  -- END IF;
  -- RETURN Temp MOD (x + Big_Unsigned_One)
  --
  -- If we do so, and if (x > Big_Unsigned_Last), the distribution won't be
  -- uniform, because all the numbers between 0 and (Big_Unsigned_Last - x)
  -- will have a greater probability than numbers between
  -- (Big_Unsigned_Last - x) and x.
  --
  -- So, we choose a random number for the index, between First_Indice and
  -- x.Last_Significant_Digit. Then, we loop in order to have a random number
  -- below or equal to x.
  --
  -- This algorithm is slower, but there is a better uniform distribution.
  --
BEGIN
  IF (x.Last_Significant_Digit = First_Indice) AND THEN
     (x.Mantissa(First_Indice) = 0) THEN
     -- If x = 0...
    RETURN Big_Unsigned_Zero;
  END IF;
  IF (Last_x = First_Indice) THEN
    Temp_Index := First_Indice;
  ELSE
    Temp_Index := Random(Last_x);
  END IF;
  IF (Temp_Index < Last_x) THEN
    Temp_Unsigned := Random(Temp_Index);
  ELSE
    -- Temp_Index = Last_x
    WHILE (Temp_Boolean = TRUE) LOOP
      Temp_MyType1 := x.Mantissa(Temp_Index);
      Temp_MyType2 := Random(Temp_MyType1);
      Temp_Unsigned.Mantissa(Temp_Index) := Temp_MyType2;
      IF (Temp_Index = First_Indice) THEN
        EXIT;
      END IF;
      Temp_Boolean := (Temp_MyType1 = Temp_MyType2);
      Temp_Index := Temp_Index - 1;
    END LOOP;
    IF (Temp_Index > First_Indice) THEN
      Temp_Unsigned := Temp_Unsigned + Random(Temp_Index - 1);
    END IF;
    Temp_Index := Last_x;
    WHILE (Temp_Index > First_Indice) AND THEN
          (Temp_Unsigned.Mantissa(Temp_Index) = 0) LOOP
      Temp_Index := Temp_Index - 1;
    END LOOP;
    Temp_Unsigned.Last_Significant_Digit := Temp_Index;
  END IF;
  RETURN Temp_Unsigned;
END Random;


FUNCTION Random(x, y : IN Big_Unsigned) RETURN Big_Unsigned IS
  Mini : CONSTANT Big_Unsigned := Min(x, y);
  Diff : CONSTANT Big_Unsigned := Max(x, y) - Mini;
BEGIN
  RETURN Mini + Random(Diff);
END Random;


BEGIN
  IF (Largest_Unsigned(Index_Type'LAST) >= B) THEN
    RAISE Instanciation_Error;
  END IF;
  Randomize;
END Big_Unsigned_Random;
