--******************************************************************************
--                                                                            **
--  File:        BINUGECO.ADB                                                 **
--  Description: Big_Number.Generic_Conversion package body                   **
--  Revision:    0.28 (beta version)                                          **
--  Date:        July 8, 2001                                                 **
--  Author:      Jrme Delcourt                                              **
--  Mail:        sikander@club-internet.fr                                    **
--                                                                            **
--  Copyright (c) Jrme Delcourt, 1998, 1999, 2000                           **
--  62, rue Ngrier                                                           **
--  59800 Lille                                                               **
--  FRANCE                                                                    **
--                                                                            **
--  Permission granted to use for any purpose, provided this copyright        **
--  remains attached and unmodified.                                          **
--                                                                            **
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR            **
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED            **
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       **
--                                                                            **
--******************************************************************************

SEPARATE(Big_Number)

PACKAGE BODY Generic_Conversion IS

--******************************************************************************
--******************************************************************************

FUNCTION Mod_Number2Big_Unsigned(x : IN Mod_Generic) RETURN Big_Unsigned IS
  Result : Big_Unsigned := Big_Unsigned_Zero;
  FUNCTION Shift_Right(Value : IN Mod_Generic; Amount : IN NATURAL)
                      RETURN Mod_Generic;
  PRAGMA IMPORT(INTRINSIC, Shift_Right);
BEGIN
  IF (Mod_Generic'SIZE <= My_Type_Size) THEN
    -- No problem : My_Type'LAST >= Mod_Generic'LAST.
    -- Note that Result.Last_Significant_Digit was already initialized
    -- to First_Indice, since Result was initialized to Big_Unsigned_Zero.
    Result.Mantissa(First_Indice) := My_Type(x);
    RETURN Result;
  END IF;
  -- Now, we must check the case where the size of x is bigger than
  -- the size of My_Type (note : the mantisse of a Big_Unsigned variable
  -- is coded by an array of My_Type elements).
  DECLARE
    Remainder : Mod_Generic := x;
    Index : Index_Type := First_Indice;
  BEGIN
    LOOP
      Result.Mantissa(Index) := My_Type(Remainder AND Mod_Generic(My_Type'LAST));
      -- Note that My_Type'LAST = Base - 1
      Remainder := Shift_Right(Remainder, My_Type_Size);
      EXIT WHEN (Remainder = 0) OR ELSE (Index = Last_Indice);
      Index := Index + 1;
    END LOOP;
    IF (Remainder > 0) THEN
      -- You must think of a stupid situation, where the total number of bits
      -- in the mantissa of the result is smaller than the number of significant
      -- bits of x.
      -- [It doesn't worth to use this package, which may be buggy, if the ADA
      -- compiler can provide a type (predefined or defined by a user modular
      -- type) with more bits in the mantissa and with bug-free related
      -- functions.]
      RAISE Conversion_Error;
    END IF;
    Result.Last_Significant_Digit := Index;
    RETURN Result;
  END;
END Mod_Number2Big_Unsigned;

--******************************************************************************
--******************************************************************************

FUNCTION Mod_Number2Big_Signed(x : IN Mod_Generic;
                               Positive : IN BOOLEAN := TRUE) RETURN Big_Signed IS
  FUNCTION My_Conversion IS NEW Mod_Number2Big_Unsigned(Mod_Generic);
  Result : Big_Signed;
BEGIN
  Result := Unsigned2Signed(My_Conversion(x), Positive);
  IF (x = 0) THEN
    Result.Positive := TRUE;
  END IF;
  RETURN Result;
END Mod_Number2Big_Signed;

--******************************************************************************
--******************************************************************************

FUNCTION Int_Number2Big_Unsigned(x : IN Int_Generic) RETURN Big_Unsigned IS
  Result : Big_Unsigned := Big_Unsigned_Zero;
  FUNCTION Shift_Right(Value : IN Int_Generic; Amount : IN NATURAL)
                       RETURN Int_Generic;
  PRAGMA IMPORT(INTRINSIC, Shift_Right);
BEGIN
  IF (x < 0) THEN
    RAISE Conversion_Error;
  ELSIF (Int_Generic'SIZE <= My_Type_Size + 1) THEN
    -- No problem : My_Type'LAST > Int_Generic'LAST, because, for integer types,
    -- there is one bit reserved to store the sign.
    Result.Mantissa(First_Indice) := My_Type(x);
    RETURN Result;
  END IF;
  -- Now, we must check the case where the size of x is bigger than
  -- the size of My_Type (note : the mantissa of a Big_Unsigned variable
  -- is coded by an array of My_Type elements).
  DECLARE
    TmpBase : CONSTANT Int_Generic := Int_Generic(Base);
    -- Remark : Base = My_Type'LAST + 1;
    -- Here, Int_Generic'LAST > My_Type'LAST,
    -- So, Int_Generic'LAST >= Base
    Remainder : Int_Generic := x;
    Index : Index_Type := First_Indice;
  BEGIN
    LOOP
      Result.Mantissa(Index) := My_Type(Remainder MOD TmpBase);
      Remainder := Shift_Right(Remainder, My_Type_Size);
      EXIT WHEN (Remainder = 0) OR ELSE (Index = Last_Indice);
      Index := Index + 1;
    END LOOP;
    IF (Remainder > 0) THEN
      -- You must think of a stupid situation, where the total number of bits
      -- in the mantisse of the result is smaller than the number of significant
      -- bits of x.
      -- [It doesn't worth to use this package, which may be buggy, if the ADA
      -- compiler can provide a type (predefined or defined by a user modular
      -- type) with more bits in the mantissa and with bug-free related
      -- functions.]
      RAISE Conversion_Error;
    END IF;
    Result.Last_Significant_Digit := Index;
    RETURN Result;
  END;
END Int_Number2Big_Unsigned;

--******************************************************************************
--******************************************************************************

FUNCTION Int_Number2Big_Signed(x : IN Int_Generic) RETURN Big_Signed IS
  FUNCTION My_Conversion IS NEW Int_Number2Big_Unsigned(Int_Generic);
BEGIN
  RETURN Unsigned2Signed(My_Conversion(ABS(x)), x>=0);
END Int_Number2Big_Signed;

--******************************************************************************
--******************************************************************************

END Generic_Conversion;
