--******************************************************************************
--                                                                            **
--  File:        BINUSTCO.ADB                                                 **
--  Description: Big_Number.String_Conversion package body                    **
--  Revision:    0.28 (beta version)                                          **
--  Date:        July 8, 2001                                                 **
--  Author:      Jrme Delcourt                                              **
--  Mail:        sikander@club-internet.fr                                    **
--                                                                            **
--  Copyright (c) Jrme Delcourt, 1998, 1999, 2000, 2001                     **
--  62, rue Ngrier                                                           **
--  59800 Lille                                                               **
--  FRANCE                                                                    **
--                                                                            **
--  Permission granted to use for any purpose, provided this copyright        **
--  remains attached and unmodified.                                          **
--                                                                            **
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR            **
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED            **
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       **
--                                                                            **
--******************************************************************************

SEPARATE(Big_Number)

PACKAGE BODY String_Conversion IS

--******************************************************************************
--******************************************************************************

  SUBTYPE Numeric_Digit IS CHARACTER RANGE '0'..'9';

--******************************************************************************
--******************************************************************************
--**           Conversion of strings to Big_Unsigned or Big_Signed            **
--******************************************************************************
--******************************************************************************

FUNCTION String_Size_Required(x : IN STRING) RETURN NATURAL IS
  Temp_Size : NATURAL := 0;
  x_First : CONSTANT POSITIVE := x'FIRST;
  Char : CHARACTER;
BEGIN
  FOR I IN x'RANGE LOOP
    Char := x(I);
    IF NOT(Char IN Numeric_Digit) THEN
      IF ((Char = '-') OR ELSE (Char = '+')) THEN
        -- A sign is authorized only for the first character
        IF (I /= x_First) THEN
          RAISE Conversion_Error;
        END IF;
      ELSIF (Char /= ' ') AND THEN (Char /= '_') THEN
        RAISE Conversion_Error;
      END IF;
    ELSE
      -- Now, Char is in '0'..'9'.
      Temp_Size := Temp_Size + 1;
    END IF;
  END LOOP;
  RETURN Temp_Size;
EXCEPTION
  WHEN OTHERS => RAISE Conversion_Error;
END String_Size_Required;

--******************************************************************************
--******************************************************************************

FUNCTION Str2Unsigned(x : IN STRING) RETURN Big_Unsigned IS
  Temp_Char : CHARACTER;
  Temp_Digit : Numeric_Digit;
  Temp : Big_Unsigned := Big_Unsigned_Zero;
  Pos0 : CONSTANT INTEGER := Numeric_Digit'POS('0');
  FUNCTION Integer2BU IS NEW Int_Number2Big_Unsigned(INTEGER);
BEGIN
  FOR I IN x'RANGE LOOP
    Temp_Char := x(I);
    Temp_Digit := Numeric_Digit(Temp_Char);
    Temp := (Temp * 10) + Integer2BU(Numeric_Digit'POS(Temp_Digit) - Pos0);
  END LOOP;
  RETURN Temp;
EXCEPTION
  WHEN OTHERS => RAISE Conversion_Error;
END Str2Unsigned;

--******************************************************************************
--******************************************************************************

PROCEDURE Del_Spaces_And_Underscore(x : IN STRING; y : OUT STRING) IS
  Dernier : CONSTANT POSITIVE := x'LAST;
  J : POSITIVE := y'FIRST;
  Temp_Char : CHARACTER;
BEGIN
  FOR I IN x'RANGE LOOP
    Temp_Char := x(I);
    IF (Temp_Char /= ' ') AND THEN (Temp_Char /= '_') THEN
      y(J) := Temp_Char;
      J := J + 1;
    END IF;
  END LOOP;
END Del_Spaces_And_Underscore;

--******************************************************************************
--******************************************************************************

FUNCTION String2Unsigned(x : IN STRING) RETURN Big_Unsigned IS
  String_Size : NATURAL;
  Temp_Positive, Sign_Present : BOOLEAN;
BEGIN
  String_Size := String_Size_Required(x);
  IF (String_Size = 0) THEN
    RETURN Big_Unsigned_Zero;
  END IF;
  DECLARE
    First_Char : CONSTANT CHARACTER := x(x'FIRST);
  BEGIN
    Sign_Present := (First_Char = '+') OR ELSE (First_Char = '-');
    Temp_Positive := (First_Char = '+') OR ELSE (First_Char IN Numeric_Digit);
  END;
  IF (Temp_Positive = FALSE) THEN
    RAISE Conversion_Error;
  END IF;
  DECLARE
    Temp_String : STRING(1..POSITIVE(String_Size));
  BEGIN
    IF (Sign_Present = TRUE) THEN
      Del_Spaces_And_Underscore(x(x'FIRST + 1..x'LAST), Temp_String);
    ELSE
      Del_Spaces_And_Underscore(x, Temp_String);
    END IF;
    RETURN Str2Unsigned(Temp_String);
  END;
EXCEPTION
  WHEN OTHERS => RAISE Conversion_Error;
END String2Unsigned;

--******************************************************************************
--******************************************************************************

FUNCTION String2Signed(x : IN STRING) RETURN Big_Signed IS
  String_Size : NATURAL;
  Sign_Present : BOOLEAN;
  Temp_Positive : BOOLEAN;
BEGIN
  String_Size := String_Size_Required(x);
  IF (String_Size = 0) THEN
    RETURN Big_Signed_Zero;
  END IF;
  DECLARE
    First_Char : CONSTANT CHARACTER := x(x'FIRST);
  BEGIN
    Sign_Present := (First_Char = '+') OR ELSE (First_Char = '-');
    Temp_Positive := (First_Char = '+') OR ELSE (First_Char IN Numeric_Digit);
  END;
  DECLARE
    Temp_String : STRING(1..POSITIVE(String_Size));
  BEGIN
    IF (Sign_Present = TRUE) THEN
      Del_Spaces_And_Underscore(x(x'FIRST + 1..x'LAST), Temp_String);
    ELSE
      Del_Spaces_And_Underscore(x, Temp_String);
    END IF;
    RETURN Unsigned2Signed(Str2Unsigned(Temp_String), Temp_Positive);
  END;
EXCEPTION
  WHEN OTHERS => RAISE Conversion_Error;
END String2Signed;

--******************************************************************************
--******************************************************************************
--**       Conversion of unbounded strings to Big_Unsigned or Big_Signed      **
--******************************************************************************
--******************************************************************************

FUNCTION UStr2Unsigned(x : IN UNBOUNDED_STRING) RETURN Big_Unsigned IS
  Temp_Char : CHARACTER;
  Temp_Digit : Numeric_Digit;
  Temp : Big_Unsigned := Big_Unsigned_Zero;
  Pos0 : CONSTANT INTEGER := Numeric_Digit'POS('0');
  FUNCTION Integer2BU IS NEW Int_Number2Big_Unsigned(INTEGER);
BEGIN
  FOR I IN 1..LENGTH(x) LOOP
    Temp_Char := ELEMENT(x, I);
    Temp_Digit := Numeric_Digit(Temp_Char);
    Temp := (Temp * 10) + Integer2BU(Numeric_Digit'POS(Temp_Digit) - Pos0);
  END LOOP;
  RETURN Temp;
EXCEPTION
  WHEN OTHERS => RAISE Conversion_Error;
END UStr2Unsigned;

--******************************************************************************
--******************************************************************************

FUNCTION UString2Unsigned(x : IN UNBOUNDED_STRING) RETURN Big_Unsigned IS
  Positive, Sign_Present : BOOLEAN;
  Temp_x : UNBOUNDED_STRING := Del_Spaces(x);
  Size1, Size2 : INTEGER;
BEGIN
  Temp_x := Del_Character(Temp_x, '_');
  IF (LENGTH(Temp_x) = 0) THEN
    RETURN Big_Unsigned_Zero;
  END IF;
  DECLARE
    First_Char : CONSTANT CHARACTER := ELEMENT(x, 1);
  BEGIN
    Sign_Present := (First_Char = '+') OR ELSE (First_Char = '-');
    Positive := (First_Char = '+') OR ELSE (First_Char IN Numeric_Digit);
  END;
  IF ((Sign_Present = TRUE) AND THEN (LENGTH(Temp_x) = 1))
    -- If x contains only a sign (+ or -)
    OR ELSE (Positive = FALSE) THEN
    -- or if the number is negative,
    -- this string can't be converted to a Big_Unsigned.
    RAISE Conversion_Error;
  END IF;
  -- Now, the number is positive.
  Size1 := LENGTH(Temp_x);
  Temp_x := Del_Character(Temp_x, '+');
  Temp_x := Del_Character(Temp_x, '-');
  Size2 := LENGTH(Temp_x);
  IF (Size1 - Size2 > 1) THEN
    -- If there are 2 or more signs in Temp_x...
    RAISE Conversion_Error;
  END IF;
  RETURN UStr2Unsigned(Temp_x);
EXCEPTION
  WHEN OTHERS => RAISE Conversion_Error;
END UString2Unsigned;

--******************************************************************************
--******************************************************************************

FUNCTION UString2Signed(x : IN UNBOUNDED_STRING) RETURN Big_Signed IS
  Positive, Sign_Present : BOOLEAN;
  Temp_x : UNBOUNDED_STRING := Del_Spaces(x);
  Size1, Size2 : INTEGER;
BEGIN
  Temp_x := Del_Character(Temp_x, '_');
  IF (LENGTH(Temp_x) = 0) THEN
    RETURN Big_Signed_Zero;
  END IF;
  DECLARE
    First_Char : CONSTANT CHARACTER := ELEMENT(x, 1);
  BEGIN
    Sign_Present := (First_Char = '+') OR ELSE (First_Char = '-');
    Positive := (First_Char = '+') OR ELSE (First_Char IN Numeric_Digit);
  END;
  IF ((Sign_Present = TRUE) AND THEN (LENGTH(Temp_x) = 1)) THEN
    -- If x contains only a sign (+ or -),
    -- this string can't be converted to a Big_Unsigned.
    RAISE Conversion_Error;
  END IF;
  -- Now, the sign is known.
  Size1 := LENGTH(Temp_x);
  Temp_x := Del_Character(Temp_x, '+');
  Temp_x := Del_Character(Temp_x, '-');
  Size2 := LENGTH(Temp_x);
  IF (Size1 - Size2 > 1) THEN
    -- If there are 2 or more signs in Temp_x...
    RAISE Conversion_Error;
  END IF;
  RETURN Unsigned2Signed(UStr2Unsigned(Temp_x), Positive);
EXCEPTION
  WHEN OTHERS => RAISE Conversion_Error;
END UString2Signed;

--******************************************************************************
--******************************************************************************
--**      Conversion of Big_Unsigned or Big_Signed to unbounded strings       **
--******************************************************************************
--******************************************************************************

FUNCTION Big_Unsigned2UString(x : IN Big_Unsigned;
                              Base : IN My_Type := 10)
                              RETURN UNBOUNDED_STRING IS
  TYPE Trans IS ARRAY(My_Type RANGE 0..35) OF CHARACTER;
  Table : CONSTANT Trans
                := ('0','1','2','3','4','5','6','7','8','9','A','B','C','D','E',
                    'F','G','H','I','J','K','L','M','N','O','P','Q','R','S','T',
                    'U','V','W','X','Y','Z');
  Result : UNBOUNDED_STRING := NULL_UNBOUNDED_STRING;
  Temp_x : Big_Unsigned := x;
  Remainder : My_Type;
BEGIN
  IF (Base > 35) OR ELSE (Base < 2) THEN
    RAISE Base_Error;
  END IF;
  IF (Temp_x.Last_Significant_Digit = First_Indice) AND THEN
     (Temp_x.Mantissa(First_Indice) = 0) THEN
    Result := TO_UNBOUNDED_STRING("0");
    RETURN Result;
  END IF;
  WHILE (Temp_x /= Big_Unsigned_Zero) LOOP
    Short_Div(Temp_x, Base, Temp_x, Remainder);
    Result := Table(Remainder) & Result;
  END LOOP;
  RETURN Result;
END Big_Unsigned2UString;

--******************************************************************************
--******************************************************************************

FUNCTION Big_Signed2UString(x : IN Big_Signed;
                            Base : IN My_Type := 10)
                            RETURN UNBOUNDED_STRING IS
  Temp : CONSTANT Big_Unsigned := Signed2Unsigned(x);
BEGIN
  IF (x.Positive = FALSE) THEN
    RETURN "-" & Big_Unsigned2UString(Temp, Base);
  ELSE
    RETURN Big_Unsigned2UString(Temp, Base);
  END IF;
END Big_Signed2UString;

--******************************************************************************
--******************************************************************************

END String_Conversion;
