--******************************************************************************
--                                                                            **
--  File:        BINUUNNU.ADB                                                 **
--  Description: Big_Number.Unsigned_Number package body                      **
--  Revision:    0.28 (beta version)                                          **
--  Date:        July 8, 2001                                                 **
--  Author:      Jrme Delcourt                                              **
--  Mail:        sikander@club-internet.fr                                    **
--                                                                            **
--  Copyright (c) Jrme Delcourt, 1998, 1999, 2000, 2001                     **
--  62, rue Ngrier                                                           **
--  59800 Lille                                                               **
--  FRANCE                                                                    **
--                                                                            **
--  Permission granted to use for any purpose, provided this copyright        **
--  remains attached and unmodified.                                          **
--                                                                            **
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR            **
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED            **
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       **
--                                                                            **
--******************************************************************************

SEPARATE(Big_Number)

PACKAGE BODY Unsigned_Number IS

--******************************************************************************
--******************************************************************************
--**                              Comparisons                                 **
--**                                                                          **
--** The comparisons ">", "<=" and ">=" can be done by a call to "<"          **
--** (see below for explanations). The advantage is that we have only one     **
--** function to write and test, the 3 others depend on it.                   **
--******************************************************************************
--******************************************************************************

FUNCTION "<"(Left, Right : IN Big_Unsigned) RETURN BOOLEAN IS
  -- Remark : in order to understand this function, you must remind that the
  -- coding of the digits of x and y, in their mantissa, looks like the
  -- "Big_Endian" coding.
  Indice : Index_Type := Left.Last_Significant_Digit;
BEGIN
  IF (Indice /= Right.Last_Significant_Digit) THEN
    RETURN (Indice < Right.Last_Significant_Digit);
  END IF;
  -- Now, Left.Last_Significant_Digit = Right.Last_Significant_Digit
  WHILE (Left.Mantissa(Indice) = Right.Mantissa(Indice)) AND THEN
        (Indice > First_Indice) LOOP
    Indice := Indice - 1;
  END LOOP;
  RETURN (Left.Mantissa(Indice) < Right.Mantissa(Indice));
END "<";

--******************************************************************************

FUNCTION ">"(Left, Right : IN Big_Unsigned) RETURN BOOLEAN IS
BEGIN
  -- (Left > Right) is the same comparison than (Right < Left)...
  RETURN (Right < Left);
END ">";

--******************************************************************************

FUNCTION "<="(Left, Right : IN Big_Unsigned) RETURN BOOLEAN IS
  -- (Left <= Right) = NOT (Right < Left)
BEGIN
  RETURN NOT (Right < Left);
END "<=";

--******************************************************************************

FUNCTION ">="(Left, Right : IN Big_Unsigned) RETURN BOOLEAN IS
BEGIN
  -- (Left >= Right) = NOT (Left < Right).
  RETURN NOT (Left < Right);
END ">=";

--******************************************************************************

FUNCTION Is_Null(x : IN Big_Unsigned) RETURN BOOLEAN IS
BEGIN
  IF (x.Last_Significant_Digit = First_Indice) AND THEN
     (x.Mantissa(First_Indice) = 0) THEN
    RETURN TRUE;
  ELSE
    RETURN FALSE;
  END IF;
END Is_Null;

--******************************************************************************

FUNCTION Min(x, y : IN Big_Unsigned) RETURN Big_Unsigned IS
BEGIN
  IF (x < y) THEN
    RETURN x;
  ELSE
    RETURN y;
  END IF;
END Min;

--******************************************************************************

FUNCTION Max(x, y : IN Big_Unsigned) RETURN Big_Unsigned IS
BEGIN
  IF (x < y) THEN
    RETURN y;
  ELSE
    RETURN x;
  END IF;
END Max;

--******************************************************************************
--******************************************************************************
--**                            Common functions                              **
--******************************************************************************
--******************************************************************************

PROCEDURE Swap1(x, y : IN OUT Big_Unsigned) IS
  Temp : CONSTANT Big_Unsigned := x;
BEGIN
  x := y;
  y := Temp;
END Swap1;

--******************************************************************************

PROCEDURE Swap2(x, y : IN OUT Big_Unsigned) IS
  Max_Index : CONSTANT Index_Type
                     := Max(x.Last_Significant_Digit, y.Last_Significant_Digit);
  Temp : My_Type;
BEGIN
  FOR I IN First_Indice..Max_Index LOOP
    Temp := x.Mantissa(I);
    x.Mantissa(I) := y.Mantissa(I);
    y.Mantissa(I) := Temp;
  END LOOP;
END Swap2;

--******************************************************************************

FUNCTION Shift_Right(Value : IN Largest_Unsigned; Amount : IN NATURAL)
                     RETURN Largest_Unsigned;
FUNCTION Shift_Left(Value : IN Largest_Unsigned; Amount : IN NATURAL)
                    RETURN Largest_Unsigned;

FUNCTION Shift_Right(Value : IN My_Type; Amount : IN NATURAL)
                     RETURN My_Type;
PRAGMA IMPORT(INTRINSIC, Shift_Right);

FUNCTION Shift_Left(Value : IN My_Type; Amount : IN NATURAL)
                    RETURN My_Type;
PRAGMA IMPORT(INTRINSIC, Shift_Left);

--******************************************************************************

PROCEDURE Inc(x : IN OUT Big_Unsigned) IS
  Temp_Index : Index_Type := First_Indice;
BEGIN
  WHILE (x.Mantissa(Temp_Index) = Greatest_Digit) LOOP
    -- If x.Mantissa(Temp_Index) = Greatest_Digit, set it to 0
    -- and try to increment the next digit.
    IF (Temp_Index = Last_Indice) THEN
      -- In this case, the next digit can't be incremented.
      RAISE Big_Number_Overflow;
    END IF;
    x.Mantissa(Temp_Index) := 0;
    Temp_Index := Temp_Index + 1;
  END LOOP;
  x.Mantissa(Temp_Index) := x.Mantissa(Temp_Index) + 1;
  IF (Temp_Index > x.Last_Significant_Digit) THEN
    x.Last_Significant_Digit := Temp_Index;
  END IF;
END Inc;

--******************************************************************************

PROCEDURE Dec(x : IN OUT Big_Unsigned) IS
  Temp_Index : Index_Type := First_Indice;
BEGIN
  IF (x.Last_Significant_Digit = First_Indice) AND THEN
     (x.Mantissa(First_Indice) = 0) THEN
     -- If x = Big_Unsigned_Zero...
    RAISE Big_Unsigned_Substraction_Error;
  END IF;
  WHILE (x.Mantissa(Temp_Index) = 0) LOOP
    -- If x.Mantissa(Temp_Index) = 0, set it to Greatest_Digit
    -- and try to decrement the next digit.
    x.Mantissa(Temp_Index) := Greatest_Digit;
    Temp_Index := Temp_Index + 1;
  END LOOP;
  x.Mantissa(Temp_Index) := x.Mantissa(Temp_Index) - 1;
  IF (Temp_Index > First_Indice) AND THEN
     (x.Mantissa(Temp_Index) = 0) AND THEN
     (Temp_Index = x.Last_Significant_Digit) THEN
    x.Last_Significant_Digit := Temp_Index - 1;
  END IF;
END Dec;

--******************************************************************************

FUNCTION "+"(x, y : IN Big_Unsigned) RETURN Big_Unsigned IS
  Min_Index : CONSTANT Index_Type
                     := Min(x.Last_Significant_Digit, y.Last_Significant_Digit);
  Max_Index : CONSTANT Index_Type
                     := Max(x.Last_Significant_Digit, y.Last_Significant_Digit);
  Temp_Index : Index_Type := Min_Index;
  Carry : Largest_Unsigned := 0;
  Temp_Elt : Largest_Unsigned;
  Temp_Result : Big_Unsigned;
BEGIN
  FOR i IN (Max_Index + 1)..Last_Indice LOOP
    Temp_Result.Mantissa(i) := 0;
  END LOOP;
  Temp_Result.Last_Significant_Digit := Max_Index;
  FOR i IN First_Indice..Min_Index LOOP
    -- Remark : if i = First_Indice, Carry is 0.
    Temp_Elt :=   Largest_Unsigned(x.Mantissa(i))
                + Largest_Unsigned(y.Mantissa(i))
                + Carry;
    Carry := Shift_Right(Temp_Elt, My_Type_Size);  -- Carry := Temp_Elt / Base
    Temp_Result.Mantissa(i) := My_Type(Temp_Elt AND Base_Minus_One);
                         -- := Temp_Elt MOD Base
  END LOOP;
  IF (Min_Index /= Max_Index) THEN
    IF (x.Last_Significant_Digit > Min_Index) THEN
      Temp_Result.Mantissa(Min_Index+1..Max_Index)
                                          := x.Mantissa(Min_Index+1..Max_Index);
    ELSE
      Temp_Result.Mantissa(Min_Index+1..Max_Index)
                                          := y.Mantissa(Min_Index+1..Max_Index);
    END IF;
  END IF;
  WHILE (Carry > 0) AND THEN (Temp_Index < Last_Indice) LOOP
    Temp_Index := Temp_Index + 1;
    IF (Temp_Result.Mantissa(Temp_Index) < Greatest_Digit) THEN
      Carry := 0;
      Temp_Result.Mantissa(Temp_Index) := Temp_Result.Mantissa(Temp_Index) + 1;
      IF (Max_Index < Temp_Index) THEN
        Temp_Result.Last_Significant_Digit := Temp_Index;
      END IF;
    ELSE
      IF (Temp_Index = Last_Indice) THEN
        RAISE Big_Number_Overflow;
      END IF;
      Temp_Result.Mantissa(Temp_Index) := 0;
    END IF;
  END LOOP;
  RETURN Temp_Result;
END "+";

--******************************************************************************

FUNCTION "-"(x, y : IN Big_Unsigned) RETURN Big_Unsigned IS
BEGIN
  IF (x < y) THEN
    RAISE Big_Unsigned_Substraction_Error;
  END IF;
  --
  -- Common situation : x > y.
  -- We must calculate x - y.
  --
  DECLARE
    Nb_x : CONSTANT Index_Type := x.Last_Significant_Digit;
    Nb_y : CONSTANT Index_Type := y.Last_Significant_Digit;
    Temp_Index : Index_Type;
    Carry : My_Type := 0;
    Temp_Result : Big_Unsigned;
  BEGIN
    FOR i IN First_Indice..Nb_y LOOP
      IF (x.Mantissa(i) > y.Mantissa(i)) OR ELSE
         (x.Mantissa(i) = y.Mantissa(i) AND THEN Carry = 0) THEN
        -- This test is equivalent to (x.Mantissa(i) >= y.Mantissa(i) + Carry),
        -- but this last expression could raise a CONSTRAINT_ERROR if
        -- y.Mantissa(i) = My_Type'LAST and Carry = 1.
        Temp_Result.Mantissa(i) := x.Mantissa(i) - y.Mantissa(i) - Carry;
        Carry := 0;
      ELSE  -- x.Mantissa(i) < y.Mantissa(i) + Carry
        -- Temp_Result.Mantissa(i):=Base + x.Mantissa(i) - y.Mantissa(i) - Carry
        -- Base = My_Type'LAST + 1
        -- ==> Temp_Result.Mantissa(i)
        --                :=x.Mantissa(i)+(My_Type'LAST-y.Mantissa(i))+(1-Carry)
        -- My_Type'LAST = Greatest_Digit
        -- My_Type'LAST-y.Mantissa(i) = Greatest_Digit XOR y.Mantissa(i)
        -- ==> Temp_Result.Mantissa(i)
        --          :=x.Mantissa(i)+(Greatest_Digit XOR y.Mantissa(i))+(1-Carry)
        -- Don't change the order of operations, because that could cause
        -- overflows.
        Temp_Result.Mantissa(i) :=   x.Mantissa(i)
                                   + (Greatest_Digit XOR y.Mantissa(i))
                                   + (1 - Carry);
        -- Then Carry must be set to 1.
        Carry := 1;
      END IF;
    END LOOP;
    Temp_Index := Nb_y + 1;
    WHILE (Carry = 1) LOOP
      IF (x.Mantissa(Temp_Index) >= Carry) THEN
        Temp_Result.Mantissa(Temp_Index) := x.Mantissa(Temp_Index) - Carry;
        Carry := 0;
      ELSE
        -- x.Mantissa(Temp_Index) < Carry.
        -- This situation can occur only if x.Mantissa(i) = 0
        -- and if Carry is already equal to 1.
        Temp_Result.Mantissa(Temp_Index) := Greatest_Digit;
        -- Greatest_Digit := My_Type'LAST;
        -- It is not necessary to add "Carry := 1;",
        -- because Carry is already set to 1.
      END IF;
      Temp_Index := Temp_Index + 1;
    END LOOP;
    FOR i IN Temp_Index..Nb_x LOOP
      Temp_Result.Mantissa(i) := x.Mantissa(i);
    END LOOP;
    FOR i IN (Nb_x + 1)..Last_Indice LOOP
      Temp_Result.Mantissa(i) := 0;
    END LOOP;
    Temp_Index := Nb_x;
    WHILE (Temp_Result.Mantissa(Temp_Index) = 0) AND THEN
          (Temp_Index > First_Indice) LOOP
      Temp_Index := Temp_Index - 1;
    END LOOP;
    Temp_Result.Last_Significant_Digit := Temp_Index;
    RETURN Temp_Result;
  END;
END "-";

--******************************************************************************

FUNCTION Shift_Left(x : IN Big_Unsigned; Nb : IN Index_Type)
                   RETURN Big_Unsigned IS
BEGIN
  -- IF (Nb > Last_Indice) OR ELSE (x.Last_Significant_Digit + Nb > Last_Indice) THEN
  --   RAISE Big_Number_Overflow;
  -- END IF;
  -- In these lines of code, (x.Last_Significant_Digit + Nb) could raise
  -- an overflow, so they are replaced by the following lines :
  IF (Nb > Last_Indice) OR ELSE (x.Last_Significant_Digit > Last_Indice - Nb) THEN
    RAISE Big_Number_Overflow;
  END IF;
  DECLARE
    Result : Big_Unsigned;
    Last_Digit_Indice : CONSTANT Index_Type := x.Last_Significant_Digit + Nb;
  BEGIN
    IF (Nb /= 0) THEN
      FOR I IN First_Indice..(Nb - 1) LOOP
        Result.Mantissa(I) := 0;
      END LOOP;
    END IF;
    IF (Last_Digit_Indice /= Last_Indice) THEN
      FOR I IN (Last_Digit_Indice + 1)..Last_Indice LOOP
        Result.Mantissa(I) := 0;
      END LOOP;
    END IF;
    Result.Mantissa(Nb..Last_Digit_Indice)
                          := x.Mantissa(First_Indice..x.Last_Significant_Digit);
    Result.Last_Significant_Digit := Last_Digit_Indice;
    RETURN Result;
  END;
END Shift_Left;

--******************************************************************************

FUNCTION "*"(x : IN Big_Unsigned; y : IN My_Type) RETURN Big_Unsigned IS
BEGIN
  IF (y = 0) THEN
    RETURN Big_Unsigned_Zero;
  ELSIF (y = 1) THEN
    RETURN x;
  END IF;
  DECLARE
    Nb_x : CONSTANT Index_Type := x.Last_Significant_Digit;
    Temp_y : CONSTANT Largest_Unsigned := Largest_Unsigned(y);
    Carry : Largest_Unsigned := 0;
    Result : Big_Unsigned;
  BEGIN
    IF (Nb_x < Last_Indice) THEN
      FOR I IN (Nb_x + 1)..Last_Indice LOOP
        Result.Mantissa(I) := 0;
      END LOOP;
    END IF;
    FOR I IN First_Indice..Nb_x LOOP
      Carry := Largest_Unsigned(x.Mantissa(I)) * Temp_y + Carry;
      Result.Mantissa(I) := My_Type(Carry AND Largest_Unsigned(My_Type'LAST));
      Carry := Shift_Right(Carry, My_Type_Size);        -- := Carry DIV Base
    END LOOP;
    IF (Carry = 0) THEN
      Result.Last_Significant_Digit := Nb_x;
    ELSE
      IF (Nb_x < Last_Indice) THEN
        Result.Last_Significant_Digit := Nb_x + 1;
        Result.Mantissa(Nb_x + 1) := My_Type(Carry);
      ELSE
        RAISE Big_Number_Overflow;
      END IF;
    END IF;
    RETURN Result;
  END;             -- END of the DECLARE block
END "*";

--******************************************************************************

FUNCTION Special_Mul(x : IN Big_Unsigned;
                     y : IN My_Type;
                     Left_Shift : IN Index_Type) RETURN Big_Unsigned IS
  -- Internal function
  -- This function returns x * y, with a left shift of Left_Shift digits.
  -- algorithm similar to the one used for "*"(x : IN Big_Unsigned; y : IN My_Type)
BEGIN
  IF (y = 0) THEN
    RETURN Big_Unsigned_Zero;
  END IF;
  DECLARE
    Nb_x : CONSTANT Index_Type := x.Last_Significant_Digit;
    Expected_Indice : CONSTANT Index_Type := Nb_x + Left_Shift;
    Temp_y : CONSTANT Largest_Unsigned := Largest_Unsigned(y);
    Carry : Largest_Unsigned := 0;
    Result : Big_Unsigned;
  BEGIN
    IF (Left_Shift > First_Indice) THEN
      FOR I IN First_Indice..(Left_Shift - 1) LOOP
        Result.Mantissa(I) := 0;
      END LOOP;
    END IF;
    IF (Expected_Indice < Last_Indice) THEN
      FOR I IN (Expected_Indice + 1)..Last_Indice LOOP
        Result.Mantissa(I) := 0;
      END LOOP;
    END IF;
    FOR I IN First_Indice..Nb_x LOOP
      Carry := Largest_Unsigned(x.Mantissa(I)) * Temp_y + Carry;
      Result.Mantissa(I + Left_Shift)
                           := My_Type(Carry AND Largest_Unsigned(My_Type'LAST));
      Carry := Shift_Right(Carry, My_Type_Size);        -- := Carry DIV Base
    END LOOP;
    IF (Carry = 0) THEN
      Result.Last_Significant_Digit := Expected_Indice;
    ELSE
      IF (Expected_Indice < Last_Indice) THEN
        Result.Last_Significant_Digit := Expected_Indice + 1;
        Result.Mantissa(Expected_Indice + 1) := My_Type(Carry);
      ELSE
        RAISE Big_Number_Overflow;
      END IF;
    END IF;
    RETURN Result;
  END;             -- END of the DECLARE block
END Special_Mul;

PRAGMA INLINE(Special_Mul);

--******************************************************************************

FUNCTION "*"(x, y : IN Big_Unsigned) RETURN Big_Unsigned IS
  Lastx : CONSTANT Index_Type := x.Last_Significant_Digit;
  Lasty : CONSTANT Index_Type := y.Last_Significant_Digit;
  Result : Big_Unsigned;
BEGIN
  IF (Lastx = First_Indice) THEN
    RETURN (y * x.Mantissa(First_Indice));
  END IF;
  IF (Lastx > Last_Indice - Lasty) THEN
    RAISE Big_Number_Overflow;
  END IF;
  Result := x * y.Mantissa(First_Indice);
  -- Result is initialized here, because I don't want to waste
  -- execution time before, during the computation of the simple cases.
  FOR I IN 1..Lasty LOOP
    Result := Result + Special_Mul(x, y.Mantissa(I), I);
  END LOOP;
  RETURN Result;
EXCEPTION
  WHEN OTHERS => RAISE Big_Number_Overflow;
END "*";

--******************************************************************************

FUNCTION "**"(x : IN Big_Unsigned; y : IN My_Type) RETURN Big_Unsigned IS
BEGIN
  IF (y = 0) THEN
    RETURN Big_Unsigned_One;
  END IF;
  IF ((x.Last_Significant_Digit = First_Indice) AND THEN
     (x.Mantissa(First_Indice) <= 1))
     OR ELSE (y = 1) THEN
    -- Here, y /= 0.
    -- So, if y = 1 or x = 0 or x = 1, the result of (x ** y) is always x.
    RETURN x;
  END IF;
  DECLARE
    Temp_Result : Big_Unsigned := x;
  BEGIN
    FOR I IN 2..(y/2) LOOP
      Temp_Result := Temp_Result * x;
    END LOOP;
    Temp_Result := Temp_Result * Temp_Result;
    IF (y MOD 2 = 1) THEN
      Temp_Result := Temp_Result * x;
    END IF;
    RETURN Temp_Result;
  END;
END "**";

--******************************************************************************

FUNCTION "**"(x : IN Big_Unsigned; y : IN Index_Type) RETURN Big_Unsigned IS
BEGIN
  IF (y = 0) THEN
    RETURN Big_Unsigned_One;
  END IF;
  IF ((x.Last_Significant_Digit = First_Indice) AND THEN
     (x.Mantissa(First_Indice) <= 1))
     OR ELSE (y = 1) THEN
    -- Here, y /= 0.
    -- So, if y = 1 or x = 0 or x = 1, the result of (x ** y) is always x.
    RETURN x;
  END IF;
  DECLARE
    Temp_Result : Big_Unsigned := x;
  BEGIN
    FOR I IN 2..(y/2) LOOP
      Temp_Result := Temp_Result * x;
    END LOOP;
    Temp_Result := Temp_Result * Temp_Result;
    IF (y MOD 2 = 1) THEN
      Temp_Result := Temp_Result * x;
    END IF;
    RETURN Temp_Result;
  END;
END "**";

--******************************************************************************

FUNCTION "**"(x, y : IN Big_Unsigned) RETURN Big_Unsigned IS
  Temp_Result : Big_Unsigned := x;
  Indice : Big_Unsigned := Big_Unsigned_One;
BEGIN
  IF (y.Last_Significant_Digit = First_Indice) THEN
    IF (y.Mantissa(First_Indice) = 0) THEN
      -- If y = 0, x^y = x^0 = 1, even if x = 0.
      RETURN Big_Unsigned_One;
    ELSE
      RETURN x ** My_Type(y.Mantissa(First_Indice));
    END IF;
  END IF;
  WHILE (Indice < y) LOOP
    Inc(Indice);
    Temp_Result := Temp_Result * x;
  END LOOP;
  RETURN Temp_Result;
END "**";

--******************************************************************************

FUNCTION "**"(x : IN Big_Unsigned;
              y : IN Big_Signed) RETURN Big_Unsigned IS
BEGIN
  IF (y.Positive = FALSE) THEN    -- Here, we test if (y < Big_Signed_Zero).
    RETURN Big_Unsigned_Zero;
  END IF;
  DECLARE
    Temp_y : CONSTANT Big_Unsigned := Signed2Unsigned(y);
  BEGIN
    RETURN (x ** Temp_y);
  END;
END "**";

--******************************************************************************
--******************************************************************************

PROCEDURE Big_Div(x, y : IN Big_Unsigned;
                  Quot, Remainder : OUT Big_Unsigned) IS
  Lasty : CONSTANT Index_Type := y.Last_Significant_Digit;
BEGIN
  -- analysis of simple cases.
  IF (Lasty = First_Indice) THEN
    CASE y.Mantissa(First_Indice) IS
      WHEN 0 => RAISE Division_By_Zero;
      WHEN 1 => Quot := x;
                Remainder := Big_Unsigned_Zero;
                RETURN;
      WHEN OTHERS => DECLARE
                       Temp_Remainder : My_Type;
                       Temp_y : CONSTANT My_Type := y.Mantissa(First_Indice);
                     BEGIN
                       -- If y has only one significant item in its mantissa,
                       -- we use the function Short_Div, which is faster.
                       -- See below for the implementation of Short_Div.
                       Short_Div(x, Temp_y, Quot, Temp_Remainder);
                       Remainder := (Last_Significant_Digit => First_Indice,
                                     Mantissa => (Temp_Remainder, OTHERS => 0));
                       RETURN;
                     END;
    END CASE;
  END IF;
  IF (x.Last_Significant_Digit = First_Indice) AND THEN
     (x.Mantissa(First_Indice) = 0) THEN
    Quot := Big_Unsigned_Zero;
    Remainder := Big_Unsigned_Zero;
    RETURN;
  ELSIF (x < y) THEN
    Quot := Big_Unsigned_Zero;
    Remainder := x;
    RETURN;
  ELSIF (x = y) THEN
    Quot := Big_Unsigned_One;
    Remainder := Big_Unsigned_Zero;
    RETURN;
  END IF;
  -- Now, there is only the case where (x > y) and (y /= 0).
  DECLARE

    FUNCTION Two_Biggest_Digits(x : IN Big_Unsigned) RETURN Largest_Unsigned IS
      Lastx : CONSTANT Index_Type := x.Last_Significant_Digit;
    BEGIN
      -- We don't have to check if there is only one digit :
      -- because of the algorithm, there are at least two digits.
      RETURN Largest_Unsigned(x.Mantissa(Lastx-1))
                + Shift_Left(Largest_Unsigned(x.Mantissa(Lastx)), My_Type_Size);
    END Two_Biggest_Digits;

    ----------------------------------------------------------------------------

    Two_Biggest_Digits_of_y : CONSTANT Largest_Unsigned:= Two_Biggest_Digits(y);
    Temp_Unsigned : Big_Unsigned;
    Temp_Quot : Big_Unsigned := Big_Unsigned_Zero;
    Temp_Remainder : Big_Unsigned := x;
    Temp_Index, Last_Indice_Quot : Index_Type;
    Carry : My_Type;
    Guess_Digit : My_Type;

    ----------------------------------------------------------------------------

    FUNCTION Last_Indice_of_the_Quotient RETURN Index_Type IS
      -- The function Last_Indice_of_the_Quotient is used to calculate the
      -- number of items of the quotient and, so, the indice of the array (used
      -- to represent the mantissa) where we have to put the guessed digit.
      Diff : CONSTANT Index_Type := Temp_Remainder.Last_Significant_Digit-Lasty;
      Indice : Index_Type := Lasty;
    BEGIN
      WHILE (Temp_Remainder.Mantissa(Indice+Diff) = y.Mantissa(Indice)) AND THEN
            (Indice > First_Indice) LOOP
        Indice := Indice - 1;
      END LOOP;
      IF (Temp_Remainder.Mantissa(Indice+Diff) < y.Mantissa(Indice)) AND THEN
         (Diff > First_Indice) THEN
        RETURN Diff - 1;
      ELSE
        RETURN Diff;
      END IF;
    END Last_Indice_of_the_Quotient;

    ----------------------------------------------------------------------------

    FUNCTION Guess RETURN My_Type IS
      Last : CONSTANT Index_Type := Temp_Remainder.Last_Significant_Digit;
      Temp_Rem : CONSTANT Largest_Unsigned := Two_Biggest_Digits(Temp_Remainder);
      Temp_Digit : Largest_Unsigned;
    BEGIN
      -- In this function, Last >= Lasty, and lasty >= 1.
      IF (Last_Indice_Quot = Last - Lasty) THEN
        Temp_Digit := Temp_Rem / Two_Biggest_Digits_of_y;
        IF (Lasty >= 2) AND THEN
           (Temp_Rem MOD Two_Biggest_Digits_of_y < Shift_Right(Largest_Unsigned(y.Mantissa(Lasty-2))*Temp_Digit,My_Type_Size)) THEN
          Temp_Digit := Temp_Digit - 1;
        END IF;
        RETURN My_Type(Temp_Digit);
      END IF;
      -- We have now to study the case where Indice_Digit = Last - Lasty - 1.
      -- Search for Temp_Digit, so Temp_Remainder / y = Temp_Digit
      DECLARE
        Temp_y1 : CONSTANT Largest_Unsigned
                                         := Largest_Unsigned(y.Mantissa(Lasty));
        Temp_y2 : CONSTANT Largest_Unsigned
                                       := Largest_Unsigned(y.Mantissa(Lasty-1));
        Max_Digit : Largest_Unsigned := Temp_Rem / Temp_y1;
        Min_Digit : Largest_Unsigned := Temp_Rem / (Temp_y1 + 1);
        Temp : Largest_Unsigned;
      BEGIN
        IF (Max_Digit > Base_Minus_One) THEN
          Max_Digit := Base_Minus_One;
        END IF;
        LOOP
          IF (Max_Digit = Min_Digit) THEN
            RETURN My_Type(Min_Digit);
          END IF;
          IF (Max_Digit - Min_Digit = 1) THEN
            Temp := Shift_Right(Temp_y2 * Max_Digit, My_Type_Size)
                    + (Temp_y1 * Max_Digit);
            IF (Temp <= Temp_Rem) THEN
              RETURN My_Type(Max_Digit);
            ELSE
              RETURN My_Type(Min_Digit);
            END IF;
          END IF;
          -- Now, Max_Digit - Min_Digit >= 2
          Temp_Digit := Shift_Right(Max_Digit + Min_Digit, 1);
            -- Temp_Digit := (Max_Digit + Min_Digit) / 2;
          Temp := Shift_Right(Temp_y2 * Temp_Digit, My_Type_Size)
                  + (Temp_y1 * Temp_Digit);
          IF (Temp = Temp_Rem) THEN
            RETURN My_Type(Temp_Digit);
          ELSIF (Temp > Temp_Rem) THEN
            Max_Digit := Temp_Digit;
            IF ((Temp - Temp_Rem) MOD Temp_y1 > 0) THEN
              Temp := (Temp - Temp_Rem) / Temp_y1 + 1;
            ELSE
              Temp := (Temp - Temp_Rem) / Temp_y1;
            END IF;
            IF (Temp <= Temp_Digit) THEN
              Min_Digit := Temp_Digit - Temp;
            END IF;
          ELSE   -- Temp < Temp_Rem
            Min_Digit := Temp_Digit;
            Temp := (Temp_Rem - Temp) / Temp_y1;
            IF (Base_Minus_One - Temp_Digit >= Temp) THEN
              Max_Digit := Temp_Digit + Temp;
            END IF;
          END IF;
        END LOOP;
      END;
    END Guess;

    ----------------------------------------------------------------------------

  BEGIN
    Temp_Quot.Last_Significant_Digit := Last_Indice_of_the_Quotient;
    WHILE (y < Temp_Remainder) LOOP
      Last_Indice_Quot := Last_Indice_of_the_Quotient;
      Guess_Digit := Guess;
      --
      -- Guess_Digit is the real digit or the real digit plus one.
      -- Temp_Unsigned := y * Guess_Digit * (Base**Last_Indice_Quot)
      --
      Temp_Unsigned := Special_mul(y, Guess_Digit, Last_Indice_Quot);
      WHILE (Temp_Remainder < Temp_Unsigned) LOOP
        -- Decrement Guess_Digit and calculate the new value of Temp_Unsigned
        -- We have not to test if Guess_Digit = 0,
        -- because Temp_Unsigned would be equal to zero,
        -- so it could not be greater than Temp_Remainder.
        Guess_Digit := Guess_Digit - 1;
        -- The following lines are a kind of inlining for
        -- calculating Temp_Unsigned - y * (Base**Last_Indice_Quot)
        Carry := 0;
        FOR i IN First_Indice..Lasty LOOP
          Temp_Index := i + Last_Indice_Quot;
          IF (Temp_Unsigned.Mantissa(Temp_Index) > y.Mantissa(i)) OR ELSE
             (Temp_Unsigned.Mantissa(Temp_Index)=y.Mantissa(i) AND THEN Carry=0) THEN
            Temp_Unsigned.Mantissa(Temp_Index)
                  := Temp_Unsigned.Mantissa(Temp_Index) - y.Mantissa(i) - Carry;
            Carry := 0;
          ELSE  -- Temp_Unsigned.Mantissa(Temp_Index) < y.Mantissa(i) + Carry
            Temp_Unsigned.Mantissa(Temp_Index)
                                         := Temp_Unsigned.Mantissa(Temp_Index)
                                            + (Greatest_Digit XOR y.Mantissa(i))
                                            + (1 - Carry);
            Carry := 1;
          END IF;
        END LOOP;
        -- During the last preceding loop, Temp_Index was equal
        -- to y.Last_Significant_Digit + Last_Indice_Quot.
        IF (Carry = 1) THEN
          Temp_Index := Temp_Index + 1;
          Temp_Unsigned.Mantissa(Temp_Index)
                                  := Temp_Unsigned.Mantissa(Temp_Index) - Carry;
        END IF;
        WHILE (Temp_Unsigned.Mantissa(Temp_Unsigned.Last_Significant_Digit) = 0)
              AND THEN (Temp_Unsigned.Last_Significant_Digit > First_Indice) LOOP
          Temp_Unsigned.Last_Significant_Digit
                                    := Temp_Unsigned.Last_Significant_Digit - 1;
        END LOOP;
      END LOOP;
      Carry := 0;
      --
      -- Calculate the new value of Temp_Remainder, which will be
      -- Temp_Remainder - Temp_Unsigned
      --
      FOR i IN Last_Indice_Quot..Temp_Unsigned.Last_Significant_Digit LOOP
        IF (Temp_Remainder.Mantissa(i) > Temp_Unsigned.Mantissa(i)) OR ELSE
           (Temp_Remainder.Mantissa(i) = Temp_Unsigned.Mantissa(i) AND THEN Carry=0) THEN
          Temp_Remainder.Mantissa(i)
              := Temp_Remainder.Mantissa(i) - Temp_Unsigned.Mantissa(i) - Carry;
          Carry := 0;
        ELSE  -- Temp_Remainder.Mantissa(i) < Temp_Unsigned.Mantissa(i) + Carry
          Temp_Remainder.Mantissa(i)
                             := Temp_Remainder.Mantissa(i)
                                + (Greatest_Digit XOR Temp_Unsigned.Mantissa(i))
                                + (1 - Carry);
          Carry := 1;
        END IF;
      END LOOP;
      IF (Carry = 1) THEN
        Temp_Index := Temp_Unsigned.Last_Significant_Digit + 1;
        Temp_Remainder.Mantissa(Temp_Index)
                                 := Temp_Remainder.Mantissa(Temp_Index) - Carry;
      END IF;
      WHILE (Temp_Remainder.Mantissa(Temp_Remainder.Last_Significant_Digit) = 0)
            AND THEN (Temp_Remainder.Last_Significant_Digit > First_Indice) LOOP
        Temp_Remainder.Last_Significant_Digit
                                   := Temp_Remainder.Last_Significant_Digit - 1;
      END LOOP;
      --
      -- Now, we have the correct values of Guess_Digit and Temp_Remainder.
      --
      Temp_Quot.Mantissa(Last_Indice_Quot) := Guess_Digit;
    END LOOP;
    Quot := Temp_Quot;
    Remainder := Temp_Remainder;
  END;        -- DECLARE
END Big_Div;

--******************************************************************************
--******************************************************************************

FUNCTION "/"(x, y : IN Big_Unsigned) RETURN Big_Unsigned IS
  Quot, Remainder : Big_Unsigned;
BEGIN
  Big_Div(x, y, Quot, Remainder);
  RETURN Quot;
END "/";

--******************************************************************************

FUNCTION "MOD"(x, y : IN Big_Unsigned) RETURN Big_Unsigned IS
  Quot, Remainder : Big_Unsigned;
BEGIN
  Big_Div(x, y, Quot, Remainder);
  RETURN Remainder;
END "MOD";

--******************************************************************************

PROCEDURE Short_Div(x : IN Big_Unsigned;
                    y : IN My_Type;
                    Quot : OUT Big_Unsigned;
                    Remainder : OUT My_Type) IS
  Lastx : CONSTANT Index_Type := x.Last_Significant_Digit;
  Temp_Quot : Big_Unsigned := Big_Unsigned_Zero;
  Carry : Largest_Unsigned := 0;
  Temp : Largest_Unsigned;
  Temp_y : CONSTANT Largest_Unsigned := Largest_Unsigned(y);
BEGIN
  IF (y = 0) THEN
    RAISE Division_By_Zero;
  END IF;
  IF (y = 1) THEN
    Quot := x;
    Remainder := 0;
    RETURN;
  END IF;
  FOR Indice IN REVERSE First_Indice..Lastx LOOP
    Temp := Largest_Unsigned(x.Mantissa(Indice))
            + Shift_Left(Carry, My_Type_Size);
    Temp_Quot.Mantissa(Indice) := My_Type(Temp / Temp_y);
    Carry := Temp MOD Temp_y;
  END LOOP;
  IF (Lastx > 0) AND THEN
     (Temp_Quot.Mantissa(Lastx) = 0) THEN
    Temp_Quot.Last_Significant_Digit := Lastx - 1;
  ELSE
    Temp_Quot.Last_Significant_Digit := Lastx;
  END IF;
  Quot := Temp_Quot;
  Remainder := My_Type(Carry);
END Short_Div;

--******************************************************************************

FUNCTION "/"(x : IN Big_Unsigned; y : IN My_Type) RETURN Big_Unsigned IS
  Quot : Big_Unsigned;
  Remainder : My_Type;
BEGIN
  Short_Div(x, y, Quot, Remainder);
  RETURN Quot;
END "/";

--******************************************************************************

FUNCTION "MOD"(x : IN Big_Unsigned; y : IN My_Type) RETURN My_Type IS
  Quot : Big_Unsigned;
  Remainder : My_Type;
BEGIN
  Short_Div(x, y, Quot, Remainder);
  RETURN Remainder;
END "MOD";

--******************************************************************************
--******************************************************************************
--**                           Special functions                              **
--******************************************************************************
--******************************************************************************

FUNCTION Number_of_Bytes(x : IN Big_Unsigned) RETURN Largest_Unsigned IS
  Temp1 : Largest_Unsigned;
  Temp2 : My_Type := x.Mantissa(x.Last_Significant_Digit);
BEGIN
  Temp1 := Largest_Unsigned(x.Last_Significant_Digit) * (My_Type_Size / 8);
  WHILE (Temp2 /= 0) LOOP
    Temp2 := Shift_Right(Temp2, 8);
    Temp1 := Temp1 + 1;
  END LOOP;
  RETURN Temp1;
END Number_of_Bytes;

--******************************************************************************

FUNCTION Number_of_Bits(x : IN Big_Unsigned) RETURN Largest_Unsigned IS
  Temp1 : Largest_Unsigned;
  Temp2 : My_Type := x.Mantissa(x.Last_Significant_Digit);
BEGIN
  Temp1 := Largest_Unsigned(x.Last_Significant_Digit) * My_Type_Size;
  WHILE (Temp2 /= 0) LOOP
    Temp2 := Shift_Right(Temp2, 1);
    Temp1 := Temp1 + 1;
  END LOOP;
  RETURN Temp1;
END Number_of_Bits;

--******************************************************************************

FUNCTION Square_Root(x : IN Big_Unsigned) RETURN Big_Unsigned IS
  Result, Temp_Result : Big_Unsigned;
  -- Remark :
  --   In order to avoid the use of a fifth variable, Diff is initialized to 3.
  Diff : Big_Unsigned := (Last_Significant_Digit => First_Indice,
                          Mantissa => (3, OTHERS => 0));
BEGIN
  IF (x.Last_Significant_Digit = First_Indice) THEN
    IF (x.Mantissa(First_Indice) = 0) THEN
      RETURN Big_Unsigned_Zero;
    ELSIF (x.Mantissa(First_Indice) <= 3) THEN
      RETURN Big_Unsigned_One;
    END IF;
  END IF;

  -- Boucle de calcul
  Result := x / 2;
  WHILE (Diff.Last_Significant_Digit > First_Indice) OR ELSE
        ((Diff.Last_Significant_Digit = First_Indice)
         AND THEN (Diff.Mantissa(First_Indice) > 1)) LOOP
    Temp_Result := Result;
    Result := (Temp_Result + x/Temp_Result) / 2;
    IF (Temp_Result > Result) THEN
      Diff := Temp_Result - Result;
    ELSE
      Diff := Result - Temp_Result;
    END IF;
  END LOOP;
  RETURN Result;
END Square_Root;

--******************************************************************************

FUNCTION Factorial(x : IN Big_Unsigned) RETURN Big_Unsigned IS
  Indice, Result : Big_Unsigned := Big_Unsigned_One;
BEGIN
  IF (x.Last_Significant_Digit = First_Indice) AND THEN
     (x.Mantissa(First_Indice) <= 1) THEN
    RETURN Big_Unsigned_One;
  END IF;
  WHILE (Indice < x) LOOP
    Inc(Indice);
    Result := Result * Indice;
  END LOOP;
  RETURN Result;
END Factorial;

--******************************************************************************

FUNCTION GCF(x, y : IN Big_Unsigned) RETURN Big_Unsigned IS
  -- GCF = greatest common factor

  FUNCTION Temp_GCF(A, B : IN Big_Unsigned) RETURN Big_Unsigned IS
    N, Temp_A, Temp : Big_Unsigned;
    Temp_B : Big_Unsigned := B;
  BEGIN
    Big_Div(A, B, Temp, N);
    LOOP
      IF (N.Last_Significant_Digit = First_Indice) THEN
        CASE (N.Mantissa(First_Indice)) IS
          WHEN 0 => RETURN Temp_B;         -- IF (N = Big_Unsigned_Zero)...
          WHEN 1 => RETURN N;              -- IF (N = Big_Unsigned_One)...
          WHEN OTHERS => NULL;
        END CASE;
      END IF;
      Temp_A := Temp_B;
      Temp_B := N;
      Big_Div(Temp_A, Temp_B, Temp, N);
        -- In this loop, temp is only a temporary variable and hasn't any
        -- effect in this algorithm.
        -- Now, N = Temp_A MOD Temp_B.
    END LOOP;
  END Temp_GCF;

BEGIN    -- Function GCF
  IF (x.Last_Significant_Digit = First_Indice) AND THEN
     (x.Mantissa(First_Indice) = 0) THEN
     -- IF x = Big_Unsigned_Zero THEN
    RETURN y;
  ELSIF (x = y) OR ELSE
        ((y.Last_Significant_Digit = First_Indice) AND THEN
         (y.Mantissa(First_Indice) = 0)) THEN
    RETURN x;
  ELSIF (x < y) THEN
    RETURN Temp_GCF(y,x);
  ELSE
    RETURN Temp_GCF(x,y);
  END IF;
END GCF;

--******************************************************************************

FUNCTION LCM(x, y : IN Big_Unsigned) RETURN Big_Unsigned IS
  -- finds the Least Common Multiple between 2 arguments : LCM
BEGIN
  RETURN (x / GCF(x,y)) * y;
END LCM;

--******************************************************************************

FUNCTION "AND"(x, y : IN Big_Unsigned) RETURN Big_Unsigned IS
  Temp : Big_Unsigned := x;
  Temp_Index : Index_Type
                     := Max(x.Last_Significant_Digit, y.Last_Significant_Digit);
BEGIN
  FOR I IN First_Indice..y.Last_Significant_Digit LOOP
    Temp.Mantissa(I) := Temp.Mantissa(I) AND y.Mantissa(I);
  END LOOP;
  WHILE (Temp_Index > First_Indice) AND THEN (Temp.Mantissa(Temp_Index)=0) LOOP
    Temp_Index := Temp_Index - 1;
  END LOOP;
  Temp.Last_Significant_Digit := Temp_Index;
  RETURN Temp;
END "AND";

--******************************************************************************

FUNCTION "XOR"(x, y : IN Big_Unsigned) RETURN Big_Unsigned IS
  Temp : Big_Unsigned := x;
  Temp_Index : Index_Type
                     := Max(x.Last_Significant_Digit, y.Last_Significant_Digit);
BEGIN
  FOR I IN First_Indice..y.Last_Significant_Digit LOOP
    Temp.Mantissa(I) := Temp.Mantissa(I) XOR y.Mantissa(I);
  END LOOP;
  WHILE (Temp_Index > First_Indice) AND THEN (Temp.Mantissa(Temp_Index)=0) LOOP
    Temp_Index := Temp_Index - 1;
  END LOOP;
  Temp.Last_Significant_Digit := Temp_Index;
  RETURN Temp;
END "XOR";

--******************************************************************************

FUNCTION NCR(N, R : IN Big_Unsigned) RETURN Big_Unsigned IS
  -- NCR(N,R) = N!/[R!(N-R)!]
  -- In this algorythm, we research of the maximum (max_R) and the
  -- minimum (Min_R) of R and (N-R).
  -- We have : NCR(N,R) = [N! / Max_R!] / Min_R!;
  -- Then, nCr(N,R) is considered as a fraction, even if its result is an
  -- integer : Temp_NCR_Numerator / Temp_NCR_Denominator
  -- with Temp_NCR_Numerator = N! / Max_R!
  -- and Temp_NCR_Denominateur = Min_R!
  Temp_N : CONSTANT Big_Unsigned := N - R;
  Max_R : CONSTANT Big_Unsigned := Max(R, Temp_N);
  Min_R : CONSTANT Big_Unsigned := Min(R, Temp_N);
  Indice : Big_Unsigned := Big_Unsigned_One;
  Temp_NCR_Numerator : Big_Unsigned := Max_R + Big_Unsigned_One;
  Temp_NCR_Denominator : Big_Unsigned := Big_Unsigned_One;
  Temp_Num, Temp_Den : Big_Unsigned;

  PROCEDURE Simplify(x, y : IN OUT Big_Unsigned) IS
    Temp : CONSTANT Big_Unsigned := GCF(x,y);
  BEGIN
    x := x / Temp;
    y := y / Temp;
  END Simplify;

BEGIN
  IF (N = R) THEN
    RETURN Big_Unsigned_One;
  END IF;
  IF (N < R) THEN
    RETURN Big_Unsigned_Zero;
  END IF;
  -- Now, N > R.
  WHILE (Indice < Min_R) LOOP
    Temp_Num := Max_R + Indice;
    Inc(Temp_Num);
    Temp_Den := Indice;
    Inc(Temp_Den);
    Simplify(Temp_Den, Temp_Num);
    Simplify(Temp_NCR_Denominator, Temp_Num);
    Simplify(Temp_NCR_Numerator, Temp_Den);
    Temp_NCR_Numerator := Temp_NCR_Numerator * Temp_Num;
    Temp_NCR_Denominator := Temp_NCR_Denominator * Temp_Den;
    Inc(Indice);
  END LOOP;
  RETURN Temp_NCR_Numerator;
END NCR;

--******************************************************************************

FUNCTION NPR(N, R : IN Big_Unsigned) RETURN Big_Unsigned IS
  Temp_Result : Big_Unsigned := Big_Unsigned_One;
  Index : Big_Unsigned;
BEGIN
  IF (N < R) THEN
    RETURN Big_Unsigned_Zero;
  END IF;
  IF (R.Last_Significant_Digit = First_Indice) AND THEN
     (R.Mantissa(First_Indice)  = 0) THEN
    RETURN Big_Unsigned_One;
  END IF;
  Index := N - R;
  WHILE (Index < N) LOOP
    Inc(Index);
    Temp_Result := Temp_Result * Index;
  END LOOP;
  RETURN Temp_Result;
END NPR;

--******************************************************************************

FUNCTION K(N, P : IN Big_Unsigned) RETURN Big_Unsigned IS
  Temp : Big_Unsigned;
BEGIN
  -- K(N,P) = NCR(N+P-1);
  -- We have to calculate Temp = N+P-1.
  -- The 3 following lines allows to avoid an overflow, in the
  -- case if N+P = Big_Unsigned_Last + 1 : in this situation,
  -- Temp will be Big_Unsigned_Last.
  Temp := Max(N,P);
  Dec(Temp);
  Temp := Temp + Min(N,P);
  -- Now, Temp = N+P-1.
  RETURN NCR(Temp,P);
END K;

--******************************************************************************

END Unsigned_Number;
