--******************************************************************************
--                                                                            **
--  File:        BIGNUMBE.ADS                                                 **
--  Description: Big_Number package specification                             **
--  Revision:    0.28 (beta version)                                          **
--  Date:        July 8, 2001                                                 **
--  Author:      Jrme Delcourt                                              **
--  Mail:        sikander@club-internet.fr                                    **
--                                                                            **
--  Copyright (c) Jrme Delcourt, 1998, 1999, 2000, 2001                     **
--  62, rue Ngrier                                                           **
--  59000 Lille                                                               **
--  FRANCE                                                                    **
--                                                                            **
--  Permission granted to use for any purpose, provided this copyright        **
--  remains attached and unmodified.                                          **
--                                                                            **
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR            **
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED            **
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       **
--                                                                            **
--******************************************************************************
--                                                                            **
-- Please, don't care of my poor english in the comments (I'm french...).     **
-- Don't hesitate to report any bug, comment, idea or improvement to my       **
-- e-mail, which is sikander@club-internet.fr.                                **
--                                                                            **
-- I tried to give a understandable name to all the functions or variables    **
-- used in this package (the beta-test functions were initially written with  **
-- some names in french). In case of problem with the choosen names, tell it  **
-- to me.                                                                     **
--                                                                            **
-- You can distribute this package, but at no charge at all, and only if      **
-- you specify where you modify it (if you do so). Don't remove the comments. **
--                                                                            **
--******************************************************************************

WITH SYSTEM;
WITH CALENDAR;
WITH TEXT_IO;
WITH ADA.STRINGS.UNBOUNDED;
WITH UStrings;


GENERIC
  TYPE Index_Type IS MOD <>;
  Nb_Item : Index_Type;


PACKAGE Big_Number IS

  --
  -- This package contains two main "sub-packages", one for big unsigned
  -- integer numbers (package Unsigned_Number), another for big signed
  -- numbers (package Signed_Number).
  --
  -- Seven other "sub-packages" are defined :
  --  * Package Conversion, for the conversion of big integers, signed or
  --    unsigned : conversion of one of this type of integers to the other or
  --    to a "LONG_LONG_FLOAT" variable;
  --  * Generic package Generic_Conversion for the conversion of a predefined
  --    type of integer in ADA (one function can be used to convert a modular
  --    type) to a big integer type (signed or unsigned);
  --  * String_Conversion, for the conversion between string of numeric
  --    characters and the types Big_Signed or Big_Unsigned;
  --  * Big_Signed_IO : for the output of a Big_Signed on the screen;
  --  * Big_Unsigned_IO : for the output of a Big_Unsigned on the screen;
  --  * Big_Fraction (complete version only);
  --  * Big_Unsigned_Random (complete version only).
  --
  -- This package is generic. It allows you to use big integers, signed or
  -- unsigned.The mantissa of these numbers is represented by an array of
  -- unsigned integers (the size of these integers is defined by My_Type_Size :
  -- see the note below).
  --
  -- For the instanciation of this package, Index_Type represents the type
  -- of array index, and Nb_Item represents the number of elements in this
  -- array. So you will be able to use numbers coded on My_Type_Size*Nb_Item
  -- bits. Even if Nb_Item can be 1, it should be greater or equal to 3, because
  -- TypeSize is set to the half of the greatest size for an integer :
  -- if Nb_Item is equal to 1 or 2, the corresponding types are usually
  -- already defined by the compiler.
  --
  -- The elements of this array can take every value allowed for a TypeSize
  -- bit number : they are not limited to a power of 10. Some other packages
  -- could use the limitation to a power of 10 in order to facilitate
  -- the IOs of these big numbers, but this has the inconvenient to waste
  -- some memory.
  --
  -- Example of instanciation :
  --
  -- WITH ModInt;     USE ModInt;
  -- WITH BigNumber;
  --
  -- PACKAGE Big_Integers IS NEW BigNumber(Byte, 64);
  --
  -- I suppose that the type Byte is defined in the package ModInt by
  -- TYPE Byte IS MOD 2**8;
  -- FOR Byte'SIZE USE 8;
  --
  -- Then, the package Big_Integers will allow you to use big integers with
  -- a mantissa of (64 * My_Type_Size / 8) bytes.
  --
  --
  -- Remark : be careful when you instanciate this package.
  --
  --  1) Choose very well the parameters of instanciation, because each
  --     big number can occupy a lot of memory. For example, if Nb_Item = 8000,
  --     the mantissa of each variable of the type Big_Unsigned will
  --     take 8000 * (My_Type_Size/8) bytes (the sign is a separate field of
  --     the record).
  --  2) Ideally, Nb_Item should be greater than or equal to 3. Otherwise,
  --     the big integers managed by this package have the same size than
  --     types predefined by your compiler : there will be only one advantage,
  --      the type Fraction.
  --

  TYPE Largest_Unsigned IS MOD SYSTEM.Max_Binary_Modulus;

  My_Type_Size : CONSTANT := Largest_Unsigned'SIZE / 2;

  TYPE My_Type IS MOD 2**My_Type_Size;
  FOR My_Type'SIZE USE My_Type_Size;


--******************************************************************************
--******************************************************************************

  TYPE Big_Signed IS PRIVATE;
  TYPE Big_Unsigned IS PRIVATE;
  TYPE Fraction IS PRIVATE;
 
  ------------------------------
  -- Definition of exceptions --
  ------------------------------

  Big_Number_Overflow : EXCEPTION;
     -- This exception is raised is raised in case of overflow during
     -- an operation on 2 big numbers.

  Division_By_Zero : EXCEPTION;
     -- This exception is raised during a tentative of division by 0 in
     -- a calculus on big numbers.

  Parameter_Error : EXCEPTION;
     -- This exception is raised in case of a bad parameter (for example,
     -- a parameter below 0 with the function Factorial(x : IN Big_Signed).

  
  --------------------------
  -- Constant definitions --
  --------------------------

  Big_Unsigned_Zero : CONSTANT Big_Unsigned;   -- = 0
  Big_Unsigned_One : CONSTANT Big_Unsigned;    -- = 1
  Big_Unsigned_Two : CONSTANT Big_Unsigned;    -- = 2
  Big_Unsigned_First : CONSTANT Big_Unsigned;  -- = 0
  Big_Unsigned_Last : CONSTANT Big_Unsigned;   -- = "Big_Unsigned'LAST"

  Big_Signed_Zero : CONSTANT Big_Signed;   -- = 0
  Big_Signed_One : CONSTANT Big_Signed;    -- = 1
  Big_Signed_First : CONSTANT Big_Signed;  -- = "Big_Signed'FIRST"
  Big_Signed_Last : CONSTANT Big_Signed;   -- = "Big_Signed'LAST"


--******************************************************************************
--******************************************************************************

PACKAGE Conversion IS

  FUNCTION Unsigned2Signed(x : IN Big_Unsigned;
                           Positive : IN BOOLEAN := TRUE) RETURN Big_Signed;

  FUNCTION Signed2Unsigned(x : IN Big_Signed) RETURN Big_Unsigned;


  FUNCTION Big_Unsigned2Long_Long_Float(x : IN Big_Unsigned)
                                        RETURN LONG_LONG_FLOAT;

  FUNCTION Big_Signed2Long_Long_Float(x : IN Big_Signed)
                                      RETURN LONG_LONG_FLOAT;

END Conversion;

 
--******************************************************************************
--******************************************************************************

USE Conversion;

PACKAGE Generic_Conversion IS

  Conversion_Error : EXCEPTION;
     -- This exception is raised if an error occurs during a conversion
     -- (for example if you try to convert an integer below 0 to a
     -- Big_Unsigned).

  ------------------------------------------------------------------------------

  GENERIC
    TYPE Mod_Generic IS MOD <>;

  FUNCTION Mod_Number2Big_Unsigned(x : IN Mod_Generic) RETURN Big_Unsigned;

  ------------------------------------------------------------------------------

  GENERIC
    TYPE Mod_Generic IS MOD <>;

  FUNCTION Mod_Number2Big_Signed(x : IN Mod_Generic;
                                 Positive : IN BOOLEAN := TRUE) RETURN Big_Signed;

  ------------------------------------------------------------------------------

  GENERIC
    TYPE Int_Generic IS RANGE <>;

  FUNCTION Int_Number2Big_Unsigned(x : IN Int_Generic) RETURN Big_Unsigned;

  ------------------------------------------------------------------------------

  GENERIC
    TYPE Int_Generic IS RANGE <>;

  FUNCTION Int_Number2Big_Signed(x : IN Int_Generic) RETURN Big_Signed;

  ------------------------------------------------------------------------------

END Generic_Conversion;

--******************************************************************************
--******************************************************************************

USE Conversion;

PACKAGE Unsigned_Number IS

  --------------------------
  -- Exception definition --
  --------------------------
  
  Big_Unsigned_Substraction_Error : EXCEPTION;
    -- This exception is raised when you try to compute x - y, where y is
    -- greater than x (a Big_Unsigned variable can't have a negative value).

  -----------------
  -- Comparisons --
  -----------------

  FUNCTION "<"(Left, Right : IN Big_Unsigned) RETURN BOOLEAN;
  FUNCTION ">"(Left, Right : IN Big_Unsigned) RETURN BOOLEAN;
  FUNCTION "<="(Left, Right : IN Big_Unsigned) RETURN BOOLEAN;
  FUNCTION ">="(Left, Right : IN Big_Unsigned) RETURN BOOLEAN;

  FUNCTION Is_Null(x : IN Big_Unsigned) RETURN BOOLEAN;

  FUNCTION Min(x, y : IN Big_Unsigned) RETURN Big_Unsigned;
  FUNCTION Max(x, y : IN Big_Unsigned) RETURN Big_Unsigned;

  ----------------------
  -- Common functions --
  ----------------------

  PROCEDURE Inc(x : IN OUT Big_Unsigned);  -- Set x to x + 1.
  PROCEDURE Dec(x : IN OUT Big_Unsigned);  -- Set x to x - 1.

  PROCEDURE Swap1(x, y : IN OUT Big_Unsigned);
  PROCEDURE Swap2(x, y : IN OUT Big_Unsigned);
    -- Swap1 and Swap2 swaps the contents of x and y.
    -- Swap2 is faster than Swap1 when x and y are slightly lower
    -- than Big_Unsigned_Last. Otherwise, you should prefer Swap1.
    -- So examine the usual conditions in your program in order
    -- choose the right function.
  FUNCTION "+"(x, y : IN Big_Unsigned) RETURN Big_Unsigned;
  FUNCTION "-"(x, y : IN Big_Unsigned) RETURN Big_Unsigned;
  FUNCTION "*"(x : IN Big_Unsigned; y : IN My_Type) RETURN Big_Unsigned;
  FUNCTION "*"(x, y : IN Big_Unsigned) RETURN Big_Unsigned;
  FUNCTION "**"(x : IN Big_Unsigned; y : IN Index_Type) RETURN Big_Unsigned;
  FUNCTION "**"(x : IN Big_Unsigned; y : IN My_Type) RETURN Big_Unsigned;
    -- This function returns (x ** y).
    -- If the result is too big, the function raises the
    -- exception Big_Number_Overflow.
    -- If y = 0, this function returns Big_Unsigned_One.
  FUNCTION "**"(x, y : IN Big_Unsigned) RETURN Big_Unsigned;
    -- This function returns (x ** y).
    -- If the result is too big, the function raises the
    -- exception Big_Number_Overflow.
    -- If y = Big_Unsigned_Zero, this function returns Big_Unsigned_One.
  FUNCTION "**"(x : IN Big_Unsigned;
                y : IN Big_Signed) RETURN Big_Unsigned;
    -- This function returns (x ** y).
    -- If the result is too big, the function raises the
    -- exception Big_Number_Overflow.
    -- If y = Big_Signed_Zero, this function returns Big_Unsigned_One.
    -- If y < Big_Signed_Zero, this function returns Big_Unsigned_Zero.
  FUNCTION "/"(x, y : IN Big_Unsigned) RETURN Big_Unsigned;
    -- Returns x/y
  FUNCTION "MOD"(x, y : IN Big_Unsigned) RETURN Big_Unsigned;
    -- Returns x MOD y
  FUNCTION "/"(x : IN Big_Unsigned; y : IN My_Type) RETURN Big_Unsigned;
    -- Returns x/y
  FUNCTION "MOD"(x : IN Big_Unsigned; y : IN My_Type) RETURN My_Type;
    -- Returns x MOD y
  PROCEDURE Big_Div(x, y : IN Big_Unsigned;
                    Quot, Remainder : OUT Big_Unsigned);
    -- This procedure returns Quot and Remainder, so :
    -- x = Quot * y + Remainder, and Quot * y < x < (Quot + 1) * y.
  PROCEDURE Short_Div(x : IN Big_Unsigned;
                      y : IN My_Type;
                      Quot : OUT Big_Unsigned;
                      Remainder : OUT My_Type);
    -- This procedure returns Quot and Remainder, so :
    -- x = Quot * y + Remainder, and Quot * y < x < (Quot + 1) * y.

  -----------------------
  -- Special functions --
  -----------------------

  FUNCTION Number_of_Bytes(x : IN Big_Unsigned) RETURN Largest_Unsigned;
    -- This function returns the number of bytes in the mantissa of x.
  FUNCTION Number_of_Bits(x : IN Big_Unsigned) RETURN Largest_Unsigned;
    -- This function returns the number of bits in the mantissa of x.

  FUNCTION Square_Root(x : IN Big_Unsigned) RETURN Big_Unsigned;
  FUNCTION Factorial(x : IN Big_Unsigned) RETURN Big_Unsigned;
    -- This function returns x!.
  FUNCTION GCF(x, y : IN Big_Unsigned) RETURN Big_Unsigned;
    -- This function returns the greatest common factor of x and y.
  FUNCTION LCM(x, y : IN Big_Unsigned) RETURN Big_Unsigned;
    -- This function returns the least common factor of x and y.

  FUNCTION "AND"(x, y : IN Big_Unsigned) RETURN Big_Unsigned;
    -- This function returns x and y.
  FUNCTION "XOR"(x, y : IN Big_Unsigned) RETURN Big_Unsigned;
    -- This function returns x xor y.


  FUNCTION NCR(N, R : IN Big_Unsigned) RETURN Big_Unsigned;
    -- NCR = N! / [R! * (N-R)!]
  FUNCTION NPR(N, R : IN Big_Unsigned) RETURN Big_Unsigned;
  FUNCTION K(N, P : IN Big_Unsigned) RETURN Big_Unsigned;
    -- K(N, P) = NCR(N+P-1,P)

END Unsigned_Number;

--******************************************************************************
--******************************************************************************

USE Conversion;


PACKAGE Signed_Number IS

  -----------------
  -- Comparisons --
  -----------------

  FUNCTION "<"(Left, Right : IN Big_Signed) RETURN BOOLEAN;
  FUNCTION ">"(Left, Right : IN Big_Signed) RETURN BOOLEAN;
  FUNCTION "<="(Left, Right : IN Big_Signed) RETURN BOOLEAN;
  FUNCTION ">="(Left, Right : IN Big_Signed) RETURN BOOLEAN;

  FUNCTION Is_Null(x : IN Big_Signed) RETURN BOOLEAN;

  FUNCTION Min(x, y : IN Big_Signed) RETURN Big_Signed;
  FUNCTION Max(x, y : IN Big_Signed) RETURN Big_Signed;

  ----------------------
  -- Common functions --
  ----------------------

  PROCEDURE Inc(x : IN OUT Big_Signed);  -- Set x to x + 1.
  PROCEDURE Dec(x : IN OUT Big_Signed);  -- Set x to x - 1.

  PROCEDURE Swap1(x, y : IN OUT Big_Signed);
  PROCEDURE Swap2(x, y : IN OUT Big_Signed);
    -- Swap1 and Swap2 swaps the contents of x and y.
    -- Swap2 is faster than Swap1 when the absolute value of x and y are
    -- slightly lower than Big_Signed_Last. Otherwise, you should prefer Swap1.
    -- So examine the usual conditions in your program in order
    -- choose the right function.
  FUNCTION "+"(x, y : IN Big_Signed) RETURN Big_Signed;
  FUNCTION "-"(x, y : IN Big_Signed) RETURN Big_Signed;
  FUNCTION "*"(x, y : IN Big_Signed) RETURN Big_Signed;
  FUNCTION "**"(x : IN Big_Signed; y : IN Index_Type) RETURN Big_Signed;
  FUNCTION "**"(x : IN Big_Signed; y : IN My_Type) RETURN Big_Signed;
    -- This function returns (x ** y).
    -- If the result is too big, the function raises the
    -- exception Big_Number_Overflow.
    -- If y = 0, this function returns Big_Signed_One.
    -- By definition, Index_Type is a modular type, so y can't be negative.
  FUNCTION "**"(x : IN Big_Signed;
                y : IN Big_Unsigned) RETURN Big_Signed;
    -- This function returns (x ** y).
    -- If the result is too big, the function raised the
    -- exception Big_Number_Overflow.
    -- If y = Big_Unsigned_Zero, this function returns Big_Signed_One.
    -- By definition, y can't be negative.
  FUNCTION "**"(x, y : IN Big_Signed) RETURN Big_Signed;
    -- This function returns (x ** y) if y is positive.
    -- If y is < 0, the function returns Big_Unsigned_Zero.
    -- If y = Big_Signed_Zero, this function returns Big_Signed_One;
    -- If the result is too big, the function raises the
    -- exception Big_Number_Overflow.
  FUNCTION "/"(x, y : IN Big_Signed) RETURN Big_Signed;
  FUNCTION "MOD"(x, y : IN Big_Signed) RETURN Big_Signed;
    -- This function returns a number, always positive, so that
    -- (x - Result) is an exact multiple of y.
  FUNCTION "/"(x : IN Big_Signed; y : IN My_Type) RETURN Big_Signed;
  FUNCTION "MOD"(x : IN Big_Signed; y : IN My_Type) RETURN My_Type;
    -- This function returns a number, always positive, so that
    -- (x - Result) is an exact multiple of y.
  FUNCTION Opposite(x : IN Big_Signed) RETURN Big_Signed;
    -- Returns -x.
  FUNCTION Absolute(x : IN Big_Signed) RETURN Big_Signed;
    -- Returns -x if x < 0.
    -- Returns x if x >=0.
  PROCEDURE Big_Div(x, y : IN Big_Signed;
                    Quot, Remainder : OUT Big_Signed);
    -- This procedure returns Quot and Remainder so that :
    -- x = Quot * y + Remainder,
    -- and  Absolute(Quot * y) < Absolute(x) < Absolute(Quot + 1) * y)

  -----------------------
  -- Special functions --
  -----------------------

  FUNCTION Number_of_Bytes(x : IN Big_Signed) RETURN Largest_Unsigned;
    -- This function returns the number of bytes in the mantissa of x.
  FUNCTION Number_of_Bits(x : IN Big_Signed) RETURN Largest_Unsigned;
    -- This function returns the number of bits in the mantissa of x.

  FUNCTION Square_Root(x : IN Big_Signed) RETURN Big_Signed;
  FUNCTION Factorial(x : IN Big_Signed) RETURN Big_Signed;
    -- This function returns x!.
    -- If x is strictly lower than 0, the exception Parameter_Error is raised.
  FUNCTION GCF(x, y : IN Big_Signed) RETURN Big_Signed;
    -- This function returns the greatest common factor of x and y.
  FUNCTION LCM(x, y : IN Big_Signed) RETURN Big_Signed;
    -- This function returns the least common factor of x and y.


END Signed_Number;

--******************************************************************************
--******************************************************************************

USE Generic_Conversion;
USE Unsigned_Number;
USE ADA.STRINGS.UNBOUNDED;
USE UStrings;


PACKAGE String_Conversion IS

  Conversion_Error : EXCEPTION;
     -- This exception is raised if an error occurs during a conversion.

  Base_Error : EXCEPTION;
  -- This exception is raised when the base parameter is greater than 36,
  -- or when this parameter is lower than 2.

  FUNCTION String2Unsigned(x : IN STRING) RETURN Big_Unsigned;
  FUNCTION String2Signed(x : IN STRING) RETURN Big_Signed;
  -- These two functions convert an integer in base 10, given as a string,
  -- respectively in a Big_Unsigned or a Big_Signed.

  FUNCTION UString2Unsigned(x : IN UNBOUNDED_STRING) RETURN Big_Unsigned;
  FUNCTION UString2Signed(x : IN UNBOUNDED_STRING) RETURN Big_Signed;
  -- These two functions convert an integer in base 10, given as a string,
  -- respectively in a Big_Unsigned or a Big_Signed.


  FUNCTION Big_Unsigned2UString(x : IN Big_Unsigned;
                               Base : IN My_Type := 10)
                               RETURN UNBOUNDED_STRING;
    -- This function returns the UNBOUNDED_STRING corresponding to
    -- the Big_Unsigned number x, using the base Base. By default, Base = 10.
  FUNCTION Big_Signed2UString(x : IN Big_Signed;
                             Base : IN My_Type := 10)
                             RETURN UNBOUNDED_STRING;
    -- This function returns the UNBOUNDED_STRING corresponding to
    -- the Big_Signed number x, using the base Base. By default, Base = 10.

END String_Conversion;


--******************************************************************************
--******************************************************************************

USE Unsigned_Number;
USE String_Conversion;
USE TEXT_IO;
USE UStrings;


PACKAGE Big_Unsigned_IO IS

  PROCEDURE GET_LINE(x : OUT Big_Unsigned);
    -- This procedure waits for a Big_Unsigned, to be entered in base 10 on the
    -- keyboard.
  PROCEDURE GET_LINE(File : IN FILE_TYPE; x : OUT Big_Unsigned);
    -- This procedure reads a Big_Unsigned in a text file.

  PROCEDURE PUT(x : IN Big_Unsigned;
                Base : IN My_Type := 10);
    -- This procedure writes on screen the variable x, using the
    -- indicated base (by default, it's base 10).
  PROCEDURE PUT(File : IN FILE_TYPE;
                x : IN Big_Unsigned;
                Base : IN My_Type := 10);
    -- This procedure writes the variable x in a text file, using the
    -- indicated base (by default, it's base 10).

  PROCEDURE PUT_LINE(x : IN Big_Unsigned;
                     Base : IN My_Type := 10);
    -- This procedure writes on screen the variable x, using the
    -- indicated base (by default, it's base 10), and then go to the next line.
  PROCEDURE PUT_LINE(File : IN FILE_TYPE;
                     x : IN Big_Unsigned;
                     Base : IN My_Type := 10);
    -- This procedure writes the variable x in a text file, using the
    -- indicated base (by default, it's base 10), and then go to the next line.


END Big_Unsigned_IO;

--******************************************************************************
--******************************************************************************

USE Signed_Number;
USE String_Conversion;
USE TEXT_IO;
USE UStrings;


PACKAGE Big_Signed_IO IS

  PROCEDURE GET_LINE(x : OUT Big_Signed);
    -- This procedure waits a Big_Signed, to be entered in base 10 on the
    -- keyboard.
  PROCEDURE GET_LINE(File : IN FILE_TYPE; x : OUT Big_Signed);
    -- This procedure reads a Big_signed in a text file.

  PROCEDURE PUT(x : IN Big_Signed;
                Base : IN My_Type := 10);
    -- This procedure writes on screen the variable x, using the
    -- indicated base (by default, it's base 10).
  PROCEDURE PUT(File : IN FILE_TYPE;
                x : IN Big_Signed;
                Base : IN My_Type := 10);
    -- This procedure writes the variable x in a text file, using the
    -- indicated base (by default, it's base 10).

  PROCEDURE PUT_LINE(x : IN Big_Signed;
                     Base : IN My_Type := 10);
    -- This procedure writes on screen the variable x, using the
    -- indicated base (by default, it's base 10), and then go to the next line.
  PROCEDURE PUT_LINE(File : IN FILE_TYPE;
                     x : IN Big_Signed;
                     Base : IN My_Type := 10);
    -- This procedure writes the variable x in a text file, using the
    -- indicated base (by default, it's base 10), and then go to the next line.

END Big_Signed_IO;

--******************************************************************************
--******************************************************************************

USE Unsigned_Number;
USE Big_Unsigned_IO;
USE ADA.STRINGS.UNBOUNDED;
USE UStrings;

PACKAGE Big_Fraction IS

-- Remark : for all the operations on fractions, the function or the procedure
-- return the simplified form of the result. For example, 1/4 + 1/4 = 2/4,
-- but the function "+" returns 1/2 (2/4 = 1/2). The simplification of the
-- result requires some more execution time, but it allows to avoid some
-- overflows.

  Fraction_Overflow : EXCEPTION;

  FUNCTION CreateFraction(Numerator, Denominator : IN Big_Unsigned;
                          Positive : IN BOOLEAN := TRUE) RETURN Fraction;
  FUNCTION CreateFraction(Numerator, Denominator : IN Big_Signed) RETURN Fraction;

  FUNCTION Get_Numerator(x : IN Fraction) RETURN Big_Unsigned;
  FUNCTION Get_Denominator(x : IN Fraction) RETURN Big_Unsigned;
  FUNCTION Is_Positive(x : IN Fraction) RETURN BOOLEAN;

  FUNCTION Inverse_Fraction(x : IN Fraction) RETURN Fraction;
    -- This function returns 1/x.

    -- No comments for the following functions...
  FUNCTION "*"(x, y : IN Fraction) RETURN Fraction;
  FUNCTION "/"(x, y : IN Fraction) RETURN Fraction;
  FUNCTION "+"(x, y : IN Fraction) RETURN Fraction;
  FUNCTION "-"(x, y : IN Fraction) RETURN Fraction;
  FUNCTION "**"(x : IN Fraction; y : IN Big_Unsigned) RETURN Fraction;
  FUNCTION "**"(x : IN Fraction; y : IN Big_Signed) RETURN Fraction;
  FUNCTION "**"(x : IN Fraction; y : IN Index_Type) RETURN Fraction;
  FUNCTION "**"(x : IN Fraction; y : IN My_Type) RETURN Fraction;
  FUNCTION "*"(x : IN Fraction; y : IN Big_Unsigned) RETURN Fraction;
  FUNCTION "*"(x : IN Fraction; y : IN Big_Signed) RETURN Fraction;
  FUNCTION "/"(x : IN Fraction; y : IN Big_Unsigned) RETURN Fraction;
  FUNCTION "/"(x : IN Fraction; y : IN Big_Signed) RETURN Fraction;

  FUNCTION "<"(Left, Right : IN Fraction) RETURN BOOLEAN;
  FUNCTION ">"(Left, Right : IN Fraction) RETURN BOOLEAN;
  FUNCTION "<="(Left, Right : IN Fraction) RETURN BOOLEAN;
  FUNCTION ">="(Left, Right : IN Fraction) RETURN BOOLEAN;

  FUNCTION Opposite(x : IN Fraction) RETURN Fraction;
    -- Returns -x.
  FUNCTION Absolute(x : IN Fraction) RETURN Fraction;
    -- Returns -x if x < 0.
    -- Returns x if x >=0.
  FUNCTION Is_Null(x : IN Fraction) RETURN BOOLEAN;

  PROCEDURE PUT(x : IN Fraction);
    -- This procedure puts x on the screen.
  PROCEDURE PUT_LINE(x : IN Fraction);
    -- This procedure puts x on the screen and then go to the next line.

  PROCEDURE PUT(File : IN FILE_TYPE; x : IN Fraction);
    -- This procedure puts x in a text file.
  PROCEDURE PUT_LINE(File : IN FILE_TYPE; x : IN Fraction);
    -- This procedure puts x in a text file and then go to the next line.

  FUNCTION UString2Fraction(x : IN UNBOUNDED_STRING) RETURN Fraction;

  PROCEDURE GET_LINE(x : OUT Fraction);
    -- This procedure gets x and then go to the next line.
    -- The exception Parameter_Error is raised in case of error
    -- (for example if there are 2 or more '/').
  PROCEDURE GET_LINE(File : IN FILE_TYPE; x : OUT Fraction);
    -- This procedure gets x in a text file and then go to the next line.
    -- The exception Parameter_Error is raised in case of error
    -- (for example if there are 2 or more '/').

  PROCEDURE Float_in_UString2Fraction(x : IN UNBOUNDED_STRING;
                                      Nb_Periods : IN POSITIVE;
                                      Result : OUT Fraction;
                                      Error : OUT BOOLEAN);
  -- This procedure tries to get the fraction corresponding to
  -- the float point number contained in the string x.
  -- INPUT :
  --    - x : the string
  --    - Nb_Periods :  the minimal number of periods to find in the string
  --                    if you want the function to consider that number as a
  --                    fraction. For example, the string "5.345345345345"
  --                    contains the period "345" four times.
  --                    The string "5.3453453453457" is considered to contain 
  --                    no period, even if there are 4 repetitions of "345",
  --                    because these repetitions are followed by "7", which
  --                    DOES NOT belong to this sequence.
  --
  -- OUTPUT :
  --    - Result : the fraction (only if Error = FALSE)
  --    - Error : TRUE if x isn't a valid string or if Nb_Periods = 1,
  --              FALSE otherwise.
  --
  -- An valid string is : "(+/-)nnn.nnnnn[(E:e(+/-))nnnnn]".
  -- The mantissa is composed of any decimal digit and may have only one sign
  -- at the beginning. The mantissa may be a real number.
  -- If there is an exponent, it must start with E or e, may have only one
  -- sign and MUST be an integer.
  -- In the mantissa or in the exponent, the spaces and the character "_"
  -- are accepted, and they will have no consequence concerning the result.
  --
  -- Remarks :
  -- 1) If Nb_Periods = 4, 1.353535 (3 repetitions of "35") will be considered
  --    as the exact value of the fraction 270_707 / 200_000, but with the
  --    number 1.35353535 (which has 4 repetitions of "35"), the result will be
  --    the fraction 134/99.
  -- 2) The parameter Nb_Periods must always be equal or greater than 2.
  --    There is no upper limit for Nb_Periods.
  -- 3) Be careful : if there is an error and if you don't check it
  --    (by regarding the boolean returned by the procedure), you
  --    may have a DIVISION_BY_ZERO exception in a further operation,
  --    because, in this case, the function sets the result to 0/0.
  --    That was not done to plague you, but just to help you to detect
  --    errors in your programs : I think that, during the tests,
  --    it's better to have an exception than a bad result which you may
  --    never detect. Furthermore, exceptions can be handled.
  --


END Big_Fraction;

--******************************************************************************
--******************************************************************************

USE CALENDAR;
USE Unsigned_Number;


PACKAGE Big_Unsigned_Random IS

Instanciation_Error : EXCEPTION;
  -- This exception is raised when Index_Type'LAST > (2**TypeSize - 1) / 125.
  -- It is due to the implementation : in the package Big_Unsigned_Random,
  -- ForMultiplication'LAST / 125 must be greater than Index_Type'LAST.

PROCEDURE Randomize;
  -- This procedure resets the pseudo-random number generator.
  -- It is automatically executed at the start of the program.

FUNCTION Random RETURN Big_Unsigned;
  -- This function returns a Big_Unsigned between Big_Unsigned_Zero
  -- and Big_Unsigned_Last.

FUNCTION Random(x : IN Big_Unsigned) RETURN Big_Unsigned;
  -- This function returns a Big_Unsigned between Big_Unsigned_Zero
  -- and x. If x = Big_Unsigned_Zero, it returns Big_Unsigned_Zero.

FUNCTION Random(x, y : IN Big_Unsigned) RETURN Big_Unsigned;
  -- This function returns a Big_Unsigned between Min(x,y) and Max(x,y).


END Big_Unsigned_Random;

--******************************************************************************
--******************************************************************************
--******************************************************************************
--******************************************************************************

PRIVATE

  TYPE Mantiss IS ARRAY(0..Nb_Item-1) OF My_Type;

  TYPE Big_Signed IS RECORD
                       Positive : BOOLEAN;
                       Last_Significant_Digit : Index_Type;
                       Mantissa : Mantiss;
                     END RECORD;

  TYPE Big_Unsigned IS RECORD
                         Last_Significant_Digit : Index_Type;
                         Mantissa : Mantiss;
                       END RECORD;

  TYPE Fraction IS RECORD
                     Positive    : BOOLEAN;
                     Numerator   : Big_Unsigned;
                     Denominator : Big_Unsigned;
                   END RECORD;

  --
  -- By convention, the lower digit is in Mantissa(0), and the greater one
  -- is in Mantissa(Nb_Elt-1). This looks like a "BigEndian" coding.
  --

  Big_Unsigned_Zero : CONSTANT Big_Unsigned
                                    := (Last_Significant_Digit => 0,
                                        Mantissa => (OTHERS => My_Type(0)));
  Big_Unsigned_One : CONSTANT Big_Unsigned
                                    := (Last_Significant_Digit => 0,
                                        Mantissa => (1, OTHERS => My_Type(0)));
  Big_Unsigned_Two : CONSTANT Big_Unsigned
                                    := (Last_Significant_Digit => 0,
                                        Mantissa => (2, OTHERS => My_Type(0)));
  Big_Unsigned_First : CONSTANT Big_Unsigned
                                    := Big_Unsigned_Zero;
  Big_Unsigned_Last : CONSTANT Big_Unsigned
                                    := (Last_Significant_Digit => Nb_Item - 1,
                                        Mantissa => (OTHERS => My_Type'LAST));

  Big_Signed_Zero : CONSTANT Big_Signed
                                    := (Positive => TRUE,
                                        Last_Significant_Digit => 0,
                                        Mantissa => (OTHERS => My_Type(0)));
  Big_Signed_One : CONSTANT Big_Signed
                                    := (Positive => TRUE,
                                        Last_Significant_Digit => 0,
                                        Mantissa => (1, OTHERS => My_Type(0)));
  Big_Signed_First : CONSTANT Big_Signed
                                    := (Positive => FALSE,
                                        Last_Significant_Digit => Nb_Item - 1,
                                        Mantissa => (OTHERS => My_Type'LAST));
  Big_Signed_Last : CONSTANT Big_Signed
                                    := (Positive => TRUE,
                                        Last_Significant_Digit => Nb_Item - 1,
                                        Mantissa => (OTHERS => My_Type'LAST));

--******************************************************************************
--******************************************************************************

END Big_Number;
