--
-- Permission to use, copy, modify, and distribute this software and its
-- documentation for any purpose and without fee is hereby granted,
-- provided that the above copyright and authorship notice appear in all
-- copies and that both that copyright notice and this permission notice
-- appear in supporting documentation.
-- 
-- The ARA makes no representations about the suitability of this software
-- for any purpose.  It is provided "as is" without express
-- or implied warranty.
-- 
--******************************************************************************
--******************************************************************************

--
-- Copyright (C) 1996 Ada Resource Association (ARA), Columbus, Ohio.
-- Author: David A. Wheeler
--

PACKAGE BODY UStrings IS

  Input_Line_Buffer_Length : CONSTANT := 1024;
    -- If an input line is longer, Get_Line will recurse to read in the line.


PROCEDURE Swap(Left, Right : IN OUT UNBOUNDED_STRING) IS
  -- Implement Swap.  This is the portable but slow approach.
  Temporary : CONSTANT UNBOUNDED_STRING := Left;
BEGIN
  Left := Right;
  Right := Temporary;
END Swap;


FUNCTION Empty(S : IN UNBOUNDED_STRING) RETURN BOOLEAN IS
  -- returns True if Length(S)=0.
BEGIN
 RETURN (LENGTH(S) = 0);
END Empty;
 

------------------------------------------------------------------------
-- Implement UNBOUNDED_STRING I/O by calling Text_IO String routines. --
------------------------------------------------------------------------


  -- Get_Line gets a line of text, limited only by the maximum number of
  -- characters in an UNBOUNDED_STRING.  It reads characters into a buffer
  -- and if that isn't enough, recurses to read the rest.

PROCEDURE Get_Line(File : IN File_Type; Item : OUT UNBOUNDED_STRING) IS

  FUNCTION More_Input RETURN UNBOUNDED_STRING IS
    Input : STRING(1..Input_Line_Buffer_Length);
    Last  : NATURAL;
  BEGIN
    GET_LINE(File, Input, Last);
    IF (Last < Input'LAST) THEN
      RETURN To_Unbounded_String(Input(1..Last));
    ELSE
      RETURN To_Unbounded_String(Input(1..Last)) & More_Input;
    END IF;
  END More_Input;

BEGIN
  Item := More_Input;
END Get_Line;


PROCEDURE Get_Line(Item : OUT UNBOUNDED_STRING) IS
BEGIN
  Get_Line(Current_Input, Item);
END Get_Line;


PROCEDURE Put(File : IN File_Type; Item : IN UNBOUNDED_STRING) IS
BEGIN
  Put(File, To_String(Item));
END Put;


PROCEDURE Put(Item : IN UNBOUNDED_STRING) IS
BEGIN
  PUT(Current_Output, To_String(Item));
END Put;


PROCEDURE Put_Line(File : IN File_Type; Item : IN UNBOUNDED_STRING) IS
BEGIN
  PUT(File, Item);
  NEW_LINE(File);
END Put_Line;


PROCEDURE Put_Line(Item : IN UNBOUNDED_STRING) IS
BEGIN
  PUT(Current_Output, Item);
  NEW_LINE;
END Put_Line;


--*******************
--* Other functions *
--*******************

FUNCTION Copy(s : IN UNBOUNDED_STRING;
              Index : IN POSITIVE;
              Size : IN NATURAL)
              RETURN UNBOUNDED_STRING IS
  Temp_Result : UNBOUNDED_STRING := NULL_UNBOUNDED_STRING;
  Source_Length : CONSTANT NATURAL := LENGTH(s);
BEGIN
  IF (Size = 0) THEN
    RETURN Temp_Result;
  END IF;
  DECLARE
    Temp_Index : POSITIVE := Index;
    Result_Size : NATURAL := 0;
    Number : POSITIVE;
  BEGIN
    IF (POSITIVE(Size) + Index > Source_Length) THEN
      Number := Source_Length - Index + 1;
    ELSE
      Number := POSITIVE(Size);
    END IF;
    FOR I IN 1..Number LOOP
      Temp_Result := Temp_Result & Element(S, Temp_Index);
      Temp_Index := Temp_Index + 1;
    END LOOP;
    RETURN Temp_Result;
  END;
END Copy;

--*******************************
--* Character related functions *
--*******************************

FUNCTION Del_Spaces(Item : IN UNBOUNDED_STRING) RETURN UNBOUNDED_STRING IS
  Result : UNBOUNDED_STRING := Null_Unbounded_String;
  Item_Length : CONSTANT NATURAL := LENGTH(Item);
  Temp_Char : CHARACTER;
BEGIN
  FOR I IN 1..Item_Length LOOP
    Temp_Char := ELEMENT(Item, I);
    IF (Temp_Char /= ' ') THEN
      Result := Result & Temp_Char;
    END IF;
  END LOOP;
  RETURN Result;
END Del_Spaces;


FUNCTION Del_Character(Item : IN UNBOUNDED_STRING;
                       Char : IN CHARACTER)
                      RETURN UNBOUNDED_STRING IS
  Result : UNBOUNDED_STRING := Null_Unbounded_String;
  Item_Length : CONSTANT NATURAL := LENGTH(Item);
  Temp_Char : CHARACTER;
BEGIN
  FOR I IN 1..Item_Length LOOP
    Temp_Char := ELEMENT(Item, I);
    IF (Temp_Char /= Char) THEN
      Result := Result & Temp_Char;
    END IF;
  END LOOP;
  RETURN Result;
END Del_Character;


FUNCTION Char_Count(S : IN UNBOUNDED_STRING;
                    Char : IN CHARACTER)
                    RETURN NATURAL IS
  Temp_Result : NATURAL := 0;
  String_Length : CONSTANT NATURAL := LENGTH(S);
BEGIN
  FOR I IN 1..String_Length LOOP
    IF (ELEMENT(S, I) = Char) THEN
      Temp_Result := Temp_Result + 1;
    END IF;
  END LOOP;
  RETURN Temp_Result;
END Char_Count;


FUNCTION First_Matching_Char_Position(S : IN UNBOUNDED_STRING;
                                      Char : IN CHARACTER)
                                      RETURN NATURAL IS
  String_Length : CONSTANT NATURAL := LENGTH(S);
  I : NATURAL := 1;
BEGIN
  LOOP
    EXIT WHEN (I > String_Length);
    IF (ELEMENT(S, I) = Char) THEN
      RETURN I;
    END IF;
    I := I + 1;
  END LOOP;
  RETURN 0;
END First_Matching_Char_Position;


FUNCTION Last_Matching_Char_Position(S : IN UNBOUNDED_STRING;
                                     Char : IN CHARACTER)
                                     RETURN NATURAL IS
  String_Length : CONSTANT NATURAL := LENGTH(S);
  I : NATURAL := String_Length;
BEGIN
  LOOP
    EXIT WHEN (I = 0);
    IF (ELEMENT(S, I) = Char) THEN
      RETURN I;
    END IF;
    I := I - 1;
  END LOOP;
  RETURN 0;
END Last_Matching_Char_Position;


FUNCTION Char_Replace(S : IN UNBOUNDED_STRING;
                      Char_to_be_replaced, New_Char : IN CHARACTER)
                      RETURN UNBOUNDED_STRING IS
  Temp_Result : UNBOUNDED_STRING := NULL_UNBOUNDED_STRING;
  Temp : UNBOUNDED_STRING := S;
  I : NATURAL;
BEGIN
  I := First_Matching_Char_Position(S, Char_to_be_replaced);
  WHILE (I /= 0) LOOP
    Temp_Result := Copy(Temp, 1, I-1) & New_Char;
    Temp := Copy(Temp, I + 1, LENGTH(Temp) - I);
    I := First_Matching_Char_Position(Temp, Char_to_be_replaced);
  END LOOP;
  Temp_Result := Temp_Result & Temp;
  RETURN Temp_Result;
END Char_Replace;

--******************************************************************************
--******************************************************************************

END Ustrings;

--
-- Permission to use, copy, modify, and distribute this software and its
-- documentation for any purpose and without fee is hereby granted,
-- provided that the above copyright and authorship notice appear in all
-- copies and that both that copyright notice and this permission notice
-- appear in supporting documentation.
-- 
-- The ARA makes no representations about the suitability of this software
-- for any purpose.  It is provided "as is" without express
-- or implied warranty.
-- 
