--******************************************************************************
--******************************************************************************

--
-- Copyright (C) 1996 Ada Resource Association (ARA), Columbus, Ohio.
-- Author: David A. Wheeler
--
-- The function COPY and all the character related functions have been
-- added by Jerome DELCOURT.
--

WITH TEXT_IO, ADA.STRINGS.UNBOUNDED;
USE TEXT_IO, ADA.STRINGS.UNBOUNDED;

PACKAGE UStrings IS

  -- This package provides a simpler way to work with type
  -- Unbounded_String, since this type will be used very often.
  -- Most users will want to ALSO with "ADA.STRINGS.UNBOUNDED".
  -- Ideally this would be a child package of "ADA.STRINGS.UNBOUNDED".
  --

  -- This package provides the following simplifications:
  --  + Shortens the type name from "UNBOUNDED_STRING" to "Ustring".
  --  + Creates shorter function names for To_Unbounded_String, i.e.
  --    To_Ustring(U) and U(S).  "U" is not a very readable name, but
  --    it's such a common operation that a short name seems appropriate
  --    (this function is needed every time a String constant is used).
  --    It also creates S(U) as the reverse of U(S).
  --  + Adds other subprograms, currently just "Swap".
  --  + Other packages can use this package to provide other simplifications.

  SUBTYPE Ustring IS UNBOUNDED_STRING;

  FUNCTION To_Ustring(Source : IN STRING) RETURN UNBOUNDED_STRING
                                          RENAMES To_Unbounded_String;
  FUNCTION U(Source : IN STRING) RETURN UNBOUNDED_STRING
                                 RENAMES To_Unbounded_String;
  FUNCTION S(Source : IN UNBOUNDED_STRING) RETURN STRING
                                           RENAMES To_String;

  -- "Swap" is important for reuse in some other packages, so we'll define it.

  PROCEDURE Swap(Left, Right : IN OUT UNBOUNDED_STRING);


  FUNCTION Empty(S : UNBOUNDED_STRING) RETURN BOOLEAN;
   -- returns True if Length(S)=0.
  PRAGMA INLINE(Empty);

  --*****************
  --* I/O Routines. *
  --*****************

  PROCEDURE Get_Line(File : IN File_Type; Item : OUT UNBOUNDED_STRING);
  PROCEDURE Get_Line(Item : OUT UNBOUNDED_STRING);

  PROCEDURE Put(File : IN File_Type; Item : IN UNBOUNDED_STRING);
  PROCEDURE Put(Item : IN UNBOUNDED_STRING);

  PROCEDURE Put_Line(File : IN File_Type; Item : IN UNBOUNDED_STRING);
  PROCEDURE Put_Line(Item : IN UNBOUNDED_STRING);


  FUNCTION Copy(s : IN UNBOUNDED_STRING;
                Index : IN POSITIVE;
                Size : IN NATURAL)
                RETURN UNBOUNDED_STRING;
  -- This function returns the unbounded string contained in s, which starts
  -- at the character number Index, and which has Size character. If the string
  -- is not long enough, the result may have less the Size character.
  -- If s is empty, or if Index if greater than the size of the string s,
  -- the result is an empty string.

  -- *******************************
  -- * Character related functions *
  -- *******************************

  FUNCTION Del_Spaces(Item : IN UNBOUNDED_STRING) RETURN UNBOUNDED_STRING;
    -- Removes all the spaces from the strings.
  FUNCTION Del_Character(Item : IN UNBOUNDED_STRING;
                         Char : IN CHARACTER)
                         RETURN UNBOUNDED_STRING;
    -- Removes all the occurences of Char from the string.

  FUNCTION Char_Count(S : IN UNBOUNDED_STRING;
                      Char : IN CHARACTER)
                      RETURN NATURAL;
    -- Counts the number of Char in this string.

  FUNCTION First_Matching_Char_Position(S : IN UNBOUNDED_STRING;
                                        Char : IN CHARACTER)
                                        RETURN NATURAL;
    -- Finds the first occurence of the character in the string.
    -- Returns 0 if none found.

  FUNCTION Last_Matching_Char_Position(S : IN UNBOUNDED_STRING;
                                       Char : IN CHARACTER)
                                       RETURN NATURAL;
    -- Finds the last occurence of the character in the string.
    -- Returns 0 if none found.

  FUNCTION Char_Replace(S : IN UNBOUNDED_STRING;
                        Char_to_be_replaced, New_char : IN CHARACTER)
                        RETURN UNBOUNDED_STRING;
    -- Counts the number of Char in this string.

END UStrings;
